#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include "buf.h"
extern int vsnprintf(char *buf, size_t bufsize, const char *__restrict fmt, va_list args);

static inline size_t av_log2(size_t x) {
    size_t ret;
    __asm__ volatile("or $1, %1 \n\t"
                     "bsr %1, %0 \n\t"
                    :"=a"(ret) : "D"(x));
    return ret;
}

#if defined(__linux)
void *mremap(void *old_address, size_t old_size, size_t new_size, int flags);
#endif

static inline void *
_mem_alloc(size_t size)
/* Allocate a "large" object */
{
  void *hdr_ptr = NULL;
#if defined(_WIN32) || defined(_WIN64)
  NtAllocateVirtualMemory(((HANDLE)-1), (void**)&hdr_ptr, 0, &size, MEM_COMMIT|MEM_RESERVE, PAGE_READWRITE);
#else
  hdr_ptr = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON|MAP_NORESERVE, -1, 0);
#endif
  return hdr_ptr;
}

static inline void *
_mem_realloc(void *ptr, size_t old_size, size_t new_size)
/* Allocate a "large" object */
{
  void *hdr_ptr = NULL;
#if defined(__linux__)
  hdr_ptr = mremap(ptr, old_size, new_size, 1);
#else
#if defined(_WIN32) || defined(_WIN64)
  NtAllocateVirtualMemory(((HANDLE)-1), (void**)&hdr_ptr, 0, &new_size, MEM_COMMIT|MEM_RESERVE, PAGE_READWRITE);
  memcpy(hdr_ptr, ptr, old_size);
  NtFreeVirtualMemory(((HANDLE)-1), (void**)&ptr, 0, MEM_RELEASE);
#else
  hdr_ptr = mmap(NULL, new_size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON|MAP_NORESERVE, -1, 0);
  memcpy(hdr_ptr, ptr, old_size);
  munmap(ptr, old_size);
#endif
#endif
  return hdr_ptr;
}

static inline void
_mem_free(void *ptr, size_t size)
{
#if defined(_WIN32) || defined(_WIN64)
    NtFreeVirtualMemory(((HANDLE)-1), (void**)&ptr, 0, MEM_RELEASE);
#else
    munmap(ptr, size);
#endif
}

static void buffer_alloc(buf_t *buf, size_t new_size)
{
    if (new_size > buf->alloc) {
        new_size = (2ULL << av_log2(new_size));
        buf->w_buffer = _mem_realloc(buf->w_buffer, buf->alloc, new_size);
        buf->alloc = new_size;
    }
}

static void buffer_alloc2(buf_t *buf, size_t append_len)
{
    if (buf->used+append_len > buf->alloc) {
        buf->w_buffer = _mem_realloc(buf->w_buffer, buf->alloc, 2*buf->alloc);
        buf->alloc <<= 1;
    }
}

static inline void
buffer_check_limits(buf_t *buf, size_t data_size, size_t pos)
{
    size_t new_size = pos + data_size;

    buffer_alloc(buf, new_size);

    if (new_size > buf->used)
        buf->used = new_size;
}

void buffer_create(buf_t *buf, size_t init_size)
{
    if(init_size < 4096) init_size = 4096; //always allocate at least a page to begin with.
    buf->w_buffer = _mem_alloc(init_size);
    buf->alloc = init_size;
    buf->used = 0;
    buf->pos = 0;
}

void buffer_create_from_data(buf_t *buf, uint8_t *data, size_t data_size)
{
    buf->w_buffer = data;
    buf->alloc = 0;
    buf->used = data_size;
    buf->pos = 0;
}

void buffer_free(buf_t *_buf)
{
    if (_buf->w_buffer)
        _mem_free(_buf->w_buffer, _buf->alloc);
}

size_t buffer_read(buf_t *_buf, void *data, size_t data_size)
{
    size_t data_to_read = data_size;
    if ((_buf->used - _buf->pos) < data_to_read) {
        data_to_read = (_buf->used - _buf->pos);
    }
    memcpy(data, _buf->w_buffer + _buf->pos, data_to_read);
    _buf->pos += data_to_read;
    return data_to_read;
}

uint16_t buffer_read_le16(buf_t *_buf)
{
    uint16_t ret = 0;
    if ((_buf->pos + 2) < _buf->used) {
        ret = *(uint16_t*)(_buf->w_buffer + _buf->pos);
        _buf->pos += 2;
    }
    return ret;
}

uint32_t buffer_read_le32(buf_t *_buf)
{
    uint32_t ret = 0;
    if ((_buf->pos + 4) < _buf->used) {
        ret = *(uint32_t*)(_buf->w_buffer + _buf->pos);
        _buf->pos += 4;
    }
    return ret;
}

uint64_t buffer_read_le64(buf_t *_buf)
{
    uint64_t ret = 0;
    if ((_buf->pos + 8) < _buf->used) {
        ret = *(uint64_t*)(_buf->w_buffer + _buf->pos);
        _buf->pos += 8;
    }
    return ret;
}

#define SEEK_SET	0	/* seek relative to beginning of file */
#define SEEK_CUR	1	/* seek relative to current file position */
#define SEEK_END	2	/* seek relative to end of file */

size_t buffer_getpos(buf_t *_buf)
{
    return _buf->pos;
}

size_t buffer_setpos(buf_t *_buf, size_t newpos, unsigned int use_curpos)
{
    if (use_curpos) {
        newpos += _buf->pos;
    }
    if (newpos < _buf->used) {
        _buf->pos = newpos;
    }
    return _buf->pos;
}

void buffer_write(buf_t *_buf, size_t pos,
                  const void *data, size_t data_size)
{
    if (pos >= _buf->used) pos = _buf->used;
    buffer_check_limits(_buf, data_size, pos);
    memcpy(_buf->w_buffer + pos, data, data_size);
}

void buffer_append_data(buf_t *buf, const void *data, size_t data_size)
{
    size_t new_size = buf->used + data_size;
    buffer_alloc(buf, new_size);
    memcpy(buf->w_buffer + buf->used, data, data_size);
    buf->used += data_size;
}

void buffer_append_cstring(buf_t *buf, const void *_data)
{
    const char *data = (const char*)_data;
    size_t data_size = strlen(data);
    size_t new_size = buf->used + data_size;
    buffer_alloc(buf, new_size);
    memcpy(buf->w_buffer + buf->used, data, data_size);
    buf->used += data_size;
}

void buffer_append_c(buf_t *buf, unsigned char chr)
{
    buffer_alloc2(buf, 1);
    buf->w_buffer[buf->used++] = chr;
}

void buffer_append_le16(buf_t *buf, uint16_t num)
{
    buffer_alloc2(buf, 8);
    *(uint16_t*)(buf->w_buffer + buf->used) = num;
    buf->used += 2;
}

void buffer_append_le32(buf_t *buf, uint32_t num)
{
    buffer_alloc2(buf, 8);
    *(uint32_t*)(buf->w_buffer + buf->used) = num;
    buf->used += 4;
}

void buffer_append_le64(buf_t *buf, uint64_t num)
{
    buffer_alloc2(buf, 8);
    *(uint64_t*)(buf->w_buffer + buf->used) = num;
    buf->used += 8;
}

void buffer_insert(buf_t *_buf, size_t pos, const void *data, size_t data_size)
{
    if (pos >= _buf->used) {
        buffer_append_data(_buf, data, data_size);
    } else {
        //buffer_copy(_buf, pos + data_size, _buf, pos, (size_t)-1);
        uint8_t *buf_offset = _buf->w_buffer + pos;
        size_t copy_size = _buf->used - pos;
        buffer_check_limits(_buf, data_size, pos);
        memmove(buf_offset + data_size, buf_offset, copy_size);
        memcpy(buf_offset, data, data_size);
    }
}

void buffer_delete(buf_t *buf, size_t pos, size_t size)
{
    size_t end_size;
    if (pos >= buf->used)
        return;
    end_size = buf->used - pos;

    if (size < end_size) {
        /* delete from between */
        end_size -= size;
        memmove(buf->w_buffer + pos, buf->w_buffer + pos + size, end_size);
    } else {
        /* delete the rest of the buffer */
        end_size = 0;
    }

    buf->used = (pos + end_size);
}

void buffer_copy(buf_t *dest, size_t dest_pos,
                 const buf_t *src, size_t src_pos, size_t copy_size)
{
    size_t max_size;

    max_size = src->used - src_pos;
    if (copy_size > max_size)
        copy_size = max_size;

    buffer_check_limits(dest, dest_pos, copy_size);
    if (src == dest) {
        memmove(dest->w_buffer + dest_pos,
                src->w_buffer + src_pos, copy_size);
    } else {
        memcpy(dest->w_buffer + dest_pos,
               src->w_buffer + src_pos, copy_size);
    }
}

void buffer_append_buf(buf_t *dest, const buf_t *src, size_t src_pos, size_t copy_size)
{
    size_t max_size = src->used - src_pos;
    if (src_pos > src->used) return;
    if (copy_size > max_size) copy_size = max_size;
    buffer_append_data(dest, src->w_buffer + src_pos, copy_size);
}

size_t buffer_get_used_size(buf_t *buf)
{
	return buf->used;
}

void buffer_set_used_size(buf_t *buf, size_t new_used_size)
{
    if (new_used_size < buf->used) buf->used = new_used_size;
}

#if 0
static inline uint16_t u8toa(unsigned char uval)
{
    static const char hex[]="0123456789abcdef";
    uint16_t _buf = 0;
    _buf |= (hex[((uval >> 4) & 0x0f)] << 0);
    _buf |= (hex[((uval >> 0) & 0x0f)] << 8);
    return _buf;
}

void buffer_itoa10(buf_t *buf, size_t uval)
{
    uint32_t digit;
    unsigned int i = 0;
    unsigned char _buf[24];
    do {
        digit = (uval % 5);
        uval = (uval / 5);
        if (uval & 1) digit += 5;
        _buf[i++] = (digit + '0');
        uval >>= 1;
    } while (uval);

    buffer_alloc2(buf, i);
    while(i-- > 0) {
        buf->w_buffer[buf->used++] = _buf[i];
    }
}
#endif

unsigned int buffer_vprintf(buf_t *buf, const char *__restrict fmt, va_list old_args)
{
    unsigned int len = 0;
    va_list args;

    do {
        size_t new_size, slack = buf->alloc - buf->used;
        if (slack > 0) {
            int ret = 0;
            va_copy(args, old_args);
            ret = vsnprintf((char *)buf->w_buffer + buf->used, slack, fmt, args);
            va_end(args);

            /* len = number of chars written, not including the terminating
             * NULL, so len+1 > slack means "truncated". MSVC differs here
             * and returns -1 for truncated. */
            if (ret < 0) {
                len = 2*slack;
            } else {
                len = ret;
            }
            if (len+1 <= slack)
                break;
        }
        /* Grow the buffer and retry */
        new_size = buf->used + len;
        buffer_alloc(buf, new_size);
    } while (1);
    buf->used += len;
    return len;
}

unsigned int buffer_printf(buf_t *buffer, const char *__restrict fmt, ...)
{
    unsigned int ret;
    va_list args;
    va_start(args, fmt);
    ret = buffer_vprintf(buffer, fmt, args);
    va_end(args);
    return ret;
}

