#ifndef BUFFER_H
#define BUFFER_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>

typedef struct real_buffer {
	unsigned char *w_buffer;
	size_t pos, used, alloc;
} buf_t;

/* WARNING: Be careful with functions that return pointers to data.
   With dynamic buffers they are valid only as long as buffer is not
   realloc()ed. You shouldn't rely on it being valid if you have modified
   buffer in any way. */

/* Creates a dynamically growing buffer. Whenever write would exceed the
   current size it's grown. */
void buffer_create(buf_t *buf, size_t init_size);
void buffer_create_from_data(buf_t *buf, unsigned char *data, size_t data_size);
/* Free the memory used by buffer. Not needed if the memory is free'd
   directly from the memory pool. */
void buffer_free(buf_t *buf);

/* Read data from buffer at current file pointer */
size_t buffer_read(buf_t *buf, void *data, size_t sz);
/* Read single character from buffer at current file pointer */
static inline uint8_t buffer_read_c(buf_t *_buf)
{
    if (_buf->pos <= _buf->used) {
        return _buf->w_buffer[_buf->pos++];
    }
    return 0;
}

static inline void buffer_unread_c(buf_t *_buf)
{
    if ((_buf->pos > 0) && (_buf->pos < _buf->used)) {
        _buf->pos--;
    }
}
/* Read a 16-bit little-endian integer from buffer */
uint16_t buffer_read_le16(buf_t *_buf);
/* Read a 32-bit little-endian integer from buffer */
uint32_t buffer_read_le32(buf_t *_buf);
/* Read a 64-bit little-endian integer from buffer */
uint64_t buffer_read_le64(buf_t *_buf);

/* Get current file pointer. */
size_t buffer_getpos(buf_t *_buf);
/* Set current file pointer.
 * use_curpos == 0 corresponds to SEEK_SET,
 * use_curpos == 1 corresponds to SEEK_CUR.
 */
size_t buffer_setpos(buf_t *_buf, size_t newpos, unsigned int use_curpos);
/* feof()-alike. */
static inline unsigned int buffer_iseof(buf_t *_buf)
{
    return ((_buf->pos < _buf->used) ? 0 : 1);
}

/* Write data to buffer at specified position. */
void buffer_write(buf_t *buf, size_t pos, const void *data, size_t data_size);
/* Append data to buffer. */
void buffer_append_data(buf_t *buf, const void *data, size_t data_size);
/* Append C string to buffer. */
void buffer_append_cstring(buf_t *buf, const void *data);
/* Append character to buffer. */
void buffer_append_c(buf_t *buf, unsigned char chr);
/* Append 16-bit little-endian integer to buffer */
void buffer_append_le16(buf_t *buf, uint16_t num);
/* Append 32-bit little-endian integer to buffer */
void buffer_append_le32(buf_t *buf, uint32_t num);
/* Append integer as string. */
void buffer_itoa10(buf_t *buf, size_t val);
/* Append hex integer as string. 0x is prepended to the string. */
void buffer_itoa16(buf_t *buf, size_t val);
/* Print formatted string to buffer. */
unsigned int buffer_vprintf(buf_t *buf, const char *__restrict fmt, va_list args);
unsigned int buffer_printf(buf_t *buffer, const char *__restrict fmt, ...);

/* Insert data to buffer. */
void buffer_insert(buf_t *buf, size_t pos, const void *data, size_t data_size);

/* Delete data from buffer. */
void buffer_delete(buf_t *buf, size_t pos, size_t size);

/* Copy data from buffer to another. The buffers may be same in which case
   it's internal copying, possibly with overlapping positions (ie. memmove()
   like functionality).
   copy_size may be set to (size_t)-1 to copy the rest of the used data in buffer. */
void buffer_copy(buf_t *dest, size_t dest_pos,
                 const buf_t *src, size_t src_pos, size_t copy_size);

/* Append data to buffer from another. */
void buffer_append_buf(buf_t *dest, const buf_t *src, size_t src_pos, size_t copy_size);

/* Returns the current buffer size. */
size_t buffer_get_used_size(buf_t *buf);

/* "Sets" the current buffer size. Primarily used to declare a buffer as empty
   but without freeing it, so that it can be reused repeatedly if needed.  */
void buffer_set_used_size(buf_t *buf, size_t new_used_size);

/* Returns pointer to beginning of buffer data. Current used size of buffer is
   stored in used_size if it's non-NULL. */
static inline const unsigned char *
buffer_get_data(buf_t *buf, size_t *used_size_r)
{
	if (used_size_r != NULL)
		*used_size_r = buf->used;
	return buf->w_buffer;
}

/* Returns 0 if buffer contents are identical. */
static inline int buffer_cmp(buf_t *buf1, buf_t *buf2)
{
    if (buf1->used != buf2->used) {
        return ((buf1->used > buf2->used) ? 1 : -1);
    }
    return memcmp(buf1->w_buffer, buf2->w_buffer, buf1->used);
}

#ifdef __cplusplus
};
#endif

#endif
