/* cmapdump.c -- parse a CMap file and dump it as a c-struct */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "buf.h"
#include "cmap.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

static fz_warn_context default_warnctx;

/* Warning context */
void fz_flush_warnings(fz_warn_context *ctx)
{
	if (ctx->count > 1) {
        char tmpbuf[1024];
		int len = snprintf(tmpbuf, 1023, "warning: ... repeated %u times ...\n", ctx->count);
        write(2, tmpbuf, len);
	}
	ctx->message[0] = 0;
    ctx->msglen = 0;
	ctx->count = 0;
}


void fz_warn(fz_warn_context *ctx, const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
    int buflen;

	va_start(ap, fmt);
	buflen = vsnprintf(buf, sizeof(buf)-2, fmt, ap);
    buf[buflen++] = '\n';
	va_end(ap);
#ifdef USE_OUTPUT_DEBUG_STRING
	OutputDebugStringA(buf);
#endif

    if (!ctx) ctx = &default_warnctx;

	if (!strcmp(buf, ctx->message)) {
		ctx->count++;
	} else {
		fz_flush_warnings(ctx);
        write(2, "warning: ", 9);
        write(2, buf, buflen);
		pdf_strlcpy(ctx->message, buf, buflen+1);
        ctx->msglen = buflen;
		ctx->count = 1;
	}
}

static uint8_t *cmapdump_open_file(const char *name, size_t *filesize)
{
    unsigned char *inbuf = NULL;
    struct stat st;
	int fd = open(name, O_BINARY | O_RDONLY, 0);
	if (fd == -1) {
		fprintf(stderr, "cmapdump: error: cannot open input CMap file %s!\n", name);
        return NULL;
    }
    fstat(fd, &st);
    inbuf = malloc(st.st_size+1);
    read(fd, inbuf, st.st_size);
    close(fd);
    if (filesize) {
        *filesize = st.st_size;
    }
    return inbuf;
}

static void
clean(char *p)
{
	while (*p)
	{
		if ((*p == '/') || (*p == '.') || (*p == '\\') || (*p == '-'))
			*p = '_';
		p ++;
	}
}

int
main(int argc, char **argv)
{
	pdf_cmap *cmap;
    uint8_t *cmap_inbuf = NULL;
    size_t cmap_filesize = 0;
	buf_t fi;
	FILE *fo;
	char name[256];
	char *realname;
	int i, k, m;

	if (argc < 3)
	{
		fprintf(stderr, "usage: cmapdump output.c lots of cmap files\n");
		return 1;
	}

#undef fopen
	fo = fopen(argv[1], "wb");
	if (!fo)
	{
		fprintf(stderr, "cmapdump: could not open output file '%s'\n", argv[1]);
		return 1;
	}

	fprintf(fo, "/* This is an automatically generated file. Do not edit. */\n");

	for (i = 2; i < argc; i++) {
		realname = strrchr(argv[i], '/');
		if (!realname)
			realname = strrchr(argv[i], '\\');
		if (realname)
			realname ++;
		else
			realname = argv[i];

		if (strlen(realname) > (sizeof name - 1)) {
			fprintf(stderr, "cmapdump: file name too long\n");
			if (fclose(fo)) {
				fprintf(stderr, "cmapdump: could not close output file '%s'\n", argv[1]);
				return 1;
			}
			return 1;
		}

		strcpy(name, realname);
		clean(name);

        cmap_inbuf = cmapdump_open_file(argv[i], &cmap_filesize);
        buffer_create_from_data(&fi, cmap_inbuf, cmap_filesize);
		cmap = pdf_load_cmap(NULL, &fi);
		buffer_free(&fi);

		fprintf(fo, "\n/* %s */\n\n", cmap->cmap_name);

		fprintf(fo, "static const pdf_range cmap_%s_ranges[] = {", name);
		if (cmap->rlen == 0) {
			fprintf(fo, " {0,%d,0}", PDF_CMAP_RANGE);
		}
		for (k = 0; k < cmap->rlen; k++) {
			if (k % 4 == 0)
				fprintf(fo, "\n");
			fprintf(fo, "{%d,%d,%d},",
				cmap->ranges[k].low, cmap->ranges[k].extent_flags, cmap->ranges[k].out);
		}
		fprintf(fo, "\n};\n\n");

		if (cmap->tlen == 0) {
			fprintf(fo, "static const unsigned short cmap_%s_table[] = { 0 };\n\n", name);
		} else {
			fprintf(fo, "static const unsigned short cmap_%s_table[%d] = {",
				name, cmap->tlen);
			for (k = 0; k < cmap->tlen; k++) {
				if (k % 12 == 0)
					fprintf(fo, "\n");
				fprintf(fo, "%d,", cmap->table[k]);
			}
			fprintf(fo, "\n};\n\n");
		}

		if (cmap->mlen > 0) {
			fprintf(fo, "static const pdf_mrange cmap_%s_mranges[] = {", name);
			for (k = 0; k < cmap->mlen; k++) {
				fprintf(fo, "\n{%uu,%uu,{", cmap->mranges[k].low, cmap->mranges[k].len);
				for (m = 0; m < PDF_MRANGE_CAP; ++m)
					fprintf(fo, "%uu,", cmap->mranges[k].out[m]);
				fprintf(fo, "}},");
			}
			fprintf(fo, "\n};\n\n");
		} else {
			fprintf(fo, "static const pdf_mrange cmap_%s_mranges[] = { 0 };\n\n", name);
        }

		fprintf(fo, "static pdf_cmap cmap_%s = {\n", name);
		fprintf(fo, "\t{-1, pdf_free_cmap_imp}, ");
		fprintf(fo, "\"%s\", ", cmap->cmap_name);
		fprintf(fo, "\"%s\", 0, ", cmap->usecmap_name);
		fprintf(fo, "%d, ", cmap->wmode);
		fprintf(fo, "%d,\n\t{ ", cmap->codespace_len);
		if (cmap->codespace_len == 0)
		{
			fprintf(fo, "{0,0,0},");
		}
		for (k = 0; k < cmap->codespace_len; k++)
		{
			fprintf(fo, "{%d,%d,%d},",
				cmap->codespace[k].n, cmap->codespace[k].low, cmap->codespace[k].high);
		}
		fprintf(fo, " },\n");

		fprintf(fo, "\t%u, %u, (pdf_range*) cmap_%s_ranges,\n", cmap->rlen, cmap->rlen, name);
		fprintf(fo, "\t%u, %u, (unsigned short*) cmap_%s_table,\n", cmap->tlen, cmap->tlen, name);
		fprintf(fo, "\t%u, %u, (pdf_mrange*) cmap_%s_mranges,\n", cmap->mlen, cmap->mlen, name);
		fprintf(fo, "};\n");
		printf("\t{\"%s\",&cmap_%s},\n", cmap->cmap_name, name);
	}

	if (fclose(fo))
	{
		fprintf(stderr, "cmapdump: could not close output file '%s'\n", argv[1]);
		return 1;
	}

	return 0;
}

