/*
 * Bidirectional text processing.
 *
 * Derived from the SmartOffice code, which is itself derived
 * from the example unicode standard code. Original copyright
 * messages follow:
 *
 * Copyright (C) Picsel, 2004-2008. All Rights Reserved.
 *
 * Processes Unicode text by arranging the characters into an order suitable
 * for display. E.g. Hebrew text will be arranged from right-to-left and
 * any English within the text will remain in the left-to-right order.
 * Characters such as parenthesis will be substituted for their mirrored
 * equivalents if they are part of text which must be reversed.
 *
 * This is an implementation of the Unicode Bidirectional Algorithm which
 * can be found here: http://www.unicode.org/reports/tr9/ and is based
 * on the reference implementation found on Unicode.org.
 */

#ifndef BIDI_BIDI_H
#define BIDI_BIDI_H

#include "mupdf/fitz/system.h"

/* Bidirectional Character Types
 * as defined by the Unicode Bidirectional Algorithm Table 3-7.
 * The list of bidirectional character types here is not grouped the
 * same way as the table 3-7, since the numeric values for the types
 * are chosen to keep the state and action tables compact.
 */
enum
{
	/* input types */
			/* ON MUST be zero, code relies on ON = N = 0 */
	BDI_ON = 0,	/**< Other Neutral  */
	BDI_L,		/**< Left-to-right Letter */
	BDI_R,		/**< Right-to-left Letter */
	BDI_AN,		/**< Arabic Number */
	BDI_EN,		/**< European Number */
	BDI_AL,		/**< Arabic Letter (Right-to-left) */
	BDI_NSM,	/**< Non-spacing Mark */
	BDI_CS,		/**< Common Separator */
	BDI_ES,		/**< European Separator */
	BDI_ET,		/**< European Terminator (post/prefix e.g. $ and %) */

	/* resolved types */
	BDI_BN,		/**< Boundary neutral (type of RLE etc after explicit levels)*/

	/* input types, */
	BDI_S,		/**< Segment Separator (TAB)		 used only in L1 */
	BDI_WS,		/**< White space					 used only in L1 */
	BDI_B,		/**< Paragraph Separator (aka as PS) */

	/* types for explicit controls */
	BDI_RLO,	/**< these are used only in X1-X9 */
	BDI_RLE,
	BDI_LRO,
	BDI_LRE,
	BDI_PDF,

	/* resolved types, also resolved directions */
	BDI_N = BDI_ON	/**< alias, where ON, WS and S are treated the same */
};

typedef enum fz_bidi_direction_e
{
	BIDI_LEFT_TO_RIGHT = 0,
	BIDI_RIGHT_TO_LEFT = 1,
	BIDI_NEUTRAL = 2
}
fz_bidi_direction;

typedef enum fz_bidi_flags_e
{
	BIDI_CLASSIFY_WHITE_SPACE = 1,
	BIDI_REPLACE_TAB = 2
}
fz_bidi_flags;

typedef uint8_t fz_bidi_chartype;
#define BIDI_LEVEL_MAX 125 /* Updated for 6.3.0 */

/**
 * Prototype for callback function supplied to Bidi_fragmentText.
 *
 * @param	 fragment	 First character in fragment
 * @param	 fragmentLen Number of characters in fragment
 * @param	 block_r2l	 True if block should concatenate with other blocks as right-to-left
 * @param	 char_r2l	 True if characters within block should be laid out as right-to-left
 * @param	 mirror		 The mirror code of the fragment if it exists
 * @param	 arg		 Data from caller of Bidi_fragmentText
 */
typedef void (fz_bidi_fragment_callback)(const uint16_t *fragment, size_t fragmentLen,
                                         int block_r2l, int char_r2l, uint16_t mirror, void *arg);

/**
 * Partitions the given Unicode sequence into one or more unidirectional
 * fragments and invokes the given callback function for each fragment.
 *
 * For example, if directionality of text is:
 *			   0123456789
 *			   rrlllrrrrr,
 * we'll invoke callback with:
 *			   &text[0], length == 2, rightToLeft ==  true
 *			   &text[2], length == 3, rightToLeft == false
 *			   &text[5], length == 5, rightToLeft ==  true.
 *
 * @param[in] text	   start of Unicode sequence
 * @param[in] textlen  number of Unicodes to analyse
 * @param[in] baseDir  direction of paragraph (specify BIDI_NEUTRAL to force auto-detection)
 * @param[in] callback function to be called for each fragment
 * @param[in] arg	   data to be passed to the callback function
 * @param[in] bidiFlag flag to be passed to the callback function
 */
void fz_bidi_fragment_text(fz_context *ctx, const uint16_t *text, size_t textlen, fz_bidi_direction *baseDir,
                           fz_bidi_fragment_callback *callback, void *arg, int bidiFlag);

#endif /* BIDI_BIDI_H */

/** @} */
