#ifndef MUPDF_FITZ_BUFFER_H
#define MUPDF_FITZ_BUFFER_H

#include "mupdf/fitz/system.h"
#include "mupdf/fitz/context.h"

/*
	fz_buffer is a wrapper around a dynamically allocated array of bytes.

	Buffers have a capacity (the number of bytes storage immediately
	available) and a current size.
*/
typedef struct fz_buffer_s fz_buffer;

/*
	fz_keep_buffer: Increment the reference count for a buffer.

	buf: The buffer to increment the reference count for.

	Returns a pointer to the buffer. Does not throw exceptions.
*/
fz_buffer *fz_keep_buffer(fz_buffer *buf);

/*
	fz_drop_buffer: Decrement the reference count for a buffer.

	buf: The buffer to decrement the reference count for.
*/
void fz_drop_buffer(fz_buffer *buf);

/*
	fz_buffer_storage: Retrieve information on the storage currently used
	by a buffer.

	data: Pointer to place to retrieve data pointer.

	Returns length of stream.
*/
unsigned int fz_buffer_storage(fz_buffer *buf, unsigned char **data);

struct fz_buffer_s
{
	int refs;
	unsigned char *data;
	unsigned int cap, len;
	int unused_bits;
};

/*
	fz_new_buffer: Create a new buffer.

	capacity: Initial capacity.

	Returns pointer to new buffer. Throws exception on allocation
	failure.
*/
fz_buffer *fz_new_buffer(unsigned int capacity);

/*
	fz_new_buffer: Create a new buffer.

	capacity: Initial capacity.

	Returns pointer to new buffer. Throws exception on allocation
	failure.
*/
fz_buffer *fz_new_buffer_from_data(unsigned char *data, unsigned int size);

/*
	fz_resize_buffer: Ensure that a buffer has a given capacity,
	truncating data if required.

	buf: The buffer to alter.

	capacity: The desired capacity for the buffer. If the current size
	of the buffer contents is smaller than capacity, it is truncated.

*/
void fz_resize_buffer(fz_buffer *buf, unsigned int capacity);
void fz_ensure_buffer(fz_buffer *buf, unsigned int min);

/*
	fz_grow_buffer: Make some space within a buffer (i.e. ensure that
	capacity > size).

	buf: The buffer to grow.

	May throw exception on failure to allocate.
*/
void fz_grow_buffer(fz_buffer *buf);

/*
	fz_trim_buffer: Trim wasted capacity from a buffer.

	buf: The buffer to trim.
*/
void fz_trim_buffer(fz_buffer *buf);

/*
	fz_buffer_cat: Concatenate buffers

	buf: first to concatenate and the holder of the result
	extra: second to concatenate

	May throw exception on failure to allocate.
*/
void fz_buffer_cat(fz_buffer *buf, fz_buffer *extra);

void fz_write_buffer(fz_buffer *buf, const void *data, unsigned int len);

void fz_write_buffer_byte(fz_buffer *buf, int val);

void fz_write_buffer_int32le(fz_buffer *buf, unsigned int val);

void fz_buffer_itoa10(fz_buffer *out, unsigned int uval);

void fz_write_buffer_rune(fz_buffer *buf, unsigned int val);

void fz_write_buffer_pad(fz_buffer *buf);

static inline void
fz_buffer_puts(fz_buffer *buf, const char *str)
{
    fz_write_buffer(buf, str, strlen(str));
}

static inline void fz_write_buffer_int32be(fz_buffer *buf, unsigned int val)
{
    __asm__("bswap   %0" : "+r" (val));
    fz_write_buffer_int32le(buf, val);
}

/*
	fz_buffer_printf: print a formatted string to a buffer.
    The buffer will grow as required.
*/
int fz_buffer_printf(fz_buffer *buffer, const char *fmt, ...);
int fz_buffer_vprintf(fz_buffer *buffer, const char *fmt, va_list args);

/*
	fz_buffer_printf: print a string formatted as a pdf string to a buffer.
	The buffer will grow.
*/
void
fz_buffer_cat_pdf_string(fz_buffer *buffer, const char *text);

#endif
