#ifndef MUPDF_PDF_OBJECT_H
#define MUPDF_PDF_OBJECT_H

typedef struct pdf_document_s pdf_document;
#define PDF_FLAGS_DIRTY 16

/*
 * Dynamic objects.
 * The same type of objects as found in PDF and PostScript.
 * Used by the filters and the mupdf parser.
 */

typedef struct pdf_obj_s pdf_obj;

typedef enum pdf_objkind_e
{
	PDF_NULL = 0,
	PDF_BOOL = 'b',
	PDF_INT = 'i',
	PDF_REAL = 'f',
	PDF_STRING = 's',
	PDF_NAME = 'n',
	PDF_ARRAY = 'a',
	PDF_DICT = 'd',
	PDF_INDIRECT = 'r'
} pdf_objkind;

typedef struct keyval
{
	pdf_obj *k;
	pdf_obj *v;
} keyval;

struct pdf_obj_s
{
	int refs;
	unsigned char kind;
	unsigned char flags;
	pdf_document *doc;
	int parent_num;
	union
	{
		int b;
		int i;
		float f;
		struct {
			unsigned short len;
			char buf[1];
		} s;
		char n[1];
		struct {
			int len;
			int cap;
			pdf_obj **items;
		} a;
		struct {
			int len;
			int cap;
			struct keyval *items;
		} d;
		struct {
			int num;
			unsigned short gen;
		} r;
	} u;
};

pdf_obj *pdf_new_obj(pdf_document *doc);
pdf_obj *pdf_resolve_indirect(pdf_obj *ref);

static inline int pdf_is_indirect(pdf_obj *obj)
{
	return obj ? obj->kind == PDF_INDIRECT : 0;
}

static inline pdf_obj *pdf_new_null(pdf_document *doc)
{
	pdf_obj *obj = pdf_new_obj(doc);
	obj->kind = PDF_NULL;
	return obj;
}

static inline pdf_obj *pdf_new_bool(pdf_document *doc, int b)
{
	pdf_obj *obj = pdf_new_obj(doc);
	obj->kind = PDF_BOOL;
	obj->u.b = b;
	return obj;
}

static inline pdf_obj *pdf_new_int(pdf_document *doc, int i)
{
	pdf_obj *obj = pdf_new_obj(doc);
	obj->kind = PDF_INT;
	obj->u.i = i;
	return obj;
}

static inline pdf_obj *pdf_new_real(pdf_document *doc, float f)
{
	pdf_obj *obj = pdf_new_obj(doc);
	obj->kind = PDF_REAL;
	obj->u.f = f;
	return obj;
}

static inline pdf_obj *pdf_new_indirect(pdf_document *doc, int num, unsigned short gen)
{
	pdf_obj *obj = pdf_new_obj(doc);
	obj->kind = PDF_INDIRECT;
	obj->u.r.num = num;
	obj->u.r.gen = gen;
	return obj;
}

pdf_obj *pdf_new_string(pdf_document *doc, const char *str, unsigned int len);
pdf_obj *pdf_new_name(pdf_document *doc, const char *str);
pdf_obj *pdf_new_array(pdf_document *doc, int initialcap);
pdf_obj *pdf_new_dict(pdf_document *doc, int initialcap);
pdf_obj *pdf_new_rect(pdf_document *doc, const fz_rect *rect);
pdf_obj *pdf_new_matrix(pdf_document *doc, const fz_matrix *mtx);
pdf_obj *pdf_copy_array(pdf_obj *array);
pdf_obj *pdf_copy_dict(pdf_obj *dict);

pdf_obj *pdf_new_obj_from_str(pdf_document *doc, const char *src);

pdf_obj *pdf_keep_obj(pdf_obj *obj);
void pdf_drop_obj(pdf_obj *obj);

/* type queries */
#define RESOLVE(obj) \
	do { \
		if (obj && obj->kind == PDF_INDIRECT) { \
			obj = pdf_resolve_indirect(obj); \
		} \
	} while (0)

static inline int pdf_is_null(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_NULL : 0;
}

static inline int pdf_is_bool(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_BOOL : 0;
}

static inline int pdf_is_int(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_INT : 0;
}

static inline int pdf_is_real(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_REAL : 0;
}

static inline int pdf_is_string(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_STRING : 0;
}

static inline int pdf_is_name(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_NAME : 0;
}

static inline int pdf_is_array(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_ARRAY : 0;
}

static inline int pdf_is_dict(pdf_obj *obj)
{
	RESOLVE(obj);
	return obj ? obj->kind == PDF_DICT : 0;
}

int pdf_is_stream(pdf_document *doc, int num, int gen);

int pdf_objcmp(pdf_obj *a, pdf_obj *b);

/* obj marking and unmarking functions - to avoid infinite recursions. */
int pdf_obj_marked(pdf_obj *obj);
int pdf_mark_obj(pdf_obj *obj);
void pdf_unmark_obj(pdf_obj *obj);

/* obj memo functions - allows us to secretly remember "a memo" (a bool) in
 * an object, and to read back whether there was a memo, and if so, what it
 * was. */
void pdf_set_obj_memo(pdf_obj *obj, int memo);
int pdf_obj_memo(pdf_obj *obj, int *memo);

/* obj dirty bit support. */
static inline int pdf_obj_is_dirty(pdf_obj *obj)
{
	if (!obj)
		return 0;
	return !!(obj->flags & PDF_FLAGS_DIRTY);
}

static inline void pdf_dirty_obj(pdf_obj *obj)
{
	if (!obj)
		return;
	obj->flags |= PDF_FLAGS_DIRTY;
}

static inline void pdf_clean_obj(pdf_obj *obj)
{
	if (!obj)
		return;
	obj->flags &= ~PDF_FLAGS_DIRTY;
}

/* safe, silent failure, no error reporting on type mismatches */
static inline int pdf_to_bool(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj)
		return 0;
	return obj->kind == PDF_BOOL ? obj->u.b : 0;
}

static inline int pdf_to_int(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj)
		return 0;
	if (obj->kind == PDF_INT)
		return obj->u.i;
	if (obj->kind == PDF_REAL)
		return (int)(obj->u.f + 0.5f); /* No roundf in MSVC */
	return 0;
}

static inline float pdf_to_real(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj)
		return 0;
	if (obj->kind == PDF_REAL)
		return obj->u.f;
	if (obj->kind == PDF_INT)
		return obj->u.i;
	return 0;
}

static inline char *pdf_to_name(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_NAME)
		return "";
	return obj->u.n;
}

static inline char *pdf_to_str_buf(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_STRING)
		return "";
	return obj->u.s.buf;
}

static inline int pdf_to_str_len(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_STRING)
		return 0;
	return obj->u.s.len;
}

static inline pdf_obj *pdf_to_dict(pdf_obj *obj)
{
	RESOLVE(obj);
	return (obj && obj->kind == PDF_DICT ? obj : NULL);
}

static inline int pdf_to_num(pdf_obj *obj)
{
	if (!obj || obj->kind != PDF_INDIRECT)
		return 0;
	return obj->u.r.num;
}

static inline int pdf_to_gen(pdf_obj *obj)
{
	if (!obj || obj->kind != PDF_INDIRECT)
		return 0;
	return obj->u.r.gen;
}

void pdf_array_put(pdf_obj *array, unsigned int i, pdf_obj *obj);
void pdf_array_push(pdf_obj *array, pdf_obj *obj);
void pdf_array_insert(pdf_obj *array, pdf_obj *obj, unsigned int index);
void pdf_array_delete(pdf_obj *array, unsigned int index);
int pdf_array_contains(pdf_obj *array, pdf_obj *obj);

static inline unsigned int pdf_array_len(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_ARRAY)
		return 0;
	return obj->u.a.len;
}

static inline pdf_obj *pdf_array_get(pdf_obj *obj, unsigned int i)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_ARRAY)
		return NULL;
	if (i >= obj->u.a.len)
		return NULL;
	return obj->u.a.items[i];
}

static inline void pdf_array_push_drop(pdf_obj *obj, pdf_obj *item)
{
	pdf_array_push(obj, item);
	pdf_drop_obj(item);
}

static inline void pdf_array_insert_drop(pdf_obj *obj, pdf_obj *item, int i)
{
	pdf_array_insert(obj, item, i);
	pdf_drop_obj(item);
}

int pdf_dict_len(pdf_obj *dict);
pdf_obj *pdf_dict_get_key(pdf_obj *dict, int idx);
pdf_obj *pdf_dict_get_val(pdf_obj *dict, int idx);
pdf_obj *pdf_dict_gets(pdf_obj *dict, const char *key);
pdf_obj *pdf_dict_getp(pdf_obj *dict, const char *key);
void pdf_dict_put(pdf_obj *dict, pdf_obj *key, pdf_obj *val);
void pdf_dict_putp(pdf_obj *dict, const char *key, pdf_obj *val);
void pdf_dict_del(pdf_obj *dict, pdf_obj *key);
void pdf_dict_dels(pdf_obj *dict, const char *key);
void pdf_sort_dict(pdf_obj *dict);

static inline pdf_obj *
pdf_dict_get(pdf_obj *obj, pdf_obj *key)
{
	if (!key || key->kind != PDF_NAME)
		return NULL;
	return pdf_dict_gets(obj, pdf_to_name(key));
}

static inline pdf_obj *
pdf_dict_getsa(pdf_obj *obj, const char *key, const char *abbrev)
{
	pdf_obj *v = pdf_dict_gets(obj, key);
	if (v)
		return v;
	return pdf_dict_gets(obj, abbrev);
}

static inline void
pdf_dict_puts(pdf_obj *obj, const char *key, pdf_obj *val)
{
	pdf_document *doc = obj->doc;
	pdf_obj *keyobj = pdf_new_name(doc, key);
	pdf_dict_put(obj, keyobj, val);
	pdf_drop_obj(keyobj);
}

static inline void
pdf_dict_puts_drop(pdf_obj *obj, const char *key, pdf_obj *val)
{
    pdf_dict_puts(obj, key, val);
	pdf_drop_obj(val);
}

static inline void
pdf_dict_putp_drop(pdf_obj *obj, const char *keys, pdf_obj *val)
{
	pdf_dict_putp(obj, keys, val);
	pdf_drop_obj(val);
}

/*
	Recurse through the object structure setting the node's parent_num to num.
	parent_num is used when a subobject is to be changed during a document edit.
	The whole containing hierarchy is moved to the incremental xref section, so
	to be later written out as an incremental file update.
*/
void pdf_set_obj_parent(pdf_obj *obj, int num);

char *pdf_objkindstr(pdf_obj *obj);
unsigned int pdf_sprint_obj(char *s, unsigned int n, pdf_obj *obj);
void pdf_bufprint_obj(fz_buffer *outbuf, pdf_obj *obj);

#ifndef NDEBUG
void pdf_print_obj(pdf_obj *obj);
void pdf_print_ref(pdf_obj *obj);
#endif

char *pdf_to_utf8(pdf_document *doc, pdf_obj *src);
unsigned short *pdf_to_ucs2(pdf_document *doc, pdf_obj *src);
pdf_obj *pdf_to_utf8_name(pdf_document *doc, pdf_obj *src);
char *pdf_from_ucs2(pdf_document *doc, unsigned short *str);
void pdf_to_ucs2_buf(unsigned short *buffer, pdf_obj *src);

fz_rect *pdf_to_rect(fz_context *ctx, pdf_obj *array, fz_rect *rect);
fz_matrix *pdf_to_matrix(fz_context *ctx, pdf_obj *array, fz_matrix *mat);

pdf_document *pdf_get_indirect_document(pdf_obj *obj);
void pdf_set_str_len(pdf_obj *obj, int newlen);
static inline void pdf_set_int(pdf_obj *obj, int i)
{
	if (!obj || obj->kind != PDF_INT)
		return;
	obj->u.i = i;
}

#endif
