/* cmapdump.c -- parse a CMap file and dump it as a c-struct */

#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include "cmap.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

static fz_warn_context default_warnctx;

/* Warning context */
void fz_flush_warnings(fz_warn_context *ctx)
{
	if (ctx->count > 1) {
        char tmpbuf[1024];
		int len = snprintf(tmpbuf, 1023, "warning: ... repeated %u times ...\n", ctx->count);
        write(2, tmpbuf, len);
	}
	ctx->message[0] = 0;
    ctx->msglen = 0;
	ctx->count = 0;
}

void fz_warn(fz_warn_context *ctx, const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
    int buflen;

	va_start(ap, fmt);
	buflen = vsnprintf(buf, sizeof(buf)-2, fmt, ap);
    buf[buflen++] = '\n';
	va_end(ap);
#ifdef USE_OUTPUT_DEBUG_STRING
	OutputDebugStringA(buf);
#endif

    if (!ctx) ctx = &default_warnctx;

	if (!strcmp(buf, ctx->message)) {
		ctx->count++;
	} else {
		fz_flush_warnings(ctx);
        write(2, "warning: ", 9);
        write(2, buf, buflen);
		memcpy(ctx->message, buf, buflen+1);
        ctx->msglen = buflen;
		ctx->count = 1;
	}
}

static void
clean(char *p)
{
	while (*p)
	{
		if ((*p == '/') || (*p == '.') || (*p == '\\') || (*p == '-'))
			*p = '_';
		p ++;
	}
}

static const char *cmap_names[4] = { "Identity1-H", "Identity2-H", "Identity1-V", "Identity2-V" };

int
main(int argc, char **argv)
{
	pdf_cmap *cmap, *cmaps[4];
	FILE *fo;
	char name[256];
	char *realname;
	int i, k, m;

	if (argc < 2)
	{
		fprintf(stderr, "usage: cmapdump output.c lots of cmap files\n");
		return 1;
	}
	cmaps[0] = pdf_new_identity_cmap(NULL, 0, 1);
	cmaps[1] = pdf_new_identity_cmap(NULL, 0, 2);
	cmaps[2] = pdf_new_identity_cmap(NULL, 1, 1);
	cmaps[3] = pdf_new_identity_cmap(NULL, 2, 2);

#undef fopen
	fo = fopen(argv[1], "wb");
	if (!fo)
	{
		fprintf(stderr, "cmapdump: could not open output file '%s'\n", argv[1]);
		return 1;
	}

	fprintf(fo, "/* This is an automatically generated file. Do not edit. */\n");

	for (i = 0; i < 4; i++) {
        cmap = cmaps[i];
		strcpy(name, cmap_names[i]);
		clean(name);

		fprintf(fo, "\n/* %s */\n\n", cmap->cmap_name);

		fprintf(fo, "static const pdf_range cmap_%s_ranges[] = {", name);
		if (cmap->rlen == 0) {
			fprintf(fo, " {0,%d,0}", PDF_CMAP_RANGE);
		}
		for (k = 0; k < cmap->rlen; k++) {
			if (k % 4 == 0)
				fprintf(fo, "\n  ");
			fprintf(fo, "{0x%04x,0x%02x,%d},", cmap->ranges[k].low, cmap->ranges[k].extent_flags, cmap->ranges[k].out);
		}
		fprintf(fo, "\n};\n\n");

		if (cmap->tlen == 0) {
			fprintf(fo, "static const unsigned short cmap_%s_table[] = { 0 };\n\n", name);
		} else {
			fprintf(fo, "static const unsigned short cmap_%s_table[%d] = {",
				name, cmap->tlen);
			for (k = 0; k < cmap->tlen; k++) {
				if (k % 12 == 0)
					fprintf(fo, "\n  ");
				fprintf(fo, "0x%04x,", cmap->table[k]);
			}
			fprintf(fo, "\n};\n\n");
		}

		fprintf(fo, "static pdf_cmap cmap_%s = {\n", name);
		fprintf(fo, "\t{-1, pdf_free_cmap_imp}, ");
		fprintf(fo, "\"%s\", \"%s\", 0, %d,\n\t%u, { ", cmap->cmap_name, cmap->usecmap_name, cmap->wmode, cmap->codespace_len);
		if (cmap->codespace_len == 0)
		{
			fprintf(fo, "{0,0,0},");
		}
		for (k = 0; k < cmap->codespace_len; k++)
		{
			fprintf(fo, "{%u,0x%x,0x%x},",
				cmap->codespace[k].n, cmap->codespace[k].low, cmap->codespace[k].high);
		}
		fprintf(fo, " },\n");

		fprintf(fo, "\t%u, %u, (pdf_range*) cmap_%s_ranges,\n", cmap->rlen, cmap->rlen, name);
		fprintf(fo, "\t%u, %u, (unsigned short*) cmap_%s_table,\n", cmap->tlen, cmap->tlen, name);
		fprintf(fo, "\t0, 0, (pdf_xrange*) NULL,\n");
		fprintf(fo, "};\n");
		printf("\t{\"%s\",&cmap_%s},\n", cmap->cmap_name, name);
	}

	if (fclose(fo))
	{
		fprintf(stderr, "cmapdump: could not close output file '%s'\n", argv[1]);
		return 1;
	}

	return 0;
}

