#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "buf.h"
#include "cmap.h"
#include "pdf-lex.h"

/*
 * CMap parser
 */

/*
 * Copy src to string dst of size siz.
 * At most siz-1 characters will be copied. Always NUL terminates.
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t pdf_strlcpy(char *dst, const char *src, size_t siz)
{
    char *d = dst;
    const char *s = src;
    size_t n = siz;

    /* Copy as many bytes as will fit */
    while (--n != 0) {
        if ((*d++ = *s++) == '\0')
            break;
    }

    /* Not enough room in dst, add NUL and return */
    if (n == 0) {
        *d = '\0';      /* NUL-terminate dst */
        return siz;
    }

    return(s - src - 1);    /* count does not include NUL */
}

static int
pdf_code_from_string(char *buf, int len)
{
	int a = 0;
	while (len--)
		a = (a << 8) | *(unsigned char *)buf++;
	return a;
}

static void
pdf_parse_cmap_name(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok = pdf_lex(file, buf);
	if (tok == PDF_TOK_NAME)
		pdf_strlcpy(cmap->cmap_name, buf->scratch, sizeof(cmap->cmap_name));
	else
		fz_warn(ctx, "expected name after CMapName in cmap");
}

static void
pdf_parse_wmode(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok = pdf_lex(file, buf);
	if (tok == PDF_TOK_INT)
		pdf_set_cmap_wmode(cmap, buf->i);
	else
		fz_warn(ctx, "expected integer after WMode in cmap");
}

static unsigned int
pdf_parse_codespace_range(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	int lo, hi;

	while (1)
	{
		tok = pdf_lex(file, buf);

		if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "endcodespacerange"))
			return 1;

		else if (tok == PDF_TOK_STRING)
		{
			lo = pdf_code_from_string(buf->scratch, buf->len);
			tok = pdf_lex(file, buf);
			if (tok == PDF_TOK_STRING)
			{
				hi = pdf_code_from_string(buf->scratch, buf->len);
				pdf_add_codespace(ctx, cmap, lo, hi, buf->len);
			}
			else break;
		}

		else break;
	}

	fz_warn(ctx, "expected string or endcodespacerange");
    return 0;
}

static unsigned int
pdf_parse_cid_range(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	int lo, hi, dst;

	while (1) {
		tok = pdf_lex(file, buf);

		if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "endcidrange")) {
			return 1;
		} else if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string or endcidrange");
            return 0;
        }

		lo = pdf_code_from_string(buf->scratch, buf->len);

		tok = pdf_lex(file, buf);
		if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string");
            return 0;
        }

		hi = pdf_code_from_string(buf->scratch, buf->len);

		tok = pdf_lex(file, buf);
		if (tok != PDF_TOK_INT) {
			fz_warn(ctx, "expected integer");
            return 0;
        }

		dst = buf->i;

		pdf_map_range_to_range(cmap, lo, hi, dst);
	}
}

static unsigned int
pdf_parse_cid_char(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	int src, dst;

	while (1) {
		tok = pdf_lex(file, buf);
		if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "endcidchar")) {
			return 1;
		} else if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string or endcidchar");
            return 0;
        }

		src = pdf_code_from_string(buf->scratch, buf->len);

		tok = pdf_lex(file, buf);
		if (tok != PDF_TOK_INT) {
			fz_warn(ctx, "expected integer");
            return 0;
        }

		dst = buf->i;

		pdf_map_range_to_range(cmap, src, src, dst);
	}
}

static unsigned int
pdf_parse_bf_range_array(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf, int lo, int hi)
{
	pdf_token tok;
	int dst[256];
	int i;

	while (1) {
		tok = pdf_lex(file, buf);
		if (tok == PDF_TOK_CLOSE_ARRAY) {
			return 1;
		/* Note: does not handle [ /Name /Name ... ] */
		} else if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string or ]");
            return 0;
        }

		if (buf->len >> 1) {
			unsigned int len = (buf->len >> 1);
            if (len > 256) len = 256;
			for (i = 0; i < len; i++)
				dst[i] = pdf_code_from_string(&buf->scratch[i * 2], 2);

			pdf_map_one_to_many(cmap, lo, dst, buf->len >> 1);
		}

		lo ++;
	}
}

static unsigned int
pdf_parse_bf_range(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	int lo, hi, dst;

	while (1) { 
		tok = pdf_lex(file, buf);
		if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "endbfrange")) {
			return 1;
		} else if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string or endbfrange");
            return 0;
        }

		lo = pdf_code_from_string(buf->scratch, buf->len);

		tok = pdf_lex(file, buf);
		if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string");
            return 0;
        }

		hi = pdf_code_from_string(buf->scratch, buf->len);
		if (lo < 0 || lo > 65535 || hi < 0 || hi > 65535 || lo > hi)
		{
			fz_warn(ctx, "bf_range limits out of range in cmap %s", cmap->cmap_name);
			return 0;
		}

		tok = pdf_lex(file, buf);

		if (tok == PDF_TOK_STRING)
		{
			if (buf->len == 2)
			{
				dst = pdf_code_from_string(buf->scratch, buf->len);
				pdf_map_range_to_range(cmap, lo, hi, dst);
			}
			else
			{
				int dststr[256];
				unsigned int i;

				if (buf->len >> 1) {
					unsigned int len = (buf->len >> 1);
                    if (len > 256) len = 256;
					for (i = 0; i < len; i++)
						dststr[i] = pdf_code_from_string(&buf->scratch[i * 2], 2);

					while (lo <= hi) {
						dststr[i-1] ++;
						pdf_map_one_to_many(cmap, lo, dststr, i);
						lo ++;
					}
				}
			}
		} else if (tok == PDF_TOK_OPEN_ARRAY) {
			pdf_parse_bf_range_array(ctx, cmap, file, buf, lo, hi);
        } else {
			fz_warn(ctx, "expected string or array or endbfrange");
            return 0;
		}
	}
}

static unsigned int
pdf_parse_bf_char(fz_warn_context *ctx, pdf_cmap *cmap, buf_t *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	int dst[256];
	int src;
	unsigned int i;

	while (1)
	{
		tok = pdf_lex(file, buf);

		if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "endbfchar")) {
			return 1;
		} else if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string or endbfchar");
            return 0;
        }

		src = pdf_code_from_string(buf->scratch, buf->len);

		tok = pdf_lex(file, buf);
		/* Note: does not handle /dstName */
		if (tok != PDF_TOK_STRING) {
			fz_warn(ctx, "expected string");
            return 0;
        }

		if (buf->len / 2) {
			unsigned int len = (buf->len >> 1);
            if (len > 256) len = 256;
			for (i = 0; i < len; i++)
				dst[i] = pdf_code_from_string(&buf->scratch[i * 2], 2);
			pdf_map_one_to_many(cmap, src, dst, i);
		}
	}
}

pdf_cmap *pdf_load_cmap(fz_warn_context *ctx, buf_t *file)
{
	pdf_cmap *cmap;
	char key[64];
	pdf_lexbuf buf;
	pdf_token tok;

	pdf_lexbuf_init(ctx, &buf, PDF_LEXBUF_SMALL);
	cmap = pdf_new_cmap();

	strcpy(key, ".notdef");

	while (1) {
			tok = pdf_lex(file, &buf);
			if (tok == PDF_TOK_EOF) {
				break;
			} else if (tok == PDF_TOK_NAME) {
				if (!strcmp(buf.scratch, "CMapName"))
					pdf_parse_cmap_name(ctx, cmap, file, &buf);
				else if (!strcmp(buf.scratch, "WMode"))
					pdf_parse_wmode(ctx, cmap, file, &buf);
				else
					pdf_strlcpy(key, buf.scratch, sizeof key);
			} else if (tok == PDF_TOK_KEYWORD) {
				if (!strcmp(buf.scratch, "endcmap")) {
					break;
				} else if (!strcmp(buf.scratch, "usecmap")) {
					pdf_strlcpy(cmap->usecmap_name, key, sizeof(cmap->usecmap_name));
				} else if (!strcmp(buf.scratch, "begincodespacerange")) {
					if (!pdf_parse_codespace_range(ctx, cmap, file, &buf))
                        goto err;
				} else if (!strcmp(buf.scratch, "beginbfchar")) {
					if (!pdf_parse_bf_char(ctx, cmap, file, &buf))
                        goto err;
				} else if (!strcmp(buf.scratch, "begincidchar")) {
					if (!pdf_parse_cid_char(ctx, cmap, file, &buf))
                        goto err;
				} else if (!strcmp(buf.scratch, "beginbfrange")) {
					if (!pdf_parse_bf_range(ctx, cmap, file, &buf))
                        goto err;
				} else if (!strcmp(buf.scratch, "begincidrange")) {
					if (!pdf_parse_cid_range(ctx, cmap, file, &buf))
                        goto err;
                }
			}

			/* ignore everything else */
		}

	pdf_sort_cmap(ctx, cmap);
	pdf_lexbuf_fin(&buf);
	return cmap;

err:
	pdf_free_cmap(cmap);
	fz_warn(ctx, "syntaxerror in cmap");
    return NULL;
}

