/**
  * Bidirectional text processing.
  *
  * Processes uint16_t text by arranging the characters into an order suitable
  * for display. E.g. Hebrew text will be arranged from right-to-left and
  * any English within the text will remain in the left-to-right order.
  * Characters such as parenthesis will be substituted for their mirrored
  * equivalents if they are part of text which must be reversed.
  *
  * This is an implementation of the uint16_t Bidirectional Algorithm which
  * can be found here: http://www.uint16_t.org/reports/tr9/ and is based
  * on the reference implementation of the algorithm found on that page.
  *
  * For a nice overview of how it works, read this...
  * http://www.w3.org/TR/REC-html40/struct/dirlang.html
  *
  * Extracted from the SmartOffice code, where it was modified by Ian
  * Beveridge.
  *
  * Copyright (C) Picsel, 2004. All Rights Reserved.
  */

/**
  * Original copyright notice from uint16_t reference implementation.
  * ----------------------------------------------------------------
  * Written by: Asmus Freytag
  *	 C++ and Windows dependencies removed, and
  *	 command line interface added by: Rick McGowan
  *
  *	 Copyright (C) 1999, ASMUS, Inc.	 All Rights Reserved
  */

/*
 * Includes...
 */

#include "mupdf/fitz.h"
#include "bidi-impl.h" /* standard bidi code interface */

/*
 * Macros...
 */

#define ODD(x) ((x) & 1)

#define REPLACEABLE_TYPE(t)\
						(((t)==BDI_ES) || ((t)==BDI_ET )|| ((t)==BDI_CS )||\
						 ((t)==BDI_NSM)|| ((t)==BDI_PDF)|| ((t)==BDI_BN )||\
						 ((t)==BDI_S)  || ((t)==BDI_WS )|| ((t)==BDI_N  ) )

#ifdef DEBUG_BIDI_VERBOSE
#define DBUGVF(params) do { fz_warn params; } while (0)
#else
#define DBUGVF(params) do {} while (0)
#endif

#ifdef DEBUG_BIDI_OUTLINE
#define DBUGH(params) do { fz_warn params; } while (0)
#else
#define DBUGH(params) do {} while (0)
#endif

#define UNICODE_EOS                                  ((uint16_t)0)
#define UNICODE_DIGIT_ZERO                           ((uint16_t)0x0030)
#define UNICODE_DIGIT_NINE                           ((uint16_t)0x0039)
#define UNICODE_RTL_START                            ((uint16_t)0x0590)
#define UNICODE_RTL_END                              ((uint16_t)0x07BF)
#define UNICODE_ZERO_WIDTH_NON_JOINER                ((uint16_t)0x200C)
#define UNICODE_FULLWIDTH_DIGIT_ZERO                 ((uint16_t)0xFF10)
#define UNICODE_FULLWIDTH_DIGIT_NINE                 ((uint16_t)0xFF19)

#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif

/*
 * Enumerations...
 */

#ifdef DEBUG
/* display support: */
static const char char_from_types[] =
{
	' ',	/* ON, */
	'>',	/* L,  */
	'<',	/* R,  */
	'9',	/* AN, */
	'1',	/* EN, */
	'a',	/* AL  */
	'@',	/* NSM */
	'.',	/* CS  */
	',',	/* ES  */
	'$',	/* ET  */
	':',	/* BN  */
	'X',	/* S   */
	'_',	/* WS  */
	'B',	/* B   */
	'+',	/* RLO */
	'+',	/* RLE */
	'+',	/* LRO */
	'+',	/* LRE */
	'-',	/* PDF */
	'='	 /* LS  */
};
#endif /* DEBUG */

/*
 * Functions and static functions...
 */

/* UCDN uses a different ordering than Bidi does. We cannot
 * change to the UCDN ordering, as the bidi-std.c code relies
 * on the exact ordering (at least that N = ON = 0). We
 * therefore map between the two using this small table. It
 * also takes care of fudging LRI, RLI, FSI and PDI, that this
 * code does not currently support. */
static const uint8_t ucdn_to_bidi[] =
{
	BDI_L,		/* UCDN_BIDI_CLASS_L = 0 */
	BDI_LRE,	/* UCDN_BIDI_CLASS_LRE = 1 */
	BDI_LRO,	/* UCDN_BIDI_CLASS_LRO = 2 */
	BDI_R,		/* UCDN_BIDI_CLASS_R = 3 */
	BDI_AL,		/* UCDN_BIDI_CLASS_AL = 4 */
	BDI_RLE,	/* UCDN_BIDI_CLASS_RLE = 5 */
	BDI_RLO,	/* UCDN_BIDI_CLASS_RLO = 6 */
	BDI_PDF,	/* UCDN_BIDI_CLASS_PDF = 7 */
	BDI_EN,		/* UCDN_BIDI_CLASS_EN = 8 */
	BDI_ES,		/* UCDN_BIDI_CLASS_ES = 9 */
	BDI_ET,		/* UCDN_BIDI_CLASS_ET = 10 */
	BDI_AN,		/* UCDN_BIDI_CLASS_AN = 11 */
	BDI_CS,		/* UCDN_BIDI_CLASS_CS = 12 */
	BDI_NSM,	/* UCDN_BIDI_CLASS_NSM = 13 */
	BDI_BN,		/* UCDN_BIDI_CLASS_BN = 14 */
	BDI_B,		/* UCDN_BIDI_CLASS_B = 15 */
	BDI_S,		/* UCDN_BIDI_CLASS_S = 16 */
	BDI_WS,		/* UCDN_BIDI_CLASS_WS = 17 */
	BDI_ON,		/* UCDN_BIDI_CLASS_ON = 18 */
	BDI_LRE,	/* UCDN_BIDI_CLASS_LRI = 19 */
	BDI_RLE,	/* UCDN_BIDI_CLASS_RLI = 20 */
	BDI_N,		/* UCDN_BIDI_CLASS_FSI = 21 */
	BDI_N,		/* UCDN_BIDI_CLASS_PDI = 22 */
};

#define class_from_ch_ws(ch) (ucdn_to_bidi[ucdn_get_bidi_class(ch)])

/** Return a direction for white-space on the second pass of the algorithm.
  */
static unsigned char class_from_ch_n(uint16_t ch)
{
	int from_ch_ws = class_from_ch_ws(ch);

	if (from_ch_ws == BDI_S || from_ch_ws == BDI_WS)
		return BDI_N;

	return from_ch_ws;
}

/*
 *The original BidiMirrorList bidiMirrorList[] array has been replaced by
 *uint32_t bidiOptMirrorList[] an array that represents a tree structure
 *formed from the original mirror list by the Perl script
 *create-mirrorlist.pl in Review #11191
 *
 *below is the first part of the tree:- representing the uint16_t values
 *0x0028, 0x0029, 0x003C, 0x003E, 0x005B, 0x005D
 *
 *               0
 *               |
 *           ----0----
 *          /    |    \
 *	   2	 3     5
 *	  / \   / \   / \
 *	 8   9 C   E B   D
 *
 *Every uint16_t value that shares nybble ancestors share parent nodes
 *on the tree.
 *The tree is of fixed depth 4 and each node holds the following values
 *{4-bits nybble, 12-bits next, 16-bits uint16_t mirror}.
 *The tree is represented as an array of bit-masks.
 *
 *The nybble value is the value in the tree above.
 *The next Offset is the index into the array which points to the next
 *sibling for the current node (i.e. the next node horizintally to the right
 *that shares the same parent).
 *The mirror value is the code to be mapped.
 */
static const uint32_t bidiOptMirrorList[] =
{
	0x00120000, 0x0FFF0000, 0x20050000,
	0x80040029, 0x9FFF0028, 0x30080000,
	0xC007003E, 0xEFFF003C, 0x500B0000,
	0xB00A005D, 0xDFFF005B, 0x700E0000,
	0xB00D007D, 0xDFFF007B, 0xA0100000,
	0xBFFF00BB, 0xBFFF0000, 0xBFFF00AB,
	0x215A0000, 0x00200000, 0x30170000,
	0x9016203A, 0xAFFF2039, 0x401A0000,
	0x50192046, 0x6FFF2045, 0x701D0000,
	0xD01C207E, 0xEFFF207D, 0x8FFF0000,
	0xD01F208E, 0xEFFF208D, 0x20A20000,
	0x00280000, 0x8023220B, 0x9024220C,
	0xA025220D, 0xB0262208, 0xC0272209,
	0xDFFF220A, 0x102A0000, 0x5FFF29F5,
	0x302D0000, 0xC02C223D, 0xDFFF223C,
	0x402F0000, 0x3FFF22CD, 0x50340000,
	0x20312253, 0x30322252, 0x40332255,
	0x5FFF2254, 0x603F0000, 0x40362265,
	0x50372264, 0x60382267, 0x70392266,
	0x803A2269, 0x903B2268, 0xA03C226B,
	0xB03D226A, 0xE03E226F, 0xFFFF226E,
	0x70500000, 0x00412271, 0x10422270,
	0x20432273, 0x30442272, 0x40452275,
	0x50462274, 0x60472277, 0x70482276,
	0x80492279, 0x904A2278, 0xA04B227B,
	0xB04C227A, 0xC04D227D, 0xD04E227C,
	0xE04F227F, 0xFFFF227E, 0x805E0000,
	0x00522281, 0x10532280, 0x20542283,
	0x30552282, 0x40562285, 0x50572284,
	0x60582287, 0x70592286, 0x805A2289,
	0x905B2288, 0xA05C228B, 0xB05D228A,
	0xFFFF2290, 0x90630000, 0x0060228F,
	0x10612292, 0x20622291, 0x8FFF29B8,
	0xA06A0000, 0x206522A3, 0x306622A2,
	0x60672ADE, 0x80682AE4, 0x90692AE3,
	0xBFFF2AE5, 0xB0730000, 0x006C22B1,
	0x106D22B0, 0x206E22B3, 0x306F22B2,
	0x407022B5, 0x507122B4, 0x607222B7,
	0x7FFF22B6, 0xC0790000, 0x907522CA,
	0xA07622C9, 0xB07722CC, 0xC07822CB,
	0xDFFF2243, 0xD0860000, 0x007B22D1,
	0x107C22D0, 0x607D22D7, 0x707E22D6,
	0x807F22D9, 0x908022D8, 0xA08122DB,
	0xB08222DA, 0xC08322DD, 0xD08422DC,
	0xE08522DF, 0xFFFF22DE, 0xE0950000,
	0x008822E1, 0x108922E0, 0x208A22E3,
	0x308B22E2, 0x408C22E5, 0x508D22E4,
	0x608E22E7, 0x708F22E6, 0x809022E9,
	0x909122E8, 0xA09222EB, 0xB09322EA,
	0xC09422ED, 0xDFFF22EC, 0xFFFF0000,
	0x009722F1, 0x109822F0, 0x209922FA,
	0x309A22FB, 0x409B22FC, 0x609C22FD,
	0x709D22FE, 0xA09E22F2, 0xB09F22F3,
	0xC0A022F4, 0xD0A122F6, 0xEFFF22F7,
	0x30AB0000, 0x00A80000, 0x80A52309,
	0x90A62308, 0xA0A7230B, 0xBFFF230A,
	0x2FFF0000, 0x90AA232A, 0xAFFF2329,
	0x70CC0000, 0x60B50000, 0x80AE2769,
	0x90AF2768, 0xA0B0276B, 0xB0B1276A,
	0xC0B2276D, 0xD0B3276C, 0xE0B4276F,
	0xFFFF276E, 0x70BC0000, 0x00B72771,
	0x10B82770, 0x20B92773, 0x30BA2772,
	0x40BB2775, 0x5FFF2774, 0xD0C10000,
	0x50BE27D6, 0x60BF27D5, 0xD0C027DE,
	0xEFFF27DD, 0xEFFF0000, 0x20C327E3,
	0x30C427E2, 0x40C527E5, 0x50C627E4,
	0x60C727E7, 0x70C827E6, 0x80C927E9,
	0x90CA27E8, 0xA0CB27EB, 0xBFFF27EA,
	0x90FD0000, 0x80DB0000, 0x30CF2984,
	0x40D02983, 0x50D12986, 0x60D22985,
	0x70D32988, 0x80D42987, 0x90D5298A,
	0xA0D62989, 0xB0D7298C, 0xC0D8298B,
	0xD0D92990, 0xE0DA298F, 0xFFFF298E,
	0x90E50000, 0x00DD298D, 0x10DE2992,
	0x20DF2991, 0x30E02994, 0x40E12993,
	0x50E22996, 0x60E32995, 0x70E42998,
	0x8FFF2997, 0xB0E70000, 0x8FFF2298,
	0xC0ED0000, 0x00E929C1, 0x10EA29C0,
	0x40EB29C5, 0x50EC29C4, 0xFFFF29D0,
	0xD0F70000, 0x00EF29CF, 0x10F029D2,
	0x20F129D1, 0x40F229D5, 0x50F329D4,
	0x80F429D9, 0x90F529D8, 0xA0F629DB,
	0xBFFF29DA, 0xFFFF0000, 0x50F92215,
	0x80FA29F9, 0x90FB29F8, 0xC0FC29FD,
	0xDFFF29FC, 0xAFFF0000, 0x21030000,
	0xB1002A2C, 0xC1012A2B, 0xD1022A2E,
	0xEFFF2A2D, 0x31080000, 0x41052A35,
	0x51062A34, 0xC1072A3D, 0xDFFF2A3C,
	0x610B0000, 0x410A2A65, 0x5FFF2A64,
	0x71110000, 0x910D2A7A, 0xA10E2A79,
	0xD10F2A7E, 0xE1102A7D, 0xFFFF2A80,
	0x81190000, 0x01132A7F, 0x11142A82,
	0x21152A81, 0x31162A84, 0x41172A83,
	0xB1182A8C, 0xCFFF2A8B, 0x91260000,
	0x111B2A92, 0x211C2A91, 0x311D2A94,
	0x411E2A93, 0x511F2A96, 0x61202A95,
	0x71212A98, 0x81222A97, 0x91232A9A,
	0xA1242A99, 0xB1252A9C, 0xCFFF2A9B,
	0xA1320000, 0x11282AA2, 0x21292AA1,
	0x612A2AA7, 0x712B2AA6, 0x812C2AA9,
	0x912D2AA8, 0xA12E2AAB, 0xB12F2AAA,
	0xC1302AAD, 0xD1312AAC, 0xFFFF2AB0,
	0xB13B0000, 0x01342AAF, 0x31352AB4,
	0x41362AB3, 0xB1372ABC, 0xC1382ABB,
	0xD1392ABE, 0xE13A2ABD, 0xFFFF2AC0,
	0xC1460000, 0x013D2ABF, 0x113E2AC2,
	0x213F2AC1, 0x31402AC4, 0x41412AC3,
	0x51422AC6, 0x61432AC5, 0xD1442ACE,
	0xE1452ACD, 0xFFFF2AD0, 0xD14F0000,
	0x01482ACF, 0x11492AD2, 0x214A2AD1,
	0x314B2AD4, 0x414C2AD3, 0x514D2AD6,
	0x614E2AD5, 0xEFFF22A6, 0xE1550000,
	0x315122A9, 0x415222A8, 0x515322AB,
	0xC1542AED, 0xDFFF2AEC, 0xFFFF0000,
	0x71572AF8, 0x81582AF7, 0x91592AFA,
	0xAFFF2AF9, 0x31700000, 0x0FFF0000,
	0x01650000, 0x815E3009, 0x915F3008,
	0xA160300B, 0xB161300A, 0xC162300D,
	0xD163300C, 0xE164300F, 0xFFFF300E,
	0x1FFF0000, 0x01673011, 0x11683010,
	0x41693015, 0x516A3014, 0x616B3017,
	0x716C3016, 0x816D3019, 0x916E3018,
	0xA16F301B, 0xBFFF301A, 0xFFFF0000,
	0xFFFF0000, 0x01750000, 0x8174FF09,
	0x9FFFFF08, 0x11780000, 0xC177FF1E,
	0xEFFFFF1C, 0x317B0000, 0xB17AFF3D,
	0xDFFFFF3B, 0x517F0000, 0xB17DFF5D,
	0xD17EFF5B, 0xFFFFFF60, 0x6FFF0000,
	0x0181FF5F, 0x2182FF63, 0x3FFFFF62
};

#define NUM_BIDI_OPTMIRRORS (sizeof(bidiOptMirrorList)/sizeof(uint32_t))

#define MIRROR_NYBBLE_SHIFT	 28
#define MIRROR_NEXT_MASK		0x0FFF0000
#define MIRROR_NEXT_SHIFT	   16
#define MIRROR_VALUE_MASK	   0x0000FFFF
#define TRIPPLE_NYBBLE_NEG	  4095

/** The Bidi_mirrorChar function tranverses the tree looking for nybble
  * matches.
  *
  * EXAMPLE:
  * If uint16_t value u=0x005D has been passed to Bidi_mirrorChar for testing,
  * it is separated into nybbles 0,0,5,D
  * The first node in the tree (index 0 in the array) is the starting point,
  * and luckily the first 4-bits hold a 0, which matches. We can therefore
  * drop down to the child to check that. The child of a node is stored in
  * the next array position, in this case index 1.
  * The nybble of index 1 is a 0 and matches the second nybble we are
  * searching for. Because of the match, the next child index 2 is moved to.
  * Now the nybble at index 2 of the tree is a 2 and doesn't match the 5 we
  * are searching for. Bits 5-16 of the array represent the offset to the
  * next sibling. In this case the offset is 5, when we move to this array
  * value we see that it holds a 3 nybble, again this doesn't match, so we
  * get the next offset, which is 8, and move to that sibling.
  * The sibling holds the nybble 5, which is the next one we want; so
  * increment the index to its child. The child doesn't match and holds a B
  * instead of a D. So we move to the next sibling ... which matches!!
  * This node holds the mirror character 0x005B in bits 17-32.
  */
uint16_t Bidi_mirrorChar(const uint16_t u)
{
    unsigned int i;

	if((u < UNICODE_RTL_START) || (u > UNICODE_RTL_END)) {
		/* uint16_t value lies outside of RTL character range
		 * and could therefore be a mirrorable charcter
		 */
		unsigned int index = 0;
		uint16_t uCpy = (uint16_t)u;
		/* look through each nybble of the uint16_t and search in tree
		 * for matches
		 */
		for(i = 0; i < 4; i++) {
			uint8_t n = (uint8_t)(uCpy >> 12); /* the current nybble of the target */
			/* the nybble of the first child node */
			uint8_t nybble = (uint8_t)(bidiOptMirrorList[index] >> MIRROR_NYBBLE_SHIFT);

			/* search through node siblings for the matching nybble */
			while(nybble != n) {
				/* index to next sibling */
				index = (bidiOptMirrorList[index] & MIRROR_NEXT_MASK) >> MIRROR_NEXT_SHIFT;
				if(index == TRIPPLE_NYBBLE_NEG) { /* 12-bit next value is -1 */
					/* no more siblings */
					return UNICODE_EOS;
				}
				nybble = (uint8_t)(bidiOptMirrorList[index] >> MIRROR_NYBBLE_SHIFT);
			}
			uCpy <<= 4;
			index++;
		}
		return (uint16_t)(bidiOptMirrorList[--index] & MIRROR_VALUE_MASK);
	}
	return u;
}

/** Subsitutes characters with their mirrored equivalents, e.g. swaps
  * '(' with ')'. Implements rule L4 of the uint16_t Bidirectional
  * algorithm. Mirrored characters are found using a binary search through
  * the lookup table bidiOptMirrorList.
  */
void fz_bidi_mirror_line(uint16_t *text, unsigned int len, const unsigned char *poslevel)
{
    unsigned int i;
	uint16_t out;

	for (i = 0; i < len; i ++) {
		if (!(poslevel[i] & 1))
			continue;
		out = Bidi_mirrorChar(text[i]);
		if(out)
			text[i] = out;
	}
}

static unsigned int
is_european_number(const uint16_t *str, unsigned int len)
{
    unsigned int i;
	for (i = 0 ; i < len; i++) {
		const uint16_t  u = str[i];
		if ((u >= UNICODE_RTL_START && u <= UNICODE_RTL_END)) {
			/* This is just a normal RTL character or accent */
			return FALSE;
		} else if(!((u >= 0x30 && u <= 0x39) || (u >= 0xff10 && u <= 0xff19) || (u == UNICODE_ZERO_WIDTH_NON_JOINER))) {
			return FALSE;
		}
	}
	return TRUE;
}

static void
do_callback(const uint16_t *fragment, size_t fragment_len, int block_r2l, uint16_t mirror,
            fz_bidi_fragment_callback *callback, void *arg)
{
	char char_r2l = block_r2l && !is_european_number(fragment, fragment_len);
	(*callback)(fragment, fragment_len, block_r2l, char_r2l, mirror, arg);
}

/** Searches a RTL fragment for a mirror character
 * When it finds one it creates a separate fragment for the
 * character and the surrounding fragments. It passes the mirrored
 * uint16_t back through the callback.
 */
static void
create_fragment_mirrors(const uint16_t *text, unsigned int len, fz_bidi_fragment_callback *callback, void *arg)
{
	unsigned int i, lastPtr = 0;
	uint16_t mirror;

	assert(text != NULL);
	assert(len > 0);
	for (i = 0; i < len; i ++) {
		mirror = Bidi_mirrorChar(text[i]);
		if(mirror != UNICODE_EOS) {
			/* create preceding fragment */
			if(i > lastPtr) {
				do_callback(&text[lastPtr], i - lastPtr, TRUE, UNICODE_EOS, callback, arg);
				DBUGVF(("create mirror fragment for %x\n",(int)text[i]));
			}
			/* create mirror fragment */
			do_callback(&text[i], 1, TRUE, mirror, callback, arg);
			lastPtr = i + 1;
		}
	}

	if(lastPtr < len) {
		/* create end fragment */
		do_callback(&text[lastPtr], len - lastPtr, TRUE, UNICODE_EOS, callback, arg);
	}
}

/** Determines the character classes for all following
  * passes of the algorithm. A character class is basically the type of Bidi
  * behaviour that the character exhibits.
  */
static void
classify_characters(const uint16_t *text, unsigned int len,
                    unsigned char *types, fz_bidi_flags flags)
{
	int i;

	if ((flags & BIDI_CLASSIFY_WHITE_SPACE)!=0)
	{
		for (i = 0; i < len; i++)
		{
			types[i] = class_from_ch_ws(text[i]);
		}
	}
	else
	{
#ifdef DEBUG_BIDI_VERBOSE
		fprintf(stderr, "Text:   ");
		for (i = 0; i < len; i++)
		{
			/* So that we can actually sort of read the debug string, any
			 * non-ascii characters are replaced with a 1-digit hash
			 * value from 0-9, making non-english characters appear
			 * as numbers
			 */
			fprintf(stderr, "%c", (text[i ]>= 32)?text[i]:(char)((text[i] % 9) + 0x30));
		}
		fprintf(stderr, "\nTypes:  ");
#endif
		for (i = 0; i < len; i++) {
			types[i] = class_from_ch_n(text[i]);
#ifdef DEBUG_BIDI_VERBOSE
			fprintf(stderr, "%c", char_from_types[types[i]]);
#endif
		}
#ifdef DEBUG_BIDI_VERBOSE
		fprintf(stderr, "\n");
#endif
	}
}

/** Determines the base level of the text.
  * Implements rule P2 of the uint16_t Bidi Algorithm.
  * Note: Ignores explicit embeddings
  */
static unsigned char base_level_from_text(unsigned char *types, unsigned int len)
{
    unsigned int i;

	for (i = 0; i < len; i++)
	{
		switch (types[i])
		{
		/* strong left */
		case BDI_L:
			return BIDI_LEFT_TO_RIGHT;

		/* strong right */
		case BDI_R:
		case BDI_AL:
			return BIDI_RIGHT_TO_LEFT;
		}
	}
	return BIDI_LEFT_TO_RIGHT;
}

static fz_bidi_direction direction_from_type(unsigned int type)
{
	switch( type )
	{
	case BDI_L:
	case BDI_EN:
		return BIDI_LEFT_TO_RIGHT;

	case BDI_R:
	case BDI_AL:
		return BIDI_RIGHT_TO_LEFT;

	default:
		return BIDI_NEUTRAL;
	}
}

static void
classify_quoted_blocks(const uint16_t *text, size_t len, unsigned char *types)
{
	size_t i;
	int inQuote = FALSE;
	int pdfNeeded = FALSE;
	int ltrFound = FALSE;
	int rtlFound = FALSE;

	/* Only do anything special here if there is mixed content
	 * (LTR *and* RTL) in the text.
	 */
	for (i = 0; i < len; i++)
	{
		switch ( direction_from_type(types[i]) )
		{
		case BIDI_LEFT_TO_RIGHT:
			ltrFound = TRUE;
			break;

		case BIDI_RIGHT_TO_LEFT:
			rtlFound = TRUE;
			break;

		default:
			break;
		}
	}

	/* Only make any changes if *both* LTR and RTL characters exist
	 * in this text.
	 */
	if ( !ltrFound || !rtlFound )
	{
		return;
	}

	for (i = 0; i < len; i++)
	{
		if (text[i]=='"')
		{
			/* If we're already in a quote then terminate it,
			 * else start a new block.
			 */
			if ( inQuote )
			{
				inQuote = FALSE;
				if ( pdfNeeded )
				{
					pdfNeeded = FALSE;
					types[i] = BDI_PDF;
				}
			}
			else
			{
				size_t j;
				int   done = FALSE;

				inQuote = TRUE;

				/* Find the first strong right or left type and
				 * use that to determine whether we should classify
				 * the quote as LRE or RLE.  Or neither, if we
				 * hit another quote before any strongly-directional
				 * character.
				 */
				for ( j = i + 1;
					  !done && (j < len) && text[j] != '"';
					  ++j )
				{
					switch( types[j] )
					{
					case BDI_RLE:
					case BDI_LRE:
						done = TRUE;
						break;

					case BDI_L:
					case BDI_EN:
						types[i]  = BDI_LRE;
						pdfNeeded = TRUE;
						done	  = TRUE;
						break;

					case BDI_R:
					case BDI_AL:
						types[i]  = BDI_RLE;
						pdfNeeded = TRUE;
						done	  = TRUE;
						break;

					default:
						break;
					}
				}
			}
		}
	}
}

// === RESOLVE IMPLLICIT =================================================

/*------------------------------------------------------------------------
	Function: resolveImplicit

	Recursively resolves implicit embedding levels.
	Implements rules I1 and I2 of the Unicode Bidirectional Algorithm.

	Input: Array of direction classes
		   Character count
		   Base level

	In/Out: Array of embedding levels

	Note: levels may exceed 15 on output.
		  Accepted subset of direction classes
		  R, L, AN, EN
------------------------------------------------------------------------*/
const unsigned char add_level[][4] =
{
	// L,  R,	AN, EN = cls
    // level =
/* even */ { 0,	1,	2,	2 }, // EVEN
/* odd	*/ { 1,	0,	1,	1 }	 // ODD
};

static inline void fz_bidi_resolve_implicit(const unsigned char *pcls, unsigned char *plevel, int cch)
{
	int ich;

	for (ich = 0; ich < cch; ich++) {
		// cannot resolve bn here, since some bn were resolved to strong
		// types in resolveWeak. To remove these we need the original
		// types, which are available again in resolveWhiteSpace
		if (pcls[ich] == BDI_BN) {
			continue;
		}
		assert(pcls[ich] > 0); // "No Neutrals allowed to survive here."
		assert(pcls[ich] < 5); // "Out of range."
		plevel[ich] += add_level[plevel[ich] & 1][pcls[ich] - 1];
	}
}

/*------------------------------------------------------------------------
	Function: fz_bidi_resolve_whitespace

	Resolves levels for WS and S
	Implements rule L1 of the Unicode bidi Algorithm.

	Input:	Base embedding level
			Character count
			Array of direction classes (for one line of text)

	In/Out: Array of embedding levels (for one line of text)

	Note: this should be applied a line at a time. The default driver
		  code supplied in this file assumes a single line of text; for
		  a real implementation, cch and the initial pointer values
		  would have to be adjusted.
------------------------------------------------------------------------*/
static inline
void set_deferred_level_run(unsigned char *pval, int cval, int iStart, unsigned char nval)
{
	int i;
	for (i = iStart - 1; i >= iStart - cval; i--) {
		pval[i] = nval;
	}
}

static void fz_bidi_resolve_whitespace(unsigned char baselevel, const unsigned char *pcls, unsigned char *plevel, int cch)
{
	int cchrun = 0;
	unsigned char oldlevel = baselevel;
	int ich;

	for (ich = 0; ich < cch; ich++)
	{
		switch(pcls[ich])
		{
		default:
			cchrun = 0; // any other character breaks the run
			break;
		case BDI_RLE:
		case BDI_LRE:
		case BDI_LRO:
		case BDI_RLO:
		case BDI_PDF:
		case BDI_BN:
			plevel[ich] = oldlevel;
			cchrun++;
			break;

		case BDI_WS:
			cchrun++;
			break;
		case BDI_S:
		case BDI_B:
			// reset levels for WS before eot
			set_deferred_level_run(plevel, cchrun, ich, baselevel);
			cchrun = 0;
			plevel[ich] = baselevel;
			break;
		}
		oldlevel = plevel[ich];
	}
	// reset level before eot
	set_deferred_level_run(plevel, cchrun, ich, baselevel);
}

// === RESOLVE NEUTAL TYPES ==============================================

// action values
enum neutral_action
{
	// action to resolve previous input
	En = 3 << 4,		// resolve neutrals run to embedding level direction
	Rn = BDI_R << 4,	// resolve neutrals run to strong right
	Ln = BDI_L << 4,	// resolved neutrals run to strong left
	In = (1<<8),		// increment count of deferred neutrals
};

// state values
enum neutral_state
{
	// new temporary class
	l = 0,	// L and characters resolved to L
	r = 1,	// R and characters resolved to R
	ln = 2,	// N preceded by left
	rn = 3,	// N preceded by right
};

/*------------------------------------------------------------------------
	Notes:

	By rule W7, whenever a EN is 'dominated' by an L (including start of
	run with embedding direction = L) it is resolved to, and further treated
	as L.

	This leads to the need for 'a' and 'na' states.
------------------------------------------------------------------------*/

const int action_neutrals[][5] =
{
    //	N,	L,	R = cls
	// state =
	{In,  0,  0},		// r	right
	{In,  0,  0},	// l	left

	{In, En, Rn},		// rn	N preceded by right
	{In, Ln, En},		// ln	N preceded by left
} ;

static const unsigned char state_neutrals[2] = {
    0x13, 0x12
} ;

/*------------------------------------------------------------------------
	Function: resolveNeutrals

	Resolves the directionality of neutral character types.

	Implements rules W7, N1 and N2 of the Unicode Bidi Algorithm.

	Input: Array of embedding levels
		   Character count
		   Baselevel

	In/Out: Array of directional classes

	Note: On input only these directional classes are expected
		  R,  L,  N, AN, EN and BN

		  W8 resolves a number of ENs to L
------------------------------------------------------------------------*/
static inline
void set_deferred_run(unsigned char *pval, int cval, int iStart, unsigned char nval)
{
	int i;
	for (i = iStart - 1; i >= iStart - cval; i--) {
		pval[i] = nval;
	}
}

static void fz_bidi_resolve_neutrals(unsigned char baselevel, unsigned char *pcls, const unsigned char *plevel, int cch)
{
	// the state at the start of text depends on the base level
	int state = (baselevel & 1) ? r : l;
	unsigned char cls, cls_new, cls_run;
	int ich, cch_run = 0, action;
	unsigned char level = baselevel;

	for (ich = 0; ich < cch; ich++) {
		// ignore boundary neutrals
		if (pcls[ich] == BDI_BN) {
			// include in the count for a deferred run
			if (cch_run)
				cch_run++;
			// skip any further processing
			continue;
		}

		assert(pcls[ich] < 3); // "Only N, L, R are allowed"
		cls = pcls[ich];
		action = action_neutrals[state][cls&0x03];

		// resolve the directionality for deferred runs
		cls_run = ((action >> 4) & 0x0F);
        if (cls_run == 3) cls_run = ((level & 1) ? BDI_R : BDI_L);
		if (cls_run != BDI_N) {
			set_deferred_run(pcls, cch_run, ich, cls_run);
			cch_run = 0;
		}

		// resolve the directionality class at the current location
		cls_new = (action & 0x0F);
        if (cls_new == In) cls_new = 0;
		if (cls_new != BDI_N)
			pcls[ich] = cls_new;

		if (In & action)
			cch_run++;

		state = ((state_neutrals[state&1] >> (2*cls)) & 0x03);
		level = plevel[ich];
	}

	// resolve any deferred runs
	cls = ((level & 1) ? BDI_R : BDI_L);	// eor has type of current level

	// resolve the directionality for deferred runs
	cls_run = ((action_neutrals[state][cls] >> 4) & 0x0F);
    if (cls_run == 3) cls_run = ((level & 1) ? BDI_R : BDI_L);
	if (cls_run != BDI_N)
		set_deferred_run(pcls, cch_run, ich, cls_run);
}

/* Creates a buffer with an embedding level for every uint16_t in the
 * given text.  Also determines the base level and returns it in
 * *baseDir if *baseDir does not initially contain a valid direction.
 */
unsigned char *
create_levels(fz_context *ctx, const uint16_t *text, size_t len, fz_bidi_direction *baseDir, int resolveWhiteSpace, int bidiFlag)
{
	unsigned char baseLevel, *levels, *types;
	levels = fz_malloc(ctx, len);
	types = fz_malloc(ctx, len);

	{
		classify_characters(text, len, types, bidiFlag);

		if (*baseDir != BIDI_LEFT_TO_RIGHT && *baseDir != BIDI_RIGHT_TO_LEFT) {
			/* Derive the base level from the text and
			 * update *baseDir in case the caller wants to know.
			 */
			baseLevel = base_level_from_text(types, len);
			*baseDir = ODD(baseLevel)==1 ? BIDI_RIGHT_TO_LEFT : BIDI_LEFT_TO_RIGHT;
		} else {
			baseLevel = (unsigned char)*baseDir;
		}

		{
			/* Replace tab with base direction, i.e. make tab appear as
			 * 'strong left' if the base direction is left-to-right and
			 * 'strong right' if base direction is right-to-left.  This
			 * allows Layout to implicitly treat tabs as 'segment separators'.
			 */
			size_t i;

			for (i = 0u; i < len; i++) {
				if (text[i]=='\t') {
					types[i] = (*baseDir == BIDI_RIGHT_TO_LEFT) ? BDI_R : BDI_L;
				}
			}
		}

		/* Look for quotation marks.  Classify them as RLE or LRE
		 * or leave them alone, depending on what follows them.
		 */
		classify_quoted_blocks(text, len, types);

		/* Work out the levels and character types... */
		(void)fz_bidi_resolve_explicit(baseLevel, BDI_N, types, levels, len, 0);
		fz_bidi_resolve_weak(ctx, baseLevel, types, levels, len);
		fz_bidi_resolve_neutrals(baseLevel,types, levels, len);
		fz_bidi_resolve_implicit(types, levels, len);

		classify_characters(text, len, types, BIDI_CLASSIFY_WHITE_SPACE);

		if (resolveWhiteSpace)
		{
			/* resolve whitespace */
			fz_bidi_resolve_whitespace(baseLevel, types, levels, len);
		}

		/* The levels buffer now has odd and even numbers indicating
		 * rtl or ltr characters, respectively.
		 */
#ifdef DEBUG_BIDI_VERBOSE
		fprintf(stderr, "Levels: ");
		{
			size_t i;
			for (i = 0; i < len; i++)
			{
				fprintf(stderr, "%d", levels[i]>9?0:levels[i]);
			}
			fprintf(stderr, "\n");
		}
#endif
	}
	fz_free(ctx, types);
	return levels;
}

/* Partitions the given uint16_t sequence into one or more unidirectional
 * fragments and invokes the given callback function for each fragment.
 */
void fz_bidi_fragment_text(fz_context *ctx,
			const uint16_t *text,
			size_t textlen,
			fz_bidi_direction *baseDir,
			fz_bidi_fragment_callback *callback,
			void *arg,
			int bidiFlag)
{
	size_t i, startOfFragment;
	unsigned char *levels;

	if (text == NULL || callback == NULL || textlen == 0)
		return;

	DBUGH(("fz_bidi_fragment_text( '%S', len = %d )\n", text, textlen ));

	levels = create_levels(ctx, text, textlen, baseDir, FALSE, bidiFlag);

	/* We now have an array with an embedding level
	 * for each uint16_t in text.
	 */
	assert(levels != NULL);

	fz_try(ctx)
	{
		startOfFragment = 0;
		for (i = 1; i < textlen; i++)
		{
			if (levels[i] != levels[i-1])
			{
				/* We've gone past the end of the fragment.
				 * Create a text object for it, then start
				 * a new fragment.
				 */
				if(ODD(levels[startOfFragment]) != 0)
				{
					/* if RTL check for mirrors and create sub-frags */
					create_fragment_mirrors(&text[startOfFragment], i - startOfFragment, callback, arg);
				}
				else
				{
					do_callback(&text[startOfFragment], i - startOfFragment, ODD(levels[startOfFragment]), UNICODE_EOS, callback, arg);
				}
				startOfFragment = i;
			}
		}
		/* Now i == textlen.  Deal with the final (or maybe only) fragment. */
		if(ODD(levels[startOfFragment]) != 0) {
			/* if RTL check for mirrors and create sub-frags */
			create_fragment_mirrors(&text[startOfFragment], i - startOfFragment, callback, arg);
		} else {
			/* otherwise create 1 fragment */
			do_callback(&text[startOfFragment], i - startOfFragment, ODD(levels[startOfFragment]), UNICODE_EOS, callback, arg);
		}
	}
	fz_always(ctx)
	{
		fz_free(ctx, levels);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}
}

