/************************************************************************
 * $Id: minibidi.c 9426 2012-03-05 18:34:40Z simon $
 *
 * ------------
 * Description:
 * ------------
 * This is an implemention of Unicode's Bidirectional Algorithm
 * (known as UAX #9).
 *
 *   http://www.unicode.org/reports/tr9/
 *
 * Author: Ahmad Khalifa
 *
 * -----------------
 * Revision Details:    (Updated by Revision Control System)
 * -----------------
 *  $Date: 2012-03-05 18:34:40 +0000 (Mon, 05 Mar 2012) $
 *  $Author: simon $
 *  $Revision: 9426 $
 *
 * (www.arabeyes.org - under MIT license)
 *
 ************************************************************************/

/*
 * TODO:
 * =====
 * - Explicit marks need to be handled (they are not 100% now)
 * - Ligatures
 */

#include <stdint.h>
#include <stdlib.h>	/* definition of wchar_t*/
#include <string.h>

#define LMASK	0x3F	/* Embedding Level mask */
#define OMASK	0xC0	/* Override mask */
#define OISL	0x80	/* Override is L */
#define OISR	0x40	/* Override is R */

/* Shaping Helpers */
#define STYPE(xh) ((((xh) >= SHAPE_FIRST) && ((xh) <= SHAPE_LAST)) ? \
shapetypes[(xh)-SHAPE_FIRST].type : SU) /*))*/
#define SISOLATED(xh) (shapetypes[(xh)-SHAPE_FIRST].form_b)
#define SFINAL(xh) ((xh)+1)
#define SINITIAL(xh) ((xh)+2)
#define SMEDIAL(ch) ((ch)+3)

#define leastGreaterOdd(x) (((x)+1) | 1)
#define leastGreaterEven(x) (((x)+2) &~ 1)

typedef struct bidi_char {
    unsigned int origwc, wc;
    unsigned short index;
} bidi_char;

/* function declarations */
void flipThisRun(bidi_char *from, unsigned char* level, int max, int count);
int findIndexOfRun(unsigned char* level , int start, int count, int tlevel);
unsigned char getType(int ch);
unsigned char setOverrideBits(unsigned char level, unsigned char override);
int getPreviousLevel(unsigned char* level, int from);
int do_shape(bidi_char *line, bidi_char *to, int count);
int do_bidi(bidi_char *line, unsigned int count);
unsigned int Bidi_mirrorChar(unsigned int u);

/* character types */
enum {
    L,
    LRE,
    LRO,
    R,
    AL,
    RLE,
    RLO,
    PDF,
    EN,
    ES,
    ET,
    AN,
    CS,
    NSM,
    BN,
    B,
    S,
    WS,
    ON
};

/*
 * Flips the text buffer, according to max level, and
 * all higher levels
 *
 * Input:
 * from: text buffer, on which to apply flipping
 * level: resolved levels buffer
 * max: the maximum level found in this line (should be unsigned char)
 * count: line size in bidi_char
 */
void flipThisRun(bidi_char *from, unsigned char *level, int max, int count)
{
    int i, j, k, tlevel;
    bidi_char temp;

    j = i = 0;
    while (i<count && j<count) {

	/* find the start of the run of level=max */
	tlevel = max;
	i = j = findIndexOfRun(level, i, count, max);
	/* find the end of the run */
	while (i<count && tlevel <= level[i]) {
	    i++;
	}
	for (k = i - 1; k > j; k--, j++) {
	    temp = from[k];
	    from[k] = from[j];
	    from[j] = temp;
	}
    }
}

/*
 * Finds the index of a run with level equals tlevel
 */
int findIndexOfRun(unsigned char* level , int start, int count, int tlevel)
{
    int i;
    for (i=start; i<count; i++) {
	if (tlevel == level[i]) {
	    return i;
	}
    }
    return count;
}

/*
 * Returns the bidi character type of ch.
 *
 * The data table in this function is constructed from the Unicode
 * Character Database, downloadable from unicode.org at the URL
 * 
 *     http://www.unicode.org/Public/UNIDATA/UnicodeData.txt
 * 
 * by the following fragment of Perl:

perl -ne 'split ";"; $num = hex $_[0]; $type = $_[4];' \
      -e '$fl = ($_[1] =~ /First/ ? 1 : $_[1] =~ /Last/ ? 2 : 0);' \
      -e 'if ($type eq $runtype and ($runend == $num-1 or ' \
      -e '    ($fl==2 and $pfl==1))) {$runend = $num;} else { &reset; }' \
      -e '$pfl=$fl; END { &reset }; sub reset {' \
      -e 'printf"        {0x%04x, 0x%04x, %s},\n",$runstart,$runend,$runtype' \
      -e '  if defined $runstart and $runtype ne "ON";' \
      -e '$runstart=$runend=$num; $runtype=$type;}' \
    UnicodeData.txt

 */
static const struct {
	    unsigned int first, last, type;
} lookup[] = {
        {0x0000, 0x0008, BN},
        {0x0009, 0x0009, S},
        {0x000a, 0x000a, B},
        {0x000b, 0x000b, S},
        {0x000c, 0x000c, WS},
        {0x000d, 0x000d, WS},
        {0x000e, 0x001b, BN},
        {0x001c, 0x001e, B},
        {0x001f, 0x001f, S},
        {0x0020, 0x0020, WS},
        {0x0023, 0x0025, ET},
        {0x002b, 0x002b, ES},
        {0x002c, 0x002c, CS},
        {0x002d, 0x002d, ES},
        {0x002e, 0x002f, CS},
        {0x0030, 0x0039, EN},
        {0x003a, 0x003a, CS},
        {0x0041, 0x005a, L},
        {0x0061, 0x007a, L},
        {0x007f, 0x0084, BN},
        {0x0085, 0x0085, B},
        {0x0086, 0x009f, BN},
        {0x00a0, 0x00a0, CS},
        {0x00a2, 0x00a5, ET},
        {0x00aa, 0x00aa, L},
        {0x00ad, 0x00ad, BN},
        {0x00b0, 0x00b1, ET},
        {0x00b2, 0x00b3, EN},
        {0x00b5, 0x00b5, L},
        {0x00b9, 0x00b9, EN},
        {0x00ba, 0x00ba, L},
        {0x00c0, 0x00d6, L},
        {0x00d8, 0x00f6, L},
        {0x00f8, 0x0236, L},
        {0x0250, 0x02b8, L},
        {0x02bb, 0x02c1, L},
        {0x02d0, 0x02d1, L},
        {0x02e0, 0x02e4, L},
        {0x02ee, 0x02ee, L},
        {0x0300, 0x0357, NSM},
        {0x035d, 0x036f, NSM},
        {0x037a, 0x037a, L},
        {0x0386, 0x0386, L},
        {0x0388, 0x038a, L},
        {0x038c, 0x038c, L},
        {0x038e, 0x03a1, L},
        {0x03a3, 0x03ce, L},
        {0x03d0, 0x03f5, L},
        {0x03f7, 0x03fb, L},
        {0x0400, 0x0482, L},
        {0x0483, 0x0486, NSM},
        {0x0488, 0x0489, NSM},
        {0x048a, 0x04ce, L},
        {0x04d0, 0x04f5, L},
        {0x04f8, 0x04f9, L},
        {0x0500, 0x050f, L},
        {0x0531, 0x0556, L},
        {0x0559, 0x055f, L},
        {0x0561, 0x0587, L},
        {0x0589, 0x0589, L},
        {0x0591, 0x05a1, NSM},
        {0x05a3, 0x05b9, NSM},
        {0x05bb, 0x05bd, NSM},
        {0x05be, 0x05be, R},
        {0x05bf, 0x05bf, NSM},
        {0x05c0, 0x05c0, R},
        {0x05c1, 0x05c2, NSM},
        {0x05c3, 0x05c3, R},
        {0x05c4, 0x05c4, NSM},
        {0x05d0, 0x05ea, R},
        {0x05f0, 0x05f4, R},
        {0x0600, 0x0603, AL},
        {0x060c, 0x060c, CS},
        {0x060d, 0x060d, AL},
        {0x0610, 0x0615, NSM},
        {0x061b, 0x061b, AL},
        {0x061f, 0x061f, AL},
        {0x0621, 0x063a, AL},
        {0x0640, 0x064a, AL},
        {0x064b, 0x0658, NSM},
        {0x0660, 0x0669, AN},
        {0x066a, 0x066a, ET},
        {0x066b, 0x066c, AN},
        {0x066d, 0x066f, AL},
        {0x0670, 0x0670, NSM},
        {0x0671, 0x06d5, AL},
        {0x06d6, 0x06dc, NSM},
        {0x06dd, 0x06dd, AL},
        {0x06de, 0x06e4, NSM},
        {0x06e5, 0x06e6, AL},
        {0x06e7, 0x06e8, NSM},
        {0x06ea, 0x06ed, NSM},
        {0x06ee, 0x06ef, AL},
        {0x06f0, 0x06f9, EN},
        {0x06fa, 0x070d, AL},
        {0x070f, 0x070f, BN},
        {0x0710, 0x0710, AL},
        {0x0711, 0x0711, NSM},
        {0x0712, 0x072f, AL},
        {0x0730, 0x074a, NSM},
        {0x074d, 0x074f, AL},
        {0x0780, 0x07a5, AL},
        {0x07a6, 0x07b0, NSM},
        {0x07b1, 0x07b1, AL},
        {0x0901, 0x0902, NSM},
        {0x0903, 0x0939, L},
        {0x093c, 0x093c, NSM},
        {0x093d, 0x0940, L},
        {0x0941, 0x0948, NSM},
        {0x0949, 0x094c, L},
        {0x094d, 0x094d, NSM},
        {0x0950, 0x0950, L},
        {0x0951, 0x0954, NSM},
        {0x0958, 0x0961, L},
        {0x0962, 0x0963, NSM},
        {0x0964, 0x0970, L},
        {0x0981, 0x0981, NSM},
        {0x0982, 0x0983, L},
        {0x0985, 0x098c, L},
        {0x098f, 0x0990, L},
        {0x0993, 0x09a8, L},
        {0x09aa, 0x09b0, L},
        {0x09b2, 0x09b2, L},
        {0x09b6, 0x09b9, L},
        {0x09bc, 0x09bc, NSM},
        {0x09bd, 0x09c0, L},
        {0x09c1, 0x09c4, NSM},
        {0x09c7, 0x09c8, L},
        {0x09cb, 0x09cc, L},
        {0x09cd, 0x09cd, NSM},
        {0x09d7, 0x09d7, L},
        {0x09dc, 0x09dd, L},
        {0x09df, 0x09e1, L},
        {0x09e2, 0x09e3, NSM},
        {0x09e6, 0x09f1, L},
        {0x09f2, 0x09f3, ET},
        {0x09f4, 0x09fa, L},
        {0x0a01, 0x0a02, NSM},
        {0x0a03, 0x0a03, L},
        {0x0a05, 0x0a0a, L},
        {0x0a0f, 0x0a10, L},
        {0x0a13, 0x0a28, L},
        {0x0a2a, 0x0a30, L},
        {0x0a32, 0x0a33, L},
        {0x0a35, 0x0a36, L},
        {0x0a38, 0x0a39, L},
        {0x0a3c, 0x0a3c, NSM},
        {0x0a3e, 0x0a40, L},
        {0x0a41, 0x0a42, NSM},
        {0x0a47, 0x0a48, NSM},
        {0x0a4b, 0x0a4d, NSM},
        {0x0a59, 0x0a5c, L},
        {0x0a5e, 0x0a5e, L},
        {0x0a66, 0x0a6f, L},
        {0x0a70, 0x0a71, NSM},
        {0x0a72, 0x0a74, L},
        {0x0a81, 0x0a82, NSM},
        {0x0a83, 0x0a83, L},
        {0x0a85, 0x0a8d, L},
        {0x0a8f, 0x0a91, L},
        {0x0a93, 0x0aa8, L},
        {0x0aaa, 0x0ab0, L},
        {0x0ab2, 0x0ab3, L},
        {0x0ab5, 0x0ab9, L},
        {0x0abc, 0x0abc, NSM},
        {0x0abd, 0x0ac0, L},
        {0x0ac1, 0x0ac5, NSM},
        {0x0ac7, 0x0ac8, NSM},
        {0x0ac9, 0x0ac9, L},
        {0x0acb, 0x0acc, L},
        {0x0acd, 0x0acd, NSM},
        {0x0ad0, 0x0ad0, L},
        {0x0ae0, 0x0ae1, L},
        {0x0ae2, 0x0ae3, NSM},
        {0x0ae6, 0x0aef, L},
        {0x0af1, 0x0af1, ET},
        {0x0b01, 0x0b01, NSM},
        {0x0b02, 0x0b03, L},
        {0x0b05, 0x0b0c, L},
        {0x0b0f, 0x0b10, L},
        {0x0b13, 0x0b28, L},
        {0x0b2a, 0x0b30, L},
        {0x0b32, 0x0b33, L},
        {0x0b35, 0x0b39, L},
        {0x0b3c, 0x0b3c, NSM},
        {0x0b3d, 0x0b3e, L},
        {0x0b3f, 0x0b3f, NSM},
        {0x0b40, 0x0b40, L},
        {0x0b41, 0x0b43, NSM},
        {0x0b47, 0x0b48, L},
        {0x0b4b, 0x0b4c, L},
        {0x0b4d, 0x0b4d, NSM},
        {0x0b56, 0x0b56, NSM},
        {0x0b57, 0x0b57, L},
        {0x0b5c, 0x0b5d, L},
        {0x0b5f, 0x0b61, L},
        {0x0b66, 0x0b71, L},
        {0x0b82, 0x0b82, NSM},
        {0x0b83, 0x0b83, L},
        {0x0b85, 0x0b8a, L},
        {0x0b8e, 0x0b90, L},
        {0x0b92, 0x0b95, L},
        {0x0b99, 0x0b9a, L},
        {0x0b9c, 0x0b9c, L},
        {0x0b9e, 0x0b9f, L},
        {0x0ba3, 0x0ba4, L},
        {0x0ba8, 0x0baa, L},
        {0x0bae, 0x0bb5, L},
        {0x0bb7, 0x0bb9, L},
        {0x0bbe, 0x0bbf, L},
        {0x0bc0, 0x0bc0, NSM},
        {0x0bc1, 0x0bc2, L},
        {0x0bc6, 0x0bc8, L},
        {0x0bca, 0x0bcc, L},
        {0x0bcd, 0x0bcd, NSM},
        {0x0bd7, 0x0bd7, L},
        {0x0be7, 0x0bf2, L},
        {0x0bf9, 0x0bf9, ET},
        {0x0c01, 0x0c03, L},
        {0x0c05, 0x0c0c, L},
        {0x0c0e, 0x0c10, L},
        {0x0c12, 0x0c28, L},
        {0x0c2a, 0x0c33, L},
        {0x0c35, 0x0c39, L},
        {0x0c3e, 0x0c40, NSM},
        {0x0c41, 0x0c44, L},
        {0x0c46, 0x0c48, NSM},
        {0x0c4a, 0x0c4d, NSM},
        {0x0c55, 0x0c56, NSM},
        {0x0c60, 0x0c61, L},
        {0x0c66, 0x0c6f, L},
        {0x0c82, 0x0c83, L},
        {0x0c85, 0x0c8c, L},
        {0x0c8e, 0x0c90, L},
        {0x0c92, 0x0ca8, L},
        {0x0caa, 0x0cb3, L},
        {0x0cb5, 0x0cb9, L},
        {0x0cbc, 0x0cbc, NSM},
        {0x0cbd, 0x0cc4, L},
        {0x0cc6, 0x0cc8, L},
        {0x0cca, 0x0ccb, L},
        {0x0ccc, 0x0ccd, NSM},
        {0x0cd5, 0x0cd6, L},
        {0x0cde, 0x0cde, L},
        {0x0ce0, 0x0ce1, L},
        {0x0ce6, 0x0cef, L},
        {0x0d02, 0x0d03, L},
        {0x0d05, 0x0d0c, L},
        {0x0d0e, 0x0d10, L},
        {0x0d12, 0x0d28, L},
        {0x0d2a, 0x0d39, L},
        {0x0d3e, 0x0d40, L},
        {0x0d41, 0x0d43, NSM},
        {0x0d46, 0x0d48, L},
        {0x0d4a, 0x0d4c, L},
        {0x0d4d, 0x0d4d, NSM},
        {0x0d57, 0x0d57, L},
        {0x0d60, 0x0d61, L},
        {0x0d66, 0x0d6f, L},
        {0x0d82, 0x0d83, L},
        {0x0d85, 0x0d96, L},
        {0x0d9a, 0x0db1, L},
        {0x0db3, 0x0dbb, L},
        {0x0dbd, 0x0dbd, L},
        {0x0dc0, 0x0dc6, L},
        {0x0dca, 0x0dca, NSM},
        {0x0dcf, 0x0dd1, L},
        {0x0dd2, 0x0dd4, NSM},
        {0x0dd6, 0x0dd6, NSM},
        {0x0dd8, 0x0ddf, L},
        {0x0df2, 0x0df4, L},
        {0x0e01, 0x0e30, L},
        {0x0e31, 0x0e31, NSM},
        {0x0e32, 0x0e33, L},
        {0x0e34, 0x0e3a, NSM},
        {0x0e3f, 0x0e3f, ET},
        {0x0e40, 0x0e46, L},
        {0x0e47, 0x0e4e, NSM},
        {0x0e4f, 0x0e5b, L},
        {0x0e81, 0x0e82, L},
        {0x0e84, 0x0e84, L},
        {0x0e87, 0x0e88, L},
        {0x0e8a, 0x0e8a, L},
        {0x0e8d, 0x0e8d, L},
        {0x0e94, 0x0e97, L},
        {0x0e99, 0x0e9f, L},
        {0x0ea1, 0x0ea3, L},
        {0x0ea5, 0x0ea5, L},
        {0x0ea7, 0x0ea7, L},
        {0x0eaa, 0x0eab, L},
        {0x0ead, 0x0eb0, L},
        {0x0eb1, 0x0eb1, NSM},
        {0x0eb2, 0x0eb3, L},
        {0x0eb4, 0x0eb9, NSM},
        {0x0ebb, 0x0ebc, NSM},
        {0x0ebd, 0x0ebd, L},
        {0x0ec0, 0x0ec4, L},
        {0x0ec6, 0x0ec6, L},
        {0x0ec8, 0x0ecd, NSM},
        {0x0ed0, 0x0ed9, L},
        {0x0edc, 0x0edd, L},
        {0x0f00, 0x0f17, L},
        {0x0f18, 0x0f19, NSM},
        {0x0f1a, 0x0f34, L},
        {0x0f35, 0x0f35, NSM},
        {0x0f36, 0x0f36, L},
        {0x0f37, 0x0f37, NSM},
        {0x0f38, 0x0f38, L},
        {0x0f39, 0x0f39, NSM},
        {0x0f3e, 0x0f47, L},
        {0x0f49, 0x0f6a, L},
        {0x0f71, 0x0f7e, NSM},
        {0x0f7f, 0x0f7f, L},
        {0x0f80, 0x0f84, NSM},
        {0x0f85, 0x0f85, L},
        {0x0f86, 0x0f87, NSM},
        {0x0f88, 0x0f8b, L},
        {0x0f90, 0x0f97, NSM},
        {0x0f99, 0x0fbc, NSM},
        {0x0fbe, 0x0fc5, L},
        {0x0fc6, 0x0fc6, NSM},
        {0x0fc7, 0x0fcc, L},
        {0x0fcf, 0x0fcf, L},
        {0x1000, 0x1021, L},
        {0x1023, 0x1027, L},
        {0x1029, 0x102a, L},
        {0x102c, 0x102c, L},
        {0x102d, 0x1030, NSM},
        {0x1031, 0x1031, L},
        {0x1032, 0x1032, NSM},
        {0x1036, 0x1037, NSM},
        {0x1038, 0x1038, L},
        {0x1039, 0x1039, NSM},
        {0x1040, 0x1057, L},
        {0x1058, 0x1059, NSM},
        {0x10a0, 0x10c5, L},
        {0x10d0, 0x10f8, L},
        {0x10fb, 0x10fb, L},
        {0x1100, 0x1159, L},
        {0x115f, 0x11a2, L},
        {0x11a8, 0x11f9, L},
        {0x1200, 0x1206, L},
        {0x1208, 0x1246, L},
        {0x1248, 0x1248, L},
        {0x124a, 0x124d, L},
        {0x1250, 0x1256, L},
        {0x1258, 0x1258, L},
        {0x125a, 0x125d, L},
        {0x1260, 0x1286, L},
        {0x1288, 0x1288, L},
        {0x128a, 0x128d, L},
        {0x1290, 0x12ae, L},
        {0x12b0, 0x12b0, L},
        {0x12b2, 0x12b5, L},
        {0x12b8, 0x12be, L},
        {0x12c0, 0x12c0, L},
        {0x12c2, 0x12c5, L},
        {0x12c8, 0x12ce, L},
        {0x12d0, 0x12d6, L},
        {0x12d8, 0x12ee, L},
        {0x12f0, 0x130e, L},
        {0x1310, 0x1310, L},
        {0x1312, 0x1315, L},
        {0x1318, 0x131e, L},
        {0x1320, 0x1346, L},
        {0x1348, 0x135a, L},
        {0x1361, 0x137c, L},
        {0x13a0, 0x13f4, L},
        {0x1401, 0x1676, L},
        {0x1680, 0x1680, WS},
        {0x1681, 0x169a, L},
        {0x16a0, 0x16f0, L},
        {0x1700, 0x170c, L},
        {0x170e, 0x1711, L},
        {0x1712, 0x1714, NSM},
        {0x1720, 0x1731, L},
        {0x1732, 0x1734, NSM},
        {0x1735, 0x1736, L},
        {0x1740, 0x1751, L},
        {0x1752, 0x1753, NSM},
        {0x1760, 0x176c, L},
        {0x176e, 0x1770, L},
        {0x1772, 0x1773, NSM},
        {0x1780, 0x17b6, L},
        {0x17b7, 0x17bd, NSM},
        {0x17be, 0x17c5, L},
        {0x17c6, 0x17c6, NSM},
        {0x17c7, 0x17c8, L},
        {0x17c9, 0x17d3, NSM},
        {0x17d4, 0x17da, L},
        {0x17db, 0x17db, ET},
        {0x17dc, 0x17dc, L},
        {0x17dd, 0x17dd, NSM},
        {0x17e0, 0x17e9, L},
        {0x180b, 0x180d, NSM},
        {0x180e, 0x180e, WS},
        {0x1810, 0x1819, L},
        {0x1820, 0x1877, L},
        {0x1880, 0x18a8, L},
        {0x18a9, 0x18a9, NSM},
        {0x1900, 0x191c, L},
        {0x1920, 0x1922, NSM},
        {0x1923, 0x1926, L},
        {0x1927, 0x192b, NSM},
        {0x1930, 0x1931, L},
        {0x1932, 0x1932, NSM},
        {0x1933, 0x1938, L},
        {0x1939, 0x193b, NSM},
        {0x1946, 0x196d, L},
        {0x1970, 0x1974, L},
        {0x1d00, 0x1d6b, L},
        {0x1e00, 0x1e9b, L},
        {0x1ea0, 0x1ef9, L},
        {0x1f00, 0x1f15, L},
        {0x1f18, 0x1f1d, L},
        {0x1f20, 0x1f45, L},
        {0x1f48, 0x1f4d, L},
        {0x1f50, 0x1f57, L},
        {0x1f59, 0x1f59, L},
        {0x1f5b, 0x1f5b, L},
        {0x1f5d, 0x1f5d, L},
        {0x1f5f, 0x1f7d, L},
        {0x1f80, 0x1fb4, L},
        {0x1fb6, 0x1fbc, L},
        {0x1fbe, 0x1fbe, L},
        {0x1fc2, 0x1fc4, L},
        {0x1fc6, 0x1fcc, L},
        {0x1fd0, 0x1fd3, L},
        {0x1fd6, 0x1fdb, L},
        {0x1fe0, 0x1fec, L},
        {0x1ff2, 0x1ff4, L},
        {0x1ff6, 0x1ffc, L},
        {0x2000, 0x200a, WS},
        {0x200b, 0x200d, BN},
        {0x200e, 0x200e, L},
        {0x200f, 0x200f, R},
        {0x2028, 0x2028, WS},
        {0x2029, 0x2029, B},
        {0x202a, 0x202a, LRE},
        {0x202b, 0x202b, RLE},
        {0x202c, 0x202c, PDF},
        {0x202d, 0x202d, LRO},
        {0x202e, 0x202e, RLO},
        {0x202f, 0x202f, WS},
        {0x2030, 0x2034, ET},
        {0x2044, 0x2044, CS},
        {0x205f, 0x205f, WS},
        {0x2060, 0x2063, BN},
        {0x206a, 0x206f, BN},
        {0x2070, 0x2070, EN},
        {0x2071, 0x2071, L},
        {0x2074, 0x2079, EN},
        {0x207a, 0x207b, ET},
        {0x207f, 0x207f, L},
        {0x2080, 0x2089, EN},
        {0x208a, 0x208b, ET},
        {0x20a0, 0x20b1, ET},
        {0x20d0, 0x20ea, NSM},
        {0x2102, 0x2102, L},
        {0x2107, 0x2107, L},
        {0x210a, 0x2113, L},
        {0x2115, 0x2115, L},
        {0x2119, 0x211d, L},
        {0x2124, 0x2124, L},
        {0x2126, 0x2126, L},
        {0x2128, 0x2128, L},
        {0x212a, 0x212d, L},
        {0x212e, 0x212e, ET},
        {0x212f, 0x2131, L},
        {0x2133, 0x2139, L},
        {0x213d, 0x213f, L},
        {0x2145, 0x2149, L},
        {0x2160, 0x2183, L},
        {0x2212, 0x2213, ET},
        {0x2336, 0x237a, L},
        {0x2395, 0x2395, L},
        {0x2488, 0x249b, EN},
        {0x249c, 0x24e9, L},
        {0x2800, 0x28ff, L},
        {0x3000, 0x3000, WS},
        {0x3005, 0x3007, L},
        {0x3021, 0x3029, L},
        {0x302a, 0x302f, NSM},
        {0x3031, 0x3035, L},
        {0x3038, 0x303c, L},
        {0x3041, 0x3096, L},
        {0x3099, 0x309a, NSM},
        {0x309d, 0x309f, L},
        {0x30a1, 0x30fa, L},
        {0x30fc, 0x30ff, L},
        {0x3105, 0x312c, L},
        {0x3131, 0x318e, L},
        {0x3190, 0x31b7, L},
        {0x31f0, 0x321c, L},
        {0x3220, 0x3243, L},
        {0x3260, 0x327b, L},
        {0x327f, 0x32b0, L},
        {0x32c0, 0x32cb, L},
        {0x32d0, 0x32fe, L},
        {0x3300, 0x3376, L},
        {0x337b, 0x33dd, L},
        {0x33e0, 0x33fe, L},
        {0x3400, 0x4db5, L},
        {0x4e00, 0x9fa5, L},
        {0xa000, 0xa48c, L},
        {0xac00, 0xd7a3, L},
        {0xd800, 0xfa2d, L},
        {0xfa30, 0xfa6a, L},
        {0xfb00, 0xfb06, L},
        {0xfb13, 0xfb17, L},
        {0xfb1d, 0xfb1d, R},
        {0xfb1e, 0xfb1e, NSM},
        {0xfb1f, 0xfb28, R},
        {0xfb29, 0xfb29, ET},
        {0xfb2a, 0xfb36, R},
        {0xfb38, 0xfb3c, R},
        {0xfb3e, 0xfb3e, R},
        {0xfb40, 0xfb41, R},
        {0xfb43, 0xfb44, R},
        {0xfb46, 0xfb4f, R},
        {0xfb50, 0xfbb1, AL},
        {0xfbd3, 0xfd3d, AL},
        {0xfd50, 0xfd8f, AL},
        {0xfd92, 0xfdc7, AL},
        {0xfdf0, 0xfdfc, AL},
        {0xfe00, 0xfe0f, NSM},
        {0xfe20, 0xfe23, NSM},
        {0xfe50, 0xfe50, CS},
        {0xfe52, 0xfe52, CS},
        {0xfe55, 0xfe55, CS},
        {0xfe5f, 0xfe5f, ET},
        {0xfe62, 0xfe63, ET},
        {0xfe69, 0xfe6a, ET},
        {0xfe70, 0xfe74, AL},
        {0xfe76, 0xfefc, AL},
        {0xfeff, 0xfeff, BN},
        {0xff03, 0xff05, ET},
        {0xff0b, 0xff0b, ET},
        {0xff0c, 0xff0c, CS},
        {0xff0d, 0xff0d, ET},
        {0xff0e, 0xff0e, CS},
        {0xff0f, 0xff0f, ES},
        {0xff10, 0xff19, EN},
        {0xff1a, 0xff1a, CS},
        {0xff21, 0xff3a, L},
        {0xff41, 0xff5a, L},
        {0xff66, 0xffbe, L},
        {0xffc2, 0xffc7, L},
        {0xffca, 0xffcf, L},
        {0xffd2, 0xffd7, L},
        {0xffda, 0xffdc, L},
        {0xffe0, 0xffe1, ET},
        {0xffe5, 0xffe6, ET},
        {0x10000, 0x1005d, L},
        {0x10080, 0x1013f, L},
        {0x10300, 0x1034a, L},
        {0x10380, 0x1039f, L},
        {0x10400, 0x1049d, L},
        {0x104a0, 0x104a9, L},
        {0x10800, 0x1083f, R},
        {0x1d000, 0x1d166, L},
        {0x1d167, 0x1d169, NSM},
        {0x1d16a, 0x1d172, L},
        {0x1d173, 0x1d17a, BN},
        {0x1d17b, 0x1d182, NSM},
        {0x1d183, 0x1d184, L},
        {0x1d185, 0x1d18b, NSM},
        {0x1d18c, 0x1d1a9, L},
        {0x1d1aa, 0x1d1ad, NSM},
        {0x1d1ae, 0x1d1dd, L},
        {0x1d400, 0x1d7ff, L},
        {0x20000, 0x2a6d6, L},
        {0x2f800, 0x2faff, L}
};

unsigned char getType(int ch)
{
    int i, j, k;

    i = -1;
    j = (sizeof(lookup) / sizeof(lookup[0]));

    while (j - i > 1) {
	    k = (i + j) / 2;
	    if (ch < lookup[k].first)
	        j = k;
	    else if (ch > lookup[k].last)
	        i = k;
	    else
	        return lookup[k].type;
    }

    /*
     * If we reach here, the character was not in any of the
     * intervals listed in the lookup table. This means we return
     * ON (`Other Neutrals'). This is the appropriate code for any
     * character genuinely not listed in the Unicode table, and
     * also the table above has deliberately left out any
     * characters _explicitly_ listed as ON (to save space!).
     */
    return ON;
}

/*
 * Function exported to front ends to allow them to identify
 * bidi-active characters (in case, for example, the platform's
 * text display function can't conveniently be prevented from doing
 * its own bidi and so special treatment is required for characters
 * that would cause the bidi algorithm to activate).
 * 
 * This function is passed a single Unicode code point, and returns
 * nonzero if the presence of this code point can possibly cause
 * the bidi algorithm to do any reordering. Thus, any string
 * composed entirely of characters for which is_rtl() returns zero
 * should be safe to pass to a bidi-active platform display
 * function without fear.
 * 
 * (is_rtl() must therefore also return true for any character
 * which would be affected by Arabic shaping, but this isn't
 * important because all such characters are right-to-left so it
 * would have flagged them anyway.)
 */
unsigned int is_rtl(unsigned int c)
{
    /*
     * After careful reading of the Unicode bidi algorithm (URL as
     * given at the top of this file) I believe that the only
     * character classes which can possibly cause trouble are R,
     * AL, RLE and RLO. I think that any string containing no
     * character in any of those classes will be displayed
     * uniformly left-to-right by the Unicode bidi algorithm.
     */
    const int mask = (1<<R) | (1<<AL) | (1<<RLE) | (1<<RLO);
    return mask & (1 << (getType(c)));
}

/*
 * The most significant 2 bits of each level are used to store
 * Override status of each character
 * This function sets the override bits of level according
 * to the value in override, and reurns the new byte.
 */
unsigned char setOverrideBits(unsigned char level, unsigned char override)
{
    if (override == ON)
	return level;
    else if (override == R)
	return level | OISR;
    else if (override == L)
	return level | OISL;
    return level;
}

/*
 * Find the most recent run of the same value in `level', and
 * return the value _before_ it. Used to process U+202C POP
 * DIRECTIONAL FORMATTING.
 */
int getPreviousLevel(unsigned char* level, int from)
{
    if (from > 0) {
        unsigned char current = level[--from];
        while (from >= 0 && level[from] == current)
            from--;
        if (from >= 0)
            return level[from];
        return -1;
    } else {
        return -1;
    }
}

/*
 * The Main Bidi Function, and the only function that should
 * be used by the outside world.
 *
 * line: a buffer of size count containing text to apply
 * the Bidirectional algorithm to.
 */

int do_bidi(bidi_char *line, unsigned int count)
{
    unsigned char* types;
    unsigned char* levels;
    unsigned char paragraphLevel;
    unsigned char currentEmbedding;
    unsigned char currentOverride;
    unsigned char tempType;
    int i, j, yes, bover;

    /* Check the presence of R or AL types as optimization */
    yes = 0;
    for (i=0; i<count; i++) {
	    int type = getType(line[i].wc);
	    if (type == R || type == AL) {
	            yes = 1;
	            break;
	    }
    }
    if (yes == 0)
	    return L;

    /* Initialize types, levels */
    types = malloc(count);
    levels = malloc(count);

    /* Rule (P1)  NOT IMPLEMENTED
     * P1. Split the text into separate paragraphs. A paragraph separator is
     * kept with the previous paragraph. Within each paragraph, apply all the
     * other rules of this algorithm.
     */

    /* Rule (P2), (P3)
     * P2. In each paragraph, find the first character of type L, AL, or R.
     * P3. If a character is found in P2 and it is of type AL or R, then set
     * the paragraph embedding level to one; otherwise, set it to zero.
     */
    paragraphLevel = 0;
    for (i=0; i<count ; i++) {
	    int type = getType(line[i].wc);
	    if (type == R || type == AL) {
	        paragraphLevel = 1;
	        break;
	    } else if (type == L) {
	        break;
        }
    }

	/* Replace tab with base direction, i.e. make tab appear as
	 * 'strong left' if the base direction is left-to-right and
	 * 'strong right' if base direction is right-to-left.  This
	 * allows Layout to implicitly treat tabs as 'segment separators'.
	 */
    for (i=0; i<count ; i++) {
	    if (text[i]=='\t') {
		    types[i] = (paragraphLevel ? BDI_R : BDI_L);
	    }
    }

    /* Rule (X1)
     * X1. Begin by setting the current embedding level to the paragraph
     * embedding level. Set the directional override status to neutral.
     */
    currentEmbedding = paragraphLevel;
    currentOverride = ON;

    /* Rule (X2), (X3), (X4), (X5), (X6), (X7), (X8)
     * X2. With each RLE, compute the least greater odd embedding level.
     * X3. With each LRE, compute the least greater even embedding level.
     * X4. With each RLO, compute the least greater odd embedding level.
     * X5. With each LRO, compute the least greater even embedding level.
     * X6. For all types besides RLE, LRE, RLO, LRO, and PDF:
     *		a. Set the level of the current character to the current
     *		    embedding level.
     *		b.  Whenever the directional override status is not neutral,
     *               reset the current character type to the directional
     *               override status.
     * X7. With each PDF, determine the matching embedding or override code.
     * If there was a valid matching code, restore (pop) the last
     * remembered (pushed) embedding level and directional override.
     * X8. All explicit directional embeddings and overrides are completely
     * terminated at the end of each paragraph. Paragraph separators are not
     * included in the embedding. (Useless here) NOT IMPLEMENTED
     */
    bover = 0;
    for (i=0; i<count; i++) {
	    tempType = getType(line[i].wc);
	    switch (tempType) {
	        case RLE:
	            currentEmbedding = levels[i] = leastGreaterOdd(currentEmbedding);
	            levels[i] = setOverrideBits(levels[i], currentOverride);
	            currentOverride = ON;
	            break;
	        case LRE:
	            currentEmbedding = levels[i] = leastGreaterEven(currentEmbedding);
	            levels[i] = setOverrideBits(levels[i], currentOverride);
	            currentOverride = ON;
	            break;

	        case RLO:
	            currentEmbedding = levels[i] = leastGreaterOdd(currentEmbedding);
	            tempType = currentOverride = R;
	            bover = 1;
	            break;

	        case LRO:
	            currentEmbedding = levels[i] = leastGreaterEven(currentEmbedding);
	            tempType = currentOverride = L;
	            bover = 1;
	            break;

	        case PDF:
            {
                int prevlevel = getPreviousLevel(levels, i);

                if (prevlevel == -1) {
                    currentEmbedding = paragraphLevel;
                    currentOverride = ON;
                } else {
                    currentOverride = currentEmbedding & OMASK;
                    currentEmbedding = currentEmbedding & ~OMASK;
                }
            }
	        levels[i] = currentEmbedding;
	        break;

	        /* Whitespace is treated as neutral for now */
	        case WS:
	        case S:
	            levels[i] = currentEmbedding;
	            tempType = ON;
	            if (currentOverride != ON)
		            tempType = currentOverride;
	            break;

	        default:
	            levels[i] = currentEmbedding;
	            if (currentOverride != ON)
		            tempType = currentOverride;
	            break;

	    }
	    types[i] = tempType;
    }

	/* Replace tab with base direction, i.e. make tab appear as
	 * 'strong left' if the base direction is left-to-right and
	 * 'strong right' if base direction is right-to-left.  This
	 * allows Layout to implicitly treat tabs as 'segment separators'.
	 */
    for (i=0; i<count; i++) {
	    if (text[i]=='\t') {
		    types[i] = (paragraphLevel ? R : L);
	    }
    }

    /* this clears out all overrides, so we can use levels safely... */
    /* checks bover first */
    if (bover)
	for (i=0; i<count; i++)
	    levels[i] = levels[i] & LMASK;

    /* Rule (X9)
     * X9. Remove all RLE, LRE, RLO, LRO, PDF, and BN codes.
     * Here, they're converted to BN.
     */
    for (i=0; i<count; i++) {
	switch (types[i]) {
	  case RLE:
	  case LRE:
	  case RLO:
	  case LRO:
	  case PDF:
	    types[i] = BN;
	    break;
	}
    }

    /* Rule (W1)
     * W1. Examine each non-spacing mark (NSM) in the level run, and change
     * the type of the NSM to the type of the previous character. If the NSM
     * is at the start of the level run, it will get the type of sor.
     */
    if (types[0] == NSM)
	types[0] = paragraphLevel;

    for (i=1; i<count; i++) {
	if (types[i] == NSM)
	    types[i] = types[i-1];
	/* Is this a safe assumption?
	 * I assumed the previous, IS a character.
	 */
    }

    /* Rule (W3)
     * W3. Change all ALs to R.
     *
     * Optimization: on Rule Xn, we might set a flag on AL type
     * to prevent this loop in L R lines only...
     */
    for (i=0; i<count; i++) {
	if (types[i] == AL)
	    types[i] = R;
    }

    /* Rule (W4)
     * W4. A single European separator between two European numbers changes
     * to a European number. A single common separator between two numbers
     * of the same type changes to that type.
     */
    for (i=1; i<(count-1); i++) {
	if (types[i] == ES) {
	    if (types[i-1] == EN && types[i+1] == EN)
		types[i] = EN;
	} else if (types[i] == CS) {
            if (types[i-1] == EN && types[i+1] == EN)
                types[i] = EN;
            else if (types[i-1] == AN && types[i+1] == AN)
                types[i] = AN;
        }
    }

    /* Rule (W5)
     * W5. A sequence of European terminators adjacent to European numbers
     * changes to all European numbers.
     *
     * Optimization: lots here... else ifs need rearrangement
     */
    for (i=0; i<count; i++) {
	if (types[i] == ET) {
	    if (i > 0 && types[i-1] == EN) {
		types[i] = EN;
		continue;
	    } else if (i < count-1 && types[i+1] == EN) {
                types[i] = EN;
                continue;
            } else if (i < count-1 && types[i+1] == ET) {
                j=i;
                while (j <count && types[j] == ET) {
                    j++;
                }
                if (types[j] == EN)
                    types[i] = EN;
            }
	}
    }

    /* Rule (W6)
     * W6. Otherwise, separators and terminators change to Other Neutral:
     */
    for (i=0; i<count; i++) {
	switch (types[i]) {
	  case ES:
	  case ET:
	  case CS:
	    types[i] = ON;
	    break;
	}
    }

    /* Rule (N1)
     * N1. A sequence of neutrals takes the direction of the surrounding
     * strong text if the text on both sides has the same direction. European
     * and Arabic numbers are treated as though they were R.
     */
    if (count >= 2 && types[0] == ON) {
	if ((types[1] == R) || (types[1] == EN) || (types[1] == AN))
	    types[0] = R;
	else if (types[1] == L)
	    types[0] = L;
    }
    for (i=1; i<(count-1); i++) {
	if (types[i] == ON) {
	    if (types[i-1] == L) {
		j=i;
		while (j<(count-1) && types[j] == ON) {
		    j++;
		}
		if (types[j] == L) {
		    while (i<j) {
			types[i] = L;
			i++;
		    }
		}

	    } else if ((types[i-1] == R)  ||
                       (types[i-1] == EN) ||
                       (types[i-1] == AN)) {
                j=i;
                while (j<(count-1) && types[j] == ON) {
                    j++;
                }
                if ((types[j] == R)  ||
                    (types[j] == EN) ||
                    (types[j] == AN)) {
                    while (i<j) {
                        types[i] = R;
                        i++;
                    }
                }
            }
	}
    }
    if (count >= 2 && types[count-1] == ON) {
	if (types[count-2] == R || types[count-2] == EN || types[count-2] == AN)
	    types[count-1] = R;
	else if (types[count-2] == L)
	    types[count-1] = L;
    }

    /* Rule (N2)
     * N2. Any remaining neutrals take the embedding direction.
     */
    for (i=0; i<count; i++) {
	if (types[i] == ON) {
	    if ((levels[i] % 2) == 0)
		types[i] = L;
	    else
		types[i] = R;
	}
    }

    /* Rule (I1)
     * I1. For all characters with an even (left-to-right) embedding
     * direction, those of type R go up one level and those of type AN or
     * EN go up two levels.
     */
    for (i=0; i<count; i++) {
	if ((levels[i] % 2) == 0) {
	    if (types[i] == R)
		levels[i] += 1;
	    else if (types[i] == AN || types[i] == EN)
		levels[i] += 2;
	}
    }

    /* Rule (I2)
     * I2. For all characters with an odd (right-to-left) embedding direction,
     * those of type L, EN or AN go up one level.
     */
    for (i=0; i<count; i++) {
	if ((levels[i] % 2) == 1) {
	    if (types[i] == L || types[i] == EN || types[i] == AN)
		levels[i] += 1;
	}
    }
    free(types);

    /* Rule (L1)
     * L1. On each line, reset the embedding level of the following characters
     * to the paragraph embedding level:
     *		(1)segment separators, (2)paragraph separators,
     *           (3)any sequence of whitespace characters preceding
     *           a segment separator or paragraph separator,
     *           (4)and any sequence of white space characters
     *           at the end of the line.
     * The types of characters used here are the original types, not those
     * modified by the previous phase.
     */
    j=count-1;
    while (j>0 && (getType(line[j].wc) == WS)) {
	    j--;
    }
    if (j < (count-1)) {
	for (j++; j<count; j++)
	    levels[j] = paragraphLevel;
    }
    for (i=0; i<count; i++) {
	tempType = getType(line[i].wc);
	if (tempType == WS) {
	    j=i;
	    while (j<count && (getType(line[j].wc) == WS)) {
		j++;
	    }
	    if (j==count || getType(line[j].wc) == B || getType(line[j].wc) == S) {
		for (j--; j>=i ; j--) {
		    levels[j] = paragraphLevel;
		}
	    }
	} else if (tempType == B || tempType == S) {
            levels[i] = paragraphLevel;
        }
    }

    /* Rule (L4)
     * L4. A character that possesses the mirrored property as specified by
     * Section 4.7, Mirrored, must be depicted by a mirrored glyph if the
     * resolved directionality of that character is R.
     */
    /* Note: this is implemented before L2 for efficiency */
    for (i=0; i<count; i++)
	if ((levels[i] % 2) == 1)
	    line[i].wc = Bidi_mirrorChar(line[i].wc);

    /* Rule (L2)
     * L2. From the highest level found in the text to the lowest odd level on
     * each line, including intermediate levels not actually present in the
     * text, reverse any contiguous sequence of characters that are at that
     * level or higher
     */
    /* we flip the character string and leave the level array */
    i=0;
    tempType = levels[0];
    while (i < count) {
	if (levels[i] > tempType)
	    tempType = levels[i];
	i++;
    }
    /* maximum level in tempType. */
    while (tempType > 0) {     /* loop from highest level to the least odd, */
        /* which i assume is 1 */
	flipThisRun(line, levels, tempType, count);
	tempType--;
    }

    /* Rule (L3) NOT IMPLEMENTED
     * L3. Combining marks applied to a right-to-left base character will at
     * this point precede their base character. If the rendering engine
     * expects them to follow the base characters in the final display
     * process, then the ordering of the marks and the base character must
     * be reversed.
     */
    free(levels);
    return R;
}

#define UNICODE_EOS                                  ((uint16_t)0)
#define UNICODE_RTL_START                            ((uint16_t)0x0590)
#define UNICODE_RTL_END                              ((uint16_t)0x07BF)
#define MIRROR_NYBBLE_SHIFT  28
#define MIRROR_NEXT_MASK    0x0FFF0000
#define MIRROR_NEXT_SHIFT    16
#define MIRROR_VALUE_MASK   0x0000FFFF
#define TRIPPLE_NYBBLE_NEG  4095

/*
 *The original BidiMirrorList bidiMirrorList[] array has been replaced by
 *uint32_t bidiOptMirrorList[] an array that represents a tree structure
 *formed from the original mirror list by the Perl script
 *create-mirrorlist.pl in Review #11191
 *
 *below is the first part of the tree:- representing the uint16_t values
 *0x0028, 0x0029, 0x003C, 0x003E, 0x005B, 0x005D
 *
 *               0
 *               |
 *           ----0----
 *          /    |    \
 *     2     3     5
 *    / \   / \   / \
 *   8   9 C   E B   D
 *
 *Every uint16_t value that shares nybble ancestors share parent nodes
 *on the tree.
 *The tree is of fixed depth 4 and each node holds the following values                                                                                                   
 *{4-bits nybble, 12-bits next, 16-bits uint16_t mirror}.
 *The tree is represented as an array of bit-masks.
 *
 *The nybble value is the value in the tree above.
 *The next Offset is the index into the array which points to the next
 *sibling for the current node (i.e. the next node horizintally to the right
 *that shares the same parent).
 *The mirror value is the code to be mapped.
 */
static const uint32_t bidiOptMirrorList[] =
{
    0x00120000, 0x0FFF0000, 0x20050000,
    0x80040029, 0x9FFF0028, 0x30080000,
    0xC007003E, 0xEFFF003C, 0x500B0000,
    0xB00A005D, 0xDFFF005B, 0x700E0000,
    0xB00D007D, 0xDFFF007B, 0xA0100000,
    0xBFFF00BB, 0xBFFF0000, 0xBFFF00AB,
    0x215A0000, 0x00200000, 0x30170000,
    0x9016203A, 0xAFFF2039, 0x401A0000,
    0x50192046, 0x6FFF2045, 0x701D0000,
    0xD01C207E, 0xEFFF207D, 0x8FFF0000,
    0xD01F208E, 0xEFFF208D, 0x20A20000,
    0x00280000, 0x8023220B, 0x9024220C,
    0xA025220D, 0xB0262208, 0xC0272209,
    0xDFFF220A, 0x102A0000, 0x5FFF29F5,
    0x302D0000, 0xC02C223D, 0xDFFF223C,
    0x402F0000, 0x3FFF22CD, 0x50340000,
    0x20312253, 0x30322252, 0x40332255,                                                                                                                                   
    0x5FFF2254, 0x603F0000, 0x40362265,
    0x50372264, 0x60382267, 0x70392266,
    0x803A2269, 0x903B2268, 0xA03C226B,
    0xB03D226A, 0xE03E226F, 0xFFFF226E,
    0x70500000, 0x00412271, 0x10422270,
    0x20432273, 0x30442272, 0x40452275,
    0x50462274, 0x60472277, 0x70482276,
    0x80492279, 0x904A2278, 0xA04B227B,
    0xB04C227A, 0xC04D227D, 0xD04E227C,
    0xE04F227F, 0xFFFF227E, 0x805E0000,
    0x00522281, 0x10532280, 0x20542283,
    0x30552282, 0x40562285, 0x50572284,
    0x60582287, 0x70592286, 0x805A2289,
    0x905B2288, 0xA05C228B, 0xB05D228A,
    0xFFFF2290, 0x90630000, 0x0060228F,
    0x10612292, 0x20622291, 0x8FFF29B8,
    0xA06A0000, 0x206522A3, 0x306622A2,
    0x60672ADE, 0x80682AE4, 0x90692AE3,
    0xBFFF2AE5, 0xB0730000, 0x006C22B1,
    0x106D22B0, 0x206E22B3, 0x306F22B2,
    0x407022B5, 0x507122B4, 0x607222B7,
    0x7FFF22B6, 0xC0790000, 0x907522CA,
    0xA07622C9, 0xB07722CC, 0xC07822CB,
    0xDFFF2243, 0xD0860000, 0x007B22D1,
    0x107C22D0, 0x607D22D7, 0x707E22D6,
    0x807F22D9, 0x908022D8, 0xA08122DB,
    0xB08222DA, 0xC08322DD, 0xD08422DC,
    0xE08522DF, 0xFFFF22DE, 0xE0950000,
    0x008822E1, 0x108922E0, 0x208A22E3,
    0x308B22E2, 0x408C22E5, 0x508D22E4,
    0x608E22E7, 0x708F22E6, 0x809022E9,
    0x909122E8, 0xA09222EB, 0xB09322EA,
    0xC09422ED, 0xDFFF22EC, 0xFFFF0000,
    0x009722F1, 0x109822F0, 0x209922FA,
    0x309A22FB, 0x409B22FC, 0x609C22FD,
    0x709D22FE, 0xA09E22F2, 0xB09F22F3,
    0xC0A022F4, 0xD0A122F6, 0xEFFF22F7,
    0x30AB0000, 0x00A80000, 0x80A52309,
    0x90A62308, 0xA0A7230B, 0xBFFF230A,
    0x2FFF0000, 0x90AA232A, 0xAFFF2329,
    0x70CC0000, 0x60B50000, 0x80AE2769,
    0x90AF2768, 0xA0B0276B, 0xB0B1276A,
    0xC0B2276D, 0xD0B3276C, 0xE0B4276F,
    0xFFFF276E, 0x70BC0000, 0x00B72771,
    0x10B82770, 0x20B92773, 0x30BA2772,
    0x40BB2775, 0x5FFF2774, 0xD0C10000,
    0x50BE27D6, 0x60BF27D5, 0xD0C027DE,
    0xEFFF27DD, 0xEFFF0000, 0x20C327E3,
    0x30C427E2, 0x40C527E5, 0x50C627E4,
    0x60C727E7, 0x70C827E6, 0x80C927E9,
    0x90CA27E8, 0xA0CB27EB, 0xBFFF27EA,
    0x90FD0000, 0x80DB0000, 0x30CF2984,
    0x40D02983, 0x50D12986, 0x60D22985,
    0x70D32988, 0x80D42987, 0x90D5298A,
    0xA0D62989, 0xB0D7298C, 0xC0D8298B,                                                                                                                                   
    0xD0D92990, 0xE0DA298F, 0xFFFF298E,
    0x90E50000, 0x00DD298D, 0x10DE2992,
    0x20DF2991, 0x30E02994, 0x40E12993,
    0x50E22996, 0x60E32995, 0x70E42998,
    0x8FFF2997, 0xB0E70000, 0x8FFF2298,
    0xC0ED0000, 0x00E929C1, 0x10EA29C0,
    0x40EB29C5, 0x50EC29C4, 0xFFFF29D0,
    0xD0F70000, 0x00EF29CF, 0x10F029D2,
    0x20F129D1, 0x40F229D5, 0x50F329D4,
    0x80F429D9, 0x90F529D8, 0xA0F629DB,
    0xBFFF29DA, 0xFFFF0000, 0x50F92215,
    0x80FA29F9, 0x90FB29F8, 0xC0FC29FD,
    0xDFFF29FC, 0xAFFF0000, 0x21030000,
    0xB1002A2C, 0xC1012A2B, 0xD1022A2E,
    0xEFFF2A2D, 0x31080000, 0x41052A35,
    0x51062A34, 0xC1072A3D, 0xDFFF2A3C,
    0x610B0000, 0x410A2A65, 0x5FFF2A64,
    0x71110000, 0x910D2A7A, 0xA10E2A79,
    0xD10F2A7E, 0xE1102A7D, 0xFFFF2A80,
    0x81190000, 0x01132A7F, 0x11142A82,
    0x21152A81, 0x31162A84, 0x41172A83,
    0xB1182A8C, 0xCFFF2A8B, 0x91260000,
    0x111B2A92, 0x211C2A91, 0x311D2A94,
    0x411E2A93, 0x511F2A96, 0x61202A95,
    0x71212A98, 0x81222A97, 0x91232A9A,
    0xA1242A99, 0xB1252A9C, 0xCFFF2A9B,
    0xA1320000, 0x11282AA2, 0x21292AA1,
    0x612A2AA7, 0x712B2AA6, 0x812C2AA9,
    0x912D2AA8, 0xA12E2AAB, 0xB12F2AAA,
    0xC1302AAD, 0xD1312AAC, 0xFFFF2AB0,
    0xB13B0000, 0x01342AAF, 0x31352AB4,
    0x41362AB3, 0xB1372ABC, 0xC1382ABB,
    0xD1392ABE, 0xE13A2ABD, 0xFFFF2AC0,
    0xC1460000, 0x013D2ABF, 0x113E2AC2,
    0x213F2AC1, 0x31402AC4, 0x41412AC3,
    0x51422AC6, 0x61432AC5, 0xD1442ACE,
    0xE1452ACD, 0xFFFF2AD0, 0xD14F0000,
    0x01482ACF, 0x11492AD2, 0x214A2AD1,
    0x314B2AD4, 0x414C2AD3, 0x514D2AD6,
    0x614E2AD5, 0xEFFF22A6, 0xE1550000,
    0x315122A9, 0x415222A8, 0x515322AB,
    0xC1542AED, 0xDFFF2AEC, 0xFFFF0000,
    0x71572AF8, 0x81582AF7, 0x91592AFA,
    0xAFFF2AF9, 0x31700000, 0x0FFF0000,
    0x01650000, 0x815E3009, 0x915F3008,
    0xA160300B, 0xB161300A, 0xC162300D,
    0xD163300C, 0xE164300F, 0xFFFF300E,
    0x1FFF0000, 0x01673011, 0x11683010,
    0x41693015, 0x516A3014, 0x616B3017,
    0x716C3016, 0x816D3019, 0x916E3018,
    0xA16F301B, 0xBFFF301A, 0xFFFF0000,
    0xFFFF0000, 0x01750000, 0x8174FF09,
    0x9FFFFF08, 0x11780000, 0xC177FF1E,
    0xEFFFFF1C, 0x317B0000, 0xB17AFF3D,
    0xDFFFFF3B, 0x517F0000, 0xB17DFF5D,                                                                                                                                   
    0xD17EFF5B, 0xFFFFFF60, 0x6FFF0000,
    0x0181FF5F, 0x2182FF63, 0x3FFFFF62
};

#define NUM_BIDI_OPTMIRRORS (sizeof(bidiOptMirrorList)/sizeof(uint32_t))

/** The Bidi_mirrorChar function tranverses the tree looking for nybble
  * matches.
  *
  * EXAMPLE:
  * If uint16_t value u=0x005D has been passed to Bidi_mirrorChar for testing,
  * it is separated into nybbles 0,0,5,D
  * The first node in the tree (index 0 in the array) is the starting point,
  * and luckily the first 4-bits hold a 0, which matches. We can therefore
  * drop down to the child to check that. The child of a node is stored in
  * the next array position, in this case index 1.
  * The nybble of index 1 is a 0 and matches the second nybble we are
  * searching for. Because of the match, the next child index 2 is moved to.
  * Now the nybble at index 2 of the tree is a 2 and doesn't match the 5 we
  * are searching for. Bits 5-16 of the array represent the offset to the
  * next sibling. In this case the offset is 5, when we move to this array
  * value we see that it holds a 3 nybble, again this doesn't match, so we
  * get the next offset, which is 8, and move to that sibling.
  * The sibling holds the nybble 5, which is the next one we want; so
  * increment the index to its child. The child doesn't match and holds a B
  * instead of a D. So we move to the next sibling ... which matches!!
  * This node holds the mirror character 0x005B in bits 17-32.
  */
uint32_t Bidi_mirrorChar(uint32_t u)
{
    uint8_t i, n, nybble;

    if((u < UNICODE_RTL_START) || (u > UNICODE_RTL_END)) {
        /* uint16_t value lies outside of RTL character range
         * and could therefore be a mirrorable charcter
         */
        uint32_t index = 0, uCpy = u;
        /* look through each nybble of the uint16_t and search in tree
         * for matches
         */
        for(i = 0; i < 4; i++) {
            n = (uint8_t)(uCpy >> 12); /* the current nybble of the target */
            /* the nybble of the first child node */
            nybble = (uint8_t)(bidiOptMirrorList[index] >> MIRROR_NYBBLE_SHIFT);

            /* search through node siblings for the matching nybble */
            while(nybble != n) {
                /* index to next sibling */
                index = (bidiOptMirrorList[index] & MIRROR_NEXT_MASK) >> MIRROR_NEXT_SHIFT;
                if(index == TRIPPLE_NYBBLE_NEG) { /* 12-bit next value is -1 */
                    /* no more siblings */
                    return 0;
                }
                nybble = (uint8_t)(bidiOptMirrorList[index] >> MIRROR_NYBBLE_SHIFT);
            }
            uCpy <<= 4;
            index++;
        }
        return (uint16_t)(bidiOptMirrorList[--index] & MIRROR_VALUE_MASK);
    }
    return u;
}

/** Subsitutes characters with their mirrored equivalents, e.g. swaps
  * '(' with ')'. Implements rule L4 of the uint16_t Bidirectional
  * algorithm. Mirrored characters are found using a binary search through
  * the lookup table bidiOptMirrorList.
  */
void Bidi_mirror(uint32_t *text, uint32_t len, const uint8_t *poslevel)
{
    uint32_t i, out;

    for (i = 0; i < len; i ++) {
        if (!(poslevel[i] & 1))
            continue;
        out = Bidi_mirrorChar((uint16_t)text[i]);
        if(out != 0)
            text[i] = out;
    }
}

#ifdef TEST_GETTYPE

#include <stdio.h>
#include <assert.h>

int main(int argc, char **argv)
{
    static const struct { int type; char *name; } typetoname[] = {
#define TYPETONAME(X) { X , #X }
	TYPETONAME(L),
	TYPETONAME(LRE),
	TYPETONAME(LRO),
	TYPETONAME(R),
	TYPETONAME(AL),
	TYPETONAME(RLE),
	TYPETONAME(RLO),
	TYPETONAME(PDF),
	TYPETONAME(EN),
	TYPETONAME(ES),
	TYPETONAME(ET),
	TYPETONAME(AN),
	TYPETONAME(CS),
	TYPETONAME(NSM),
	TYPETONAME(BN),
	TYPETONAME(B),
	TYPETONAME(S),
	TYPETONAME(WS),
	TYPETONAME(ON),
#undef TYPETONAME
    };
    int i;

    for (i = 1; i < argc; i++) {
	unsigned long chr = strtoul(argv[i], NULL, 0);
	int type = getType(chr);
	assert(typetoname[type].type == type);
	printf("U+%04x: %s\n", chr, typetoname[type].name);
    }

    return 0;
}

#endif
