#include "mupdf/fitz.h"

fz_buffer *
fz_new_buffer(unsigned int size)
{
	fz_buffer *b;
	size = size > 1 ? size : 4096;
	b = malloc(sizeof(fz_buffer));
	memset(b, 0, sizeof(fz_buffer));
	b->refs = 1;
	b->data = malloc(size);
	b->cap = size;
	b->len = 0;
	return b;
}

fz_buffer *
fz_new_buffer_from_data(unsigned char *data, unsigned int size)
{
	fz_buffer *b;

	b = malloc(sizeof(fz_buffer));
	memset(b, 0, sizeof(fz_buffer));
	b->refs = 1;
	b->data = data;
	b->cap = size;
	b->len = size;

	return b;
}

fz_buffer *
fz_keep_buffer(fz_buffer *buf)
{
	if (buf)
		buf->refs ++;
	return buf;
}

void
fz_drop_buffer(fz_buffer *buf)
{
	if (!buf)
		return;
	if (--buf->refs == 0) {
		free(buf->data);
		free(buf);
	}
}

static inline unsigned int nearest_power(unsigned int x) {
    unsigned int ret = 2;
    __asm__ volatile("or $1, %1 \n\t"
             "bsr %1, %%ecx \n\t"
             "movl $2, %0 \n\t"
             "shl %%cl, %0 \n\t"
             :"=a"(ret) : "D"(x) : "%rcx");
    return ret;
}

void
fz_resize_buffer(fz_buffer *buf, unsigned int size)
{
	buf->data = realloc(buf->data, size);
	buf->cap = size;
	if (buf->len > buf->cap)
		buf->len = buf->cap;
}

void
fz_grow_buffer(fz_buffer *buf)
{
	unsigned int newsize = (buf->cap << 1);
	if (newsize == 0)
		newsize = 256;
	fz_resize_buffer(buf, newsize);
}

void
fz_ensure_buffer(fz_buffer *buf, unsigned int min)
{
	unsigned int newsize = ((min > buf->cap) ? min : buf->cap);
    newsize = nearest_power(newsize);
	fz_resize_buffer(buf, newsize);
}

void
fz_trim_buffer(fz_buffer *buf)
{
#if 0
	if (buf->cap > buf->len+1)
		fz_resize_buffer(buf, buf->len);
#endif
}

unsigned int
fz_buffer_storage(fz_buffer *buf, unsigned char **datap)
{
	if (datap)
		*datap = (buf ? buf->data : NULL);
	return (buf ? buf->len : 0);
}

void
fz_buffer_cat(fz_buffer *buf, fz_buffer *extra)
{
	if (buf->cap - buf->len < extra->len) {
		buf->data = realloc(buf->data, buf->len + extra->len);
		buf->cap = buf->len + extra->len;
	}

	memcpy(buf->data + buf->len, extra->data, extra->len);
	buf->len += extra->len;
}

void fz_write_buffer(fz_buffer *buf, const void *data, unsigned int len)
{
	if (buf->len + len >= buf->cap)
		fz_ensure_buffer(buf, buf->len + len);
	memcpy(buf->data + buf->len, data, len);
	buf->len += len;
}

void fz_write_buffer_byte(fz_buffer *buf, int val)
{
	if (buf->len >= buf->cap)
		fz_grow_buffer(buf);
	buf->data[buf->len++] = val;
}

void fz_write_buffer_int32le(fz_buffer *buf, unsigned int val)
{
	if (buf->len + 4 >= buf->cap)
		fz_grow_buffer(buf);
	memcpy(buf->data + buf->len, &val, 4);
	buf->len += 4;
}

void fz_write_buffer_rune(fz_buffer *buf, unsigned int c)
{
	char data[10];
	unsigned int len = fz_runetochar(data, c);
	if (buf->len + len >= buf->cap)
		fz_grow_buffer(buf);
	memcpy(buf->data + buf->len, data, len);
	buf->len += len;
}

void fz_buffer_itoa10(fz_buffer *buf, unsigned int uval)
{
	char data[25];
	unsigned int i = 0;
	while (uval) {
		data[i++] = ((uval % 10) + 0x30);
		uval /= 10;
	}
	if (buf->len + i >= buf->cap)
		fz_grow_buffer(buf);
    while (i--) buf->data[buf->len++] = data[i];
}

int fz_buffer_vprintf(fz_buffer *buffer, const char *fmt, va_list old_args)
{
    va_list args;
	int len, slack = buffer->cap - buffer->len;
	va_copy(args, old_args);
	len = fz_vsnprintf(NULL, slack, fmt, args);
    va_end(args);
	/* Grow the buffer and retry */
    if (!buffer->cap) buffer->cap = 4096;
	while (buffer->len + len >= buffer->cap) {
        buffer->cap <<= 1;
        printf("len: %u, buf: cap: %u, len: %u\n", len, buffer->cap, buffer->len);
    }
	fz_resize_buffer(buffer, buffer->cap);
	va_copy(args, old_args);
	len = fz_vsnprintf((char *)buffer->data + buffer->len, len, fmt, args);
    va_end(args);
	buffer->len += len;
	return len;
}

int fz_buffer_printf(fz_buffer *buffer, const char *fmt, ...)
{
	int ret;
	va_list args;
	va_start(args, fmt);
	ret = fz_buffer_vprintf(buffer, fmt, args);
	va_end(args);
	return ret;
}

#ifdef TEST_BUFFER_WRITE

#define TEST_LEN 1024

void
fz_test_buffer_write(void)
{
	fz_buffer *master = fz_new_buffer(TEST_LEN);
	fz_buffer *copy = fz_new_buffer(TEST_LEN);
	fz_stream *stm;
	int i, j, k;

	/* Make us a dummy buffer */
	for (i = 0; i < TEST_LEN; i++) { 
		master->data[i] = rand();
	}
	master->len = TEST_LEN;

	/* Now copy that buffer several times, checking it for validity */
	stm = fz_open_buffer(master);
	for (i = 0; i < 256; i++)
	{
		memset(copy->data, i, TEST_LEN);
		copy->len = 0;
		j = TEST_LEN * 8;
		do
		{
			k = (rand() & 31)+1;
			if (k > j)
				k = j;
			fz_write_buffer_bits(copy, fz_read_bits(stm, k), k);
			j -= k;
		}
		while (j);

		if (memcmp(copy->data, master->data, TEST_LEN) != 0)
			fprintf(stderr, "Copied buffer is different!\n");
		fz_seek(stm, 0, 0);
	}
	fz_close(stm);
	fz_drop_buffer(master);
	fz_drop_buffer(copy);
}
#endif
