/*! \file
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.  This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 */
#include "mupdf/fitz.h"

void fz_md5_stringify(char *buf, unsigned char *digest)
{
	unsigned int i;
    char *p = buf;
    static const char hex[]="0123456789abcdef";

    for (i = 0; i < 16; i++) {
        p[i + i] = hex[digest[i] >> 4];
        p[i + i + 1] = hex[digest[i] & 0x0f];
    }
    p[i + i] = '\0';
}

/*! \file
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.  This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 */
#define ROTATE(a,n) ((a)<<(n)|(a)>>(32-n))

/*! The four core functions - F1 is optimized somewhat (F1 is the same as F1) */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (y ^ (x ^  z))
#define F4(x, y, z) (y ^ (x | ~z))

/*! This is the central step in the MD5 algorithm. */
#define MD5STEP(f,w,x,y,z,in,s) \
     (w += f(x,y,z) + in, w = ROTATE(w, s) + x)

static const uint32_t md5constants[64] =
{
    0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee, 0xf57c0faf, 0x4787c62a, 0xa8304613, 0xfd469501,
    0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be, 0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821,
    0xf61e2562, 0xc040b340, 0x265e5a51, 0xe9b6c7aa, 0xd62f105d, 0x02441453, 0xd8a1e681, 0xe7d3fbc8,
    0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed, 0xa9e3e905, 0xfcefa3f8, 0x676f02d9, 0x8d2a4c8a,
    0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c, 0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70,
    0x289b7ec6, 0xeaa127fa, 0xd4ef3085, 0x04881d05, 0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665,
    0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039, 0x655b59c3, 0x8f0ccc92, 0xffeff47d, 0x85845dd1,
    0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1, 0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391,
};

/*!
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void
transform(uint32_t buf[4], uint32_t const in[16]) {
    register uint32_t a, b, c, d, i;
    const uint32_t *p = md5constants;

    a = buf[0];
    b = buf[1];
    c = buf[2];
    d = buf[3];

    for (i = 0; i < 4; i++) {
        MD5STEP(F1, a, b, c, d, in[4*i + 0] + *p++, 7);
        MD5STEP(F1, d, a, b, c, in[4*i + 1] + *p++, 12);
        MD5STEP(F1, c, d, a, b, in[4*i + 2] + *p++, 17);
        MD5STEP(F1, b, c, d, a, in[4*i + 3] + *p++, 22);
    }

    for (i = 0; i < 4; i++) {
        MD5STEP(F2, a, b, c, d, in[(4*i +  1)   ] + *p++,  5);
        MD5STEP(F2, d, a, b, c, in[(4*i +  6)&15] + *p++,  9);
        MD5STEP(F2, c, d, a, b, in[(4*i + 11)&15] + *p++, 14);
        MD5STEP(F2, b, c, d, a, in[(4*i +  0)   ] + *p++, 20);
    }

    for (i = 0; i < 4; i++) {
        MD5STEP(F3, a, b, c, d, in[( 5 - 4*i)&15] + *p++, 4);
        MD5STEP(F3, d, a, b, c, in[( 8 - 4*i)&15] + *p++, 11);
        MD5STEP(F3, c, d, a, b, in[(11 - 4*i)&15] + *p++, 16);
        MD5STEP(F3, b, c, d, a, in[(14 - 4*i)&15] + *p++, 23);
    }

    for (i = 0; i < 4; i++) {
        MD5STEP(F4, a, b, c, d, in[( 0 - 4*i)&15] + *p++, 6);
        MD5STEP(F4, d, a, b, c, in[( 7 - 4*i)&15] + *p++, 10);
        MD5STEP(F4, c, d, a, b, in[(14 - 4*i)&15] + *p++, 15);
        MD5STEP(F4, b, c, d, a, in[( 5 - 4*i)&15] + *p++, 21);
    }

    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
}

void fz_md5_init(fz_md5 *ctx)
{
    ctx->state[0] = 0x67452301;
    ctx->state[1] = 0xefcdab89;
    ctx->state[2] = 0x98badcfe;
    ctx->state[3] = 0x10325476;
    ctx->count = 0;
}

void fz_md5_update(fz_md5 *ctx, const unsigned char *buf, unsigned int len)
{
	uint32_t t;

	/* Update byte count */
	t = (ctx->count & 0x3f);
	ctx->count += len;
	t = 64 - t;	/* Space available in ctx->buffer (at least 1) */

	if (t > len) {
		memcpy((unsigned char *)ctx->buffer + 64 - t, buf, len);
		return;
	}

	/* First chunk is an odd size */
	memcpy((unsigned char *)ctx->buffer + 64 - t, buf, t);
	transform(ctx->state, ctx->buffer);
	buf += t;
	len -= t;

	/* Process data in 64-byte chunks */
	while (len >= 64) {
		memcpy(ctx->buffer, buf, 64);
		transform(ctx->state, ctx->buffer);
		buf += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */
	memcpy(ctx->buffer, buf, len);
}

void fz_md5_final(fz_md5 *ctx, unsigned char digest[16])
{
    uint64_t finalcount = (ctx->count << 3);
	unsigned char *p = (unsigned char *)ctx->buffer + (ctx->count & 0x3f);
    uint32_t *digest32 = (uint32_t*)digest;

	*p = 0x80;
    ctx->count++;
    while ((ctx->count & 63) != 56)
        fz_md5_update(ctx, "", 1);
    fz_md5_update(ctx, (uint8_t *)&finalcount, 8); /* Should cause a transform() */

    digest32[0] = ctx->state[0];
    digest32[1] = ctx->state[1];
    digest32[2] = ctx->state[2];
    digest32[3] = ctx->state[3];
}

void fz_md5_sum(unsigned char *dst, const unsigned char *src, unsigned int len)
{
    fz_md5 ctx;

    fz_md5_init(&ctx);
    fz_md5_update(&ctx, src, len);
    fz_md5_final(&ctx, dst);
}

