#include "mupdf/fitz.h"
unsigned int fz_ftoa(char *s, float v, int* K);

static const char *fz_hex_digits = "0123456789abcdef";

struct fmtbuf
{
	char *p;
	unsigned int s, n;
};

static void fmtputc(struct fmtbuf *out, int c)
{
	if (out->p && (out->n < out->s))
		out->p[out->n] = c;
	++(out->n);
}

/*
 * Convert float to shortest possible string that won't lose precision, except:
 * NaN to 0, +Inf to FLT_MAX, -Inf to -FLT_MAX.
 */
static void fmtfloat(struct fmtbuf *out, float f)
{
	char digits[40], *s = digits;
	int exp, ndigits;

	if (f != f) {
		fmtputc(out, 'N');
		fmtputc(out, 'a');
		fmtputc(out, 'N');
		return;
	}

    if (f < 0.0f) {
        fmtputc(out, '-');
        f = -f;
    }

	ndigits = fz_ftoa(digits, f, &exp);
	while (ndigits-- > 0)
		fmtputc(out, *s++);
}

static void fmtuint10(struct fmtbuf *out, uint64_t a, int z)
{
	char buf[25];
	unsigned int i = 0;

	while (a) {
		buf[i++] = ((a % 10U) + 0x30);
		a /= 10U;
	}
	while (i < z)
		buf[i++] = '0';
	buf[i] = '\0';
	while (i > 0)
		fmtputc(out, buf[--i]);
}

static void fmtuint16(struct fmtbuf *out, uint64_t a, int z)
{
	char buf[25];
	unsigned int i = 0;

	while (a) {
		buf[i++] = fz_hex_digits[a & 0x0f];
		a >>= 4;
	}
	while (i < z)
		buf[i++] = '0';
	buf[i] = '\0';
	while (i > 0)
		fmtputc(out, buf[--i]);
}

#if 0
static void fmtquote(struct fmtbuf *out, const char *s, int sq, int eq)
{
	int c;
	fmtputc(out, sq);
	while ((c = *s++) != 0) {
		switch (c) {
		default:
			if (c < 32) {
				fmtputc(out, '\\');
				fmtputc(out, '0' + ((c >> 6) & 7));
				fmtputc(out, '0' + ((c >> 3) & 7));
				fmtputc(out, '0' + ((c) & 7));
			} else {
				if (c == sq || c == eq)
					fmtputc(out, '\\');
				fmtputc(out, c);
			}
			break;
		case '\\': fmtputc(out, '\\'); fmtputc(out, '\\'); break;
		case '\b': fmtputc(out, '\\'); fmtputc(out, 'b'); break;
		case '\f': fmtputc(out, '\\'); fmtputc(out, 'f'); break;
		case '\n': fmtputc(out, '\\'); fmtputc(out, 'n'); break;
		case '\r': fmtputc(out, '\\'); fmtputc(out, 'r'); break;
		case '\t': fmtputc(out, '\\'); fmtputc(out, 't'); break;
		}
	}
	fmtputc(out, eq);
}
#endif

int fz_vsnprintf(char *buffer, size_t space, const char *fmt, va_list args)
{
	struct fmtbuf out;
	fz_matrix *m;
	fz_rect *r;
	int c, i, n, z;
	unsigned int k;
	uint64_t k64;
	double f;
	char *s;
	int length;

	out.p = buffer;
	out.s = space;
	out.n = 0;

	while ((c = *fmt++) != 0)
	{
		if (c == '%') {
			c = *fmt++;
			if (c == 0)
				break;
			z = 1;
			if (c == '0' && fmt[0] && fmt[1]) {
				z = *fmt++ - '0';
				c = *fmt++;
				while (c >= '0' && c <= '9' && fmt[0]) {
					z = z*10 + c - '0';
					c = *fmt++;
				}
			}
			/* Check for lengths */
			length = 0;
			switch (c) {
			case 'l':
				c = *fmt++;
				if (c == 'l')
					length = 64;
				else
					fmt--;
				break;
			case 'z':
				if (sizeof(size_t) >= 8)
					length = 64;
				break;
			case 'Z':
				if (sizeof(size_t) >= 8)
					length = 64;
				else
					length = 32;
				break;
			}
			if (length != 0) {
				c = *fmt++;
				if (c == 0)
					break; /* Can't warn :( */
			}
			switch (c) {
			default:
				fmtputc(&out, '%');
				fmtputc(&out, c);
				break;
			case '%':
				fmtputc(&out, '%');
				break;
			case 'M': /* fz_matrix * */
				m = va_arg(args, fz_matrix*);
				fmtfloat(&out, m->x[0]); fmtputc(&out, ' ');
				fmtfloat(&out, m->x[1]); fmtputc(&out, ' ');
				fmtfloat(&out, m->x[2]); fmtputc(&out, ' ');
				fmtfloat(&out, m->x[3]); fmtputc(&out, ' ');
				fmtfloat(&out, m->e); fmtputc(&out, ' ');
				fmtfloat(&out, m->f);
				break;
			case 'R': /* fz_rect * */
				r = va_arg(args, fz_rect*);
				fmtfloat(&out, r->x0); fmtputc(&out, ' ');
				fmtfloat(&out, r->y0); fmtputc(&out, ' ');
				fmtfloat(&out, r->x1); fmtputc(&out, ' ');
				fmtfloat(&out, r->y1);
				break;
			case 'C': /* unicode char */
				c = va_arg(args, int);
				if (c < 128) {
					fmtputc(&out, c);
				} else {
					char buf[10];
					n = fz_runetochar(buf, c);
					for (i=0; i < n; ++i)
						fmtputc(&out, buf[i]);
				}
				break;
			case 'c':
				c = va_arg(args, int);
				fmtputc(&out, c);
				break;
			case 'f':
			case 'g':
				f = va_arg(args, double);
				fmtfloat(&out, (float)f);
				break;
			case 'x':
				if (length == 64) {
					k64 = va_arg(args, uint64_t);
				} else {
					k64 = (uint64_t)va_arg(args, uint32_t);
				}
				fmtuint16(&out, k64, z);
				break;
			case 'd':
				if (length == 64) {
					k64 = va_arg(args, int64_t);
					if ((int64_t)k64 < 0) {
						fmtputc(&out, '-');
						k64 = -(int64_t)k64;
					}
				} else {
					k = va_arg(args, int);
					if ((int32_t)k < 0) {
						fmtputc(&out, '-');
						k = -(int32_t)k;
					}
                    k64 = (uint64_t)k;
				}
				fmtuint10(&out, k64, z);
				break;
			case 'u':
				if (length == 64) {
					k64 = va_arg(args, uint64_t);
				} else {
					k64 = (uint64_t)va_arg(args, uint32_t);
				}
				fmtuint10(&out, k64, z);
				break;
			case 's':
				s = va_arg(args, char*);
				if (!s)
					s = "(null)";
				while ((c = *s++) != 0)
					fmtputc(&out, c);
				break;
#if 0
			case 'q':
				s = va_arg(args, char*);
				if (!s) s = "";
				fmtquote(&out, s, '"', '"');
				break;
			case '(':
				s = va_arg(args, char*);
				if (!s) s = "";
				fmtquote(&out, s, '(', ')');
				break;
#endif
			}
		} else {
			fmtputc(&out, c);
		}
	}

	fmtputc(&out, 0);
	return out.n - 1;
}

int fz_snprintf(char *buffer, size_t space, const char *fmt, ...)
{
	int n;
	va_list ap;
	va_start(ap, fmt);
	n = fz_vsnprintf(buffer, space, fmt, ap);
	va_end(ap);
	return n;
}

