#include "mupdf/fitz.h"
#include <zlib.h>
#define MIN_BOMB (100 << 20)

int
fz_read(fz_stream *stm, unsigned char *buf, int len)
{
    int count, n;

    count = fz_mini(len, stm->wp - stm->rp);
    if (count) {
        memcpy(buf, stm->rp, count);
        stm->rp += count;
    }

    if (count == len || stm->error || stm->eof)
        return count;

    if (len - count < stm->ep - stm->bp) {
        n = stm->read(stm, stm->bp, stm->ep - stm->bp);
        if (n == 0) {
            stm->eof = 1;
        } else if (n > 0) {
            stm->rp = stm->bp;
            stm->wp = stm->bp + n;
            stm->pos += n;
        }

        n = fz_mini(len - count, stm->wp - stm->rp);
        if (n) {
            memcpy(buf + count, stm->rp, n);
            stm->rp += n;
            count += n;
        }
    } else {
        n = stm->read(stm, buf + count, len - count);
        if (n == 0) {
            stm->eof = 1;
        } else if (n > 0) {
            stm->pos += n;
            count += n;
        }
    }

    return count;
}

void
fz_fill_buffer(fz_stream *stm)
{
    int n;

    if (stm->error || stm->eof)
        return;

    fz_try(stm->ctx)
    {
        n = stm->read(stm, stm->bp, stm->ep - stm->bp);
        if (n == 0) {
            stm->eof = 1;
        } else if (n > 0) {
            stm->rp = stm->bp;
            stm->wp = stm->bp + n;
            stm->pos += n;
        }
    }
    fz_catch(stm->ctx)
    {
        fz_warn(stm->ctx->warn, "read error; treating as end of file");
        stm->error = 1;
    }
}

fz_buffer *
fz_read_best(fz_stream *stm, int initial, int *truncated)
{
    fz_buffer *buf = NULL;
    int n;
    fz_context *ctx = stm->ctx;

    if (truncated)
        *truncated = 0;

    fz_try(ctx)
    {
        if (initial < 1024)
            initial = 1024;

        buf = fz_new_buffer(initial+1);
        while (1) {
            if (buf->len == buf->cap)
                fz_grow_buffer(buf);

            if (buf->len >= MIN_BOMB && buf->len / 200 > initial) {
                fz_throw(ctx, FZ_ERROR_GENERIC, "compression bomb detected");
            }

            n = fz_read(stm, buf->data + buf->len, buf->cap - buf->len);
            if (n == 0)
                break;

            buf->len += n;
        }
    }
    fz_catch(ctx)
    {
        if (truncated) {
            *truncated = 1;
        } else {
            fz_drop_buffer(buf);
            fz_rethrow(ctx);
        }
    }
    return buf;
}

fz_buffer *
fz_read_all(fz_stream *stm, int initial)
{
    return fz_read_best(stm, initial, NULL);
}

void
fz_read_line(fz_stream *stm, char *mem, int n)
{
    char *s = mem;
    int c = EOF;
    while (n > 1) {
        c = fz_read_byte(stm);
        if (c == EOF)
            break;
        if (c == '\r') {
            c = fz_peek_byte(stm);
            if (c == '\n')
                fz_read_byte(stm);
            break;
        }
        if (c == '\n')
            break;
        *s++ = c;
        n--;
    }
    if (n) {
        *s = '\0';
    }
}

int
fz_tell(fz_stream *stm)
{
    return stm->pos - (stm->wp - stm->rp);
}

void
fz_seek(fz_stream *stm, int offset, int whence)
{
    stm->avail = 0; /* Reset bit reading */
    if (stm->seek) {
        if (whence == 1) {
            offset = fz_tell(stm) + offset;
            whence = 0;
        }
        if (whence == 0) {
            int dist = stm->pos - offset;
            if (dist >= 0 && dist <= stm->wp - stm->bp) {
                stm->rp = stm->wp - dist;
                stm->eof = 0;
                return;
            }
        }
        stm->seek(stm, offset, whence);
        stm->eof = 0;
    } else if (whence != 2) {
        if (whence == 0)
            offset -= fz_tell(stm);
        if (offset < 0)
            fz_warn(stm->ctx->warn, "cannot seek backwards");
        /* dog slow, but rare enough */
        while (offset-- > 0)
            fz_read_byte(stm);
    } else { 
        fz_warn(stm->ctx->warn, "cannot seek");
    }
}

int fz_stream_meta(fz_stream *stm, int key, int size, void *ptr)
{
    if (!stm || !stm->meta)
        return -1;
    return stm->meta(stm, key, size, ptr);
}

fz_stream *
fz_new_stream(fz_context *ctx, void *state,
    int(*read)(fz_stream *stm, unsigned char *buf, int len),
    void(*close)(fz_context *ctx, void *state))
{
    fz_stream *stm = fz_calloc_no_throw(ctx, 1, sizeof(fz_stream));

    stm->refs = 1;
    stm->error = 0;
    stm->eof = 0;
    stm->pos = 0;

    stm->bits = 0;
    stm->avail = 0;

    stm->bp = stm->buf;
    stm->rp = stm->bp;
    stm->wp = stm->bp;
    stm->ep = stm->buf + sizeof(stm->buf);

    stm->state = state;
    stm->read = read;
    stm->close = close;
    stm->seek = NULL;
    stm->ctx = ctx;

    return stm;
}

fz_stream *
fz_keep_stream(fz_stream *stm)
{
    if (stm)
        stm->refs ++;
    return stm;
}

void
fz_close(fz_stream *stm)
{
    if (!stm)
        return;
    stm->refs --;
    if (stm->refs == 0) {
        if (stm->close) {
            stm->close(stm->ctx, stm->state);
        }
        fz_free(stm->ctx, stm);
    }
}

/* Memory stream */

static int read_buffer(fz_stream *stm, unsigned char *buf, int len)
{
    return 0;
}

static void seek_buffer(fz_stream *stm, int offset, int whence)
{
    if (whence == 0)
        stm->rp = stm->bp + offset;
    if (whence == 1)
        stm->rp += offset;
    if (whence == 2)
        stm->rp = stm->ep - offset;
    stm->rp = fz_clampp(stm->rp, stm->bp, stm->ep);
    stm->wp = stm->ep;
}

static void close_buffer(fz_context *ctx, void *state_)
{
    fz_buffer *state = (fz_buffer *)state_;
    if (state)
        fz_drop_buffer(state);
}

fz_stream *
fz_open_buffer(fz_context *ctx, fz_buffer *buf)
{
    fz_stream *stm;

    fz_keep_buffer(buf);
    stm = fz_new_stream(ctx, buf, read_buffer, close_buffer);
    stm->seek = seek_buffer;

    stm->bp = buf->data;
    stm->rp = buf->data;
    stm->wp = buf->data + buf->len;
    stm->ep = buf->data + buf->len;

    stm->pos = buf->len;

    return stm;
}

fz_stream *
fz_open_memory(fz_context *ctx, unsigned char *data, unsigned int len)
{
    fz_stream *stm;

    stm = fz_new_stream(ctx, NULL, read_buffer, close_buffer);
    stm->seek = seek_buffer;

    stm->bp = data;
    stm->rp = data;
    stm->wp = data + len;
    stm->ep = data + len;

    stm->pos = len;

    return stm;
}

/* Pretend we have a filter that just copies data forever */
fz_stream *
fz_open_copy(fz_stream *chain)
{
    return fz_keep_stream(chain);
}

/* Null filter copies a specified amount of data */
struct null_filter
{
    fz_stream *chain;
    int remain;
    int pos;
};

static int
read_null(fz_stream *stm, unsigned char *buf, int len)
{
    struct null_filter *state = stm->state;
    int amount = fz_mini(len, state->remain);
    int n;

    fz_seek(state->chain, state->pos, 0);
    n = fz_read(state->chain, buf, amount);
    state->remain -= n;
    state->pos += n;
    return n;
}

static void
close_null(fz_context *ctx, void *state_)
{
    struct null_filter *state = (struct null_filter *)state_;
    fz_stream *chain = state->chain;
    fz_free(ctx, state);
    fz_close(chain);
}

fz_stream *
fz_open_null(fz_stream *chain, int len, int offset)
{
    struct null_filter *state;
    fz_context *ctx = chain->ctx;

    if (len < 0)
        len = 0;

    state = fz_calloc_no_throw(ctx, 1, sizeof(struct null_filter));
    state->chain = chain;
    state->remain = len;
    state->pos = offset;
    return fz_new_stream(ctx, state, read_null, close_null);
}

/* Concat filter concatenates several streams into one */
struct concat_filter
{
    int max;
    int count;
    int current;
    int pad; /* 1 if we should add whitespace padding between streams */
    int ws; /* 1 if we should send a whitespace padding byte next */
    fz_stream *chain[1];
};

static int
read_concat(fz_stream *stm, unsigned char *buf, int len)
{
    struct concat_filter *state = (struct concat_filter *)stm->state;
    int n;
    int read = 0;

    if (len <= 0)
        return 0;

    while (state->current != state->count && len > 0) {
        /* If we need to send a whitespace char, do that */
        if (state->ws) {
            *buf++ = 32;
            read++;
            len--;
            state->ws = 0;
            continue;
        }
        /* Otherwise, read as much data as will fit in the buffer */
        n = fz_read(state->chain[state->current], buf, len);
        read += n;
        buf += n;
        len -= n;
        /* If we didn't read any, then we must have hit the end of
         * our buffer space. Move to the next stream, and remember to
         * pad. */
        if (n == 0) {
            fz_close(state->chain[state->current]);
            state->current++;
            state->ws = state->pad;
        }
    }

    return read;
}

static void
close_concat(fz_context *ctx, void *state_)
{
    struct concat_filter *state = (struct concat_filter *)state_;
    int i;

    for (i = state->current; i < state->count; i++) {
        fz_close(state->chain[i]);
    }
    fz_free(ctx, state);
}

fz_stream *
fz_open_concat(fz_context *ctx, int len, int pad)
{
    struct concat_filter *state;

    state = fz_calloc_no_throw(ctx, 1, sizeof(struct concat_filter) + (len-1)*sizeof(fz_stream *));
    state->max = len;
    state->count = 0;
    state->current = 0;
    state->pad = pad;
    state->ws = 0; /* We never send padding byte at the start */

    return fz_new_stream(ctx, state, read_concat, close_concat);
}

void
fz_concat_push(fz_stream *concat, fz_stream *chain)
{
    struct concat_filter *state = (struct concat_filter *)concat->state;

    if (state->count == state->max)
        fz_throw(concat->ctx, FZ_ERROR_GENERIC, "Concat filter size exceeded");

    state->chain[state->count++] = chain;
}

/* ASCII Hex Decode */
typedef struct fz_ahxd_s fz_ahxd;

struct fz_ahxd_s
{
    fz_stream *chain;
    int eod;
};

static inline int iswhite(int a)
{
    switch (a) {
    case '\n': case '\r': case '\t': case ' ':
    case '\0': case '\f': case '\b': case 0177:
        return 1;
    }
    return 0;
}

static inline int ishex(int a)
{
    return (a >= 'A' && a <= 'F') ||
        (a >= 'a' && a <= 'f') ||
        (a >= '0' && a <= '9');
}

static inline int unhex(int a)
{
    if (a >= 'A' && a <= 'F') return a - 'A' + 0xA;
    if (a >= 'a' && a <= 'f') return a - 'a' + 0xA;
    if (a >= '0' && a <= '9') return a - '0';
    return 0;
}

static int
read_ahxd(fz_stream *stm, unsigned char *buf, int len)
{
    fz_ahxd *state = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;
    int a = 0, b = 0, c = 0, odd = 0;

    while (p < ep)
    {
        if (state->eod)
            return p - buf;

        c = fz_read_byte(state->chain);
        if (c < 0)
            return p - buf;

        if (ishex(c))
        {
            if (!odd)
            {
                a = unhex(c);
                odd = 1;
            }
            else
            {
                b = unhex(c);
                *p++ = (a << 4) | b;
                odd = 0;
            }
        }
        else if (c == '>')
        {
            if (odd)
                *p++ = (a << 4);
            state->eod = 1;
        }
        else if (!iswhite(c))
        {
            fz_throw(stm->ctx, FZ_ERROR_GENERIC, "bad data in ahxd: '%c'", c);
        }
    }

    return p - buf;
}

static void
close_ahxd(fz_context *ctx, void *state_)
{
    fz_ahxd *state = (fz_ahxd *)state_;
    fz_stream *chain = state->chain;
    fz_free(ctx, state);
    fz_close(chain);
}

fz_stream *
fz_open_ahxd(fz_stream *chain)
{
    fz_ahxd *state;
    fz_context *ctx = chain->ctx;

    state = fz_calloc_no_throw(ctx, 1, sizeof(fz_ahxd));
    state->chain = chain;
    state->eod = 0;
    return fz_new_stream(ctx, state, read_ahxd, close_ahxd);
}

/* ASCII 85 Decode */
typedef struct fz_a85d_s fz_a85d;

struct fz_a85d_s
{
    fz_stream *chain;
    unsigned char bp[4];
    unsigned char *rp, *wp;
    int eod;
};

static int
read_a85d(fz_stream *stm, unsigned char *buf, int len)
{
    fz_a85d *state = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;
    int count = 0;
    int word = 0;
    int c;

    while (state->rp < state->wp && p < ep)
        *p++ = *state->rp++;

    while (p < ep)
    {
        if (state->eod)
            return p - buf;

        c = fz_read_byte(state->chain);
        if (c < 0)
            return p - buf;

        if (c >= '!' && c <= 'u')
        {
            if (count == 4)
            {
                word = word * 85 + (c - '!');

                state->bp[0] = (word >> 24) & 0xff;
                state->bp[1] = (word >> 16) & 0xff;
                state->bp[2] = (word >> 8) & 0xff;
                state->bp[3] = (word) & 0xff;
                state->rp = state->bp;
                state->wp = state->bp + 4;

                word = 0;
                count = 0;
            }
            else
            {
                word = word * 85 + (c - '!');
                count ++;
            }
        }

        else if (c == 'z' && count == 0)
        {
            state->bp[0] = 0;
            state->bp[1] = 0;
            state->bp[2] = 0;
            state->bp[3] = 0;
            state->rp = state->bp;
            state->wp = state->bp + 4;
        }

        else if (c == '~')
        {
            c = fz_read_byte(state->chain);
            if (c != '>')
                fz_warn(stm->ctx->warn, "bad eod marker in a85d");

            switch (count) {
            case 0:
                break;
            case 1:
                /* Specifically illegal in the spec, but adobe
                 * and gs both cope. See normal_87.pdf for a
                 * case where this matters. */
                fz_warn(stm->ctx->warn, "partial final byte in a85d");
                break;
            case 2:
                word = word * (85 * 85 * 85) + 0xffffff;
                state->bp[0] = word >> 24;
                state->rp = state->bp;
                state->wp = state->bp + 1;
                break;
            case 3:
                word = word * (85 * 85) + 0xffff;
                state->bp[0] = word >> 24;
                state->bp[1] = word >> 16;
                state->rp = state->bp;
                state->wp = state->bp + 2;
                break;
            case 4:
                word = word * 85 + 0xff;
                state->bp[0] = word >> 24;
                state->bp[1] = word >> 16;
                state->bp[2] = word >> 8;
                state->rp = state->bp;
                state->wp = state->bp + 3;
                break;
            }
            state->eod = 1;
        }

        else if (!iswhite(c))
        {
            fz_throw(stm->ctx, FZ_ERROR_GENERIC, "bad data in a85d: '%c'", c);
        }

        while (state->rp < state->wp && p < ep)
            *p++ = *state->rp++;
    }

    return p - buf;
}

static void
close_a85d(fz_context *ctx, void *state_)
{
    fz_a85d *state = (fz_a85d *)state_;
    fz_stream *chain = state->chain;

    fz_free(ctx, state);
    fz_close(chain);
}

fz_stream *
fz_open_a85d(fz_stream *chain)
{
    fz_a85d *state;
    fz_context *ctx = chain->ctx;

    state = fz_calloc_no_throw(ctx, 1, sizeof(fz_a85d));
    state->chain = chain;
    state->rp = state->bp;
    state->wp = state->bp;
    state->eod = 0;
    return fz_new_stream(ctx, state, read_a85d, close_a85d);
}

/* Run Length Decode */
typedef struct fz_rld_s fz_rld;

struct fz_rld_s
{
    fz_stream *chain;
    int run, n, c;
};

static int
read_rld(fz_stream *stm, unsigned char *buf, int len)
{
    fz_rld *state = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;

    while (p < ep)
    {
        if (state->run == 128)
            return p - buf;

        if (state->n == 0)
        {
            state->run = fz_read_byte(state->chain);
            if (state->run < 0)
                state->run = 128;
            if (state->run < 128)
                state->n = state->run + 1;
            if (state->run > 128)
            {
                state->n = 257 - state->run;
                state->c = fz_read_byte(state->chain);
                if (state->c < 0)
                    fz_throw(stm->ctx, FZ_ERROR_GENERIC, "premature end of data in run length decode");
            }
        }

        if (state->run < 128)
        {
            while (p < ep && state->n)
            {
                int c = fz_read_byte(state->chain);
                if (c < 0)
                    fz_throw(stm->ctx, FZ_ERROR_GENERIC, "premature end of data in run length decode");
                *p++ = c;
                state->n--;
            }
        }

        if (state->run > 128)
        {
            while (p < ep && state->n)
            {
                *p++ = state->c;
                state->n--;
            }
        }
    }

    return p - buf;
}

static void
close_rld(fz_context *ctx, void *state_)
{
    fz_rld *state = (fz_rld *)state_;
    fz_stream *chain = state->chain;

    fz_free(ctx, state);
    fz_close(chain);
}

fz_stream *
fz_open_rld(fz_stream *chain)
{
    fz_rld *state;
    fz_context *ctx = chain->ctx;

    state = fz_calloc_no_throw(ctx, 1, sizeof(fz_rld));
    state->chain = chain;
    state->run = 0;
    state->n = 0;
    state->c = 0;
    return fz_new_stream(ctx, state, read_rld, close_rld);
}

/* RC4 Filter */
typedef struct fz_arc4c_s fz_arc4c;

struct fz_arc4c_s
{
    fz_stream *chain;
    fz_arc4 arc4;
};

static int
read_arc4(fz_stream *stm, unsigned char *buf, int len)
{
    fz_arc4c *state = stm->state;
    unsigned int i, n = fz_read(state->chain, buf, len);
	for (i = 0; i < n; i++) {
		unsigned char x = fz_arc4_next(&state->arc4);
		buf[i] ^= x;
	}
    return n;
}

static void
close_arc4(fz_context *ctx, void *state_)
{
    fz_arc4c *state = (fz_arc4c *)state_;
    fz_stream *chain = state->chain;

    free(state);
    fz_close(chain);
}

fz_stream *
fz_open_arc4(fz_stream *chain, unsigned char *key, unsigned keylen)
{
    fz_arc4c *state;
    fz_context *ctx = chain->ctx;

    state = fz_malloc(ctx, sizeof(fz_arc4c));
    state->chain = chain;
    fz_arc4_init(&state->arc4, key, keylen);
    return fz_new_stream(ctx, state, read_arc4, close_arc4);
}

/* zlib/DEFLATE filter */
typedef struct fz_flate_s fz_flate;

struct fz_flate_s
{
    fz_stream *chain;
    z_stream z;
};

static void *zalloc(void *opaque, unsigned int items, unsigned int size)
{
    return fz_malloc_no_throw(opaque, items*size);
}

static void zfree(void *opaque, void *ptr)
{
    fz_free(opaque, ptr);
}

static int
read_flated(fz_stream *stm, unsigned char *outbuf, int outlen)
{
    fz_flate *state = stm->state;
    fz_stream *chain = state->chain;
    z_streamp zp = &state->z;
    int code;

    zp->next_out = outbuf;
    zp->avail_out = outlen;

    while (zp->avail_out > 0)
    {
        if (chain->rp == chain->wp)
            fz_fill_buffer(chain);

        zp->next_in = chain->rp;
        zp->avail_in = chain->wp - chain->rp;

        code = inflate(zp, Z_SYNC_FLUSH);

        chain->rp = chain->wp - zp->avail_in;

        if (code == Z_STREAM_END) {
            return outlen - zp->avail_out;
        } else if (code == Z_BUF_ERROR) {
            fz_warn(stm->ctx->warn, "premature end of data in flate filter");
            return outlen - zp->avail_out;
        } else if (code == Z_DATA_ERROR && zp->avail_in == 0) {
            fz_warn(stm->ctx->warn, "ignoring zlib error: %s", zp->msg);
            return outlen - zp->avail_out;
        } else if (code != Z_OK) {
            fz_throw(stm->ctx, FZ_ERROR_GENERIC, "zlib error: %s", zp->msg);
        }
    }

    return outlen - zp->avail_out;
}

static void
close_flated(fz_context *ctx, void *state_)
{
    fz_flate *state = (fz_flate *)state_;
    int code;

    code = inflateEnd(&state->z);
    if (code != Z_OK)
        fz_warn(ctx->warn, "zlib error: inflateEnd: %s", state->z.msg);

    fz_close(state->chain);
    fz_free(ctx, state);
}

fz_stream *
fz_open_flated(fz_stream *chain, int windowbits)
{
    fz_flate *state = NULL;
    int code = Z_OK;
    fz_context *ctx = chain->ctx;

    state = fz_calloc_no_throw(ctx, 1, sizeof(fz_flate));
    state->chain = chain;

    state->z.zalloc = zalloc;
    state->z.zfree = zfree;
    state->z.opaque = ctx;
    state->z.next_in = NULL;
    state->z.avail_in = 0;

    code = inflateInit2(&state->z, windowbits);
    if (code != Z_OK) {
        fz_free(ctx, state);
        fz_close(chain);
    	fz_throw(ctx, FZ_ERROR_GENERIC, "zlib error: inflateInit: %s", state->z.msg);
	}
    return fz_new_stream(ctx, state, read_flated, close_flated);
}

/* Fax G3/G4 decoder */
/* TODO: uncompressed */

/*
<raph> the first 2^(initialbits) entries map bit patterns to decodes
<raph> let's say initial_bits is 8 for the sake of example
<raph> and that the code is 1001
<raph> that means that entries 0x90 .. 0x9f have the entry { val, 4 }
<raph> because those are all the bytes that start with the code
<raph> and the 4 is the length of the code
... if (n_bits > initial_bits) ...
<raph> anyway, in that case, it basically points to a mini table
<raph> the n_bits is the maximum length of all codes beginning with that byte
<raph> so 2^(n_bits - initial_bits) is the size of the mini-table
<raph> peter came up with this, and it makes sense
*/

typedef struct cfd_node_s cfd_node;

struct cfd_node_s
{
    short val;
    short nbits;
};

enum
{
    cfd_white_initial_bits = 8,
    cfd_black_initial_bits = 7,
    cfd_2d_initial_bits = 7,
    cfd_uncompressed_initial_bits = 6 /* must be 6 */
};

/* non-run codes in tables */
enum
{
    ERROR = -1,
    ZEROS = -2, /* EOL follows, possibly with more padding first */
    UNCOMPRESSED = -3
};

/* semantic codes for cf_2d_decode */
enum
{
    P = -4,
    H = -5,
    VR3 = 0,
    VR2 = 1,
    VR1 = 2,
    V0 = 3,
    VL1 = 4,
    VL2 = 5,
    VL3 = 6
};

/* White decoding table. */
static const cfd_node cf_white_decode[] = {
    {256,12},{272,12},{29,8},{30,8},{45,8},{46,8},{22,7},{22,7},
    {23,7},{23,7},{47,8},{48,8},{13,6},{13,6},{13,6},{13,6},{20,7},
    {20,7},{33,8},{34,8},{35,8},{36,8},{37,8},{38,8},{19,7},{19,7},
    {31,8},{32,8},{1,6},{1,6},{1,6},{1,6},{12,6},{12,6},{12,6},{12,6},
    {53,8},{54,8},{26,7},{26,7},{39,8},{40,8},{41,8},{42,8},{43,8},
    {44,8},{21,7},{21,7},{28,7},{28,7},{61,8},{62,8},{63,8},{0,8},
    {320,8},{384,8},{10,5},{10,5},{10,5},{10,5},{10,5},{10,5},{10,5},
    {10,5},{11,5},{11,5},{11,5},{11,5},{11,5},{11,5},{11,5},{11,5},
    {27,7},{27,7},{59,8},{60,8},{288,9},{290,9},{18,7},{18,7},{24,7},
    {24,7},{49,8},{50,8},{51,8},{52,8},{25,7},{25,7},{55,8},{56,8},
    {57,8},{58,8},{192,6},{192,6},{192,6},{192,6},{1664,6},{1664,6},
    {1664,6},{1664,6},{448,8},{512,8},{292,9},{640,8},{576,8},{294,9},
    {296,9},{298,9},{300,9},{302,9},{256,7},{256,7},{2,4},{2,4},{2,4},
    {2,4},{2,4},{2,4},{2,4},{2,4},{2,4},{2,4},{2,4},{2,4},{2,4},{2,4},
    {2,4},{2,4},{3,4},{3,4},{3,4},{3,4},{3,4},{3,4},{3,4},{3,4},{3,4},
    {3,4},{3,4},{3,4},{3,4},{3,4},{3,4},{3,4},{128,5},{128,5},{128,5},
    {128,5},{128,5},{128,5},{128,5},{128,5},{8,5},{8,5},{8,5},{8,5},
    {8,5},{8,5},{8,5},{8,5},{9,5},{9,5},{9,5},{9,5},{9,5},{9,5},{9,5},
    {9,5},{16,6},{16,6},{16,6},{16,6},{17,6},{17,6},{17,6},{17,6},
    {4,4},{4,4},{4,4},{4,4},{4,4},{4,4},{4,4},{4,4},{4,4},{4,4},{4,4},
    {4,4},{4,4},{4,4},{4,4},{4,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},
    {5,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},
    {14,6},{14,6},{14,6},{14,6},{15,6},{15,6},{15,6},{15,6},{64,5},
    {64,5},{64,5},{64,5},{64,5},{64,5},{64,5},{64,5},{6,4},{6,4},
    {6,4},{6,4},{6,4},{6,4},{6,4},{6,4},{6,4},{6,4},{6,4},{6,4},{6,4},
    {6,4},{6,4},{6,4},{7,4},{7,4},{7,4},{7,4},{7,4},{7,4},{7,4},{7,4},
    {7,4},{7,4},{7,4},{7,4},{7,4},{7,4},{7,4},{7,4},{-2,3},{-2,3},
    {-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},
    {-1,0},{-1,0},{-1,0},{-1,0},{-3,4},{1792,3},{1792,3},{1984,4},
    {2048,4},{2112,4},{2176,4},{2240,4},{2304,4},{1856,3},{1856,3},
    {1920,3},{1920,3},{2368,4},{2432,4},{2496,4},{2560,4},{1472,1},
    {1536,1},{1600,1},{1728,1},{704,1},{768,1},{832,1},{896,1},
    {960,1},{1024,1},{1088,1},{1152,1},{1216,1},{1280,1},{1344,1},
    {1408,1}
};

/* Black decoding table. */
static const cfd_node cf_black_decode[] = {
    {128,12},{160,13},{224,12},{256,12},{10,7},{11,7},{288,12},{12,7},
    {9,6},{9,6},{8,6},{8,6},{7,5},{7,5},{7,5},{7,5},{6,4},{6,4},{6,4},
    {6,4},{6,4},{6,4},{6,4},{6,4},{5,4},{5,4},{5,4},{5,4},{5,4},{5,4},
    {5,4},{5,4},{1,3},{1,3},{1,3},{1,3},{1,3},{1,3},{1,3},{1,3},{1,3},
    {1,3},{1,3},{1,3},{1,3},{1,3},{1,3},{1,3},{4,3},{4,3},{4,3},{4,3},
    {4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},
    {4,3},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},
    {3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},
    {3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},{3,2},
    {2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},
    {2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},
    {2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},{2,2},
    {-2,4},{-2,4},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},
    {-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-3,5},{1792,4},
    {1792,4},{1984,5},{2048,5},{2112,5},{2176,5},{2240,5},{2304,5},
    {1856,4},{1856,4},{1920,4},{1920,4},{2368,5},{2432,5},{2496,5},
    {2560,5},{18,3},{18,3},{18,3},{18,3},{18,3},{18,3},{18,3},{18,3},
    {52,5},{52,5},{640,6},{704,6},{768,6},{832,6},{55,5},{55,5},
    {56,5},{56,5},{1280,6},{1344,6},{1408,6},{1472,6},{59,5},{59,5},
    {60,5},{60,5},{1536,6},{1600,6},{24,4},{24,4},{24,4},{24,4},
    {25,4},{25,4},{25,4},{25,4},{1664,6},{1728,6},{320,5},{320,5},
    {384,5},{384,5},{448,5},{448,5},{512,6},{576,6},{53,5},{53,5},
    {54,5},{54,5},{896,6},{960,6},{1024,6},{1088,6},{1152,6},{1216,6},
    {64,3},{64,3},{64,3},{64,3},{64,3},{64,3},{64,3},{64,3},{13,1},
    {13,1},{13,1},{13,1},{13,1},{13,1},{13,1},{13,1},{13,1},{13,1},
    {13,1},{13,1},{13,1},{13,1},{13,1},{13,1},{23,4},{23,4},{50,5},
    {51,5},{44,5},{45,5},{46,5},{47,5},{57,5},{58,5},{61,5},{256,5},
    {16,3},{16,3},{16,3},{16,3},{17,3},{17,3},{17,3},{17,3},{48,5},
    {49,5},{62,5},{63,5},{30,5},{31,5},{32,5},{33,5},{40,5},{41,5},
    {22,4},{22,4},{14,1},{14,1},{14,1},{14,1},{14,1},{14,1},{14,1},
    {14,1},{14,1},{14,1},{14,1},{14,1},{14,1},{14,1},{14,1},{14,1},
    {15,2},{15,2},{15,2},{15,2},{15,2},{15,2},{15,2},{15,2},{128,5},
    {192,5},{26,5},{27,5},{28,5},{29,5},{19,4},{19,4},{20,4},{20,4},
    {34,5},{35,5},{36,5},{37,5},{38,5},{39,5},{21,4},{21,4},{42,5},
    {43,5},{0,3},{0,3},{0,3},{0,3}
};

/* 2-D decoding table. */
static const cfd_node cf_2d_decode[] = {
    {128,11},{144,10},{6,7},{0,7},{5,6},{5,6},{1,6},{1,6},{-4,4},
    {-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-4,4},{-5,3},{-5,3},
    {-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{-5,3},
    {-5,3},{-5,3},{-5,3},{-5,3},{-5,3},{4,3},{4,3},{4,3},{4,3},{4,3},
    {4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},{4,3},
    {2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{2,3},
    {2,3},{2,3},{2,3},{2,3},{2,3},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},{3,1},
    {3,1},{3,1},{3,1},{-2,4},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},
    {-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},
    {-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-1,0},{-3,3}
};

/* Uncompressed decoding table. */
static const cfd_node cf_uncompressed_decode[] = {
    {64,12},{5,6},{4,5},{4,5},{3,4},{3,4},{3,4},{3,4},{2,3},{2,3},
    {2,3},{2,3},{2,3},{2,3},{2,3},{2,3},{1,2},{1,2},{1,2},{1,2},{1,2},
    {1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},
    {0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},
    {0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},
    {0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},{0,1},
    {-1,0},{-1,0},{8,6},{9,6},{6,5},{6,5},{7,5},{7,5},{4,4},{4,4},
    {4,4},{4,4},{5,4},{5,4},{5,4},{5,4},{2,3},{2,3},{2,3},{2,3},{2,3},
    {2,3},{2,3},{2,3},{3,3},{3,3},{3,3},{3,3},{3,3},{3,3},{3,3},{3,3},
    {0,2},{0,2},{0,2},{0,2},{0,2},{0,2},{0,2},{0,2},{0,2},{0,2},{0,2},
    {0,2},{0,2},{0,2},{0,2},{0,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},
    {1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2},{1,2}
};

/* bit magic */

static inline int getbit(const unsigned char *buf, int x)
{
    return ( buf[x >> 3] >> ( 7 - (x & 7) ) ) & 1;
}

static const unsigned char mask[8] = {
    0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01, 0
};

static inline unsigned int clz(unsigned int x) {
    unsigned int ret;
    x &= 0xFF;
    __asm__ volatile("or $1, %1 \n\t"
                     "bsr %1, %0 \n\t"
                     :"=a"(ret) :"D"(x));
    return (7 - ret);
}

static inline int
find_changing(const unsigned char *line, int x, int w)
{
    int a, b, m, W;

    if (!line)
        return w;

    /* We assume w > 0, -1 <= x < w */
    if (x < 0)
    {
        x = 0;
        m = 0xFF;
    }
    else
    {
        /* Mask out the bits we've already used (including the one we started from) */
        m = mask[x & 7];
    }
    /* We have 'w' pixels (bits) in line. The last pixel that can be
     * safely accessed is the (w-1)th bit of line.
     * By taking W = w>>3, we know that the first W bytes of line are
     * full, with w&7 stray bits following. */
    W = w>>3;
    x >>= 3;
    a = line[x]; /* Safe as x < w => x <= w-1 => x>>3 <= (w-1)>>3 */
    b = a ^ (a>>1);
    b &= m;
    if (x >= W)
    {
        /* Within the last byte already */
        x = (x<<3) + clz(b);
        if (x > w)
            x = w;
        return x;
    }
    while (b == 0)
    {
        if (++x >= W)
            goto nearend;
        b = a & 1;
        a = line[x];
        b = (b<<7) ^ a ^ (a>>1);
    }
    return (x<<3) + clz(b);
nearend:
    /* We have less than a byte to go. If no stray bits, exit now. */
    if ((x<<3) == w)
        return w;
    b = a&1;
    a = line[x];
    b = (b<<7) ^ a ^ (a>>1);
    x = (x<<3) + clz(b);
    if (x > w)
        x = w;
    return x;
}

static inline int
find_changing_color(const unsigned char *line, int x, int w, int color)
{
    if (!line || x >= w)
        return w;

    x = find_changing(line, (x > 0 || !color) ? x : -1, w);

    if (x < w && getbit(line, x) != color)
        x = find_changing(line, x, w);

    return x;
}

static const unsigned char lm[8] = {
    0xFF, 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01
};

static const unsigned char rm[8] = {
    0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE
};

static inline void setbits(unsigned char *line, int x0, int x1)
{
    int a0, a1, b0, b1, a;

    if (x1 <= x0)
        return;

    a0 = x0 >> 3;
    a1 = x1 >> 3;

    b0 = x0 & 7;
    b1 = x1 & 7;

    if (a0 == a1)
    {
        if (b1)
            line[a0] |= lm[b0] & rm[b1];
    }
    else
    {
        line[a0] |= lm[b0];
        for (a = a0 + 1; a < a1; a++)
            line[a] = 0xFF;
        if (b1)
            line[a1] |= rm[b1];
    }
}

typedef struct fz_faxd_s fz_faxd;

enum
{
    STATE_NORMAL,   /* neutral state, waiting for any code */
    STATE_MAKEUP,   /* got a 1d makeup code, waiting for terminating code */
    STATE_EOL,      /* at eol, needs output buffer space */
    STATE_H1, STATE_H2, /* in H part 1 and 2 (both makeup and terminating codes) */
    STATE_DONE      /* all done */
};

struct fz_faxd_s
{
    fz_context *ctx;
    fz_stream *chain;

    int k;
    int end_of_line;
    int encoded_byte_align;
    int columns;
    int rows;
    int end_of_block;
    int black_is_1;

    int stride;
    int ridx;

    int bidx;
    unsigned int word;

    int stage;

    int a, c, dim, eolc;
    unsigned char *ref;
    unsigned char *dst;
    unsigned char *rp, *wp;
};

static inline void eat_bits(fz_faxd *fax, int nbits)
{
    fax->word <<= nbits;
    fax->bidx += nbits;
}

static int
fill_bits(fz_faxd *fax)
{
    while (fax->bidx >= 8)
    {
        int c = fz_read_byte(fax->chain);
        if (c == EOF)
            return EOF;
        fax->bidx -= 8;
        fax->word |= c << fax->bidx;
    }
    return 0;
}

static int
get_code(fz_faxd *fax, const cfd_node *table, int initialbits)
{
    unsigned int word = fax->word;
    int tidx = word >> (32 - initialbits);
    int val = table[tidx].val;
    int nbits = table[tidx].nbits;

    if (nbits > initialbits)
    {
        int mask = (1 << (32 - initialbits)) - 1;
        tidx = val + ((word & mask) >> (32 - nbits));
        val = table[tidx].val;
        nbits = initialbits + table[tidx].nbits;
    }

    eat_bits(fax, nbits);

    return val;
}

/* decode one 1d code */
static void
dec1d(fz_context *ctx, fz_faxd *fax)
{
    int code;

    if (fax->a == -1)
        fax->a = 0;

    if (fax->c)
        code = get_code(fax, cf_black_decode, cfd_black_initial_bits);
    else
        code = get_code(fax, cf_white_decode, cfd_white_initial_bits);

    if (code == UNCOMPRESSED)
        fz_throw(ctx, FZ_ERROR_GENERIC, "uncompressed data in faxd");

    if (code < 0)
        fz_throw(ctx, FZ_ERROR_GENERIC, "negative code in 1d faxd");

    if (fax->a + code > fax->columns)
        fz_throw(ctx, FZ_ERROR_GENERIC, "overflow in 1d faxd");

    if (fax->c)
        setbits(fax->dst, fax->a, fax->a + code);

    fax->a += code;

    if (code < 64)
    {
        fax->c = !fax->c;
        fax->stage = STATE_NORMAL;
    }
    else
        fax->stage = STATE_MAKEUP;
}

/* decode one 2d code */
static void
dec2d(fz_context *ctx, fz_faxd *fax)
{
    int code, b1, b2;

    if (fax->stage == STATE_H1 || fax->stage == STATE_H2)
    {
        if (fax->a == -1)
            fax->a = 0;

        if (fax->c)
            code = get_code(fax, cf_black_decode, cfd_black_initial_bits);
        else
            code = get_code(fax, cf_white_decode, cfd_white_initial_bits);

        if (code == UNCOMPRESSED)
            fz_throw(ctx, FZ_ERROR_GENERIC, "uncompressed data in faxd");

        if (code < 0)
            fz_throw(ctx, FZ_ERROR_GENERIC, "negative code in 2d faxd");

        if (fax->a + code > fax->columns)
            fz_throw(ctx, FZ_ERROR_GENERIC, "overflow in 2d faxd");

        if (fax->c)
            setbits(fax->dst, fax->a, fax->a + code);

        fax->a += code;

        if (code < 64)
        {
            fax->c = !fax->c;
            if (fax->stage == STATE_H1)
                fax->stage = STATE_H2;
            else if (fax->stage == STATE_H2)
                fax->stage = STATE_NORMAL;
        }

        return;
    }

    code = get_code(fax, cf_2d_decode, cfd_2d_initial_bits);

    switch (code)
    {
    case H:
        fax->stage = STATE_H1;
        break;

    case P:
        b1 = find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 >= fax->columns)
            b2 = fax->columns;
        else
            b2 = find_changing(fax->ref, b1, fax->columns);
        if (fax->c) setbits(fax->dst, fax->a, b2);
        fax->a = b2;
        break;

    case V0:
        b1 = find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VR1:
        b1 = 1 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 >= fax->columns) b1 = fax->columns;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VR2:
        b1 = 2 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 >= fax->columns) b1 = fax->columns;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VR3:
        b1 = 3 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 >= fax->columns) b1 = fax->columns;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VL1:
        b1 = -1 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 < 0) b1 = 0;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VL2:
        b1 = -2 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 < 0) b1 = 0;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case VL3:
        b1 = -3 + find_changing_color(fax->ref, fax->a, fax->columns, !fax->c);
        if (b1 < 0) b1 = 0;
        if (fax->c) setbits(fax->dst, fax->a, b1);
        fax->a = b1;
        fax->c = !fax->c;
        break;

    case UNCOMPRESSED:
        fz_throw(ctx, FZ_ERROR_GENERIC, "uncompressed data in faxd");

    case ERROR:
        fz_throw(ctx, FZ_ERROR_GENERIC, "invalid code in 2d faxd");

    default:
        fz_throw(ctx, FZ_ERROR_GENERIC, "invalid code in 2d faxd (%d)", code);
    }
}

static int
read_faxd(fz_stream *stm, unsigned char *buf, int len)
{
    fz_faxd *fax = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;
    unsigned char *tmp;

    if (fax->stage == STATE_DONE)
        return 0;

    if (fax->stage == STATE_EOL)
        goto eol;

loop:

    if (fill_bits(fax))
    {
        if (fax->bidx > 31)
        {
            if (fax->a > 0)
                goto eol;
            goto rtc;
        }
    }

    if ((fax->word >> (32 - 12)) == 0)
    {
        eat_bits(fax, 1);
        goto loop;
    }

    if ((fax->word >> (32 - 12)) == 1)
    {
        eat_bits(fax, 12);
        fax->eolc ++;

        if (fax->k > 0)
        {
            if (fax->a == -1)
                fax->a = 0;
            if ((fax->word >> (32 - 1)) == 1)
                fax->dim = 1;
            else
                fax->dim = 2;
            eat_bits(fax, 1);
        }
    }
    else if (fax->k > 0 && fax->a == -1)
    {
        fax->a = 0;
        if ((fax->word >> (32 - 1)) == 1)
            fax->dim = 1;
        else
            fax->dim = 2;
        eat_bits(fax, 1);
    }
    else if (fax->dim == 1)
    {
        fax->eolc = 0;
        dec1d(stm->ctx, fax);
    }
    else if (fax->dim == 2)
    {
        fax->eolc = 0;
        dec2d(stm->ctx, fax);
    }

    /* no eol check after makeup codes nor in the middle of an H code */
    if (fax->stage == STATE_MAKEUP || fax->stage == STATE_H1 || fax->stage == STATE_H2)
        goto loop;

    /* check for eol conditions */
    if (fax->eolc || fax->a >= fax->columns)
    {
        if (fax->a > 0)
            goto eol;
        if (fax->eolc == (fax->k < 0 ? 2 : 6))
            goto rtc;
    }

    goto loop;

eol:
    fax->stage = STATE_EOL;

    if (fax->black_is_1)
    {
        while (fax->rp < fax->wp && p < ep)
            *p++ = *fax->rp++;
    }
    else
    {
        while (fax->rp < fax->wp && p < ep)
            *p++ = *fax->rp++ ^ 0xff;
    }

    if (fax->rp < fax->wp)
        return p - buf;

    tmp = fax->ref;
    fax->ref = fax->dst;
    fax->dst = tmp;
    memset(fax->dst, 0, fax->stride);

    fax->rp = fax->dst;
    fax->wp = fax->dst + fax->stride;

    fax->stage = STATE_NORMAL;
    fax->c = 0;
    fax->a = -1;
    fax->ridx ++;

    if (!fax->end_of_block && fax->rows)
    {
        if (fax->ridx >= fax->rows)
            goto rtc;
    }

    /* we have not read dim from eol, make a guess */
    if (fax->k > 0 && !fax->eolc && fax->a == -1)
    {
        if (fax->ridx % fax->k == 0)
            fax->dim = 1;
        else
            fax->dim = 2;
    }

    /* if end_of_line & encoded_byte_align, EOLs are *not* optional */
    if (fax->encoded_byte_align)
    {
        if (fax->end_of_line)
            eat_bits(fax, (12 - fax->bidx) & 7);
        else
            eat_bits(fax, (8 - fax->bidx) & 7);
    }

    /* no more space in output, don't decode the next row yet */
    if (p == buf + len)
        return p - buf;

    goto loop;

rtc:
    fax->stage = STATE_DONE;
    return p - buf;
}

static void
close_faxd(fz_context *ctx, void *state_)
{
    fz_faxd *fax = (fz_faxd *)state_;
    int i;

    /* if we read any extra bytes, try to put them back */
    i = (32 - fax->bidx) / 8;
    while (i--)
        fz_unread_byte(fax->chain);

    fz_close(fax->chain);
    fz_free(ctx, fax->ref);
    fz_free(ctx, fax->dst);
    fz_free(ctx, fax);
}

/* Default: columns = 1728, end_of_block = 1, the rest = 0 */
fz_stream *
fz_open_faxd(fz_stream *chain,
    int k, int end_of_line, int encoded_byte_align,
    int columns, int rows, int end_of_block, int black_is_1)
{
    fz_context *ctx = chain->ctx;
    fz_faxd *fax = NULL;

    fax = fz_calloc_no_throw(ctx, 1, sizeof(fz_faxd));
    fax->chain = chain;

    fax->ref = NULL;
    fax->dst = NULL;

    fax->k = k;
    fax->end_of_line = end_of_line;
    fax->encoded_byte_align = encoded_byte_align;
    fax->columns = columns;
    fax->rows = rows;
    fax->end_of_block = end_of_block;
    fax->black_is_1 = black_is_1;

    fax->stride = ((fax->columns - 1) >> 3) + 1;
    fax->ridx = 0;
    fax->bidx = 32;
    fax->word = 0;

    fax->stage = STATE_NORMAL;
    fax->a = -1;
    fax->c = 0;
    fax->dim = fax->k < 0 ? 2 : 1;
    fax->eolc = 0;

    fax->ref = fz_calloc_no_throw(ctx, 1, fax->stride);
    fax->dst = fz_calloc_no_throw(ctx, 1, fax->stride);
    fax->rp = fax->dst;
    fax->wp = fax->dst + fax->stride;
    return fz_new_stream(ctx, fax, read_faxd, close_faxd);
}

/* LZW filter */

#if 0

/* TODO: error checking */
#define LZW_CLEAR(lzw)  (1 << ((lzw)->min_bits - 1))
#define LZW_EOD(lzw)    (LZW_CLEAR(lzw) + 1)
#define LZW_FIRST(lzw)  (LZW_CLEAR(lzw) + 2)

enum
{
    MAX_BITS = 12,
    NUM_CODES = (1 << MAX_BITS),
    MAX_LENGTH = 4097
};

typedef struct lzw_code_s
{
    int prev;           /* prev code (in string) */
    unsigned short length;      /* string len, including this token */
    unsigned char value;        /* data value */
    unsigned char first_char;   /* first token of string */
} lzw_code;

typedef struct fz_lzwd_s
{
    fz_stream *chain;
    int eod;

    int early_change;

    int reverse_bits;
    int min_bits;           /* minimum num bits/code */
    int code_bits;          /* num bits/code */
    int code;           /* current code */
    int old_code;           /* previously recognized code */
    int next_code;          /* next free entry */

    lzw_code table[NUM_CODES];

    unsigned char bp[MAX_LENGTH];
    unsigned char *rp, *wp;

    unsigned char buffer[4096];
} fz_lzwd;

static int
read_lzwd(fz_stream *stm, unsigned char *buf, int len)
{
    fz_lzwd *lzw = stm->state;
    lzw_code *table = lzw->table;
    unsigned char *p = buf;
    unsigned char *ep;
	unsigned char *s;
	int codelen;

    int code_bits = lzw->code_bits;
    int code = lzw->code;
    int old_code = lzw->old_code;
    int next_code = lzw->next_code;

    if (len > sizeof(lzw->buffer))
        len = sizeof(lzw->buffer);
    ep = buf + len;

    while (lzw->rp < lzw->wp && p < ep)
        *p++ = *lzw->rp++;

    while (p < ep) {
        if (lzw->eod)
            return 0;

        if (lzw->reverse_bits)
            code = fz_read_rbits(lzw->chain, code_bits);
        else
            code = fz_read_bits(lzw->chain, code_bits);

        if (fz_is_eof_bits(lzw->chain)) {
            lzw->eod = 1;
            break;
        }

        if (code == LZW_EOD(lzw)) {
            lzw->eod = 1;
            break;
        }

        if (next_code > NUM_CODES && code != LZW_CLEAR(lzw)) {
            fz_warn(stm->ctx->warn, "missing clear code in lzw decode");
            code = LZW_CLEAR(lzw);
        }

        if (code == LZW_CLEAR(lzw)) {
            code_bits = lzw->min_bits;
            next_code = LZW_FIRST(lzw);
            old_code = -1;
            continue;
        }

        /* if stream starts without a clear code, old_code is undefined... */
        if (old_code == -1) {
            old_code = code;
        } else if (next_code == NUM_CODES) {
            /* TODO: Ghostscript checks for a following clear code before tolerating */
            fz_warn(stm->ctx->warn, "tolerating a single out of range code in lzw decode");
            next_code++;
        } else if (code > next_code || next_code >= NUM_CODES) {
            fz_warn(stm->ctx->warn, "out of range code encountered in lzw decode");
        } else {
            /* add new entry to the code table */
            table[next_code].prev = old_code;
            table[next_code].first_char = table[old_code].first_char;
            table[next_code].length = table[old_code].length + 1;
            if (code < next_code)
                table[next_code].value = table[code].first_char;
            else if (code == next_code)
                table[next_code].value = table[next_code].first_char;
            else
                fz_warn(stm->ctx->warn, "out of range code encountered in lzw decode");
            next_code ++;
            if (next_code > (1 << code_bits) - lzw->early_change - 1) {
                code_bits ++;
                if (code_bits > MAX_BITS)
                    code_bits = MAX_BITS;
            }
            old_code = code;
        }

        /* code maps to a string, copy to output (in reverse...) */
        if (code >= LZW_CLEAR(lzw)) {
            codelen = table[code].length;
            lzw->rp = lzw->bp;
            lzw->wp = lzw->bp + codelen;
            s = lzw->wp;
            do {
                *(--s) = table[code].value;
                code = table[code].prev;
            } while (code >= 0 && s > lzw->bp);
        /* ... or just a single character */
        } else {
            lzw->bp[0] = code;
            lzw->rp = lzw->bp;
            lzw->wp = lzw->bp + 1;
        }

        /* copy to output */
        while (lzw->rp < lzw->wp && p < ep)
            *p++ = *lzw->rp++;
    }

    lzw->code_bits = code_bits;
    lzw->code = code;
    lzw->old_code = old_code;
    lzw->next_code = next_code;

    stm->rp = buf;
    stm->wp = p;
    if (buf == p)
        return EOF;
    stm->pos += p - buf;
    return *stm->rp++;
}

static void
close_lzwd(fz_context *ctx, void *state_)
{
    fz_lzwd *lzw = (fz_lzwd *)state_;
    fz_sync_bits(lzw->chain);
    fz_close(lzw->chain);
    fz_free(ctx, lzw);
}

/* Default: early_change = 1 */
fz_stream *fz_open_lzwd(fz_stream *chain, int early_change, int min_bits, int reverse_bits)
{
    fz_context *ctx = chain->ctx;
    fz_lzwd *lzw = NULL;
    int i;

    if (min_bits > MAX_BITS) {
        fz_warn(ctx->warn, "out of range initial lzw code size");
        min_bits = MAX_BITS;
    }

    lzw = fz_malloc_struct(ctx, fz_lzwd);
    lzw->chain = chain;
    lzw->eod = 0;
    lzw->early_change = early_change;
    lzw->reverse_bits = reverse_bits;
    lzw->min_bits = min_bits;
    lzw->code_bits = lzw->min_bits;
    lzw->code = -1;
    lzw->next_code = LZW_FIRST(lzw);
    lzw->old_code = -1;
    lzw->rp = lzw->bp;
    lzw->wp = lzw->bp;

    for (i = 0; i < LZW_CLEAR(lzw); i++) {
        lzw->table[i].value = i;
        lzw->table[i].first_char = i;
        lzw->table[i].length = 1;
        lzw->table[i].prev = -1;
    }
    for (i = LZW_CLEAR(lzw); i < NUM_CODES; i++) {
        lzw->table[i].value = 0;
        lzw->table[i].first_char = 0;
        lzw->table[i].length = 0;
        lzw->table[i].prev = -1;
    }
    return fz_new_stream(ctx, lzw, read_lzwd, close_lzwd);
}

#endif

/* TODO: error checking */

enum
{
	MIN_BITS = 9,
	MAX_BITS = 12,
	NUM_CODES = (1 << MAX_BITS),
	LZW_CLEAR = 256,
	LZW_EOD = 257,
	LZW_FIRST = 258,
	MAX_LENGTH = 4097
};

typedef struct lzw_code_s lzw_code;

struct lzw_code_s
{
	int prev;			/* prev code (in string) */
	unsigned short length;		/* string len, including this token */
	unsigned char value;		/* data value */
	unsigned char first_char;	/* first token of string */
};

typedef struct fz_lzwd_s fz_lzwd;

struct fz_lzwd_s
{
	fz_stream *chain;
	int eod;

	int early_change;
    int reverse_bits;

	int code_bits;			/* num bits/code */
	int code;			/* current code */
	int old_code;			/* previously recognized code */
	int next_code;			/* next free entry */

	lzw_code table[NUM_CODES];

	unsigned char bp[MAX_LENGTH];
	unsigned char *rp, *wp;
};

static int
read_lzwd(fz_stream *stm, unsigned char *buf, int len)
{
	fz_lzwd *lzw = stm->state;
	lzw_code *table = lzw->table;
	unsigned char *p = buf;
	unsigned char *ep = buf + len;
	unsigned char *s;
	int codelen;

	int code_bits = lzw->code_bits;
	int code = lzw->code;
	int old_code = lzw->old_code;
	int next_code = lzw->next_code;

	while (lzw->rp < lzw->wp && p < ep)
		*p++ = *lzw->rp++;

	while (p < ep)
	{
		if (lzw->eod)
			return 0;

        if (lzw->reverse_bits)
            code = fz_read_rbits(lzw->chain, code_bits);
        else
            code = fz_read_bits(lzw->chain, code_bits);

		if (fz_is_eof_bits(lzw->chain))
		{
			lzw->eod = 1;
			break;
		}

		if (code == LZW_EOD)
		{
			lzw->eod = 1;
			break;
		}

		if (next_code >= NUM_CODES && code != LZW_CLEAR)
		{
			fz_warn(stm->ctx->warn, "missing clear code in lzw decode");
			code = LZW_CLEAR;
		}

		if (code == LZW_CLEAR)
		{
			code_bits = MIN_BITS;
			next_code = LZW_FIRST;
			old_code = -1;
			continue;
		}

		/* if stream starts without a clear code, old_code is undefined... */
		if (old_code == -1)
		{
			old_code = code;
		}
		else if (code > next_code || next_code >= NUM_CODES)
		{
			fz_warn(stm->ctx->warn, "out of range code encountered in lzw decode");
		}
		else
		{
			/* add new entry to the code table */
			table[next_code].prev = old_code;
			table[next_code].first_char = table[old_code].first_char;
			table[next_code].length = table[old_code].length + 1;
			if (code < next_code)
				table[next_code].value = table[code].first_char;
			else if (code == next_code)
				table[next_code].value = table[next_code].first_char;
			else
				fz_warn(stm->ctx->warn, "out of range code encountered in lzw decode");

			next_code ++;

			if (next_code > (1 << code_bits) - lzw->early_change - 1)
			{
				code_bits ++;
				if (code_bits > MAX_BITS)
					code_bits = MAX_BITS;
			}

			old_code = code;
		}

		/* code maps to a string, copy to output (in reverse...) */
		if (code > 255)
		{
			codelen = table[code].length;
			lzw->rp = lzw->bp;
			lzw->wp = lzw->bp + codelen;

			s = lzw->wp;
			do {
				*(--s) = table[code].value;
				code = table[code].prev;
			} while (code >= 0 && s > lzw->bp);
		}

		/* ... or just a single character */
		else
		{
			lzw->bp[0] = code;
			lzw->rp = lzw->bp;
			lzw->wp = lzw->bp + 1;
		}

		/* copy to output */
		while (lzw->rp < lzw->wp && p < ep)
			*p++ = *lzw->rp++;
	}

	lzw->code_bits = code_bits;
	lzw->code = code;
	lzw->old_code = old_code;
	lzw->next_code = next_code;

	return p - buf;
}

static void
close_lzwd(fz_context *ctx, void *state_)
{
	fz_lzwd *lzw = (fz_lzwd *)state_;
	fz_sync_bits(lzw->chain);
	fz_close(lzw->chain);
	fz_free(ctx, lzw);
}

/* Default: early_change = 1 */
fz_stream *fz_open_lzwd(fz_stream *chain, int early_change, int min_bits, int reverse_bits)
{
	fz_context *ctx = chain->ctx;
	fz_lzwd *lzw = NULL;
	int i;

	fz_var(lzw);

	fz_try(ctx)
	{
		lzw = fz_malloc_struct(ctx, fz_lzwd);
		lzw->chain = chain;
		lzw->eod = 0;
		lzw->early_change = early_change;
        lzw->reverse_bits = reverse_bits;

		for (i = 0; i < 256; i++)
		{
			lzw->table[i].value = i;
			lzw->table[i].first_char = i;
			lzw->table[i].length = 1;
			lzw->table[i].prev = -1;
		}

		for (i = 256; i < NUM_CODES; i++)
		{
			lzw->table[i].value = 0;
			lzw->table[i].first_char = 0;
			lzw->table[i].length = 0;
			lzw->table[i].prev = -1;
		}

		lzw->code_bits = MIN_BITS;
		lzw->code = -1;
		lzw->next_code = LZW_FIRST;
		lzw->old_code = -1;
		lzw->rp = lzw->bp;
		lzw->wp = lzw->bp;
	}
	fz_catch(ctx)
	{
		fz_free(ctx, lzw);
		fz_close(chain);
		fz_rethrow(ctx);
	}

	return fz_new_stream(ctx, lzw, read_lzwd, close_lzwd);
}

/* PNG/TIFF predictor streams */
/* TODO: check if this works with 16bpp images */

enum { MAXC = 32 };

typedef struct fz_predict_s fz_predict;

struct fz_predict_s
{
    fz_stream *chain;

    int predictor;
    int columns;
    int colors;
    int bpc;

    int stride;
    int bpp;
    unsigned char *in;
    unsigned char *out;
    unsigned char *ref;
    unsigned char *rp, *wp;
};

static inline int getcomponent(unsigned char *line, int x, int bpc)
{
    switch (bpc) {
        case 1: return (line[x >> 3] >> ( 7 - (x & 7) ) ) & 1;
        case 2: return (line[x >> 2] >> ( ( 3 - (x & 3) ) << 1 ) ) & 3;
        case 4: return (line[x >> 1] >> ( ( 1 - (x & 1) ) << 2 ) ) & 15;
        case 8: return line[x];
        case 16: return (line[x<<1]<<8)+line[(x<<1)+1];
    }
    return 0;
}

static inline void putcomponent(unsigned char *buf, int x, int bpc, int value)
{
    switch (bpc) {
        case 1: buf[x >> 3] |= value << (7 - (x & 7)); break;
        case 2: buf[x >> 2] |= value << ((3 - (x & 3)) << 1); break;
        case 4: buf[x >> 1] |= value << ((1 - (x & 1)) << 2); break;
        case 8: buf[x] = value; break;
        case 16: buf[x<<1] = value>>8; buf[(x<<1)+1] = value; break;
    }
}

static inline int paeth(int a, int b, int c)
{
    /* The definitions of ac and bc are correct, not a typo. */
    int ac = b - c, bc = a - c, abcc = ac + bc;
    int pa = fz_absi(ac);
    int pb = fz_absi(bc);
    int pc = fz_absi(abcc);
    return pa <= pb && pa <= pc ? a : pb <= pc ? b : c;
}

static void
fz_predict_tiff(fz_predict *state, unsigned char *out, unsigned char *in, int len)
{
    int left[MAXC];
    int i, k;
    const int mask = (1 << state->bpc)-1;

    for (k = 0; k < state->colors; k++)
        left[k] = 0;

	/* special fast case */
	if (state->bpc == 8) {
		for (i = 0; i < state->columns; i++)
			for (k = 0; k < state->colors; k++)
				*out++ = left[k] = (*in++ + left[k]) & 0xFF;
		return;
	}

	/* putcomponent assumes zeroed memory for bpc < 8 */
	if (state->bpc < 8)
    	memset(out, 0, state->stride);

    for (i = 0; i < state->columns; i++) {
        for (k = 0; k < state->colors; k++) {
            int a = getcomponent(in, i * state->colors + k, state->bpc);
            int b = a + left[k];
            int c = b & mask;
            putcomponent(out, i * state->colors + k, state->bpc, c);
            left[k] = c;
        }
    }
}

static void
fz_predict_png(fz_predict *state, unsigned char *out, unsigned char *in, int len, int predictor)
{
    int bpp = state->bpp;
    int i;
    unsigned char *ref = state->ref;

    switch (predictor)
    {
        case 0:
            memcpy(out, in, len);
            break;
        case 1:
            for (i = bpp; i > 0; i--) {
                *out++ = *in++;
            }
            for (i = len - bpp; i > 0; i--) {
                *out = *in++ + out[-bpp];
                out++;
            }
            break;
        case 2:
            for (i = len; i > 0; i--) {
                *out++ = *in++ + *ref++;
            }
            break;
        case 3:
            for (i = bpp; i > 0; i--) {
                *out++ = *in++ + (*ref++) / 2;
            }
            for (i = len - bpp; i > 0; i--) {
                *out = *in++ + (out[-bpp] + *ref++) / 2;
                out++;
            }
            break;
        case 4:
            for (i = bpp; i > 0; i--) {
                *out++ = *in++ + paeth(0, *ref++, 0);
            }
            for (i = len - bpp; i > 0; i --) {
                *out = *in++ + paeth(out[-bpp], *ref, ref[-bpp]);
                ref++;
                out++;
            }
            break;
    }
}

static int
read_predict(fz_stream *stm, unsigned char *buf, int len)
{
    fz_predict *state = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;
    int ispng = state->predictor >= 10;
    int n;

    while (state->rp < state->wp && p < ep)
        *p++ = *state->rp++;

    while (p < ep) {
        n = fz_read(state->chain, state->in, state->stride + ispng);
        if (n == 0)
            return p - buf;

        if (state->predictor == 1) {
            memcpy(state->out, state->in, n);
        } else if (state->predictor == 2) {
            fz_predict_tiff(state, state->out, state->in, n);
        } else {   
            fz_predict_png(state, state->out, state->in + 1, n - 1, state->in[0]);
            memcpy(state->ref, state->out, state->stride);
        }

        state->rp = state->out;
        state->wp = state->out + n - ispng;

        while (state->rp < state->wp && p < ep)
            *p++ = *state->rp++;
    }

    return p - buf;
}

static void
close_predict(fz_context *ctx, void *state_)
{
    fz_predict *state = (fz_predict *)state_;
    fz_close(state->chain);
    fz_free(ctx, state->in);
    fz_free(ctx, state->out);
    fz_free(ctx, state->ref);
    fz_free(ctx, state);
}

/* Default values: predictor = 1, columns = 1, colors = 1, bpc = 8 */
fz_stream *
fz_open_predict(fz_stream *chain, int predictor, int columns, int colors, int bpc)
{
    fz_context *ctx = chain->ctx;
    fz_predict *state = NULL;

    if (predictor < 1)
        predictor = 1;
    if (columns < 1)
        columns = 1;
    if (colors < 1)
        colors = 1;
    if (bpc < 1)
        bpc = 8;

    state = fz_calloc_no_throw(ctx, 1, sizeof(fz_predict));
    state->in = NULL;
    state->out = NULL;
    state->chain = chain;

    state->predictor = predictor;
    state->columns = columns;
    state->colors = colors;
    state->bpc = bpc;

    if (state->predictor != 1 && state->predictor != 2 &&
        state->predictor != 10 && state->predictor != 11 &&
        state->predictor != 12 && state->predictor != 13 &&
        state->predictor != 14 && state->predictor != 15)
    {
        fz_warn(ctx->warn, "invalid predictor: %d", state->predictor);
        state->predictor = 1;
    }

    state->stride = (state->bpc * state->colors * state->columns + 7) / 8;
    state->bpp = (state->bpc * state->colors + 7) / 8;

    state->in = fz_malloc_no_throw(ctx, state->stride + 1);
    state->out = fz_malloc_no_throw(ctx, state->stride);
    state->ref = fz_calloc_no_throw(ctx, 1, state->stride);
    state->rp = state->out;
    state->wp = state->out;
    return fz_new_stream(ctx, state, read_predict, close_predict);
}

