#include "mupdf/fitz.h"
#include <wchar.h>

char *fz_strsep(char **stringp, const char *delim)
{
    char *ret = *stringp;
    if (!ret) return NULL;
    if ((*stringp = strpbrk(*stringp, delim)))
        *((*stringp)++) = '\0';
    return ret;
}

wchar_t *fz_wstrsep(wchar_t **stringp, const wchar_t *delim)
{
    wchar_t *ret = *stringp;
    if (!ret) return NULL;
    if ((*stringp = wcspbrk(*stringp, delim)))
        *((*stringp)++) = '\0';
    return ret;
}

int fz_strlcpy(char *dst, const char *src, int siz)
{
	register char *d = dst;
	register const char *s = src;
	register int n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

int fz_strlcat(char *dst, const char *src, int siz)
{
	register char *d = dst;
	register const char *s = src;
	register int n = siz;
	int dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (*d != '\0' && n-- != 0)
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return dlen + strlen(s);
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return dlen + (s - src);	/* count does not include NUL */
}

char *fz_dirname(const char *path)
{
	static const char null_or_empty_or_noslash[] = ".";
	register char *s;
	register char *last;
	char *first;

	last = s = (char*)path;

	if (!path || !path[0]) {
	    return (char *) null_or_empty_or_noslash;
	}

LOOP:
	while (*s && (*s != '/')) ++s;
	first = s;
	while (*s == '/') ++s;
	if (*s) {
		last = first;
		goto LOOP;
	}

	if (last == path) {
		if (*last != '/') {
	        return (char *) null_or_empty_or_noslash;
		}
		if ((*++last == '/') && (last[1] == 0)) {
				++last;
		}
	}
	*last = 0;
	return (char*)path;
}

#define SEP(x) ((x)=='/' || (x) == 0)

char *fz_cleanname(char *name)
{
	char *p, *q, *dotdot;
	int rooted;

	rooted = name[0] == '/';

	/*
	 * invariants:
	 *		p points at beginning of path element we're considering.
	 *		q points just past the last path element we wrote (no slash).
	 *		dotdot points just past the point where .. cannot backtrack
	 *				any further (no slash).
	 */
	p = q = dotdot = name + rooted;
	while (*p)
	{
		if(p[0] == '/') /* null element */
			p++;
		else if (p[0] == '.' && SEP(p[1]))
			p += 1; /* don't count the separator in case it is nul */
		else if (p[0] == '.' && p[1] == '.' && SEP(p[2]))
		{
			p += 2;
			if (q > dotdot) /* can backtrack */
			{
				while(--q > dotdot && *q != '/')
					;
			}
			else if (!rooted) /* /.. is / but ./../ is .. */
			{
				if (q != name)
					*q++ = '/';
				*q++ = '.';
				*q++ = '.';
				dotdot = q;
			}
		}
		else /* real path element */
		{
			if (q != name+rooted)
				*q++ = '/';
			while ((*q = *p) != '/' && *q != 0)
				p++, q++;
		}
	}

	if (q == name) /* empty string is really "." */
		*q++ = '.';
	*q = '\0';
	return name;
}

static const char *fz_blendmode_names[] =
{
	"Normal",
	"Multiply",
	"Screen",
	"Overlay",
	"Darken",
	"Lighten",
	"ColorDodge",
	"ColorBurn",
	"HardLight",
	"SoftLight",
	"Difference",
	"Exclusion",
	"Hue",
	"Saturation",
	"Color",
	"Luminosity",
};

int fz_lookup_blendmode(char *name)
{
	int i;
	for (i = 0; i < nelem(fz_blendmode_names); i++)
		if (!strcmp(name, fz_blendmode_names[i]))
			return i;
	return FZ_BLEND_NORMAL;
}

char *fz_blendmode_name(unsigned int blendmode)
{
	if (blendmode < nelem(fz_blendmode_names))
		return (char*)fz_blendmode_names[blendmode];
	return "Normal";
}

enum
{
	UTFmax = 6, /* maximum bytes per rune */
	Runesync = 0x80, /* cannot represent part of a UTF sequence (<) */
	Runeself = 0x80, /* rune and UTF sequences are the same (<) */
	Runeerror = 0xFFFD, /* decoding error in UTF */
	Runemax = 0xFFFFFFFF, /* maximum rune value */
};

enum
{
	Bit1 = 7,
	Bitx = 6,
	Bit2 = 5,
	Bit3 = 4,
	Bit4 = 3,
	Bit5 = 2,
	Bit6 = 1,
	Bit7 = 0,

	T1 = ((1<<(Bit1+1))-1) ^ 0xFF, /* 0000 0000 */
	Tx = ((1<<(Bitx+1))-1) ^ 0xFF, /* 1000 0000 */
	T2 = ((1<<(Bit2+1))-1) ^ 0xFF, /* 1100 0000 */
	T3 = ((1<<(Bit3+1))-1) ^ 0xFF, /* 1110 0000 */
	T4 = ((1<<(Bit4+1))-1) ^ 0xFF, /* 1111 0000 */
	T5 = ((1<<(Bit5+1))-1) ^ 0xFF, /* 1111 1000 */
	T6 = ((1<<(Bit6+1))-1) ^ 0xFF, /* 1111 1100 */
	T7 = 0xFE, /* 1111 1110 */

	Rune1 = (1<<(Bit1+0*Bitx))-1, /*                          0000 0000 0111 1111 */
	Rune2 = (1<<(Bit2+1*Bitx))-1, /*                          0000 0111 1111 1111 */
	Rune3 = (1<<(Bit3+2*Bitx))-1, /*                          1111 1111 1111 1111 */
	Rune4 = (1<<(Bit4+3*Bitx))-1, /* 0000 0000 0000 0001 1111 1111 1111 1111 1111 */
	Rune5 = (1<<(Bit5+4*Bitx))-1, /* 0000 0001 1111 1111 1111 1111 1111 1111 1111 */
	Rune6 =           0x7fffffff, /* 1111 1111 1111 1111 1111 1111 1111 1111 1111 */

	Maskx = (1<<Bitx)-1,	/* 0011 1111 */
	Testx = Maskx ^ 0xFF,	/* 1100 0000 */

	Bad = Runeerror,
};

unsigned int fz_chartorune(unsigned int *rune, const char *str)
{
	unsigned int c, c1, c2, c3, c4, c5, l;

	/*
	 * one character sequence
	 *	00000-0007F => T1
	 */
	c = *(const unsigned char*)str;
	if(c < Tx) {
		*rune = c;
		return 1;
	}

	/*
	 * two character sequence
	 *	0080-07FF => T2 Tx
	 */
	c1 = *(const unsigned char*)(str+1) ^ Tx;
	if(c1 & Testx)
		goto bad;
	if(c < T3) {
		if(c < T2)
			goto bad;
		l = ((c << Bitx) | c1) & Rune2;
		if(l <= Rune1)
			goto bad;
		*rune = l;
		return 2;
	}

	/*
	 * three character sequence
	 *	0800-FFFF => T3 Tx Tx
	 */
	c2 = *(const unsigned char*)(str+2) ^ Tx;
	if(c2 & Testx)
		goto bad;
	if(c < T4) {
		l = ((((c << Bitx) | c1) << Bitx) | c2) & Rune3;
		if(l <= Rune2)
			goto bad;
		*rune = l;
		return 3;
	}

	/*
	 * four character sequence (21-bit value)
	 *	10000-1FFFFF => T4 Tx Tx Tx
	 */
	c3 = *(const unsigned char*)(str+3) ^ Tx;
	if (c3 & Testx)
		goto bad;
	if (c < T5) {
		l = ((((((c << Bitx) | c1) << Bitx) | c2) << Bitx) | c3) & Rune4;
		if (l <= Rune3)
			goto bad;
		*rune = l;
		return 4;
	}

	/*
	 * five character sequence (21-bit value)
	 *	10000-1FFFFF => T5 Tx Tx Tx Tx
	 */
	c4 = *(const unsigned char*)(str+4) ^ Tx;
	if (c4 & Testx)
		goto bad;
	if (c < T6) {
		l = ((((((((c << Bitx) | c1) << Bitx) | c2) << Bitx) | c3) << Bitx) | c4) & Rune5;
		if (l <= Rune4)
			goto bad;
		*rune = l;
		return 5;
	}

	/*
	 * six character sequence (21-bit value)
	 *	10000-1FFFFF => T6 Tx Tx Tx Tx Tx
	 */
	c5 = *(const unsigned char*)(str+5) ^ Tx;
	if (c5 & Testx)
		goto bad;
	if (c < T7) {
		l = ((((((((((c << Bitx) | c1) << Bitx) | c2) << Bitx) | c3) << Bitx) | c4) << Bitx) | c5) & Rune6;
		if (l <= Rune5)
			goto bad;
		*rune = l;
		return 6;
	}

	/*
	 * bad decoding
	 */
bad:
	*rune = Bad;
	return 1;
}

unsigned int fz_runetochar(char *str, unsigned int rune)
{
	unsigned int c = rune;

	/*
	 * one character sequence
	 *	00000-0007F => 00-7F
	 */
	if(c <= Rune1) {
		str[0] = c;
		return 1;
	}

	/*
	 * two character sequence
	 *	0080-07FF => T2 Tx
	 */
	if(c <= Rune2) {
		str[0] = T2 | (c >> 1*Bitx);
		str[1] = Tx | (c & Maskx);
		return 2;
	}

	/*
	 * three character sequence
	 *	0800-FFFF => T3 Tx Tx
	 */
	if (c <= Rune3) {
		str[0] = T3 | (c >> 2*Bitx);
		str[1] = Tx | ((c >> 1*Bitx) & Maskx);
		str[2] = Tx | (c & Maskx);
		return 3;
	}

	/*
	 * four character sequence (21-bit value)
	 *	10000-1FFFFF => T4 Tx Tx Tx
	 */
	str[0] = T4 | (c >> 3*Bitx);
	str[1] = Tx | ((c >> 2*Bitx) & Maskx);
	str[2] = Tx | ((c >> 1*Bitx) & Maskx);
	str[3] = Tx | (c & Maskx);
	return 4;
}

unsigned int fz_runelen(unsigned int c)
{
	if(c <= Rune1) {
		return 1;
	}
	if(c <= Rune2) {
		return 2;
	}
	if (c <= Rune3) {
		return 3;
	}
	if (c <= Rune4) {
		return 4;
	}
	if (c <= Rune5) {
		return 5;
	}
	return 6;
}

unsigned int
fz_rune_from_utf16be(unsigned int *out, unsigned char *s, unsigned char *end)
{
	if (s + 2 <= end) {
		unsigned int a = s[0] << 8 | s[1];
		if (a >= 0xD800 && a <= 0xDFFF && s + 4 <= end) {
			unsigned int b = s[2] << 8 | s[3];
			*out = ((a - 0xD800) << 10) + (b - 0xDC00) + 0x10000;
			return 4;
		}
		*out = a;
		return 2;
	}
	*out = 0xFFFD;
	return 1;
}

static inline int iswhite(int c)
{
	return c == ' ' || c == '\n' || c == '\t';
}

int fz_atoi(const char *s)
{
	unsigned int acc = 0, neg = 0;

	if (s == NULL)
		return 0;

    while (iswhite(*s)) s++;
	if ((*s == '-') || (*s == '+')) {
		if (*s == '-') neg = 1;
		s++;
	}
    while (1) {
        unsigned char c = (*s++ - 0x30);
        if(c > 9)
            break;
		acc *= 10;
		acc += c;
	}
	return (neg ? -acc : acc);
}

#if defined(_WIN32) || defined(_WIN64)

char *
fz_utf8_from_utf16le(fz_context *ctx, const uint16_t *s)
{
	const uint16_t *src = s;
	char *d, *dst;
	unsigned int len = 1;

	while (*src) {
		len += fz_runelen(*src++);
	}

	d = fz_malloc(ctx, len+1);
	if (d != NULL) {
		dst = d;
		src = s;
		while (*src) {
			dst += fz_runetochar(dst, *src++);
		}
		*dst = 0;
	}
	return d;
}

uint16_t *
fz_utf16le_from_utf8(fz_context *ctx, const char *s)
{
	unsigned int c, i = 0, len = strlen(s);
    const char *end = s + len;
	uint16_t *r = fz_malloc(ctx, (len + 1) * sizeof(wchar_t));
	if (!r)
		return NULL;
	while (s < end) {
		s += fz_chartorune(&c, s);
		r[i++] = c;
	}
	r[i] = 0;
	return r;
}

#endif

