#include "mupdf/fitz.h"

typedef struct trace_device_s {
    unsigned char pad[1024];
    fz_buffer *buf;
} trace_device;

static void
fz_trace_matrix(trace_device *tdev, const fz_matrix *ctm)
{
	fz_buffer_printf(tdev->buf," matrix=\"%M\"", ctm);
}

static void
fz_trace_color(trace_device *tdev, fz_colorspace *colorspace, float *color, float alpha)
{
	int i;
	fz_buffer_printf(tdev->buf," colorspace=\"%s\" color=\"%g", colorspace->name, color[0]);
	for (i = 1; i < colorspace->n; i++)
		fz_buffer_printf(tdev->buf," %g", color[i]);
	fz_write_buffer_byte(tdev->buf, '\'');
	if (alpha < 1)
		fz_buffer_printf(tdev->buf," alpha=\"%g\"", alpha);
}

static void
fz_trace_path(trace_device *tdev, fz_path *path, int indent)
{
	float x, y;
	int i = 0;
	int n;
	while (i < path->len)
	{
        fz_ensure_buffer(tdev->buf, tdev->buf->len + indent);
		for (n = 0; n < indent; n++) {
			tdev->buf->data[tdev->buf->len++] = ' ';
        }
		switch (path->items[i++].k)
		{
		case FZ_MOVETO:
			x = path->items[i++].v;
			y = path->items[i++].v;
			fz_buffer_printf(tdev->buf,"<moveto x=\"%g\" y=\"%g\"/>\n", x, y);
			break;
		case FZ_LINETO:
			x = path->items[i++].v;
			y = path->items[i++].v;
			fz_buffer_printf(tdev->buf,"<lineto x=\"%g\" y=\"%g\"/>\n", x, y);
			break;
		case FZ_CURVETO:
			x = path->items[i++].v;
			y = path->items[i++].v;
			fz_buffer_printf(tdev->buf,"<curveto x1=\"%g\" y1=\"%g\"", x, y);
			x = path->items[i++].v;
			y = path->items[i++].v;
			fz_buffer_printf(tdev->buf," x2=\"%g\" y2=\"%g\"", x, y);
			x = path->items[i++].v;
			y = path->items[i++].v;
			fz_buffer_printf(tdev->buf," x3=\"%g\" y3=\"%g\"/>\n", x, y);
			break;
		case FZ_CLOSE_PATH:
			fz_write_buffer(tdev->buf, "<closepath/>\n", 13);
			break;
		}
	}
}

static void
fz_trace_begin_page(fz_context *ctx, fz_device *dev, const fz_rect *rect, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf, "<page mediabox=\"%R\" matrix=\"%M\">\n", rect, ctm);
}

static void
fz_trace_end_page(fz_context *ctx, fz_device *dev)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf, "</page>\n", 8);
}

static void
fz_trace_fill_path(fz_context *ctx, fz_device *dev, fz_path *path, int even_odd, const fz_matrix *ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf,"<fill_path", 12);
	if (even_odd)
		fz_write_buffer(tdev->buf," winding=\"eofill\"", 18);
	else
		fz_write_buffer(tdev->buf," winding=\"nonzero\"", 19);
	fz_trace_color(tdev, colorspace, color, alpha);
	fz_trace_matrix(tdev, ctm);
	fz_write_buffer(tdev->buf, ">\n", 2);
	fz_trace_path(tdev, path, 0);
	fz_write_buffer(tdev->buf,"</fill_path>\n", 13);
}

static void
fz_trace_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, fz_stroke_state *stroke, const fz_matrix *ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	int i;

	fz_write_buffer(tdev->buf,"<stroke_path", 14);
	fz_buffer_printf(tdev->buf," linewidth=\"%g\"", stroke->linewidth);
	fz_buffer_printf(tdev->buf," miterlimit=\"%g\"", stroke->miterlimit);
	fz_buffer_printf(tdev->buf," linecap=\"%d,%d,%d\"", stroke->start_cap, stroke->dash_cap, stroke->end_cap);
	fz_buffer_printf(tdev->buf," linejoin=\"%d\"", stroke->linejoin);

	if (stroke->dash_len)
	{
		fz_buffer_printf(tdev->buf," dash_phase=\"%g\" dash=\"", stroke->dash_phase);
		for (i = 0; i < stroke->dash_len; i++)
			fz_buffer_printf(tdev->buf,"%s%g", i > 0 ? " " : "", stroke->dash_list[i]);
		fz_write_buffer_byte(tdev->buf, '\'');
	}

	fz_trace_color(tdev, colorspace, color, alpha);
	fz_trace_matrix(tdev, ctm);
	fz_write_buffer(tdev->buf, ">\n", 2);

	fz_trace_path(tdev, path, 0);

	fz_write_buffer(tdev->buf,"</stroke_path>\n", 15);
}

static void
fz_trace_clip_path(fz_context *ctx, fz_device *dev, fz_path *path, const fz_rect *rect, int even_odd, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<clip_path winding=\"%s\" matrix=\"%M\"", (even_odd ? "eofill" : "nonzero"), ctm);
	if (rect)
		fz_buffer_printf(tdev->buf," contentbbox=\"%g %g %g %g\">\n", rect->x0, rect->y0, rect->x1, rect->y1);
	else
	    fz_write_buffer(tdev->buf, ">\n", 2);
	fz_trace_path(tdev, path, 0);
	fz_write_buffer(tdev->buf,"</clip_path>\n", 13);
}

static void
fz_trace_clip_stroke_path(fz_context *ctx, fz_device *dev, fz_path *path, const fz_rect *rect, fz_stroke_state *stroke, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<clip_stroke_path matrix=\"%M\">\n", ctm);
	fz_trace_path(tdev, path, 0);
	fz_write_buffer(tdev->buf,"</clip_stroke_path>\n", 20);
}

static void
fz_trace_text(fz_context *ctx, trace_device *tdev, fz_text *text)
{
	fz_buffer_printf(tdev->buf," trm=\"%g %g %g %g\">\n",
                     text->trm.x[0], text->trm.x[1], text->trm.x[2], text->trm.x[3]);
	fz_append_text_to_buffer(ctx, tdev->buf, text);
}

static void
fz_trace_fill_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<fill_text font=\"%s\" wmode=\"%d\"", text->font->name, text->wmode);
	fz_trace_color(tdev, colorspace, color, alpha);
	fz_trace_matrix(tdev, ctm);
    fz_trace_text(ctx, tdev, text);
	fz_write_buffer(tdev->buf,"</fill_text>\n", 13);
}

static void
fz_trace_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm,
	fz_colorspace *colorspace, float *color, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<stroke_text font=\"%s\" wmode=\"%d\"", text->font->name, text->wmode);
	fz_trace_color(tdev, colorspace, color, alpha);
	fz_trace_matrix(tdev, ctm);
    fz_trace_text(ctx, tdev, text);
	fz_write_buffer(tdev->buf,"</stroke_text>\n", 15);
}

static void
fz_trace_clip_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm, int accumulate)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<clip_text font=\"%s\" wmode=\"%d\" accumulate=\"%d\" matrix=\"%M\"", text->font->name, text->wmode, accumulate, ctm);
    fz_trace_text(ctx, tdev, text);
	fz_write_buffer(tdev->buf,"</clip_text>\n", 13);
}

static void
fz_trace_clip_stroke_text(fz_context *ctx, fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<clip_stroke_text font=\"%s\" wmode=\"%d\" matrix=\"%M\"", text->font->name, text->wmode, ctm);
    fz_trace_text(ctx, tdev, text);
	fz_write_buffer(tdev->buf,"</clip_stroke_text>\n", 20);
}

static void
fz_trace_ignore_text(fz_context *ctx, fz_device *dev, fz_text *text, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<ignore_text font=\"%s\" wmode=\"%d\" matrix=\"%M\"", text->font->name, text->wmode, ctm);
    fz_trace_text(ctx, tdev, text);
	fz_write_buffer(tdev->buf, "</ignore_text>\n", 15);
}

static void
fz_trace_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<fill_image alpha=\"%g\" matrix=\"%M\" width=\"%d\" height=\"%d\">\n", alpha, ctm, image->w, image->h);
}

static void
fz_trace_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, const fz_matrix *ctm, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<fill_shade alpha=\"%g\" matrix=\"%M\">\n", alpha, ctm);
}

static void
fz_trace_fill_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_matrix *ctm,
fz_colorspace *colorspace, float *color, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf, "<fill_image_mask matrix=\"%M\"", ctm);
	fz_trace_color(tdev, colorspace, color, alpha);
	fz_buffer_printf(tdev->buf," width=\"%d\" height=\"%d\">\n", image->w, image->h);
}

static void
fz_trace_clip_image_mask(fz_context *ctx, fz_device *dev, fz_image *image, const fz_rect *rect, const fz_matrix *ctm)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<clip_image_mask matrix=\"%M\" width=\"%d\" height=\"%d\">\n", image->w, image->h, ctm);
}

static void
fz_trace_pop_clip(fz_context *ctx, fz_device *dev)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf, "<pop_clip/>\n", 12);
}

static void
fz_trace_begin_mask(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int luminosity, fz_colorspace *colorspace, float *color)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<mask bbox=\"%R\" s=\"%s\">\n", bbox, luminosity ? "luminosity" : "alpha");
}

static void
fz_trace_end_mask(fz_context *ctx, fz_device *dev)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf, "</mask>\n", 8);
}

static void
fz_trace_begin_group(fz_context *ctx, fz_device *dev, const fz_rect *bbox, int isolated, int knockout, int blendmode, float alpha)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<group bbox=\"%R\" isolated=\"%d\" knockout=\"%d\" blendmode=\"%s\" alpha=\"%g\">\n",
		             bbox, isolated, knockout, fz_blendmode_name(blendmode), alpha);
}

static void
fz_trace_end_group(fz_context *ctx, fz_device *dev)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf, "</group>\n", 9);
}

static int
fz_trace_begin_tile(fz_context *ctx, fz_device *dev, const fz_rect *area, const fz_rect *view, float xstep, float ystep, const fz_matrix *ctm, int id)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_buffer_printf(tdev->buf,"<tile area=\"%R\" view=\"%R\" xstep=\"%g\" ystep=\"%g\" matrix=\"%M\">\n", area, view, xstep, ystep, ctm);
	return 0;
}

static void
fz_trace_end_tile(fz_context *ctx, fz_device *dev)
{
    trace_device *tdev = (trace_device *)dev->user;
	fz_write_buffer(tdev->buf, "</tile>\n", 8);
}

fz_device *fz_new_trace_device(fz_context *ctx, fz_buffer *buf)
{
	trace_device *tdev = fz_malloc_struct(ctx, trace_device);
	fz_device *dev;

    tdev->buf = buf;
    dev = fz_new_device(ctx, tdev);

	dev->begin_page = fz_trace_begin_page;
	dev->end_page = fz_trace_end_page;

	dev->fill_path = fz_trace_fill_path;
	dev->stroke_path = fz_trace_stroke_path;
	dev->clip_path = fz_trace_clip_path;
	dev->clip_stroke_path = fz_trace_clip_stroke_path;

	dev->fill_text = fz_trace_fill_text;
	dev->stroke_text = fz_trace_stroke_text;
	dev->clip_text = fz_trace_clip_text;
	dev->clip_stroke_text = fz_trace_clip_stroke_text;
	dev->ignore_text = fz_trace_ignore_text;

	dev->fill_shade = fz_trace_fill_shade;
	dev->fill_image = fz_trace_fill_image;
	dev->fill_image_mask = fz_trace_fill_image_mask;
	dev->clip_image_mask = fz_trace_clip_image_mask;

	dev->pop_clip = fz_trace_pop_clip;

	dev->begin_mask = fz_trace_begin_mask;
	dev->end_mask = fz_trace_end_mask;
	dev->begin_group = fz_trace_begin_group;
	dev->end_group = fz_trace_end_group;

	dev->begin_tile = fz_trace_begin_tile;
	dev->end_tile = fz_trace_end_tile;

	return dev;
}

