#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define BI_RGB 0
#define BI_RLE8 1
#define BI_RLE4 2
#define BI_BITFIELDS 3

/************* IMPLEMENTATION HELPERS ******************/

typedef struct _BITMAP
{
    uint32_t bmType, bmWidth, bmHeight, bmWidthBytes;
    uint16_t bmPlanes, bmBitsPixel;
    void *bmBits;
} BITMAP;

typedef struct _BITMAPCOREHEADER
{
    uint32_t bcSize;
    uint32_t bcWidth, bcHeight;
    uint16_t bcPlanes, bcBitCount;
} BITMAPCOREHEADER;

typedef struct _BITMAPINFOHEADER
{
    uint32_t biSize; // 4
    uint32_t biWidth, biHeight; // 12
    uint16_t biPlanes, biBitCount; // 16
    uint32_t biCompression, biSizeImage, biXPelsPerMeter, biYPelsPerMeter; // 32
    uint32_t biClrUsed, biClrImportant; // 40
} BITMAPINFOHEADER;

typedef struct tagCIEXYZ {
        uint32_t ciexyzX, ciexyzY, ciexyzZ;
} __attribute__((packed)) CIEXYZ;

typedef struct _BITMAPV4HEADER {
        uint32_t bV4Size, bV4Width, bV4Height;
        uint16_t bV4Planes, bV4BitCount;
        uint32_t bV4Compression, bV4SizeImage;
        int32_t bV4XPelsPerMeter, bV4YPelsPerMeter;
        uint32_t bV4ClrUsed, bV4ClrImportant;
        uint32_t bV4RedMask, bV4GreenMask, bV4BlueMask, bV4AlphaMask, bV4CSType;
        CIEXYZ ciexyzRed, ciexyzGreen, ciexyzBlue;
        uint32_t bV4Gamma[3];
} __attribute__((packed)) BITMAPV4HEADER;

typedef struct _BITMAPV5HEADER {
        uint32_t bV5Size, bV5Width, bV5Height;
        uint16_t bV5Planes, bV5BitCount;
        uint32_t bV5Compression, bV5SizeImage;
        int32_t bV5XPelsPerMeter, bV5YPelsPerMeter;
        uint32_t bV5ClrUsed, bV5ClrImportant;
        uint32_t bV5RedMask, bV5GreenMask, bV5BlueMask, bV5AlphaMask, bV5CSType;
        CIEXYZ ciexyzRed, ciexyzGreen, ciexyzBlue;
        uint32_t bV5Gamma[3];
        uint32_t bV5Intent, bV5ProfileData, bV5ProfileSize, bV5Reserved;
} __attribute__((packed)) BITMAPV5HEADER;

/* Structures for reading icon/cursor files and resources */
#pragma pack(push,1)
typedef struct _ICONIMAGE
{
    BITMAPINFOHEADER icHeader;      // DIB header
    uint32_t icColors[1];            // Color table
    uint8_t icXOR[1];                  // DIB bits for XOR mask
    uint8_t icAND[1];                  // DIB bits for AND mask
} ICONIMAGE;

typedef struct {
    uint8_t bWidth, bHeight, bColorCount, bReserved;
    uint16_t xHotspot, yHotspot;
    uint32_t dwDIBSize, dwDIBOffset;
} __attribute((packed)) CURSORICONFILEDIRENTRY;

typedef struct
{
    uint16_t idReserved, idType, idCount;
    CURSORICONFILEDIRENTRY  idEntries[1];
} __attribute((packed)) CURSORICONFILEDIR;

typedef struct CursorIconContext
{
    uint32_t width, height, bpp, compr;
    uint8_t *data;
    uint8_t *maskdata;
} CursorIconContext;

/************* IMPLEMENTATION CORE ****************/

static int get_dib_image_size( int width, int height, int depth )
{
    return (((width * depth + 31) / 8) & ~3) * abs( height );
}

static int bitmap_info_size(const BITMAPINFOHEADER *info, uint16_t coloruse)
{
    unsigned int colors, size, masks = 0;

    if (info->biSize == sizeof(BITMAPCOREHEADER)) {
        const BITMAPCOREHEADER *core = (const BITMAPCOREHEADER *)info;
        colors = (core->bcBitCount <= 8) ? 1 << core->bcBitCount : 0;
        return sizeof(BITMAPCOREHEADER) + colors * ((coloruse == 0) ? 3 : 2);
    } else { /* assume BITMAPINFOHEADER */
        colors = info->biClrUsed;
        if (colors > 256) /* buffer overflow otherwise */
            colors = 256;
        if (!colors && (info->biBitCount <= 8))
            colors = 1 << info->biBitCount;
        if (info->biCompression == BI_BITFIELDS) masks = 3;
        size = (sizeof(BITMAPINFOHEADER) + masks * sizeof(uint32_t));
        if (size < info->biSize)
            size = info->biSize;
        return size + colors * ((coloruse == 0) ? 4 : 2);
    }
}

int DIB_GetBitmapInfo(const BITMAPINFOHEADER *header, uint32_t *width, uint32_t *height, uint32_t *bpp, uint32_t *compr, uint32_t *sz)
{
    if (header->biSize == sizeof(BITMAPCOREHEADER)) {
        const BITMAPCOREHEADER *core = (const BITMAPCOREHEADER *)header;
        *width  = core->bcWidth;
        *height = core->bcHeight;
        *bpp    = core->bcBitCount;
        *compr  = 0;
        *sz     = get_dib_image_size(core->bcWidth, core->bcHeight, core->bcBitCount);
        return 0;
    } else if (header->biSize == sizeof(BITMAPINFOHEADER) ||
               header->biSize == sizeof(BITMAPV4HEADER) ||
               header->biSize == sizeof(BITMAPV5HEADER)) {
        *width  = header->biWidth;
        *height = header->biHeight;
        *bpp    = header->biBitCount;
        *compr  = header->biCompression;
        *sz     = header->biSizeImage;
        return 1;
    }
    printf("(%d): unknown/wrong size for header\n", header->biSize );
    return -1;
}

/***********************************************************************/
unsigned int BITMAP_LoadImageW(CursorIconContext *s, unsigned char *pvMapping, unsigned int pvMappingSize)
{
    const BITMAPINFOHEADER *pbmi;
    uint32_t dwOffset = 0, ret = 0;
    int iBMISize;
    uint32_t width, height, bpp, compr, sz;

    /* Map the bitmap info */
    {
        const uint16_t *pbmfh = (unsigned short*)pvMapping;
        if (pbmfh[0] != 0x4d42 /* 'BM' */) {
            printf("Invalid/unsupported bitmap format!\n");
            goto end;
        }
        pbmi = (const BITMAPINFOHEADER*)(pbmfh + 7);
        
        /* Get the image bits */
        if(pbmfh[6])
            dwOffset = pbmfh[6] - 14;
    }
    
    /* Fix up values */
    if(DIB_GetBitmapInfo(pbmi, &width, &height, &bpp, &compr, &sz) == -1)
        goto end;
    if((width > 65535) || (height > 65535))
        goto end;
    
    iBMISize = bitmap_info_size(pbmi, 0);
    
    /* Get a pointer to the image data */
    ret = 1;
    s->data = (char*)pbmi + (dwOffset ? dwOffset : iBMISize);
    s->width = width;
    s->height = height;
    s->bpp = bpp;
    s->compr = compr;
end:
    return ret;
}

/************* PUBLIC FUNCTIONS *******************/

#include <unistd.h>

typedef struct _bmpheader_t
{
    unsigned char magic[2];
    unsigned int data_size;
    unsigned int reserved;
    unsigned int data_offset;
} __attribute__((packed)) bmpheader_t;

void gen_bmpheader(bmpheader_t *h, unsigned int w, unsigned int height, unsigned int bpp)
{
    unsigned int data_size = height * (((w + 31) & ~31) >> 3); /* padding to 32 bit boundary */
    h->magic[0] = 'B';
    h->magic[1] = 'M';
    h->data_size = 40+12+data_size;
    h->reserved = 0;
    h->data_offset = 54 + ((bpp <= 8) ? (4 << bpp) : 0);
}

void ExtractIcon(unsigned char *presbits, unsigned int idx)
{
    CURSORICONFILEDIR* dir = (CURSORICONFILEDIR*)presbits;
    CURSORICONFILEDIRENTRY* entry;
    bmpheader_t h;
    unsigned int wBitCount;
    if (idx >= dir->idCount) {
        fprintf(stderr, "Index %u out of range ([0,%u))\n", idx, dir->idCount);
        return;
    }
    entry = &dir->idEntries[idx];
    if (((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biSize == sizeof(BITMAPCOREHEADER))
        ((BITMAPCOREHEADER *)((char *)dir + entry->dwDIBOffset))->bcHeight >>= 1;
    else
        ((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biHeight >>= 1;
    fprintf(stderr, "entry %u (of %u): dib: offset: %u, size: %u, width: %u, height: %u\n",
           idx, dir->idCount, entry->dwDIBOffset, entry->dwDIBSize, entry->bWidth, entry->bHeight);
    if (((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biSize == sizeof(BITMAPCOREHEADER))
        wBitCount = ((BITMAPCOREHEADER *)((char *)dir + entry->dwDIBOffset))->bcBitCount;
    else
        wBitCount = ((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biBitCount;
    gen_bmpheader(&h, entry->bWidth, entry->bHeight, wBitCount);
    write(1, &h, 14);
    write(1, presbits+entry->dwDIBOffset, entry->dwDIBSize);
}

void EnumerateIconDirectory(unsigned char *presbits)
{
    CURSORICONFILEDIR* dir = (CURSORICONFILEDIR*)presbits;
    CURSORICONFILEDIRENTRY* entry;
    unsigned int i, width, height, wBitCount;

    if(!(dir && !dir->idReserved && (dir->idType & 3))) {
        printf("Invalid resource.\n");
        return;
    }

    /* Find the best match for the desired size */
    for(i = 0; i < dir->idCount; i++) {
        entry = &dir->idEntries[i];
        width = entry->bWidth;
        height = entry->bHeight;
        /* We must get the bitcount from the BITMAPINFOHEADER itself */
        if (((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biSize == sizeof(BITMAPCOREHEADER))
            wBitCount = ((BITMAPCOREHEADER *)((char *)dir + entry->dwDIBOffset))->bcBitCount;
        else
            wBitCount = ((BITMAPINFOHEADER *)((char *)dir + entry->dwDIBOffset))->biBitCount;
        printf("entry %u (of %u): dib: offset: %u, size: %u, width: %u, height: %u, bpp: %u\n",
               i, dir->idCount, entry->dwDIBOffset, entry->dwDIBSize, width, height, wBitCount);
    }
}

#include <stdint.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

uint8_t *slurp_file(const char *path, size_t *filesize) {
    struct stat st;
    uint8_t *ret = NULL;
    int fd = -1;
    stat(path, &st);
    *filesize = st.st_size;
    ret = malloc(st.st_size+1);
    if((fd = open(path, O_RDONLY)) < 0) {
        return NULL;
    }
    read(fd, ret, st.st_size);
    close(fd);
    return ret;
}

int main(int argc, char **argv) {
    size_t sz = 0;
    unsigned char *buf = slurp_file(argv[1], &sz);
    if (argc < 3) {
        EnumerateIconDirectory(buf);
    } else {
        ExtractIcon(buf, atoi(argv[2]));
    }
    return 0;
}

