#include "mupdf/fitz.h"

typedef enum {			/* JPEG marker codes */
  M_SOF0  = 0xc0,
  M_SOF1  = 0xc1,
  M_SOF2  = 0xc2,
  M_SOF3  = 0xc3,
  
  M_SOF5  = 0xc5,
  M_SOF6  = 0xc6,
  M_SOF7  = 0xc7,
  
  M_JPG   = 0xc8,
  M_SOF9  = 0xc9,
  M_SOF10 = 0xca,
  M_SOF11 = 0xcb,
  
  M_SOF13 = 0xcd,
  M_SOF14 = 0xce,
  M_SOF15 = 0xcf,
  
  M_DHT   = 0xc4,
  
  M_DAC   = 0xcc,
  
  M_RST0  = 0xd0,
  M_RST1  = 0xd1,
  M_RST2  = 0xd2,
  M_RST3  = 0xd3,
  M_RST4  = 0xd4,
  M_RST5  = 0xd5,
  M_RST6  = 0xd6,
  M_RST7  = 0xd7,
  
  M_SOI   = 0xd8,
  M_EOI   = 0xd9,
  M_SOS   = 0xda,
  M_DQT   = 0xdb,
  M_DNL   = 0xdc,
  M_DRI   = 0xdd,
  M_DHP   = 0xde,
  M_EXP   = 0xdf,
  
  M_APP0  = 0xe0,
  M_APP1  = 0xe1,
  M_APP2  = 0xe2,
  M_APP3  = 0xe3,
  M_APP4  = 0xe4,
  M_APP5  = 0xe5,
  M_APP6  = 0xe6,
  M_APP7  = 0xe7,
  M_APP8  = 0xe8,
  M_APP9  = 0xe9,
  M_APP10 = 0xea,
  M_APP11 = 0xeb,
  M_APP12 = 0xec,
  M_APP13 = 0xed,
  M_APP14 = 0xee,
  M_APP15 = 0xef,
  
  M_JPG0  = 0xf0,
  M_JPG13 = 0xfd,
  M_COM   = 0xfe,
  
  M_TEM   = 0x01,
  
  M_ERROR = 0x100
} JPEG_MARKER;

/* Private state */
typedef struct _jpeg_marker_reader {
  /* Source of compressed data */
  const unsigned char *buf; /* => next byte to read from buffer */
  size_t buf_remaining;	/* # of bytes remaining in buffer */

  unsigned char saw_SOI;		/* found SOI? */
  unsigned char saw_SOF;		/* found SOF? */
  unsigned int discarded_bytes;	/* # of bytes skipped looking for a marker */

  unsigned int image_width;	/* input image width */
  unsigned int image_height;	/* input image height */
  int data_precision;		/* bits of precision in image data */
  int num_components;		/* # of color components in JPEG image */

  /* This field is shared between entropy decoder and marker parser.
   * It is either zero or the code of a JPEG marker that has been
   * read from the data source, but has not yet been processed.
   */
  unsigned int unread_marker;

  /* These three values are not used by the JPEG code, merely copied */
  /* into the JFIF APP0 marker.  density_unit can be 0 for unknown, */
  /* 1 for dots/inch, or 2 for dots/cm.  Note that the pixel aspect */
  /* ratio is defined by X_density/Y_density even when density_unit=0. */
  uint8_t  density_unit;		/* JFIF code for pixel size units */
  uint16_t X_density;		/* Horizontal pixel density */
  uint16_t Y_density;		/* Vertical pixel density */
} jpeg_marker_reader;

/*
 * Read markers until SOS or EOI.
 *
 * Returns same codes as are defined for jpeg_consume_input:
 * JPEG_SUSPENDED, JPEG_REACHED_SOS, or JPEG_REACHED_EOI.
 */

unsigned int read_markers (fz_context *ctx, jpeg_marker_reader *cinfo)
{
  unsigned int c, c2, length;

  /* Outer loop repeats once for each marker. */
  for (;;) {
    /* Collect the marker proper, unless we already did. */
    /* NB: first_marker() enforces the requirement that SOI appear first. */
    if (cinfo->unread_marker == 0) {
      if (! cinfo->saw_SOI) {
        c = *cinfo->buf++;
        c2 = *cinfo->buf++;
        cinfo->buf_remaining -= 2;
        if (c != 0xFF || c2 != (unsigned int)M_SOI) {
            fz_warn(ctx, "Not a JPEG file: starts with 0x%02x 0x%02x", c, c2);
            //return JPEG_REACHED_EOI;
        }
        cinfo->unread_marker = c2;
      } else {
        for (;;) {
          c = *cinfo->buf++, cinfo->buf_remaining--;
          /* Skip any non-FF bytes.
           * This may look a bit inefficient, but it will not occur in a valid file.
           * We sync after each discarded byte so that a suspending data source
           * can discard the byte from its buffer.
           */
          while (c != 0xFF) {
            cinfo->discarded_bytes++;
            c = *cinfo->buf++, cinfo->buf_remaining--;
          }
          /* This loop swallows any duplicate FF bytes.  Extra FFs are legal as
           * pad bytes, so don't count them in discarded_bytes.  We assume there
           * will not be so many consecutive FF bytes as to overflow a suspending
           * data source's input buffer.
           */
          do {
            c = *cinfo->buf++, cinfo->buf_remaining--;
          } while (c == 0xFF);
          if (c != 0)
            break;			/* found a valid marker, exit loop */
          /* Reach here if we found a stuffed-zero data sequence (FF/00).
           * Discard it and loop back to try again.
           */
          cinfo->discarded_bytes += 2;
        }
        if (cinfo->discarded_bytes != 0) {
          cinfo->discarded_bytes = 0;
        }
        cinfo->unread_marker = c;
      }
    }
    if (cinfo->buf_remaining < 2) {
        return 0;
    }
    length  = ((cinfo->buf[0] << 8) | (cinfo->buf[1]));

    /* At this point cinfo->unread_marker contains the marker code and the
     * input point is just past the marker proper, but before any parameters.
     * A suspension will cause us to return with this state still true.
     */
    switch (cinfo->unread_marker) {
    case M_SOI:
      if (cinfo->saw_SOI) {
          fz_warn(ctx, "Invalid JPEG file structure: two SOI markers");
          return 0;
      }
      cinfo->saw_SOI = 1;
      break;

    case M_SOF0:		/* Baseline */
    case M_SOF1:		/* Extended sequential, Huffman */
    case M_SOF2:		/* Progressive, Huffman */
    case M_SOF5:		/* Differential sequential, Huffman */
    case M_SOF6:		/* Differential progressive, Huffman */
    case M_SOF9:		/* Extended sequential, arithmetic */
    case M_SOF10:		/* Progressive, arithmetic */
    case M_SOF13:		/* Differential sequential, arithmetic */
    case M_SOF14:		/* Differential progressive, arithmetic */
      if (cinfo->buf_remaining < 8) {
        return 0;
      }
      cinfo->data_precision = cinfo->buf[2];
      cinfo->image_height   = ((cinfo->buf[3] << 8) | cinfo->buf[4]);
      cinfo->image_width    = ((cinfo->buf[5] << 8) | cinfo->buf[6]);
      cinfo->num_components = cinfo->buf[7]; 
      if ((length > 0) && (cinfo->buf_remaining > length)) {
        cinfo->buf += (size_t) length;
        cinfo->buf_remaining -= (size_t) length;
      }
      if (cinfo->saw_SOF) {
        fz_warn(ctx, "Invalid JPEG file structure: two SOF markers");
        return 0;
      }
      cinfo->saw_SOF = 1;
      break;

    case M_SOS:
      if ((length > 0) && (cinfo->buf_remaining > length)) {
        cinfo->buf += (size_t) length;
        cinfo->buf_remaining -= (size_t) length;
      }
      cinfo->unread_marker = 0;	/* processed the marker */
      return 1;
    
    case M_EOI:
      cinfo->unread_marker = 0;	/* processed the marker */
      return 0;
      
    /* Currently unsupported SOFn types */
    case M_SOF3:		/* Lossless, Huffman */
    case M_SOF7:		/* Differential lossless, Huffman */
    case M_SOF11:		/* Lossless, arithmetic */
    case M_SOF15:		/* Differential lossless, arithmetic */
      fz_warn(ctx, "Unsupported JPEG process: SOF type 0x%02x", cinfo->unread_marker);
      /* fall through */ 
    case M_JPG:			/* Reserved for JPEG extensions */
    case M_DAC:
    case M_DHT:
    case M_DQT:
    case M_DRI:
    case M_APP1:
    case M_APP2:
    case M_APP3:
    case M_APP4:
    case M_APP5:
    case M_APP6:
    case M_APP7:
    case M_APP8:
    case M_APP9:
    case M_APP10:
    case M_APP11:
    case M_APP12:
    case M_APP13:
    case M_APP14:
    case M_APP15:
    case M_COM:
    case M_DNL:			/* Ignore DNL ... perhaps the wrong thing */
      if ((length > 0) && (cinfo->buf_remaining > length)) {
        cinfo->buf += (size_t) length;
        cinfo->buf_remaining -= (size_t) length;
      }
      break;

    case M_APP0:
      {
        cinfo->buf += 2;
        cinfo->buf_remaining -= 2;
        length -= 2;
        if ((length >= 14) &&
            (((uint32_t*)cinfo->buf)[0] == 0x4649464A)) {
            /* Found JFIF APP0 marker: save info */
            cinfo->density_unit = cinfo->buf[7];
            cinfo->X_density  = ((cinfo->buf[8] << 8) | (cinfo->buf[9]));
            cinfo->Y_density  = ((cinfo->buf[10] << 8) | (cinfo->buf[11]));
        }
        if ((length > 0) && (cinfo->buf_remaining > length)) {
            cinfo->buf += (size_t) length;
            cinfo->buf_remaining -= (size_t) length;
        }
      }
      break;
      
    case M_RST0:		/* these are all parameterless */
    case M_RST1:
    case M_RST2:
    case M_RST3:
    case M_RST4:
    case M_RST5:
    case M_RST6:
    case M_RST7:
    case M_TEM:
      break;

    default:			/* must be DHP, EXP, JPGn, or RESn */
      fz_warn(ctx, "Unsupported marker type 0x%02x", cinfo->unread_marker);
      if ((length > 0) && (cinfo->buf_remaining > length)) {
        cinfo->buf += (size_t) length;
        cinfo->buf_remaining -= (size_t) length;
      }
      break;
    }
    /* Successfully processed marker, so reset state variable */
    cinfo->unread_marker = 0;
  } /* end loop */
}

void
fz_load_jpeg_info(fz_context *ctx, unsigned char *rbuf, int rlen, int *xp, int *yp, int *xresp, int *yresp, fz_colorspace **cspacep)
{
	fz_colorspace *colorspace = fz_device_rgb(ctx);
    jpeg_marker_reader marker;
    memset(&marker, 0, sizeof(jpeg_marker_reader));

	{
        unsigned int retcode = 0;

        /* Reset marker processing state */
        marker.unread_marker = 0;		/* no pending marker */
        marker.saw_SOI = 0;		/* set internal state too */
        marker.saw_SOF = 0;
        marker.discarded_bytes = 0;
        marker.buf = rbuf;
        marker.buf_remaining = rlen;

        retcode = read_markers(ctx, &marker);
        if (!retcode) {
			fz_throw(ctx, FZ_ERROR_GENERIC, "error reading jpeg header info!");
        }

		if (marker.num_components == 1) {
			colorspace = fz_device_gray(ctx);
		} else if (marker.num_components == 3) {
		} else if (marker.num_components == 4) {
            colorspace = fz_device_cmyk(ctx);
		} else {
			fz_warn(ctx, "bad number of components in jpeg: %d", marker.num_components);
        }
        *cspacep = colorspace;

		*xp = marker.image_width;
		*yp = marker.image_height;

		if (marker.density_unit == 1) {
			*xresp = marker.X_density;
			*yresp = marker.Y_density;
		} else if (marker.density_unit == 2) {
			*xresp = marker.X_density * 254 / 100;
			*yresp = marker.Y_density * 254 / 100;
        } else {
			*xresp = 0;
			*yresp = 0;
		}

		if (*xresp <= 0) *xresp = 72;
		if (*yresp <= 0) *yresp = 72;
	}
}

