#include "mupdf/fitz.h"

struct tiff
{
	fz_context *ctx;

	/* "file" */
	unsigned char *bp, *rp, *ep;

	/* byte order */
	unsigned order;

	/* offset of first ifd */
	unsigned ifd_offset;

	/* assorted tags */
	unsigned subfiletype;
	unsigned imagewidth;
	unsigned imagelength;
	unsigned photometric;
	unsigned xresolution;
	unsigned yresolution;
	unsigned resolutionunit;
};

enum
{
	TII = 0x4949, /* 'II' */
	TMM = 0x4d4d, /* 'MM' */
	TBYTE = 1,
	TASCII = 2,
	TSHORT = 3,
	TLONG = 4,
	TRATIONAL = 5
};

#define NewSubfileType 254
#define ImageWidth 256
#define ImageLength 257
#define PhotometricInterpretation 262
#define XResolution 282
#define YResolution 283
#define ResolutionUnit 296

static inline int readbyte(struct tiff *tiff)
{
	if (tiff->rp < tiff->ep)
		return *tiff->rp++;
	return EOF;
}

static inline unsigned readshort(struct tiff *tiff)
{
	unsigned a = readbyte(tiff);
	unsigned b = readbyte(tiff);
	if (tiff->order == TII)
		return (b << 8) | a;
	return (a << 8) | b;
}

static inline unsigned readint(struct tiff *tiff)
{
	unsigned a = readbyte(tiff);
	unsigned b = readbyte(tiff);
	unsigned c = readbyte(tiff);
	unsigned d = readbyte(tiff);
	if (tiff->order == TII)
		return (d << 24) | (c << 16) | (b << 8) | a;
	return (a << 24) | (b << 16) | (c << 8) | d;
}

static void
fz_read_tiff_rational_tag_value(unsigned *p, struct tiff *tiff)
{
	*p = readint(tiff);
	*p = *p / readint(tiff);
}

static void
fz_read_tiff_tag(struct tiff *tiff, unsigned offset)
{
	unsigned tag;
	unsigned type;
	unsigned count;
	unsigned value;

	tiff->rp = tiff->bp + offset;

	tag = readshort(tiff);
	type = readshort(tiff);
	count = readint(tiff);

	if ((type == TBYTE && count <= 4) ||
			(type == TSHORT && count <= 2) ||
			(type == TLONG && count <= 1))
		value = tiff->rp - tiff->bp;
	else
		value = readint(tiff);

	tiff->rp = tiff->bp + value;
	if (tiff->rp > tiff->ep)
	    tiff->rp = tiff->bp;

	switch (tag)
	{
	case NewSubfileType: // LONG
		tiff->subfiletype = readint(tiff);
		break;
	case ImageWidth: // SHORT or LONG
		tiff->imagewidth = ((type == TSHORT) ? readshort(tiff) : readint(tiff));
		break;
	case ImageLength: // SHORT or LONG
		tiff->imagelength = ((type == TSHORT) ? readshort(tiff) : readint(tiff));
		break;
	case PhotometricInterpretation: // SHORT
		tiff->photometric = readshort(tiff);
		break;
	case XResolution: // RATIONAL
		fz_read_tiff_rational_tag_value(&tiff->xresolution, tiff);
		break;
	case YResolution: // RATIONAL
		fz_read_tiff_rational_tag_value(&tiff->yresolution, tiff);
		break;
	case ResolutionUnit: // SHORT
		tiff->resolutionunit = readshort(tiff);
		break;
	default:
		/* fz_warn(tiff->ctx, "unknown tag: %d t=%d n=%d", tag, type, count); */
		break;
	}
}

static unsigned
fz_next_ifd(fz_context *ctx, struct tiff *tiff, unsigned offset)
{
	unsigned count;

	tiff->rp = tiff->bp + offset;

	if (tiff->rp <= tiff->bp || tiff->rp > tiff->ep)
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid IFD offset %u", offset);

	count = readshort(tiff);
	if (count * 12 > (unsigned)(tiff->ep - tiff->rp))
		fz_throw(ctx, FZ_ERROR_GENERIC, "overlarge IFD entry count %u", count);

	tiff->rp += count * 12;
	offset = readint(tiff);
	return offset;
}

static void
fz_seek_ifd(fz_context *ctx, struct tiff *tiff, int subimage)
{
	unsigned offset = tiff->ifd_offset;

	while (subimage--) {
		offset = fz_next_ifd(ctx, tiff, offset);
		if (offset == 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "subimage index %i out of range", subimage);
	}

	tiff->rp = tiff->bp + offset;
	if (tiff->rp < tiff->bp || tiff->rp > tiff->ep)
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid IFD offset %u", tiff->ifd_offset);
}

static void
fz_decode_tiff_ifd(fz_context *ctx, struct tiff *tiff)
{
	unsigned int i, offset = tiff->rp - tiff->bp;
	unsigned int count = readshort(tiff);

	if (count * 12 > (unsigned)(tiff->ep - tiff->rp))
		fz_throw(ctx, FZ_ERROR_GENERIC, "overlarge IFD entry count %u", count);

	offset += 2;
	for (i = 0; i < count; i++) {
		fz_read_tiff_tag(tiff, offset);
		offset += 12;
	}
}

void
fz_load_tiff_info_subimage(fz_context *ctx, unsigned char *buf, int len, int *wp, int *hp, int *xresp, int *yresp, fz_colorspace **cspacep, int subimage)
{
	fz_colorspace *colorspace = fz_device_rgb(ctx);
    unsigned int version;
	struct tiff tiff = { 0 };

	tiff.ctx = ctx;
	tiff.bp = buf;
	tiff.rp = buf;
	tiff.ep = buf + len;

	/* tag defaults, where applicable */
	tiff.xresolution = 96;
	tiff.yresolution = 96;
	tiff.resolutionunit = 2;
	tiff.subfiletype = 0;

	/*
	 * Read IFH
	 */

	/* get byte order marker */
	tiff.order = TII;
	tiff.order = readshort(&tiff);
	if (tiff.order != TII && tiff.order != TMM)
		fz_throw(ctx, FZ_ERROR_GENERIC, "not a TIFF file, wrong magic marker");

	/* check version */
	version = readshort(&tiff);
	if (version != 42)
		fz_throw(ctx, FZ_ERROR_GENERIC, "not a TIFF file, wrong version marker");

	/* get offset of IFD */
	tiff.ifd_offset = readint(&tiff);

	{
		fz_seek_ifd(ctx, &tiff, subimage);
		fz_decode_tiff_ifd(ctx, &tiff);

        if (tiff.resolutionunit == 3) { /* unit conversion needed */
		    tiff.xresolution = tiff.xresolution * 254 / 100;
		    tiff.yresolution = tiff.yresolution * 254 / 100;
        }

		*wp = tiff.imagewidth;
		*hp = tiff.imagelength;
	    /* Note xres and yres could be 0 even if unit was set. If so default to 96dpi. */
		*xresp = (tiff.xresolution ? tiff.xresolution : 96);
		*yresp = (tiff.yresolution ? tiff.yresolution : 96);

        /* from tiff->photometric: 0/1 -> gray (whiteiszero/blackiszero), 2/3 -> rgb (rgb/rgbpal), 5 -> cmyk, 6 -> yuv */
	    switch (tiff.photometric) {
	        case 0: /* WhiteIsZero -- inverted */
	        case 1: /* BlackIsZero */
		        colorspace = fz_device_gray(ctx);
		        break;
	        case 2: /* RGB */
	        case 3: /* RGBPal */
		        colorspace = fz_device_rgb(ctx);
		        break;
	        case 5: /* CMYK */
		        colorspace = fz_device_cmyk(ctx);
		        break;
	        case 6: /* YCbCr */
		        colorspace = fz_device_rgb(ctx);
		        break;
	        default:
		        fz_warn(ctx, "unknown photometric: %d", tiff.photometric);
                break;
	    }
		*cspacep = colorspace;
	}
}

void
fz_load_tiff_info(fz_context *ctx, unsigned char *buf, int len, int *wp, int *hp, int *xresp, int *yresp, fz_colorspace **cspacep)
{
	fz_load_tiff_info_subimage(ctx, buf, len, wp, hp, xresp, yresp, cspacep, 0);
}

