#include "mupdf/fitz.h"

struct info
{
	unsigned int width, height, depth, type, maxval;
	unsigned char *samples;
	unsigned int size;
	int xres, yres;
};

static inline unsigned int getuint(unsigned char *p)
{
    unsigned int *p32 = (unsigned int*)p;
	return av_bswap32(p32[0]);
}

static inline int getcomp(unsigned char *line, int x, int bpc)
{
	switch (bpc)
	{
	case 1: return (line[x >> 3] >> ( 7 - (x & 7) ) ) & 1;
	case 2: return (line[x >> 2] >> ( ( 3 - (x & 3) ) << 1 ) ) & 3;
	case 4: return (line[x >> 1] >> ( ( 1 - (x & 1) ) << 2 ) ) & 15;
	case 8: return line[x];
	case 16: return line[x << 1] << 8 | line[(x << 1) + 1];
	}
	return 0;
}

static inline int pnm_space(int c)
{
    return c == ' ' || c == '\f' || c == '\r' || c == '\t';
}

static void pnm_get(PNMContext *sc, char *str, int buf_size)
{
    char *s = str;
    int c;

    /* skip spaces and comments */
    for (;;) {
        c = *sc->bytestream++;
        if (c == '#')  {
            do {
                c = *sc->bytestream++;
            } while (c != '\n' && sc->bytestream < sc->bytestream_end);
        } else if (!pnm_space(c)) {
            break;
        }
    }

    while (sc->bytestream < sc->bytestream_end && !pnm_space(c)) {
        if ((s - str)  < buf_size - 1)
            *s++ = c;
        c = *sc->bytestream++;
    }
    *s = '\0';
}

int pam_decode_header(PNMContext * const s)
{
    char buf1[32], tuple_type[32];
    int h = 0, w = 0, depth = 0, maxval = 0;

    if (s->type==7) {
        for (;;) {
            pnm_get(s, buf1, sizeof(buf1));
            if (!strcmp(buf1, "WIDTH")) {
                pnm_get(s, buf1, sizeof(buf1));
                w = strtoul(buf1, NULL, 10);
            } else if (!strcmp(buf1, "HEIGHT")) {
                pnm_get(s, buf1, sizeof(buf1));
                h = strtoul(buf1, NULL, 10);
            } else if (!strcmp(buf1, "DEPTH")) {
                pnm_get(s, buf1, sizeof(buf1));
                depth = strtoul(buf1, NULL, 10);
            } else if (!strcmp(buf1, "MAXVAL")) {
                pnm_get(s, buf1, sizeof(buf1));
                maxval = strtoul(buf1, NULL, 10);
            } else if (!strcmp(buf1, "TUPLETYPE")) {
                pnm_get(s, tuple_type, sizeof(tuple_type));
            } else if (!strcmp(buf1, "ENDHDR")) {
                break;
            } else {
                return -1;
            }
        }
        /* check that all tags are present */
        if (w <= 0 || h <= 0 || maxval <= 0 || depth <= 0 || tuple_type[0] == '\0')
            return -1;

        info->width  = w;
        info->height = h;
        info->depth = depth;
        info->maxval = maxval;
        return 0;
    } else {
        return -1;
    }
}

fz_pixmap *fz_load_pnm(fz_context *ctx, unsigned char *p, unsigned int total)
{
    char buf1[32];
	fz_pixmap *image;
	fz_colorspace *colorspace;
	struct info pnm;
    unsigned int i = 2;

    if (p[0] != 'P') {
        fz_warn(ctx, "pnm: first byte of header was not 'P'...\n");
    }
    pnm.type = ((unsigned int)p[1] - 0x30);

    while(p[i] != '\n') i++;
    i++;
    pnm.bytestream = p+i;
    pnm.bytestream_end = p+total-i;

    if (pnm.type==1 || pnm.type==4) {
        pnm.depth = 1;
        pnm.maxval = 1;
    } else if (pnm.type==2 || pnm.type==5) {
        pnm.depth = 1;
    } else if (pnm.type==3 || pnm.type==6) {
        pnm.depth = 3;
    } else if (pnm.type==7) {
        pam_decode_header(&pnm);
        goto decode_data;
    } else {
        fz_warn(ctx, "Unsupported pixel format\n");
        return -1;
    }

    pnm_get(s, buf1, sizeof(buf1));
    pnm.width = atoi(buf1);
    pnm_get(s, buf1, sizeof(buf1));
    pnm.height = atoi(buf1);
    if ((pnm.type != 1) && (pnm.type != 4)) {
        pnm_get(s, buf1, sizeof(buf1));
        pnm.maxval = atoi(buf1);
    }

	/* Prepare output buffer */
	pnm.size = (pnm.width * pnm.height * (pnm.depth+1));
	pnm.samples = fz_malloc(ctx, info->size);

    if (pnm.type == 5 || pnm.type == 6) {
        if (pnm.maxval == 0xFF) {
            memcpy(pnm.samples, pnm.bytestream, pnm.width * pnm.height * pnm.depth);
        } else {
            unsigned int n = pnm.width * pnm.depth;
            unsigned int j, f = (255 * 128 + pnm.maxval / 2) / pnm.maxval;
            for (i = 0; i < pnm.height; i++) {
                for (j = 0; j < n; j++)
                    pnm.samples[i*n+j] = (pnm.bytestream[j] * f + 64) >> 7;
                pnm.bytestream += n;
            }
        }
    } else if (pnm.type == 4) {
        unsigned char *tmp = fz_malloc(ctx, (pnm.size+7)>>3);
        memcpy(tmp, pnm.bytestream, ((pnm.size+7)>>3));
        fz_unpack_file(pnm.samples, tmp, 1, 1, pnm.width, 0);
    } else if (pnm.type == 7) {
    } else {
    }


	r = png_read_image(ctx, &png, p, total);
	if (r < 0) {
		return NULL;
	}

    if (pnm.maxval <= 0 || pnm.maxval > 255) {
        fz_warn(ctx, "pnm: Invalid maxval: %d\n", pnm.maxval);
        pnm.maxval = 255;
    }

	if (pnm.depth == 3 || pnm.depth == 4) {
		colorspace = fz_device_rgb(ctx);
	} else {
        if (pnm.maxval == 1)
            pnm.is_monoblack = 1;
		colorspace = fz_device_gray(ctx);
    }

	stride = (png.width * png.n * png.depth + 7) / 8;

	image = fz_new_pixmap(ctx, colorspace, png.width, png.height);
	image->n = 1 + colorspace->n;
	image->xres = png.xres;
	image->yres = png.yres;

	fz_unpack_tile(image, png.samples, png.n, png.depth, stride, png.indexed);

	if (png.indexed) {
	    fz_pixmap *dst = fz_new_pixmap(ctx, fz_device_rgb(ctx), image->w, image->h);
		image = png_expand_palette(&png, dst, image);
	    fz_drop_pixmap(ctx, image);
	} else if (png.transparency) {
		png_mask_transparency(&png, image);
	}

	if (png.transparency || png.n == 2 || png.n == 4)
		fz_premultiply_pixmap(ctx, image);

	free(png.samples);
	return image;
}

