#include "mupdf/pdf.h"

/* Pretty printing objects */
struct fmt
{
    fz_buffer *buf;
	int indent;
};

static void fmt_obj(struct fmt *fmt, pdf_obj *obj);

static inline int iswhite(int ch)
{
	return ch == '\000' || ch == '\011' || ch == '\012' || ch == '\014' || ch == '\015' || ch == '\040';
}

static inline int isdelim(int ch)
{
	return
		ch == '(' || ch == ')' ||
		ch == '<' || ch == '>' ||
		ch == '[' || ch == ']' ||
		ch == '{' || ch == '}' ||
		ch == '/' ||
		ch == '%';
}

static inline void fmt_indent(struct fmt *fmt)
{
	int i = fmt->indent;
	while (i--) {
		fz_write_buffer_byte(fmt->buf, ' ');
		fz_write_buffer_byte(fmt->buf, ' ');
	}
}

static void fmt_str(struct fmt *fmt, pdf_obj *obj)
{
	char *s = pdf_to_str_buf(obj);
	int n = pdf_to_str_len(obj);
	int i, c;

	fz_write_buffer_byte(fmt->buf, '(');
	for (i = 0; i < n; i++)
	{
		c = (unsigned char)s[i];
        if (strchr("\n\t\r\f\b\\()", c))
	        fz_write_buffer_byte(fmt->buf, '\\');
		if (c == '\n')
	        fz_write_buffer_byte(fmt->buf, 'n');
		else if (c == '\r')
	        fz_write_buffer_byte(fmt->buf, 'r');
		else if (c == '\t')
	        fz_write_buffer_byte(fmt->buf, 't');
		else if (c == '\b')
	        fz_write_buffer_byte(fmt->buf, 'b');
		else if (c == '\f')
	        fz_write_buffer_byte(fmt->buf, 'f');
		else if (c == '(')
	        fz_write_buffer_byte(fmt->buf, '(');
		else if (c == ')')
	        fz_write_buffer_byte(fmt->buf, ')');
		else if (c == '\\')
	        fz_write_buffer_byte(fmt->buf, '\\');
		else if (c < 32) {
		}
		else
	        fz_write_buffer_byte(fmt->buf, c);
	}
	fz_write_buffer_byte(fmt->buf, ')');
}

static void fmt_uint10(struct fmt *out, unsigned int uval)
{
	char buf[25];
	unsigned int i = 0;
	while (uval) {
		buf[i++] = ((uval % 10) + 0x30);
		uval /= 10;
	}
    while (i--) fz_write_buffer_byte(out->buf, buf[i]);
}

static void fmt_hex(struct fmt *fmt, pdf_obj *obj)
{
    static const unsigned char hex[16] = "0123456789abcdef";
	char *s = pdf_to_str_buf(obj);
	unsigned int i, n = pdf_to_str_len(obj);

    fz_write_buffer_byte(fmt->buf, '<');
	for (i = 0; i < n; i++) {
		unsigned char b = (unsigned char) s[i];
		fz_write_buffer_byte(fmt->buf, hex[b>>4]);
		fz_write_buffer_byte(fmt->buf, hex[b&0x0f]);
	}
    fz_write_buffer_byte(fmt->buf, '>');
}

static void fmt_name(struct fmt *fmt, pdf_obj *obj)
{
    static const unsigned char hex[16] = "0123456789ABCDEF";
	unsigned char *s = (unsigned char *) pdf_to_name(obj);
	int i;

	fz_write_buffer_byte(fmt->buf, '/');

	for (i = 0; s[i]; i++)
	{
		if (isdelim(s[i]) || iswhite(s[i]) ||
			s[i] == '#' || s[i] < 32 || s[i] >= 127)
		{
            fz_write_buffer_byte(fmt->buf, '#');
			fz_write_buffer_byte(fmt->buf, hex[s[i] >> 4]);
			fz_write_buffer_byte(fmt->buf, hex[s[i] & 0x0f]);
		}
		else
		{
			fz_write_buffer_byte(fmt->buf, s[i]);
		}
	}
}

static void fmt_array(struct fmt *fmt, pdf_obj *obj)
{
	unsigned int i, n = pdf_array_len(obj);
    fz_write_buffer_byte(fmt->buf, '[');
	for (i = 0; i < n; i++) {
		fmt_obj(fmt, pdf_array_get(obj, i));
        fz_write_buffer_byte(fmt->buf, ' ');
	}
    fz_write_buffer_byte(fmt->buf, ']');
    fz_write_buffer_byte(fmt->buf, '\n');
}

static void fmt_dict(struct fmt *fmt, pdf_obj *obj)
{
	unsigned int i, n = pdf_dict_len(obj);
    fz_write_buffer(fmt->buf, "<<", 2);
	for (i = 0; i < n; i++) {
		fmt_obj(fmt, pdf_dict_get_key(obj, i));
        fz_write_buffer_byte(fmt->buf, ' ');
		fmt_obj(fmt, pdf_dict_get_val(obj, i));
	}
    fz_write_buffer(fmt->buf, ">>", 2);
    fz_write_buffer_byte(fmt->buf, '\n');
}

static void fmt_obj(struct fmt *fmt, pdf_obj *obj)
{
	if (!obj)
		fz_write_buffer(fmt->buf, "<NULL>", 6);
	else if (pdf_is_indirect(obj))
	{
		fmt_uint10(fmt, pdf_to_num(obj)); fz_write_buffer_byte(fmt->buf, ' ');
		fmt_uint10(fmt, pdf_to_gen(obj)); fz_write_buffer_byte(fmt->buf, ' ');
        fz_write_buffer_byte(fmt->buf, 'R');
	}
	else if (pdf_is_null(obj))
		fz_write_buffer(fmt->buf, "null", 4);
	else if (pdf_is_bool(obj))
		fz_buffer_puts(fmt->buf, pdf_to_bool(obj) ? "true" : "false");
	else if (pdf_is_int(obj))
	{
		fmt_uint10(fmt, pdf_to_int(obj));
	}
	else if (pdf_is_real(obj))
	{
		fz_buffer_ftoa(fmt->buf, pdf_to_real(obj));
	}
	else if (pdf_is_string(obj))
	{
		char *str = pdf_to_str_buf(obj);
		int len = pdf_to_str_len(obj);
		int added = 0;
		int i, c;
		for (i = 0; i < len; i++) {
			c = (unsigned char)str[i];
			if (strchr("()\\\n\r\t\b\f", c))
				added ++;
			else if (c < 32)
				added += 3;
		}
		if (added < len)
			fmt_str(fmt, obj);
		else
			fmt_hex(fmt, obj);
	}
	else if (pdf_is_name(obj))
		fmt_name(fmt, obj);
	else if (pdf_is_array(obj))
		fmt_array(fmt, obj);
	else if (pdf_is_dict(obj))
		fmt_dict(fmt, obj);
	else
		fz_buffer_puts(fmt->buf, "<unknown object>");
}

void
pdf_bufprint_obj(fz_buffer *outbuf, pdf_obj *obj)
{
	struct fmt fmt;
	fmt.indent = 0;
	fmt.buf = outbuf;
	fmt_obj(&fmt, obj);
}

void
pdf_bufprint_ref(fz_buffer *outbuf, pdf_obj *ref)
{
	return pdf_bufprint_obj(outbuf, pdf_resolve_indirect(ref));
}

