#include "mupdf/pdf.h"
#include <zlib.h>

/* #define DEBUG_WRITING */

typedef struct pdf_write_options_s pdf_write_options;

struct pdf_write_options_s
{
    fz_buffer *out;
	int do_expand;
	int do_deflate;
	int do_garbage;
	int *use_list;
	int *ofs_list;
	int *gen_list;
	int continue_on_error;
	int *errors;
	int file_len;
	int page_count;
};

void heap_sort(int *list, int n, const int *val, int (*ge)(int, int))
{
	int i, j;

	/* Step 1: Make a heap */
	/* Invariant: Valid heap in [0..i), unsorted elements in [i..n) */
	for (i = 1; i < n; i++) {
		/* Now bubble backwards to maintain heap invariant */
		j = i;
		while (j != 0)
		{
			int tmp;
			int k = (j-1)>>1;
			if (ge(val[list[k]], val[list[j]]))
				break;
			tmp = list[k];
			list[k] = list[j];
			list[j] = tmp;
			j = k;
		}
	}

	/* Step 2: Heap sort */
	/* Invariant: valid heap in [0..i), sorted list in [i..n) */
	/* Initially: i = n */
	for (i = n-1; i > 0; i--) {
		/* Swap the maximum (0th) element from the page_objects into its place
		 * in the sorted list (position i). */
		int tmp = list[0];
		list[0] = list[i];
		list[i] = tmp;
		/* Now, the page_objects is invalid because the 0th element is out
		 * of place. Bubble it until the page_objects is valid. */
		j = 0;
		while (1) {
			/* Children are k and k+1 */
			int k = (j+1)*2-1;
			/* If both children out of the page_objects, we're done */
			if (k > i-1)
				break;
			/* If both are in the page_objects, pick the larger one */
			if (k < i-1 && ge(val[list[k+1]], val[list[k]]))
				k++;
			/* If j is bigger than k (i.e. both of it's children),
			 * we're done */
			if (ge(val[list[j]], val[list[k]]))
				break;
			tmp = list[k];
			list[k] = list[j];
			list[j] = tmp;
			j = k;
		}
	}
}

/*
 * Garbage collect objects not reachable from the trailer.
 */

static pdf_obj *sweepref(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj)
{
	int num = pdf_to_num(obj);
	int gen = pdf_to_gen(obj);
	fz_context *ctx = doc->ctx;

	if (num < 0 || num >= pdf_xref_len(doc))
		return NULL;
	if (opts->use_list[num])
		return NULL;

	opts->use_list[num] = 1;

	/* Bake in /Length in stream objects */
	fz_try(ctx)
	{
		if (pdf_is_stream(doc, num, gen))
		{
			pdf_obj *len = pdf_dict_gets(obj, "Length");
			if (pdf_is_indirect(len))
			{
				opts->use_list[pdf_to_num(len)] = 0;
				len = pdf_resolve_indirect(len);
				pdf_dict_puts(obj, "Length", len);
			}
		}
	}
	fz_catch(ctx)
	{
		//fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
		/* Leave broken */
	}

	return pdf_resolve_indirect(obj);
}

static void sweepobj(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj)
{
	int i;

	if (pdf_is_indirect(obj))
		obj = sweepref(doc, opts, obj);

	if (pdf_is_dict(obj))
	{
		int n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			sweepobj(doc, opts, pdf_dict_get_val(obj, i));
	}

	else if (pdf_is_array(obj))
	{
		int n = pdf_array_len(obj);
		for (i = 0; i < n; i++)
			sweepobj(doc, opts, pdf_array_get(obj, i));
	}
}

/*
 * Make sure we have loaded objects from object streams.
 */

static void preloadobjstms(pdf_document *doc)
{
	pdf_obj *obj;
	int num;
	int xref_len = pdf_xref_len(doc);

	for (num = 0; num < xref_len; num++)
	{
		if (pdf_get_xref_entry(doc, num)->type == 'o')
		{
			obj = pdf_load_object(doc, num, 0);
			pdf_drop_obj(obj);
		}
	}
}

/*
 * Save streams and objects to the output
 */

static fz_buffer *deflatebuf(fz_context *ctx, unsigned char *p, int n)
{
	fz_buffer *buf;
    uLong csize;
    int t;

    buf = fz_new_buffer(compressBound(n));
    csize = buf->cap;
    t = compress(buf->data, &csize, p, n);
    if (t != Z_OK) {
        fz_drop_buffer(buf);
        fz_throw(ctx, FZ_ERROR_GENERIC, "cannot deflate buffer");
    }
    buf->len = csize;
    return buf;
}

static void copystream(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj_orig, int num, int gen)
{
 	fz_buffer *buf, *tmp;
	pdf_obj *obj;

	buf = pdf_load_raw_renumbered_stream(doc, num, gen, num, gen);
	obj = pdf_copy_dict(obj_orig);

    if (opts->do_deflate && !pdf_dict_gets(obj, "Filter")) {
        pdf_dict_puts(obj, "Filter", pdf_new_name(doc, "FlateDecode"));
        tmp = deflatebuf(doc->ctx, buf->data, buf->len);
        fz_drop_buffer(buf);
        buf = tmp;
    }

	fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	pdf_bufprint_obj(opts->out, obj);
	fz_buffer_puts(opts->out, "stream\n");
	fz_write_buffer(opts->out, buf->data, buf->len);
	fz_buffer_puts(opts->out, "endstream\nendobj\n\n");

	fz_drop_buffer(buf);
	pdf_drop_obj(obj);
}

static void expandstream(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj_orig, int num, int gen)
{
 	fz_buffer *buf, *tmp;
	pdf_obj *newlen;
	pdf_obj *obj;
	fz_context *ctx = doc->ctx;
	int truncated = 0;

	buf = pdf_load_renumbered_stream(doc, num, gen, num, gen, (opts->continue_on_error ? &truncated : NULL));
	if (truncated && opts->errors)
		(*opts->errors)++;

	obj = pdf_copy_dict(obj_orig);
	pdf_dict_dels(obj, "Filter");
	pdf_dict_dels(obj, "DecodeParms");

    if (opts->do_deflate && !pdf_dict_gets(obj, "Filter")) {
        pdf_dict_puts(obj, "Filter", pdf_new_name(doc, "FlateDecode"));
        tmp = deflatebuf(ctx, buf->data, buf->len);
        fz_drop_buffer(buf);
        buf = tmp;
    }

	newlen = pdf_new_int(doc, buf->len);
	pdf_dict_puts(obj, "Length", newlen);
	pdf_drop_obj(newlen);

	fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	pdf_bufprint_obj(opts->out, obj);
	fz_buffer_puts(opts->out, "stream\n");
	fz_write_buffer(opts->out, buf->data, buf->len);
	fz_buffer_puts(opts->out, "endstream\nendobj\n\n");

	fz_drop_buffer(buf);
	pdf_drop_obj(obj);
}

static int is_image_filter(char *s)
{
	if (!strcmp(s, "CCITTFaxDecode") || !strcmp(s, "CCF") ||
		!strcmp(s, "DCTDecode") || !strcmp(s, "DCT") ||
		!strcmp(s, "RunLengthDecode") || !strcmp(s, "RL") ||
		!strcmp(s, "JBIG2Decode"))
		return 1;
	return 0;
}

static int filter_implies_image(pdf_document *doc, pdf_obj *o)
{
	if (!o)
		return 0;
	if (pdf_is_name(o))
		return is_image_filter(pdf_to_name(o));
	if (pdf_is_array(o))
	{
		int i, len;
		len = pdf_array_len(o);
		for (i = 0; i < len; i++)
			if (is_image_filter(pdf_to_name(pdf_array_get(o, i))))
				return 1;
	}
	return 0;
}

static void writeobject(pdf_document *doc, pdf_write_options *opts, int num, int gen, int skip_xrefs)
{
	pdf_xref_entry *entry;
	pdf_obj *obj;
	pdf_obj *type;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		obj = pdf_load_object(doc, num, gen);
	}
	fz_catch(ctx)
	{
		if (opts->continue_on_error)
		{
			fz_buffer_printf(opts->out, "%d %d obj\nnull\nendobj\n", num, gen);
			if (opts->errors)
				(*opts->errors)++;
			fz_warn(ctx->warn, "caught error: %d", fz_caught(ctx));
			return;
		}
		else
			fz_rethrow(ctx);
	}

	/* skip ObjStm and XRef objects */
	if (pdf_is_dict(obj))
	{
		type = pdf_dict_gets(obj, "Type");
		if (pdf_is_name(type) && !strcmp(pdf_to_name(type), "ObjStm"))
		{
			opts->use_list[num] = 0;
			pdf_drop_obj(obj);
			return;
		}
		if (skip_xrefs && pdf_is_name(type) && !strcmp(pdf_to_name(type), "XRef"))
		{
			opts->use_list[num] = 0;
			pdf_drop_obj(obj);
			return;
		}
	}

	entry = pdf_get_xref_entry(doc, num);
	if (!pdf_is_stream(doc, num, gen))
	{
		fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	    pdf_bufprint_obj(opts->out, obj);
		fz_buffer_puts(opts->out, "endobj\n\n");
	}
	else if (entry->stm_ofs < 0 && entry->stm_buf == NULL)
	{
		fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	    pdf_bufprint_obj(opts->out, obj);
		fz_buffer_puts(opts->out, "stream\nendstream\nendobj\n\n");
	}
	else
	{
		int dontexpand = 0;
		if (opts->do_expand != 0 && opts->do_expand != fz_expand_all)
		{
			pdf_obj *o;

			if ((o = pdf_dict_gets(obj, "Type"), !strcmp(pdf_to_name(o), "XObject")) &&
				(o = pdf_dict_gets(obj, "Subtype"), !strcmp(pdf_to_name(o), "Image")))
				dontexpand = !(opts->do_expand & fz_expand_images);
			if (o = pdf_dict_gets(obj, "Filter"), filter_implies_image(doc, o))
				dontexpand = !(opts->do_expand & fz_expand_images);
			if (pdf_dict_gets(obj, "Width") != NULL && pdf_dict_gets(obj, "Height") != NULL)
				dontexpand = !(opts->do_expand & fz_expand_images);
		}
		fz_try(ctx)
		{
			if (opts->do_expand && !dontexpand)
				expandstream(doc, opts, obj, num, gen);
			else
				copystream(doc, opts, obj, num, gen);
		}
		fz_catch(ctx)
		{
			if (opts->continue_on_error)
			{
				fz_buffer_printf(opts->out, "%d %d obj\nnull\nendobj\n", num, gen);
				if (opts->errors)
					(*opts->errors)++;
				fz_warn(ctx->warn, "caught error: %d", fz_caught(ctx));
			}
			else
			{
				pdf_drop_obj(obj);
				fz_rethrow(ctx);
			}
		}
	}

	pdf_drop_obj(obj);
}

static void writexrefsubsect(fz_context *ctx, pdf_write_options *opts, int from, int to)
{
	int num;

	fz_buffer_printf(opts->out, "%d %d\n", from, to - from);
	for (num = from; num < to; num++)
	{
		if (opts->use_list[num])
			fz_buffer_printf(opts->out, "%010d %05d n \n", opts->ofs_list[num], opts->gen_list[num]);
		else
			fz_buffer_printf(opts->out, "%010d %05d f \n", opts->ofs_list[num], opts->gen_list[num]);
	}
}

static void buffer_itoa10(fz_buffer *out, unsigned int uval)
{
	char buf[25];
	unsigned int i = 0;
	while (uval) {
		buf[i++] = ((uval % 10) + 0x30);
		uval /= 10;
	}
    while (i--) fz_write_buffer_byte(out, buf[i]);
}

static void writexref(pdf_document *doc, pdf_write_options *opts, int to)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *obj, *nobj = NULL, *trailer = NULL;
	fz_md5 md5;

	fz_buffer_puts(opts->out, "xref\n");
	writexrefsubsect(doc->ctx, opts, 0, to);
	fz_write_buffer_byte(opts->out, '\n');

	fz_try(ctx)
	{
		{
            fz_buffer *idbuf = fz_new_buffer(1024);
            unsigned char md5buf[17];
			trailer = pdf_new_dict(doc, 5);

			nobj = pdf_new_int(doc, to);
			pdf_dict_puts(trailer, "Size", nobj);
			pdf_drop_obj(nobj);
			nobj = NULL;

            buffer_itoa10(idbuf, time(NULL));
            buffer_itoa10(idbuf, to);
			obj = pdf_dict_gets(pdf_trailer(doc), "Info");
	        if (obj)
                pdf_bufprint_obj(idbuf, pdf_resolve_indirect(obj));

			fz_md5_init(&md5);
			fz_md5_update(&md5, idbuf->data, idbuf->len);
			fz_md5_final(&md5, md5buf);
            fz_md5_stringify(idbuf->data, md5buf);
            idbuf->len = 32;

			{
				obj = pdf_dict_gets(pdf_trailer(doc), "Info");
				if (obj)
					pdf_dict_puts(trailer, "Info", obj);

				obj = pdf_dict_gets(pdf_trailer(doc), "Root");
				if (obj)
					pdf_dict_puts(trailer, "Root", obj);

				obj = pdf_dict_gets(pdf_trailer(doc), "ID");
				if (obj) {
                    pdf_obj *obj2 = pdf_new_string(doc, idbuf->data, idbuf->len);
                    pdf_array_insert(obj, obj2, 1);
					pdf_dict_puts(trailer, "ID", obj);
                    pdf_drop_obj(obj2);
                }
			}
            fz_drop_buffer(idbuf);
		}
	}
	fz_always(ctx)
	{
		pdf_drop_obj(nobj);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	fz_buffer_printf(opts->out, "trailer\n");
	pdf_bufprint_obj(opts->out, trailer);
    fz_write_buffer_byte(opts->out, '\n');

	pdf_drop_obj(trailer);

    write(1, opts->out->data, opts->out->len);

	//doc->has_xref_streams = 0;
}

static void
dowriteobject(pdf_document *doc, pdf_write_options *opts, int num, int pass)
{
	pdf_xref_entry *entry = pdf_get_xref_entry(doc, num);
	if (entry->type == 'f')
		opts->gen_list[num] = entry->gen;
	if (entry->type == 'n')
		opts->gen_list[num] = entry->gen;
	if (entry->type == 'o')
		opts->gen_list[num] = 0;

	if (opts->do_garbage && !opts->use_list[num])
		return;

	if (entry->type == 'n' || entry->type == 'o') {
		opts->ofs_list[num] = opts->out->len;
		writeobject(doc, opts, num, opts->gen_list[num], 1);
	} else {
		opts->use_list[num] = 0;
    }
}

static void
writeobjects(pdf_document *doc, pdf_write_options *opts)
{
	int num;
	int xref_len = pdf_xref_len(doc);

	fz_buffer_printf(opts->out, "%%PDF-%d.%d\n", doc->version / 10, doc->version % 10);
	fz_buffer_puts(opts->out, "%%\316\274\341\277\246\n\n");
	for (num = 0; num < xref_len; num++)
		dowriteobject(doc, opts, num, 0);
}

fz_buffer *pdf_write_document(pdf_document *doc, fz_write_options *fz_opts)
{
	int lastfree, num, xref_offs = 0, xref_len = 0;
	pdf_write_options opts = { 0 };
	fz_context *ctx;
	fz_write_options fz_opts_defaults = { 0 };

	if (!doc)
		return NULL;

	if (!fz_opts)
		fz_opts = &fz_opts_defaults;

	ctx = doc->ctx;

	pdf_finish_edit(doc);

	xref_len = pdf_xref_len(doc);

    opts.out = fz_new_buffer(1048576);

	fz_try(ctx)
	{
		opts.do_expand = fz_opts->do_expand;
		opts.do_garbage = fz_opts->do_garbage;
        opts.do_deflate = fz_opts->do_deflate;
		/* We deliberately make these arrays long enough to cope with
		 * 1 to n access rather than 0..n-1, and add space for 2 new
		 * extra entries that may be required for linearization. */
		opts.use_list = fz_malloc_array(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.ofs_list = fz_malloc_array(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.gen_list = fz_calloc(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.continue_on_error = fz_opts->continue_on_error;
		opts.errors = fz_opts->errors;

		for (num = 0; num < xref_len; num++)
		{
			opts.use_list[num] = 0;
			opts.ofs_list[num] = 0;
		}

		/* Make sure any objects hidden in compressed streams have been loaded */
		preloadobjstms(doc);

		/* Sweep & mark objects from the trailer */
		if (opts.do_garbage >= 1)
			sweepobj(doc, &opts, pdf_trailer(doc));
		else
			for (num = 0; num < xref_len; num++)
				opts.use_list[num] = 1;

		writeobjects(doc, &opts);

			/* Construct linked list of free object slots */
			lastfree = 0;
			for (num = 0; num < xref_len; num++)
			{
				if (!opts.use_list[num])
				{
					opts.gen_list[num]++;
					opts.ofs_list[lastfree] = num;
					lastfree = num;
				}
			}

#if 0
        fz_buffer_printf(opts.out, "%d 0 obj\n<< /Type /Catalog\n   /Pages %d 0 R\n>>\nendobj\n",
				         catalog.id, surface->pages_resource.id);
#endif
        xref_offs = opts.out->len;
		writexref(doc, &opts, xref_len);
	    fz_buffer_printf(opts.out, "startxref\n%d\n%%%%EOF\n", xref_offs);
		doc->dirty = 0;
	}
	fz_always(ctx)
	{
		fz_free(ctx, opts.use_list);
		fz_free(ctx, opts.ofs_list);
		fz_free(ctx, opts.gen_list);
		doc->freeze_updates = 0;
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}
    return opts.out;
}

#define KIDS_PER_LEVEL 32

#if 0

// TODO: pdf_rebalance_page_tree(doc);

static pdf_obj *
make_page_tree_node(pdf_document *doc, int l, int r, pdf_obj *parent_ref, int root)
{
	fz_context *ctx = doc->ctx;
	int count_per_kid, spaces;
	pdf_obj *a = NULL;
	pdf_obj *me = NULL;
	pdf_obj *o = NULL;
	pdf_obj *me_ref = NULL;

	count_per_kid = 1;
	while(count_per_kid * KIDS_PER_LEVEL < r-l)
		count_per_kid *= KIDS_PER_LEVEL;

	fz_var(o);
	fz_var(me);
	fz_var(a);
	fz_var(me_ref);

	fz_try(ctx)
	{
		me = pdf_new_dict(doc, 2);
		pdf_dict_puts_drop(me, "Type", pdf_new_name(doc, "Pages"));
		pdf_dict_puts_drop(me, "Count", pdf_new_int(doc, r-l));
		if (!root)
			pdf_dict_puts(me, "Parent", parent_ref);
		a = pdf_new_array(doc, KIDS_PER_LEVEL);
		me_ref = pdf_new_ref(doc, me);

		for (spaces = KIDS_PER_LEVEL; l < r; spaces--)
		{
			if (spaces >= r-l)
			{
				o = pdf_keep_obj(doc->page_refs[l++]);
				pdf_dict_puts(o, "Parent", me_ref);
			}
			else
			{
				int j = l+count_per_kid;
				if (j > r)
					j = r;
				o = make_page_tree_node(doc, l, j, me_ref, 0);
				l = j;
			}
			pdf_array_push(a, o);
			pdf_drop_obj(o);
			o = NULL;
		}
		pdf_dict_puts_drop(me, "Kids", a);
		a = NULL;
	}
	fz_always(ctx)
	{
		pdf_drop_obj(me);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(a);
		pdf_drop_obj(o);
		pdf_drop_obj(me);
		fz_rethrow_message(ctx, "Failed to synthesize new page tree");
	}
	return me_ref;
}

static void
pdf_rebalance_page_tree(pdf_document *doc)
{
	pdf_obj *catalog;
	pdf_obj *pages;

	if (!doc || !doc->needs_page_tree_rebuild)
		return;

	catalog = pdf_dict_gets(pdf_trailer(doc), "Root");
	pages = make_page_tree_node(doc, 0, doc->page_len, catalog, 1);
	pdf_dict_puts_drop(catalog, "Pages", pages);

	doc->needs_page_tree_rebuild = 0;
}

void
pdf_insert_page(pdf_document *doc, pdf_page *page, int at)
{
	if (!doc || !page)
		return;
	if (at < 0)
		at = 0;
	if (at > doc->page_len)
		at = doc->page_len;

	if (doc->page_len + 1 >= doc->page_cap)
	{
		int newmax = doc->page_cap * 2;
		if (newmax == 0)
			newmax = 4;
		doc->page_refs = fz_resize_array(doc->ctx, doc->page_refs, newmax, sizeof(pdf_page *));
		doc->page_objs = fz_resize_array(doc->ctx, doc->page_objs, newmax, sizeof(pdf_page *));
		doc->page_cap = newmax;
	}
	if (doc->page_len > at)
	{
		memmove(&doc->page_objs[at+1], &doc->page_objs[at], doc->page_len - at);
		memmove(&doc->page_refs[at+1], &doc->page_refs[at], doc->page_len - at);
	}

	doc->page_len++;
	doc->page_objs[at] = pdf_keep_obj(page->me);
	doc->page_refs[at] = NULL;
	doc->page_refs[at] = pdf_new_ref(doc, page->me);
	doc->needs_page_tree_rebuild = 1;
}

pdf_page *
pdf_create_page(pdf_document *doc, fz_rect mediabox, int res, int rotate)
{
	pdf_page *page = NULL;
	pdf_obj *pageobj, *obj;
	float userunit = 1;
	fz_context *ctx = doc->ctx;
	fz_matrix ctm, tmp;
	fz_rect realbox;

	page = fz_malloc_struct(ctx, pdf_page);
	obj = NULL;
	fz_var(obj);

	fz_try(ctx)
	{
		page->resources = NULL;
		page->contents = NULL;
		page->transparency = 0;
		page->links = NULL;
		page->me = pageobj = pdf_new_dict(doc, 4);

		pdf_dict_puts_drop(pageobj, "Type", pdf_new_name(doc, "Page"));

		page->mediabox.x0 = fz_min(mediabox.x0, mediabox.x1) * userunit;
		page->mediabox.y0 = fz_min(mediabox.y0, mediabox.y1) * userunit;
		page->mediabox.x1 = fz_max(mediabox.x0, mediabox.x1) * userunit;
		page->mediabox.y1 = fz_max(mediabox.y0, mediabox.y1) * userunit;
		pdf_dict_puts_drop(pageobj, "MediaBox", pdf_new_rect(doc, &page->mediabox));

		/* Snap page->rotate to 0, 90, 180 or 270 */
		if (page->rotate < 0)
			page->rotate = 360 - ((-page->rotate) % 360);
		if (page->rotate >= 360)
			page->rotate = page->rotate % 360;
		page->rotate = 90*((page->rotate + 45)/90);
		if (page->rotate > 360)
			page->rotate = 0;
		pdf_dict_puts_drop(pageobj, "Rotate", pdf_new_int(doc, page->rotate));

		fz_pre_rotate(fz_scale(&ctm, 1, -1), -page->rotate);
		realbox = page->mediabox;
		fz_transform_rect(&realbox, &ctm);
		fz_pre_scale(fz_translate(&tmp, -realbox.x0, -realbox.y0), userunit, userunit);
		fz_concat(&ctm, &ctm, &tmp);
		page->ctm = ctm;
		obj = pdf_new_dict(doc, 4);
		page->contents = pdf_new_ref(doc, obj);
		pdf_drop_obj(obj);
		obj = NULL;
		pdf_dict_puts(pageobj, "Contents", page->contents);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(page->me);
		pdf_drop_obj(obj);
		fz_free(ctx, page);
		fz_rethrow_message(ctx, "Failed to create page");
	}

	return page;
}

#endif

static void
pdf_rebalance_page_tree(pdf_document *doc)
{
}

void pdf_finish_edit(pdf_document *doc)
{
	if (!doc)
		return;
	pdf_rebalance_page_tree(doc);
}

