/*
 * pdfshow -- the ultimate pdf debugging tool
 */

#include "mupdf/pdf.h"

static pdf_document *doc = NULL;
static fz_context *ctx = NULL;
static int showbinary = 0;
static int showdecode = 1;
static int showcolumn;
unsigned char *slurp_file(const char *filename, size_t *filesize);

struct fz_font_context_s {
	int ctx_refs;
};

void fz_new_font_context(fz_context *ctx)
{
	ctx->font = fz_malloc_struct(ctx, fz_font_context);
	ctx->font->ctx_refs = 1;
}

fz_font_context *
fz_keep_font_context(fz_context *ctx)
{
	if (!ctx || !ctx->font)
		return NULL;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	ctx->font->ctx_refs++;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	return ctx->font;
}

void fz_drop_font_context(fz_context *ctx)
{
	int drop;
	if (!ctx || !ctx->font)
		return;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	drop = --ctx->font->ctx_refs;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	if (drop == 0) {
		fz_free(ctx, ctx->font);
    }
}

static void usage(void)
{
	printf("usage: pdfshowstreams [options] file.pdf [object numbers]\n");
	printf("\t-b\tprint streams as binary data\n");
	printf("\t-e\tprint encoded streams (don't decode)\n");
	printf("\t-p\tpassword\n");
}

static void showsafe(unsigned char *buf, int n)
{
	int i;
	for (i = 0; i < n; i++) {
		if (buf[i] == '\n') {
			write(1, "\n", 1);
			showcolumn = 0;
		} else if (buf[i] < 32) {
			write(1, ".", 1);
			showcolumn ++;
		} else {
            write(1, buf, 1);
			showcolumn ++;
		}
		if (showcolumn == 179) {
			write(1, "\n", 1);
			showcolumn = 0;
		}
	}
}

static void showstream(int num, int gen)
{
	fz_stream *stm;
	unsigned char buf[2048];
	int n;

	showcolumn = 0;

	if (showdecode)
		stm = pdf_open_stream(doc, num, gen);
	else
		stm = pdf_open_raw_stream(doc, num, gen);

	while (1)
	{
		n = fz_read(stm, buf, sizeof buf);
		if (n == 0)
			break;
		if (showbinary || !isatty(1))
            write(1, buf, n);
		else
			showsafe(buf, n);
	}

	fz_close(stm);
}

static void showobject(int num, int gen)
{
	pdf_obj *obj = pdf_load_object(doc, num, gen);
	if (pdf_is_stream(doc, num, gen)) {
		showstream(num, gen);
	}
	pdf_drop_obj(obj);
}

int main(int argc, char **argv)
{
	char *password = NULL; /* don't throw errors if encrypted */
	char *filename;
	int c;
    unsigned char *inbuf;
    size_t insz;
    fz_warn_context wctx;
    unsigned int k;

	while ((c = fz_getopt(argc, argv, "p:be")) != -1)
	{
		switch (c)
		{
		case 'p': password = fz_optarg; break;
		case 'b': showbinary = 1; break;
		case 'e': showdecode = 0; break;
		case '?': fz_warn(&wctx, "%s: unknown option -%c\n", argv[0], c); usage(); break;
        case ':': fz_warn(&wctx, "%s: option requires argument -%c\n", argv[0], c); usage(); break;
		default: usage(); return 1; break;
		}
	}

	if (fz_optind == argc) {
		usage();
        return 1;
    }

	filename = argv[fz_optind++];

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_UNLIMITED);
	if (!ctx) {
		write(2, "Cannot initialise context\n", 27);
        return 1;
	}

	fz_try(ctx)
	{
        inbuf = slurp_file(filename, &insz);
	    doc = pdf_open_document_no_run(ctx, inbuf, insz);
		if (pdf_needs_password(doc))
			if (!pdf_authenticate_password(doc, password))
				fz_warn(ctx->warn, "cannot authenticate password: %s", filename);

		if (fz_optind == argc) {
            for (k = 0; k < pdf_xref_len(doc); k++) {
                showobject(k, 0);
            }
        } else {
		    while (fz_optind < argc) {
                showobject(fz_atoi(argv[fz_optind++]), 0);
            }
		}
	}
	fz_catch(ctx)
	{
	}

	pdf_close_document(doc);
	fz_free_context(ctx);
	return 0;
}

