#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "mupdf/pdf.h"
#undef FLT_MIN
#undef FLT_MAX
#undef FLT_EPSILON
#define FLT_MIN 1.17549435082228750797e-38F
#define FLT_MAX 3.40282346638528859812e+38F
#define FLT_EPSILON 0x1p-23f
#define FF_ARRAY_ELEMS(a) (sizeof(a) / sizeof((a)[0]))

typedef struct psobj_s psobj;

#ifdef TEST

#define FZ_FN_MAXM 32
#define FZ_FN_MAXN 32
#define DIV_BY_ZERO(a, b, min, max) (((a) < 0) ^ ((b) < 0) ? (min) : (max))

typedef struct fz_function_s
{
    int m, n;
} fz_function;

typedef struct pdf_function_s
{
	fz_function base;
	int type;				/* 0=sample 2=exponential 3=stitching 4=postscript */
	float range[FZ_FN_MAXN][2];	/* even index : min value, odd index : max value */
	int has_range;
	psobj *psobj_code;
	int psobj_cap;
} pdf_function;

static inline float fz_clamp(float val, float min, float max)
{
    if (val < min) val = min;
    else if (val > max) val = max;
    return val;
}

#else

struct pdf_function_s
{
	fz_function base;
	int type;				/* 0=sample 2=exponential 3=stitching 4=postscript */
	float domain[FZ_FN_MAXM][2];	/* even index : min value, odd index : max value */
	float range[FZ_FN_MAXN][2];	/* even index : min value, odd index : max value */
	int has_range;

	union
	{
		struct {
			unsigned short bps;
			int size[FZ_FN_MAXM];
			float encode[FZ_FN_MAXM][2];
			float decode[FZ_FN_MAXN][2];
			float *samples;
		} sa;

		struct {
			float n;
			float c0[FZ_FN_MAXN];
			float c1[FZ_FN_MAXN];
		} e;

		struct {
			int k;
			fz_function **funcs; /* k */
			float *bounds; /* k - 1 */
			float *encode; /* k * 2 */
		} st;
	} u;
	psobj *psobj_code;
	int psobj_cap;
};

#endif

#define RADIAN 57.2957795f

static inline float lerp(float x, float xmin, float xmax, float ymin, float ymax)
{
	if (xmin == xmax)
		return ymin;
	if (ymin == ymax)
		return ymin;
	return ymin + (x - xmin) * (ymax - ymin) / (xmax - xmin);
}

static unsigned int
ps_idiv(unsigned int num, unsigned int den)
{
    unsigned int quo = 0, bit = 0;

    if (den > num)
        return 0;

    /* normalize to make den >= num, but not overflow */
    while ((den < num) && !(den >> 31)) { bit++; den <<= 1; }

    /* generate quotient, one bit at a time */
    do {
        if (den <= num) { num -= den; quo += (1U << bit); }
        den >>= 1;
    } while(bit--);
    return quo;
}

static unsigned int
ps_imod(unsigned int num, unsigned int den)
{
    unsigned int bit = 0;

    if (den > num)
        return num;

    /* normalize to make den >= num, but not overflow */
    while ((den < num) && !(den >> 31)) { bit++; den <<= 1; }

    /* generate remainder, one bit at a time */
    do {
        if (den <= num) { num -= den; }
        den >>= 1;
    } while(bit--);
    return num;
}

/*
 * PostScript calculator
 */

enum { PS_BOOL, PS_INT, PS_REAL, PS_OPERATOR, PS_BLOCK };

enum
{
	PS_OP_ABS, PS_OP_ADD, PS_OP_AND, PS_OP_ATAN, PS_OP_BITSHIFT,
	PS_OP_CEILING, PS_OP_COPY, PS_OP_COS, PS_OP_CVI, PS_OP_CVR,
	PS_OP_DIV, PS_OP_DUP, PS_OP_EQ, PS_OP_EXCH, PS_OP_EXP,
	PS_OP_FALSE, PS_OP_FLOOR, PS_OP_GE, PS_OP_GT, PS_OP_IDIV, PS_OP_IF,
	PS_OP_IFELSE, PS_OP_INDEX, PS_OP_LE, PS_OP_LN, PS_OP_LOG, PS_OP_LT,
	PS_OP_MOD, PS_OP_MUL, PS_OP_NE, PS_OP_NEG, PS_OP_NOT, PS_OP_OR,
	PS_OP_POP, PS_OP_RETURN, PS_OP_ROLL, PS_OP_ROUND, PS_OP_SIN,
	PS_OP_SQRT, PS_OP_SUB, PS_OP_TRUE, PS_OP_TRUNCATE, PS_OP_XOR
};

static char *ps_op_names[] =
{
	"abs", "add", "and", "atan", "bitshift", "ceiling", "copy",
	"cos", "cvi", "cvr", "div", "dup", "eq", "exch", "exp",
	"false", "floor", "ge", "gt", "idiv", "if", "ifelse", "index", "le", "ln",
	"log", "lt", "mod", "mul", "ne", "neg", "not", "or", "pop", "return",
	"roll", "round", "sin", "sqrt", "sub", "true", "truncate", "xor"
};

struct psobj_s
{
	int type;
	union
	{
		int i;				/* integer (stack and code) or boolean (code only) */
		float f;			/* real (stack and code) */
		int op;				/* operator (code only) */
		int block;			/* if/ifelse block pointer (code only) */
	} u;
};

typedef struct ps_stack_s ps_stack;

struct ps_stack_s
{
	psobj stack[100];
	int sp;
};

#ifndef NDEBUG
void
pdf_debug_ps_stack(ps_stack *st)
{
	int i;

	printf("stack: ");

	for (i = 0; i < st->sp; i++)
	{
		switch (st->stack[i].type)
		{
		case PS_BOOL:
			if (st->stack[i].u.i)
				printf("true ");
			else
				printf("false ");
			break;

		case PS_INT:
			printf("%d ", st->stack[i].u.i);
			break;

		case PS_REAL:
			printf("%g ", st->stack[i].u.f);
			break;
		}
	}
	printf("\n");

}
#endif

static inline float rcpss(float in)
{
    float out;
    __asm__ __volatile__("rcpss %1, %0\n" :"=x"(out) :"x"(in));
    return (out*(2.0f - out*in));
}

static void
ps_init_stack(ps_stack *st)
{
	memset(st->stack, 0, sizeof(st->stack));
	st->sp = 0;
}

static inline int ps_overflow1(ps_stack *st)
{
	return st->sp + 1 >= FF_ARRAY_ELEMS(st->stack);
}

static inline int ps_underflow1(ps_stack *st)
{
	return st->sp <= 0;
}

static inline int ps_underflow2(ps_stack *st)
{
	return st->sp - 2 < 0;
}

static inline int ps_overflow(ps_stack *st, int n)
{
	return n < 0 || st->sp + n >= FF_ARRAY_ELEMS(st->stack);
}

static inline int ps_underflow(ps_stack *st, int n)
{
	return n < 0 || st->sp - n < 0;
}

static inline int ps_is_type(ps_stack *st, int t)
{
	return !ps_underflow1(st) && st->stack[st->sp - 1].type == t;
}

static inline int ps_is_type2(ps_stack *st, int t)
{
	return !ps_underflow2(st) && st->stack[st->sp - 1].type == t && st->stack[st->sp - 2].type == t;
}

static void
ps_push_bool(ps_stack *st, int b)
{
	if (!ps_overflow(st, 1))
	{
		st->stack[st->sp].type = PS_BOOL;
		st->stack[st->sp].u.i = b;
		st->sp++;
	}
}

static void
ps_push_int(ps_stack *st, int n)
{
	if (!ps_overflow1(st))
	{
		st->stack[st->sp].type = PS_INT;
		st->stack[st->sp].u.i = n;
		st->sp++;
	}
}

static void
ps_push_real(ps_stack *st, float n)
{
	if (!ps_overflow1(st))
	{
		st->stack[st->sp].type = PS_REAL;
		if (n != n)
		{
			/* Push 1.0, as it's a small known value that won't
			 * cause a divide by 0. Same reason as in fz_atof. */
			n = 1.0;
		}
		st->stack[st->sp].u.f = fz_clamp(n, -FLT_MAX, FLT_MAX);
		st->sp++;
	}
}

static int
ps_pop_bool(ps_stack *st)
{
	if (!ps_underflow(st, 1))
	{
		if (ps_is_type(st, PS_BOOL))
			return st->stack[--st->sp].u.i;
	}
	return 0;
}

static int
ps_pop_int(ps_stack *st)
{
	if (!ps_underflow1(st))
	{
		if (ps_is_type(st, PS_INT))
			return st->stack[--st->sp].u.i;
		if (ps_is_type(st, PS_REAL))
			return st->stack[--st->sp].u.f;
	}
	return 0;
}

static float
ps_pop_real(ps_stack *st)
{
	if (!ps_underflow1(st))
	{
		if (ps_is_type(st, PS_INT))
			return st->stack[--st->sp].u.i;
		if (ps_is_type(st, PS_REAL))
			return st->stack[--st->sp].u.f;
	}
	return 0;
}

static void
ps_copy(ps_stack *st, int n)
{
	if (!ps_underflow(st, n) && !ps_overflow(st, n))
	{
		memcpy(st->stack + st->sp, st->stack + st->sp - n, n * sizeof(psobj));
		st->sp += n;
	}
}

static void
ps_roll2(ps_stack *st)
{
	psobj tmp;

	if (ps_underflow2(st))
		return;

	tmp = st->stack[st->sp - 1];
	memmove(st->stack + st->sp - 1, st->stack + st->sp - 2, 2 * sizeof(psobj));
	st->stack[st->sp - 2] = tmp;
}

static void
ps_roll(ps_stack *st, int n, int j)
{
	psobj tmp;
	int i;
    unsigned int k = ((j < 0) ? -j : j);

	if (ps_underflow(st, n) || j == 0 || n == 0)
		return;

    k = ps_imod(k, n);
	if (j >= 0)
	{
        j = k;
	}
	else
	{
        j = k;
		if (j != 0)
			j = n - j;
	}

	for (i = 0; i < j; i++)
	{
		tmp = st->stack[st->sp - 1];
		memmove(st->stack + st->sp - n + 1, st->stack + st->sp - n, n * sizeof(psobj));
		st->stack[st->sp - n] = tmp;
	}
}

static void
ps_index(ps_stack *st, int n)
{
	if (!ps_overflow1(st) && !ps_underflow(st, n))
	{
		st->stack[st->sp] = st->stack[st->sp - n - 1];
		st->sp++;
	}
}
void pdf_debug_postscript_func(psobj *funccode, psobj *code, int level);

static void
ps_run(fz_context *ctx, psobj *code, ps_stack *st, int pc)
{
	int i1, i2;
	float r1, r2;
	int b1, b2;

	while (1)
	{
		switch (code[pc].type)
		{
		case PS_INT:
			ps_push_int(st, code[pc++].u.i);
			break;

		case PS_REAL:
			ps_push_real(st, code[pc++].u.f);
			break;

		case PS_OPERATOR:
			switch (code[pc++].u.op)
			{
			case PS_OP_ABS:
				if (ps_is_type(st, PS_INT)) {
                    int t = ps_pop_int(st);
					ps_push_int(st, ((t < 0) ? -t : t));
				} else {
					ps_push_real(st, fabsf(ps_pop_real(st)));
                }
				break;

			case PS_OP_ADD:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 + i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_real(st, r1 + r2);
				}
				break;

			case PS_OP_AND:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 & i2);
				}
				else {
					b2 = ps_pop_bool(st);
					b1 = ps_pop_bool(st);
					ps_push_bool(st, b1 && b2);
				}
				break;

			case PS_OP_ATAN:
				r2 = ps_pop_real(st);
				r1 = ps_pop_real(st);
				r1 = atan2f(r1, r2) * RADIAN;
				if (r1 < 0)
					r1 += 360;
				ps_push_real(st, r1);
				break;

			case PS_OP_BITSHIFT:
				i2 = ps_pop_int(st);
				i1 = ps_pop_int(st);
				if (i2 > 0 && i2 < 8 * sizeof (i2))
					ps_push_int(st, i1 << i2);
				else if (i2 < 0 && i2 > -8 * (int)sizeof (i2))
					ps_push_int(st, (int)((unsigned int)i1 >> -i2));
				else
					ps_push_int(st, i1);
				break;

			case PS_OP_CEILING:
				r1 = ps_pop_real(st);
				ps_push_real(st, ceilf(r1));
				break;

			case PS_OP_COPY:
				ps_copy(st, ps_pop_int(st));
				break;

			case PS_OP_COS:
				r1 = ps_pop_real(st);
				ps_push_real(st, cosf(r1/RADIAN));
				break;

			case PS_OP_CVI:
				ps_push_int(st, ps_pop_int(st));
				break;

			case PS_OP_CVR:
				ps_push_real(st, ps_pop_real(st));
				break;

			case PS_OP_DIV:
				r2 = ps_pop_real(st);
				r1 = ps_pop_real(st);
				//ps_push_real(st, r1 / r2);
				ps_push_real(st, r1 * rcpss(r2));
				break;

			case PS_OP_DUP:
				ps_copy(st, 1);
				break;

			case PS_OP_EQ:
				if (ps_is_type2(st, PS_BOOL)) {
					b2 = ps_pop_bool(st);
					b1 = ps_pop_bool(st);
					ps_push_bool(st, b1 == b2);
				}
				else if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 == i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 == r2);
				}
				break;

			case PS_OP_EXCH:
				ps_roll2(st);
				break;

			case PS_OP_EXP:
				r2 = ps_pop_real(st);
				r1 = ps_pop_real(st);
				ps_push_real(st, powf(r1, r2));
				break;

			case PS_OP_FALSE:
				ps_push_bool(st, 0);
				break;

			case PS_OP_FLOOR:
				r1 = ps_pop_real(st);
				ps_push_real(st, floorf(r1));
				break;

			case PS_OP_GE:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 >= i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 >= r2);
				}
				break;

			case PS_OP_GT:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 > i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 > r2);
				}
				break;

			case PS_OP_IDIV:
				i2 = ps_pop_int(st);
				i1 = ps_pop_int(st);
				ps_push_int(st, 0);
#if 0
				if (i2 != 0)
					ps_push_int(st, ps_idiv(i1, i2));
				else
					ps_push_int(st, DIV_BY_ZERO(i1, i2, INT32_MIN, INT32_MAX));
#endif
				break;

			case PS_OP_INDEX:
				ps_index(st, ps_pop_int(st));
				break;

			case PS_OP_LE:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 <= i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 <= r2);
				}
				break;

			case PS_OP_LN:
				r1 = ps_pop_real(st);
				/* Bug 692941 - logf as separate statement */
				r2 = logf(r1);
				ps_push_real(st, r2);
				break;

			case PS_OP_LOG:
				r1 = ps_pop_real(st);
				ps_push_real(st, log10f(r1));
				break;

			case PS_OP_LT:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 < i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 < r2);
				}
				break;

			case PS_OP_MOD:
				i2 = ps_pop_int(st);
				i1 = ps_pop_int(st);
				if (i2 != 0)
					ps_push_int(st, ps_imod(i1, i2));
				else
					ps_push_int(st, DIV_BY_ZERO(i1, i2, INT32_MIN, INT32_MAX));
				break;

			case PS_OP_MUL:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 * i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_real(st, r1 * r2);
				}
				break;

			case PS_OP_NE:
				if (ps_is_type2(st, PS_BOOL)) {
					b2 = ps_pop_bool(st);
					b1 = ps_pop_bool(st);
					ps_push_bool(st, b1 != b2);
				}
				else if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_bool(st, i1 != i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_bool(st, r1 != r2);
				}
				break;

			case PS_OP_NEG:
				if (ps_is_type(st, PS_INT))
					ps_push_int(st, -ps_pop_int(st));
				else
					ps_push_real(st, -ps_pop_real(st));
				break;

			case PS_OP_NOT:
				if (ps_is_type(st, PS_BOOL))
					ps_push_bool(st, !ps_pop_bool(st));
				else
					ps_push_int(st, ~ps_pop_int(st));
				break;

			case PS_OP_OR:
				if (ps_is_type2(st, PS_BOOL)) {
					b2 = ps_pop_bool(st);
					b1 = ps_pop_bool(st);
					ps_push_bool(st, b1 || b2);
				}
				else {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 | i2);
				}
				break;

			case PS_OP_POP:
				if (!ps_underflow1(st))
					st->sp--;
				break;

			case PS_OP_ROLL:
				i2 = ps_pop_int(st);
				i1 = ps_pop_int(st);
				ps_roll(st, i1, i2);
				break;

			case PS_OP_ROUND:
				if (!ps_is_type(st, PS_INT)) {
					r1 = ps_pop_real(st);
					ps_push_real(st, rintf(r1));
				}
				break;

			case PS_OP_SIN:
				r1 = ps_pop_real(st);
				ps_push_real(st, sinf(r1/RADIAN));
				break;

			case PS_OP_SQRT:
				r1 = ps_pop_real(st);
				ps_push_real(st, sqrtf(r1));
				break;

			case PS_OP_SUB:
				if (ps_is_type2(st, PS_INT)) {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 - i2);
				}
				else {
					r2 = ps_pop_real(st);
					r1 = ps_pop_real(st);
					ps_push_real(st, r1 - r2);
				}
				break;

			case PS_OP_TRUE:
				ps_push_bool(st, 1);
				break;

			case PS_OP_TRUNCATE:
				if (!ps_is_type(st, PS_INT)) {
					r1 = ps_pop_real(st);
					ps_push_real(st, truncf(r1));
				}
				break;

			case PS_OP_XOR:
				if (ps_is_type2(st, PS_BOOL)) {
					b2 = ps_pop_bool(st);
					b1 = ps_pop_bool(st);
					ps_push_bool(st, b1 ^ b2);
				}
				else {
					i2 = ps_pop_int(st);
					i1 = ps_pop_int(st);
					ps_push_int(st, i1 ^ i2);
				}
				break;

			case PS_OP_IF:
				b1 = ps_pop_bool(st);
				if (b1)
					ps_run(ctx, code, st, code[pc + 1].u.block);
				pc = code[pc + 2].u.block;
				break;

			case PS_OP_IFELSE:
				b1 = ps_pop_bool(st);
				if (b1)
					ps_run(ctx, code, st, code[pc + 1].u.block);
				else
					ps_run(ctx, code, st, code[pc + 0].u.block);
				pc = code[pc + 2].u.block;
				break;

			case PS_OP_RETURN:
				return;

			default:
				fz_warn(ctx->warn, "foreign operator in calculator function");
				return;
			}
			break;

		default:
			fz_warn(ctx->warn, "foreign object in calculator function");
			return;
		}
	}
}

void pdf_eval_postscript_func(fz_context *ctx, pdf_function *func, float *in, float *out)
{
	ps_stack st;
	float x;
	int i;

	ps_init_stack(&st);

	for (i = 0; i < func->base.m; i++) {
		ps_push_real(&st, in[i]);
	}

	ps_run(ctx, func->psobj_code, &st, 0);

	for (i = func->base.n - 1; i >= 0; i--) {
		x = ps_pop_real(&st);
		out[i] = fz_clamp(x, func->range[i][0], func->range[i][1]);
	}
}

typedef struct psfunc_lexbuf_s
{
	char buffer[256];
	char *scratch;
    unsigned int size, len;
} psfunc_lexbuf;

#define PS_TOK_EOF 0
#define PS_TOK_REAL 1
#define PS_TOK_INT 2
#define PS_TOK_KEYWORD 3
#define PS_TOK_TRUE 4
#define PS_TOK_FALSE 5
#define PS_TOK_OPEN_BRACE '{'
#define PS_TOK_CLOSE_BRACE '}'

static void psfunc_lexbuf_init(psfunc_lexbuf *lb, unsigned int size)
{
	lb->size = size;
	lb->len = 0;
	lb->scratch = &lb->buffer[0];
}

static void
resize_code(pdf_function *func, unsigned int newsize)
{
	if (newsize >= func->psobj_cap) {
		unsigned int new_cap = func->psobj_cap + 64;
		func->psobj_code = realloc(func->psobj_code, new_cap*sizeof(psobj));
		func->psobj_cap = new_cap;
	}
}

#define IS_NUMBER \
	'.':case'0':case'1':case'2':case'3':\
	case'4':case'5':case'6':case'7':case'8':case'9'

static inline int isspace(int ch)
{
	return ch == 0x09 || ch == 0x0A || ch == 0x0C || ch == 0x0D || ch == 0x20;
}

static void
lex_name(fz_stream *f, psfunc_lexbuf *buf)
{
	char *s = buf->scratch;
	int n = buf->size;

	while (n > 1) {
		int c = fz_read_byte(f);
        if (c == 0x0b || isspace(c) || c == '{' || c == '}') {
            if (c != 0x0b)
		        fz_unread_byte(f);
	        *s = '\0';
	        buf->len = s - buf->scratch;
            return;
        }
		*s++ = c;
		n--;
	}
}

unsigned int
psfunc_lex(fz_stream *f, psfunc_lexbuf *buf)
{
    char *s;
    unsigned int is_real = 0;

	while (1)
	{
		int c = fz_read_byte(f);
		switch (c)
		{
		case 0x0b:
			return PS_TOK_EOF;
		case '\0':
        case '\t':
        case '\n':
        case '\f':
        case '\r':
        case ' ':
	        do {
		        c = fz_read_byte(f);
	        } while (!c || isspace(c));
            if (c != 0x0b)
		        fz_unread_byte(f);
			break;
		case '%':
	        do {
		        c = fz_read_byte(f);
	        } while ((c != '\n') && (c != 0x0b));
			break;
		case '{':
		case '}':
            return c;
		case '+':
		case '-':
		case IS_NUMBER:
            s = buf->scratch;
            is_real = 0;
		    *s++ = c;
	        do {
		        c = fz_read_byte(f);
		        *s++ = c;
                if (c == '.') is_real = 1;
	        } while ((((unsigned char)c - 0x30) < 10) || c == '.');
            if (c != 0x0b)
		        fz_unread_byte(f);
	        *s = '\0';
	        buf->len = s - buf->scratch;
            return is_real ? PS_TOK_REAL : PS_TOK_INT;
		default: /* isregular: !isdelim && !iswhite && c != EOF */
		    fz_unread_byte(f);
			lex_name(f, buf);
			return PS_TOK_KEYWORD;
		}
	}
}

static int
parse_code(pdf_function *func, fz_stream *stream, unsigned int *codeptr, psfunc_lexbuf *buf)
{
	unsigned int tok;
	int opptr, elseptr, ifptr;
	int a, b, mid, cmp;

	while (1)
	{
		tok = psfunc_lex(stream, buf);

		switch (tok)
		{
		case PS_TOK_EOF:
            fz_warn(NULL, "Unexpected end of PostScript function stream");
            return -1;

		case PS_TOK_REAL:
			resize_code(func, *codeptr);
			func->psobj_code[*codeptr].type = PS_REAL;
			func->psobj_code[*codeptr].u.f = atof(buf->scratch);
			++*codeptr;
			break;

		case PS_TOK_INT:
			resize_code(func, *codeptr);
			func->psobj_code[*codeptr].type = PS_INT;
			func->psobj_code[*codeptr].u.i = fz_atoi(buf->scratch);
			++*codeptr;
			break;

		case '{':
			opptr = *codeptr;
			*codeptr += 4;
			resize_code(func, *codeptr);
			ifptr = *codeptr;
			parse_code(func, stream, codeptr, buf);
			tok = psfunc_lex(stream, buf);
			if (tok == '{') {
				elseptr = *codeptr;
				parse_code(func, stream, codeptr, buf);
				tok = psfunc_lex(stream, buf);
			} else {
				elseptr = -1;
			}

#if 0
			if (tok != PDF_TOK_KEYWORD)
				fz_throw(ctx, FZ_ERROR_GENERIC, "missing keyword in 'if-else' context");
#endif

			if (!strcmp(buf->scratch, "if")) {
				if (elseptr >= 0) {
		            fz_warn(NULL, "Got 'if' operator with two blocks in PostScript function");
                    return -1;
                }
				func->psobj_code[opptr].type = PS_OPERATOR;
				func->psobj_code[opptr].u.op = PS_OP_IF;
				func->psobj_code[opptr+2].type = PS_BLOCK;
				func->psobj_code[opptr+2].u.block = ifptr;
				func->psobj_code[opptr+3].type = PS_BLOCK;
				func->psobj_code[opptr+3].u.block = *codeptr;
			} else if (!strcmp(buf->scratch, "ifelse")) {
				if (elseptr < 0) {
		            fz_warn(NULL, "Got 'ifelse' operator with one block in PostScript function");
                    return -1;
                }
				func->psobj_code[opptr].type = PS_OPERATOR;
				func->psobj_code[opptr].u.op = PS_OP_IFELSE;
				func->psobj_code[opptr+1].type = PS_BLOCK;
				func->psobj_code[opptr+1].u.block = elseptr;
				func->psobj_code[opptr+2].type = PS_BLOCK;
				func->psobj_code[opptr+2].u.block = ifptr;
				func->psobj_code[opptr+3].type = PS_BLOCK;
				func->psobj_code[opptr+3].u.block = *codeptr;
			} else {
	            fz_warn(NULL, "Expected if/ifelse operator in PostScript function, got: '%s'", buf->scratch);
                return -1;
			}
			break;

		case '}':
			resize_code(func, *codeptr);
			func->psobj_code[*codeptr].type = PS_OPERATOR;
			func->psobj_code[*codeptr].u.op = PS_OP_RETURN;
			++*codeptr;
			return 0;

		case PS_TOK_TRUE:
		case PS_TOK_FALSE:
			resize_code(func, *codeptr);
			func->psobj_code[*codeptr].type = PS_BOOL;
			func->psobj_code[*codeptr].u.i = (tok == PS_TOK_TRUE ? 1 : 0);
			++*codeptr;
			break;

		case PS_TOK_KEYWORD:
			cmp = -1;
			a = -1;
			b = FF_ARRAY_ELEMS(ps_op_names);
			while (b - a > 1)
			{
				mid = (a + b) / 2;
				cmp = strcmp(buf->scratch, ps_op_names[mid]);
				if (cmp > 0)
					a = mid;
				else if (cmp < 0)
					b = mid;
				else
					a = b = mid;
			}
			if (cmp != 0) {
				fz_warn(NULL, "unknown operator: '%s'", buf->scratch);
                return -1;
            }
            if (a == PS_OP_IF || a == PS_OP_IFELSE) {
                fz_warn(NULL, "illegally positioned if/ifelse operator in function");
                return -1;
            }

			resize_code(func, *codeptr);
			func->psobj_code[*codeptr].type = PS_OPERATOR;
			func->psobj_code[*codeptr].u.op = a;
			++*codeptr;
			break;

		default:
			fz_warn(NULL, "calculator function syntax error: token %u, scratch %s", tok, buf->scratch);
            return -1;
		}
	}
    return 0;
}

int pdf_load_postscript_func(pdf_function *func, pdf_document *doc, pdf_obj *dict, int num, int gen)
{
	fz_stream *stream = NULL;
	unsigned int codeptr;
    int ret = 0;
	psfunc_lexbuf buf;
	unsigned int tok;

	psfunc_lexbuf_init(&buf, 256);

	stream = pdf_open_stream(doc, num, gen);
	tok = psfunc_lex(stream, &buf);
	if (tok != '{') {
	    fz_warn(NULL, "stream is not a postscript function");
        ret = -1;
	} else {
		func->psobj_code = NULL;
		func->psobj_cap = 0;
		codeptr = 0;
		if (parse_code(func, stream, &codeptr, &buf) < 0) {
		    fz_warn(NULL, "cannot parse postscript function");
            ret = -1;
        }
    }
	fz_close(stream);
	func->base.size += func->psobj_cap * sizeof(psobj);
#ifdef DEBUG_PSFUNC
    pdf_debug_postscript_function(func);
#endif
    return ret;
}

/*
 * Debugging prints
 */

#ifdef DEBUG_PSFUNC
static void
pdf_debug_indent(char *prefix, int level, char *suffix)
{
	int i;

	printf("%s", prefix);

	for (i = 0; i < level; i++)
		printf("\t");

	printf("%s", suffix);
}

void pdf_debug_postscript_func(psobj *funccode, psobj *code, int level)
{
	int eof, wasop;

	pdf_debug_indent("", level, "{");

	/* Print empty blocks as { }, instead of separating braces on different lines. */
	if (code->type == PS_OPERATOR && code->u.op == PS_OP_RETURN)
	{
		printf(" } ");
		return;
	}

	pdf_debug_indent("\n", ++level, "");

	eof = 0;
	wasop = 0;
	while (!eof)
	{
		switch (code->type)
		{
		case PS_INT:
			if (wasop)
				pdf_debug_indent("\n", level, "");

			printf("%d ", code->u.i);
			wasop = 0;
			code++;
			break;

		case PS_REAL:
			if (wasop)
				pdf_debug_indent("\n", level, "");

			printf("%g ", code->u.f);
			wasop = 0;
			code++;
			break;

		case PS_OPERATOR:
			if (code->u.op == PS_OP_RETURN)
			{
				printf("\n");
				eof = 1;
			}
			else if (code->u.op == PS_OP_IF)
			{
				printf("\n");
				pdf_debug_postscript_func(funccode, &funccode[(code + 2)->u.block], level);

				printf("%s", ps_op_names[code->u.op]);
				code = &funccode[(code + 3)->u.block];
				if (code->type != PS_OPERATOR || code->u.op != PS_OP_RETURN)
					pdf_debug_indent("\n", level, "");

				wasop = 0;
			}
			else if (code->u.op == PS_OP_IFELSE)
			{
				printf("\n");
				pdf_debug_postscript_func(funccode, &funccode[(code + 2)->u.block], level);

				printf("\n");
				pdf_debug_postscript_func(funccode, &funccode[(code + 1)->u.block], level);

				printf("%s", ps_op_names[code->u.op]);
				code = &funccode[(code + 3)->u.block];
				if (code->type != PS_OPERATOR || code->u.op != PS_OP_RETURN)
					pdf_debug_indent("\n", level, "");

				wasop = 0;
			}
			else
			{
				printf("%s ", ps_op_names[code->u.op]);
				code++;
				wasop = 1;
			}
			break;
		}
	}

	pdf_debug_indent("", --level, "} ");
}

static void
pdf_debug_psfunction_imp(pdf_function *func, int level)
{
	int i;

	pdf_debug_indent("", level, "function {\n");
	pdf_debug_indent("", ++level, "");
	printf("postscript");
	pdf_debug_indent("\n", level, "");
	printf("%d input -> %d output\n", func->base.m, func->base.n);

	pdf_debug_indent("", level, "range ");
	for (i = 0; i < func->base.n; i++)
		printf("%g %g ", func->range[i][0], func->range[i][1]);
	printf("\n");
	pdf_debug_postscript_func(func->psobj_code, func->psobj_code, level);
	printf("\n");
	pdf_debug_indent("", --level, "}\n");
}

void
pdf_debug_postscript_function(pdf_function *func)
{
	pdf_debug_psfunction_imp(func, 0);
}
#endif


#ifdef TEST

#include <stdio.h>

static const char *teststr = 
"{ dup 0.85098 ge { pop 0.972549 0.929412 0.54902 } { dup 0.478431 le { pop 0.996078 0.835294 0.015686 }"
"{ 0.996078 0.972549 0.835294 0.929412 0.015686 0.54902 0.478431 0.85098"
"dup 2 index sub 9 index 3 index sub 1 index div 2 index 11 index sub 2 index div dup 11 index mul 2 index 11 index mul add 1 index 10 index mul 3 index 10 index mul add 2 index 9 index mul 4 index 9 index mul add 15 3 roll pop pop pop pop pop pop pop pop pop pop pop pop } ifelse } ifelse }";

int main(void) {
    psstream_t p;
    pdf_function f;
    float in[1], out[3];
    psfunc_initstream(&p, teststr, strlen(teststr));
    f.base.m = 1;
    f.base.n = 3;
    f.range[0][0] = 0;
    f.range[1][0] = 0;
    f.range[2][0] = 0;
    f.range[0][1] = 1;
    f.range[1][1] = 1;
    f.range[2][1] = 1;
    pdf_load_postscriptfunc(&p, &f);
    pdf_debug_function(&f);
    in[0] = (float)M_PI*0.25f;
    out[0] = out[1] = out[2] = 0.0f;
    pdf_eval_postscript_func(&f, in, out);
    printf("out: %.7f, %.7f, %.7f\n", out[0], out[1], out[2]);
    return 0;
}

#endif

