#include "mupdf/pdf.h"

/*
 * Check if an object is a stream or not.
 */
int
pdf_is_stream(pdf_document *doc, int num, int gen)
{
	pdf_xref_entry *entry;

	if (num < 0 || num >= pdf_xref_len(doc))
		return 0;

	pdf_cache_object(doc, num, gen);

	entry = pdf_get_xref_entry(doc, num);
	return entry->stm_ofs != 0 || entry->stm_buf;
}

/*
 * Scan stream dictionary for an explicit /Crypt filter
 */
static int
pdf_stream_has_crypt(fz_context *ctx, pdf_obj *stm)
{
	pdf_obj *filters;
	pdf_obj *obj;
	int i;

	filters = pdf_dict_getsa(stm, "Filter", "F");
	if (filters)
	{
		if (!strcmp(pdf_to_name(filters), "Crypt"))
			return 1;
		if (pdf_is_array(filters))
		{
			int n = pdf_array_len(filters);
			for (i = 0; i < n; i++)
			{
				obj = pdf_array_get(filters, i);
				if (!strcmp(pdf_to_name(obj), "Crypt"))
					return 1;
			}
		}
	}
	return 0;
}

static fz_jbig2_globals *
pdf_load_jbig2_globals(pdf_document *doc, pdf_obj *dict)
{
	fz_jbig2_globals *globals;
	fz_context *ctx = doc->ctx;
	fz_buffer *buf = NULL;

	if ((globals = pdf_find_item(ctx, fz_free_jbig2_globals_imp, dict)) != NULL) {
		return globals;
	}

	fz_try(ctx)
	{
		buf = pdf_load_stream(doc, pdf_to_num(dict), pdf_to_gen(dict));
		globals = fz_load_jbig2_globals(ctx, buf->data, buf->len);
		pdf_store_item(ctx, dict, globals, buf->len);
	}
	fz_always(ctx)
	{
		fz_drop_buffer(buf);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	return globals;
}

/*
 * Create a filter given a name and param dictionary.
 */
static fz_stream *
build_filter(fz_stream *chain, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params)
{
	fz_context *ctx = chain->ctx;
	char *s = pdf_to_name(f);

	int predictor = pdf_to_int(pdf_dict_gets(p, "Predictor"));
	pdf_obj *columns_obj = pdf_dict_gets(p, "Columns");
	int columns = pdf_to_int(columns_obj);
	int colors = pdf_to_int(pdf_dict_gets(p, "Colors"));
	int bpc = pdf_to_int(pdf_dict_gets(p, "BitsPerComponent"));

	if (!strcmp(s, "ASCIIHexDecode") || !strcmp(s, "AHx"))
		return fz_open_ahxd(chain);
	else if (!strcmp(s, "ASCII85Decode") || !strcmp(s, "A85"))
		return fz_open_a85d(chain);
	else if (!strcmp(s, "CCITTFaxDecode") || !strcmp(s, "CCF"))
	{
		pdf_obj *k = pdf_dict_gets(p, "K");
		pdf_obj *eol = pdf_dict_gets(p, "EndOfLine");
		pdf_obj *eba = pdf_dict_gets(p, "EncodedByteAlign");
		pdf_obj *rows = pdf_dict_gets(p, "Rows");
		pdf_obj *eob = pdf_dict_gets(p, "EndOfBlock");
		pdf_obj *bi1 = pdf_dict_gets(p, "BlackIs1");
		if (params)
		{
			/* We will shortstop here */
			params->type = FZ_IMAGE_FAX;
			params->u.fax.k = (k ? pdf_to_int(k) : 0);
			params->u.fax.eol = (eol ? pdf_to_bool(eol) : 0);
			params->u.fax.eba = (eba ? pdf_to_bool(eba) : 0);
			params->u.fax.columns = (columns_obj ? columns : 1728);
			params->u.fax.rows = (rows ? pdf_to_int(rows) : 0);
			params->u.fax.eob = (eob ? pdf_to_bool(eob) : 1);
			params->u.fax.black_is_1 = (bi1 ? pdf_to_bool(bi1) : 0);
			return chain;
		}
		return fz_open_faxd(chain,
				k ? pdf_to_int(k) : 0,
				eol ? pdf_to_bool(eol) : 0,
				eba ? pdf_to_bool(eba) : 0,
				columns_obj ? columns : 1728,
				rows ? pdf_to_int(rows) : 0,
				eob ? pdf_to_bool(eob) : 1,
				bi1 ? pdf_to_bool(bi1) : 0);
	}
	else if (!strcmp(s, "DCTDecode") || !strcmp(s, "DCT"))
	{
		pdf_obj *ct = pdf_dict_gets(p, "ColorTransform");
		if (params)
		{
			/* We will shortstop here */
			params->type = FZ_IMAGE_JPEG;
			params->u.jpeg.color_transform = (ct ? pdf_to_int(ct) : -1);
			return chain;
		}
		return fz_open_dctd(chain, ct ? pdf_to_int(ct) : -1);
	}
	else if (!strcmp(s, "RunLengthDecode") || !strcmp(s, "RL"))
	{
		if (params)
		{
			/* We will shortstop here */
			params->type = FZ_IMAGE_RLD;
			return chain;
		}
		return fz_open_rld(chain);
	}
	else if (!strcmp(s, "FlateDecode") || !strcmp(s, "Fl"))
	{
		if (params)
		{
			/* We will shortstop here */
			params->type = FZ_IMAGE_FLATE;
			params->u.flate.predictor = predictor;
			params->u.flate.columns = columns;
			params->u.flate.colors = colors;
			params->u.flate.bpc = bpc;
			return chain;
		}
		chain = fz_open_flated(chain, 15);
		if (predictor > 1)
			chain = fz_open_predict(chain, predictor, columns, colors, bpc);
		return chain;
	}
	else if (!strcmp(s, "LZWDecode") || !strcmp(s, "LZW"))
	{
		pdf_obj *ec = pdf_dict_gets(p, "EarlyChange");
		if (params)
		{
			/* We will shortstop here */
			params->type = FZ_IMAGE_LZW;
			params->u.lzw.predictor = predictor;
			params->u.lzw.columns = columns;
			params->u.lzw.colors = colors;
			params->u.lzw.bpc = bpc;
			params->u.lzw.early_change = (ec ? pdf_to_int(ec) : 1);
			return chain;
		}
		chain = fz_open_lzwd(chain, ec ? pdf_to_int(ec) : 1, 9, 0);
		if (predictor > 1)
			chain = fz_open_predict(chain, predictor, columns, colors, bpc);
		return chain;
	}
	else if (!strcmp(s, "JBIG2Decode"))
	{
		fz_jbig2_globals *globals = NULL;
		pdf_obj *obj = pdf_dict_gets(p, "JBIG2Globals");
		if (obj)
			globals = pdf_load_jbig2_globals(doc, obj);
		if (params) {
			/* We will shortstop here */
			params->type = FZ_IMAGE_JBIG2;
			params->u.jbig2.globals = globals;
			return chain;
		}
		/* fz_open_jbig2d takes possession of globals */
		return fz_open_jbig2d(chain, globals);
    }
	else if (!strcmp(s, "Crypt"))
	{
		pdf_obj *name;

		if (!doc->crypt)
		{
			fz_warn(ctx->warn, "crypt filter in unencrypted document");
			return chain;
		}

		name = pdf_dict_gets(p, "Name");
		if (pdf_is_name(name))
			return pdf_open_crypt_with_filter(chain, doc->crypt, pdf_to_name(name), num, gen);

		return chain;
	}

	fz_warn(ctx->warn, "unknown filter name (%s)", s);
	return chain;
}

/*
 * Build a chain of filters given filter names and param dicts.
 * If head is given, start filter chain with it.
 * Assume ownership of head.
 */
static fz_stream *
build_filter_chain(fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params)
{
	pdf_obj *f;
	pdf_obj *p;
	int i, n;

	n = pdf_array_len(fs);
	for (i = 0; i < n; i++) {
		f = pdf_array_get(fs, i);
		p = pdf_array_get(ps, i);
		chain = build_filter(chain, doc, f, p, num, gen, (i == n-1 ? params : NULL));
	}

	return chain;
}

/*
 * Build a filter for reading raw stream data.
 * This is a null filter to constrain reading to the stream length (and to
 * allow for other people accessing the file), followed by a decryption
 * filter.
 *
 * orig_num and orig_gen are used purely to seed the encryption.
 */
static fz_stream *
pdf_open_raw_filter(fz_stream *chain, pdf_document *doc, pdf_obj *stmobj, int num, int orig_num, int orig_gen, int offset)
{
	fz_context *ctx = chain->ctx;
	int hascrypt;
	int len;

	if (num > 0 && num < pdf_xref_len(doc))
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, num);
		if (entry->stm_buf)
			return fz_open_buffer(ctx, entry->stm_buf);
	}

	/* don't close chain when we close this filter */
	fz_keep_stream(chain);

	len = pdf_to_int(pdf_dict_gets(stmobj, "Length"));
	chain = fz_open_null(chain, len, offset);

	fz_try(ctx)
	{
		hascrypt = pdf_stream_has_crypt(ctx, stmobj);
		if (doc->crypt && !hascrypt)
			chain = pdf_open_crypt(chain, doc->crypt, orig_num, orig_gen);
	}
	fz_catch(ctx)
	{
		fz_close(chain);
		fz_rethrow(ctx);
	}

	return chain;
}

/*
 * Construct a filter to decode a stream, constraining
 * to stream length and decrypting.
 */
static fz_stream *
pdf_open_filter(fz_stream *chain, pdf_document *doc, pdf_obj *stmobj, int num, int gen, int offset, fz_compression_params *imparams)
{
	pdf_obj *filters;
	pdf_obj *params;

	filters = pdf_dict_getsa(stmobj, "Filter", "F");
	params = pdf_dict_getsa(stmobj, "DecodeParms", "DP");

	chain = pdf_open_raw_filter(chain, doc, stmobj, num, num, gen, offset);

	fz_try(doc->ctx)
	{
		if (pdf_is_name(filters))
			chain = build_filter(chain, doc, filters, params, num, gen, imparams);
		else if (pdf_array_len(filters) > 0)
			chain = build_filter_chain(chain, doc, filters, params, num, gen, imparams);
	}
	fz_catch(doc->ctx)
	{
		fz_close(chain);
		fz_rethrow(doc->ctx);
	}

	return chain;
}

/*
 * Construct a filter to decode a stream, without
 * constraining to stream length, and without decryption.
 */
fz_stream *
pdf_open_inline_stream(pdf_document *doc, pdf_obj *stmobj, int length, fz_stream *chain, fz_compression_params *imparams)
{
	pdf_obj *filters;
	pdf_obj *params;

	filters = pdf_dict_getsa(stmobj, "Filter", "F");
	params = pdf_dict_getsa(stmobj, "DecodeParms", "DP");

	/* don't close chain when we close this filter */
	fz_keep_stream(chain);

	if (pdf_is_name(filters))
		return build_filter(chain, doc, filters, params, 0, 0, imparams);
	if (pdf_array_len(filters) > 0)
		return build_filter_chain(chain, doc, filters, params, 0, 0, imparams);

	return fz_open_null(chain, length, fz_tell(chain));
}

/*
 * Open a stream for reading the raw (compressed but decrypted) data.
 */
fz_stream *
pdf_open_raw_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen)
{
	pdf_xref_entry *x;

	if (num < 0 || num >= pdf_xref_len(doc))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object id out of range (%d %d R)", num, gen);

	x = pdf_get_xref_entry(doc, num);

	pdf_cache_object(doc, num, gen);

	if (x->stm_ofs == 0)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object is not a stream");

	return pdf_open_raw_filter(doc->file, doc, x->obj, num, orig_num, orig_gen, x->stm_ofs);
}

static fz_stream *
pdf_open_image_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen, fz_compression_params *params)
{
	pdf_xref_entry *x;

	if (num < 0 || num >= pdf_xref_len(doc))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object id out of range (%d %d R)", num, gen);

	x = pdf_get_xref_entry(doc, num);

	pdf_cache_object(doc, num, gen);

	if (x->stm_ofs == 0 && x->stm_buf == NULL)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object is not a stream");

	return pdf_open_filter(doc->file, doc, x->obj, orig_num, orig_gen, x->stm_ofs, params);
}

/*
 * Open a stream for reading uncompressed data.
 * Put the opened file in doc->stream.
 * Using doc->file while a stream is open is a Bad idea.
 */
fz_stream *
pdf_open_stream(pdf_document *doc, int num, int gen)
{
	return pdf_open_image_stream(doc, num, gen, num, gen, NULL);
}

fz_stream *
pdf_open_stream_with_offset(pdf_document *doc, int num, int gen, pdf_obj *dict, int stm_ofs)
{
	if (stm_ofs == 0)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object is not a stream");

	return pdf_open_filter(doc->file, doc, dict, num, gen, stm_ofs, NULL);
}

fz_stream *
pdf_open_stream_decryptor(pdf_document *doc, int num, int gen)
{
	pdf_xref_entry *x;
	pdf_obj *filters;
    fz_stream *chain;
    unsigned int len = 0;

	if (num < 0 || num >= pdf_xref_len(doc))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object id out of range (%d %d R)", num, gen);

	x = pdf_get_xref_entry(doc, num);

	pdf_cache_object(doc, num, gen);

	if (x->stm_ofs == 0 && x->stm_buf == NULL)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "object is not a stream");

	chain = pdf_open_raw_filter(doc->file, doc, x->obj, num, num, gen, x->stm_ofs);

	/* don't close chain when we close this filter */
	fz_keep_stream(chain);

	filters = pdf_dict_getsa(x->obj, "Filter", "F");
	if (pdf_is_name(filters)) {
	    char *s = pdf_to_name(filters);
        if (!strcmp(s, "Crypt")) {
		    pdf_obj *name;

		    if (!doc->crypt) {
			    fz_warn(doc->ctx->warn, "crypt filter in unencrypted document");
		    } else {
	            pdf_obj *p = pdf_dict_getsa(x->obj, "DecodeParms", "DP");
		        name = pdf_dict_gets(p, "Name");
		        if (pdf_is_name(name))
			        return pdf_open_crypt_with_filter(chain, doc->crypt, pdf_to_name(name), num, gen);
            }
	    }
    }
	len = pdf_to_int(pdf_dict_gets(x->obj, "Length"));
	return fz_open_null(chain, len, fz_tell(chain));
}

/*
 * Load raw (compressed but decrypted) contents of a stream into buf.
 */
fz_buffer *
pdf_load_raw_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen)
{
	fz_stream *stm;
	pdf_obj *dict;
	int len;
	fz_buffer *buf;

	if (num > 0 && num < pdf_xref_len(doc))
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, num);
		if (entry->stm_buf)
			return fz_keep_buffer(entry->stm_buf);
	}

	dict = pdf_load_object(doc, num, gen);

	len = pdf_to_int(pdf_dict_gets(dict, "Length"));

	pdf_drop_obj(dict);

	stm = pdf_open_raw_renumbered_stream(doc, num, gen, orig_num, orig_gen);

	buf = fz_read_all(stm, len);

	fz_close(stm);
	return buf;
}

static int
pdf_guess_filter_length(int len, char *filter)
{
	if (!strcmp(filter, "ASCIIHexDecode"))
		return len / 2;
	if (!strcmp(filter, "ASCII85Decode"))
		return len * 4 / 5;
	if (!strcmp(filter, "FlateDecode"))
		return len * 3;
	if (!strcmp(filter, "RunLengthDecode"))
		return len * 3;
	if (!strcmp(filter, "LZWDecode"))
		return len * 2;
	return len;
}

static fz_buffer *
pdf_load_image_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen, fz_compression_params *params, int *truncated)
{
	fz_context *ctx = doc->ctx;
	fz_stream *stm = NULL;
	pdf_obj *dict, *obj;
	int i, len, n;
	fz_buffer *buf;

	if (num > 0 && num < pdf_xref_len(doc))
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, num);
		if (entry->stm_buf)
			return fz_keep_buffer(entry->stm_buf);
	}

	dict = pdf_load_object(doc, num, gen);

	len = pdf_to_int(pdf_dict_gets(dict, "Length"));
	obj = pdf_dict_gets(dict, "Filter");
	len = pdf_guess_filter_length(len, pdf_to_name(obj));
	n = pdf_array_len(obj);
	for (i = 0; i < n; i++)
		len = pdf_guess_filter_length(len, pdf_to_name(pdf_array_get(obj, i)));

	pdf_drop_obj(dict);

	stm = pdf_open_image_stream(doc, num, gen, orig_num, orig_gen, params);

	fz_try(ctx)
	{
		if (truncated)
			buf = fz_read_best(stm, len, truncated);
		else
			buf = fz_read_all(stm, len);
	}
	fz_always(ctx)
	{
		fz_close(stm);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot read raw stream (%d %d R)", num, gen);
	}

	return buf;
}

/*
 * Load uncompressed contents of a stream into buf.
 */
fz_buffer *
pdf_load_stream(pdf_document *doc, int num, int gen)
{
	return pdf_load_image_stream(doc, num, gen, num, gen, NULL, NULL);
}

fz_buffer *
pdf_load_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen, int *truncated)
{
	return pdf_load_image_stream(doc, num, gen, orig_num, orig_gen, NULL, truncated);
}

fz_compressed_buffer *
pdf_load_compressed_stream(pdf_document *doc, int num, int gen)
{
	fz_context *ctx = doc->ctx;
	fz_compressed_buffer *bc = fz_malloc_struct(ctx, fz_compressed_buffer);

	fz_try(ctx)
	{
		bc->buffer = pdf_load_image_stream(doc, num, gen, num, gen, &bc->params, NULL);
	}
	fz_catch(ctx)
	{
		fz_free(ctx, bc);
		fz_rethrow(ctx);
	}
	return bc;
}

static fz_stream *
pdf_open_object_array(pdf_document *doc, pdf_obj *list)
{
	int i, n;
	fz_context *ctx = doc->ctx;
	fz_stream *stm;

	n = pdf_array_len(list);
	stm = fz_open_concat(ctx, n, 1);

	for (i = 0; i < n; i++)
	{
		pdf_obj *obj = pdf_array_get(list, i);
		fz_try(ctx)
		{
			fz_concat_push(stm, pdf_open_stream(doc, pdf_to_num(obj), pdf_to_gen(obj)));
		}
		fz_catch(ctx)
		{
			fz_warn(ctx->warn, "cannot load content stream part %d/%d", i + 1, n);
			continue;
		}
	}

	return stm;
}

fz_stream *
pdf_open_contents_stream(pdf_document *doc, pdf_obj *obj)
{
	fz_context *ctx = doc->ctx;
	int num, gen;

	if (pdf_is_array(obj))
		return pdf_open_object_array(doc, obj);

	num = pdf_to_num(obj);
	gen = pdf_to_gen(obj);
	if (pdf_is_stream(doc, num, gen))
		return pdf_open_image_stream(doc, num, gen, num, gen, NULL);

	fz_warn(ctx->warn, "pdf object stream missing (%d %d R)", num, gen);
	return NULL;
}

/*
 * PDF 1.7 algorithm 3.1 and ExtensionLevel 3 algorithm 3.1a
 *
 * Using the global encryption key that was generated from the
 * password, create a new key that is used to decrypt individual
 * objects and streams. This key is based on the object and
 * generation numbers.
 */

static int
pdf_compute_object_key(pdf_crypt *crypt, pdf_crypt_filter *cf, int num, int gen, unsigned char *key, int max_len)
{
	fz_md5 md5;
	unsigned char message[5];
	int key_len = crypt->length / 8;

	if (key_len > max_len)
		key_len = max_len;

	if (cf->method == PDF_CRYPT_AESV3)
	{
		memcpy(key, crypt->key, key_len);
		return key_len;
	}

	fz_md5_init(&md5);
	fz_md5_update(&md5, crypt->key, key_len);
	message[0] = (num) & 0xFF;
	message[1] = (num >> 8) & 0xFF;
	message[2] = (num >> 16) & 0xFF;
	message[3] = (gen) & 0xFF;
	message[4] = (gen >> 8) & 0xFF;
	fz_md5_update(&md5, message, 5);

	if (cf->method == PDF_CRYPT_AESV2)
		fz_md5_update(&md5, (unsigned char *)"sAlT", 4);

	fz_md5_final(&md5, key);

	if (key_len + 5 > 16)
		return 16;
	return key_len + 5;
}

/*
 * PDF 1.7 algorithm 3.1 and ExtensionLevel 3 algorithm 3.1a
 *
 * Create filter suitable for de/encrypting a stream.
 */
static fz_stream *
pdf_open_crypt_imp(fz_stream *chain, pdf_crypt *crypt, pdf_crypt_filter *stmf, int num, int gen)
{
	unsigned char key[32];
	int len;

	len = pdf_compute_object_key(crypt, stmf, num, gen, key, 32);

	if (stmf->method == PDF_CRYPT_RC4)
		return fz_open_arc4(chain, key, len);

	if (stmf->method == PDF_CRYPT_AESV2 || stmf->method == PDF_CRYPT_AESV3)
		return fz_open_aesd(chain, key, len);

	return fz_open_copy(chain);
}

fz_stream *
pdf_open_crypt(fz_stream *chain, pdf_crypt *crypt, int num, int gen)
{
	return pdf_open_crypt_imp(chain, crypt, &crypt->stmf, num, gen);
}

/*
 * Parse a CF dictionary entry (PDF 1.7 table 3.22)
 */

void
pdf_parse_crypt_filter(fz_context *ctx, pdf_crypt_filter *cf, pdf_crypt *crypt, char *name)
{
	pdf_obj *obj;
	pdf_obj *dict;
	int is_identity = (strcmp(name, "Identity") == 0);
	int is_stdcf = (!is_identity && (strcmp(name, "StdCF") == 0));

	if (!is_identity && !is_stdcf)
		fz_throw(ctx, FZ_ERROR_GENERIC, "Crypt Filter not Identity or StdCF (%d %d R)", pdf_to_num(crypt->cf), pdf_to_gen(crypt->cf));

	cf->method = PDF_CRYPT_NONE;
	cf->length = crypt->length;

	if (!crypt->cf)
	{
		cf->method = (is_identity ? PDF_CRYPT_NONE : PDF_CRYPT_RC4);
		return;
	}

	dict = pdf_dict_gets(crypt->cf, name);
	if (!pdf_is_dict(dict))
		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse crypt filter (%d %d R)", pdf_to_num(crypt->cf), pdf_to_gen(crypt->cf));

	obj = pdf_dict_gets(dict, "CFM");
	if (pdf_is_name(obj))
	{
		if (!strcmp(pdf_to_name(obj), "None"))
			cf->method = PDF_CRYPT_NONE;
		else if (!strcmp(pdf_to_name(obj), "V2"))
			cf->method = PDF_CRYPT_RC4;
		else if (!strcmp(pdf_to_name(obj), "AESV2"))
			cf->method = PDF_CRYPT_AESV2;
		else if (!strcmp(pdf_to_name(obj), "AESV3"))
			cf->method = PDF_CRYPT_AESV3;
		else
			fz_warn(ctx->warn, "unknown encryption method: %s", pdf_to_name(obj));
	}

	obj = pdf_dict_gets(dict, "Length");
	if (pdf_is_int(obj))
		cf->length = pdf_to_int(obj);

	/* the length for crypt filters is supposed to be in bytes not bits */
	if (cf->length < 40)
		cf->length = cf->length * 8;

	if ((cf->length % 8) != 0)
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid key length: %d", cf->length);

	if ((crypt->r == 1 || crypt->r == 2 || crypt->r == 4) &&
		(cf->length < 0 || cf->length > 128))
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid key length: %d", cf->length);
	if ((crypt->r == 5 || crypt->r == 6) && cf->length != 256)
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid key length: %d", cf->length);
}

fz_stream *
pdf_open_crypt_with_filter(fz_stream *chain, pdf_crypt *crypt, char *name, int num, int gen)
{
	if (strcmp(name, "Identity"))
	{
		pdf_crypt_filter cf;
		pdf_parse_crypt_filter(chain->ctx, &cf, crypt, name);
		return pdf_open_crypt_imp(chain, crypt, &cf, num, gen);
	}
	return chain;
}

/* AES Filter */
typedef struct fz_aesd_s fz_aesd;

struct fz_aesd_s
{
    fz_stream *chain;
    fz_aes aes;
    unsigned char iv[16];
    int ivcount;
    unsigned char bp[323232];
    unsigned char *rp, *wp;
};

static void aes_cbc_decrypt(fz_aes *ctx, unsigned char iv[16], unsigned char *output,
				     const unsigned char *input, unsigned int length)
{
    unsigned int i;
	unsigned char temp[16];

	while(length > 0) {
		memcpy(temp, input, 16);
		aes_ecb_decrypt(ctx, input, output);
		for(i = 0; i < 16; i++)
			output[i] = (unsigned char)(output[i] ^ iv[i]);
		memcpy(iv, temp, 16);
		input += 16;
		output += 16;
		length -= 16;
	}
}

static int
read_aesd(fz_stream *stm, unsigned char *buf, int len)
{
    fz_aesd *state = stm->state;
    unsigned char *p = buf;
    unsigned char *ep = buf + len;

    while (state->ivcount < 16)
    {
        int c = fz_read_byte(state->chain);
        if (c < 0)
            fz_throw(stm->ctx, FZ_ERROR_GENERIC, "premature end in aes filter");
        state->iv[state->ivcount++] = c;
    }

    while (state->rp < state->wp && p < ep)
        *p++ = *state->rp++;

    while (p < ep)
    {
        int n = fz_read(state->chain, state->bp, 16);
        if (n == 0)
            return p - buf;
        else if (n < 16)
            fz_throw(stm->ctx, FZ_ERROR_GENERIC, "partial block in aes filter");

        aes_cbc_decrypt(&state->aes, state->iv, state->bp, state->bp, 16);
        state->rp = state->bp;
        state->wp = state->bp + 16;

        /* strip padding at end of file */
        if (fz_is_eof(state->chain))
        {
            int pad = state->bp[15];
            if (pad < 1 || pad > 16)
                fz_throw(stm->ctx, FZ_ERROR_GENERIC, "aes padding out of range: %d", pad);
            state->wp -= pad;
        }

        while (state->rp < state->wp && p < ep)
            *p++ = *state->rp++;
    }

    return p - buf;
}

static void
close_aesd(fz_context *ctx, void *state_)
{
    fz_aesd *state = (fz_aesd *)state_;
    fz_stream *chain = state->chain;
    fz_free(ctx, state);
    fz_close(chain);
}

fz_stream *
fz_open_aesd(fz_stream *chain, unsigned char *key, unsigned keylen)
{
    fz_aesd *state;
    fz_context *ctx = chain->ctx;
    state = fz_malloc_struct(ctx, fz_aesd);
    state->chain = chain;
    state->ivcount = 0;
    state->rp = state->bp;
    state->wp = state->bp;
    if (aes_setup(&state->aes, key, keylen)) {
		fz_close(chain);
    	fz_throw(ctx, FZ_ERROR_GENERIC, "AES key init failed (keylen=%d)", keylen * 8);
    }
    return fz_new_stream(ctx, state, read_aesd, close_aesd);
}

