/*
 * mudraw -- command line tool for drawing pdf/xps/cbz documents
 */

#include "mupdf/fitz.h"
#include "mupdf/pdf.h" /* for mujstest */
#include <sys/time.h>
unsigned char *slurp_file(const char *filename, size_t *filesize);

enum { TEXT_PLAIN = 1, TEXT_HTML = 2, TEXT_XML = 3 };

enum { OUT_PNG, OUT_PPM, OUT_PNM, OUT_PAM, OUT_PGM, OUT_PBM, OUT_TGA, OUT_PWG, OUT_PCL, OUT_SVG, OUT_PDF };

enum { CS_INVALID, CS_UNSET, CS_MONO, CS_GRAY, CS_GRAYALPHA, CS_RGB, CS_RGBA };

typedef struct
{
	char *suffix;
	int format;
} suffix_t;

static const suffix_t suffix_table[] =
{
	{ ".png", OUT_PNG },
	{ ".pgm", OUT_PGM },
	{ ".ppm", OUT_PPM },
	{ ".pnm", OUT_PNM },
	{ ".pam", OUT_PAM },
	{ ".pbm", OUT_PBM },
	{ ".tga", OUT_TGA},
	{ ".pwg", OUT_PWG },
	{ ".pcl", OUT_PCL },
	{ ".svg", OUT_SVG },
	{ ".pdf", OUT_PDF },
};

typedef struct
{
	char *name;
	int colorspace;
} cs_name_t;

static const cs_name_t cs_name_table[] =
{
	{ "m", CS_MONO },
	{ "mono", CS_MONO },
	{ "g", CS_GRAY },
	{ "gray", CS_GRAY },
	{ "grey", CS_GRAY },
	{ "ga", CS_GRAYALPHA },
	{ "grayalpha", CS_GRAYALPHA },
	{ "greyalpha", CS_GRAYALPHA },
	{ "rgb", CS_RGB },
	{ "rgba", CS_RGBA },
	{ "rgbalpha", CS_RGBA }
};

typedef struct
{
	int format;
	int default_cs;
	int permitted_cs[6];
} format_cs_table_t;

static const format_cs_table_t format_cs_table[] =
{
	{ OUT_PNG, CS_RGB, { CS_GRAY, CS_GRAYALPHA, CS_RGB, CS_RGBA } },
	{ OUT_PPM, CS_RGB, { CS_GRAY, CS_RGB } },
	{ OUT_PNM, CS_GRAY, { CS_GRAY, CS_RGB } },
	{ OUT_PAM, CS_RGBA, { CS_RGBA } },
	{ OUT_PGM, CS_GRAY, { CS_GRAY, CS_RGB } },
	{ OUT_PBM, CS_MONO, { CS_MONO } },
	{ OUT_TGA, CS_RGB, { CS_GRAY, CS_GRAYALPHA, CS_RGB, CS_RGBA } },
	{ OUT_PWG, CS_RGB, { CS_MONO, CS_GRAY, CS_RGB } },
	{ OUT_PCL, CS_MONO, { CS_MONO } },
	{ OUT_SVG, CS_RGB, { CS_RGB } },
	{ OUT_PDF, CS_RGB, { CS_RGB } }
};

/*
	A useful bit of bash script to call this to generate mjs files:
	for f in tests_private/pdf/forms/v1.3/ *.pdf ; do g=${f%.*} ; echo $g ; ../mupdf.git/win32/debug/mudraw.exe -j $g.mjs $g.pdf ; done

	Remove the space from "/ *.pdf" before running - can't leave that
	in here, as it causes a warning about a possibly malformed comment.
*/

static char lorem[] =
"Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum "
"vehicula augue id est lobortis mollis. Aenean vestibulum metus sed est "
"gravida non tempus lacus aliquet. Nulla vehicula lobortis tincidunt. "
"Donec malesuada nisl et lacus condimentum nec tincidunt urna gravida. "
"Sed dapibus magna eu velit ultrices non rhoncus risus lacinia. Fusce "
"vitae nulla volutpat elit dictum ornare at eu libero. Maecenas felis "
"enim, tempor a tincidunt id, commodo consequat lectus.\n"
"Morbi tincidunt adipiscing lacus eu dignissim. Pellentesque augue elit, "
"ultrices vitae fermentum et, faucibus et purus. Nam ante libero, lacinia "
"id tincidunt at, ultricies a lorem. Donec non neque at purus condimentum "
"eleifend quis sit amet libero. Sed semper, mi ut tempus tincidunt, lacus "
"eros pellentesque lacus, id vehicula est diam eu quam. Integer tristique "
"fringilla rhoncus. Phasellus convallis, justo ut mollis viverra, dui odio "
"euismod ante, nec fringilla nisl mi ac diam.\n"
"Maecenas mi urna, ornare commodo feugiat id, cursus in massa. Vivamus "
"augue augue, aliquam at varius eu, venenatis fermentum felis. Sed varius "
"turpis a felis ultrices quis aliquet nunc tincidunt. Suspendisse posuere "
"commodo nunc non viverra. Praesent condimentum varius quam, vel "
"consectetur odio volutpat in. Sed malesuada augue ut lectus commodo porta. "
"Vivamus eget mauris sit amet diam ultrices sollicitudin. Cras pharetra leo "
"non elit lacinia vulputate.\n"
"Donec ac enim justo, ornare scelerisque diam. Ut vel ante at lorem "
"placerat bibendum ultricies mattis metus. Phasellus in imperdiet odio. "
"Proin semper lacinia libero, sed rutrum eros blandit non. Duis tincidunt "
"ligula est, non pellentesque mauris. Aliquam in erat scelerisque lacus "
"dictum suscipit eget semper magna. Nullam luctus imperdiet risus a "
"semper.\n"
"Curabitur sit amet tempor sapien. Quisque et tortor in lacus dictum "
"pulvinar. Nunc at nisl ut velit vehicula hendrerit. Mauris elementum "
"sollicitudin leo ac ullamcorper. Proin vel leo nec justo tempus aliquet "
"nec ut mi. Pellentesque vel nisl id dui hendrerit fermentum nec quis "
"tortor. Proin eu sem luctus est consequat euismod. Vestibulum ante ipsum "
"primis in faucibus orci luctus et ultrices posuere cubilia Curae; Fusce "
"consectetur ultricies nisl ornare dictum. Cras sagittis consectetur lorem "
"sed posuere. Mauris accumsan laoreet arcu, id molestie lorem faucibus eu. "
"Vivamus commodo, neque nec imperdiet pretium, lorem metus viverra turpis, "
"malesuada vulputate justo eros sit amet neque. Nunc quis justo elit, non "
"rutrum mauris. Maecenas blandit condimentum nibh, nec vulputate orci "
"pulvinar at. Proin sed arcu vel odio tempus lobortis sed posuere ipsum. Ut "
"feugiat pellentesque tortor nec ornare.\n";

static char *output = NULL;
static float resolution = 72;
static int res_specified = 0;
static float rotation = 0;

static int showxml = 0;
static int showtext = 0;
static int showtime = 0;
static int showmd5 = 0;
static int showoutline = 0;
static int alphabits = 8;
static float gamma_value = 1;
static int invert = 0;
static int width = 0;
static int height = 0;
static int fit = 0;
static int errored = 0;
static int ignore_errors = 0;
static int output_format;
static int append = 0;
static int out_cs = CS_UNSET;

static pdf_document *pdfout = NULL;
static fz_stext_sheet *sheet = NULL;
static fz_device *dev = NULL;
static fz_buffer *outbuf = NULL;
static fz_colorspace *colorspace;
static char *filename;
static int files = 0;

static struct {
	int count, total;
	int min, max;
	int minpage, maxpage;
	char *minfilename;
	char *maxfilename;
} timing;

static void usage(void)
{
    const char *usage_str = "usage: mudraw [options] input [pages]\n"
		"\t-o -\toutput filename (%%d for page number)\n"
		"\t\tsupported formats: pgm, ppm, pam, png, pbm\n"
		"\t-p -\tpassword\n"
		"\t-r -\tresolution in dpi (default: 72)\n"
		"\t-w -\twidth (in pixels) (maximum width if -r is specified)\n"
		"\t-h -\theight (in pixels) (maximum height if -r is specified)\n"
		"\t-f -\tfit width and/or height exactly (ignore aspect)\n"
		"\t-c -\tcolorspace {mono,gray,grayalpha,rgb,rgba}\n"
		"\t-b -\tnumber of bits of antialiasing (0 to 8)\n"
		"\t-g\trender in grayscale\n"
		"\t-m\tshow timing information\n"
		"\t-t\tshow text (-tt for xml, -ttt for more verbose xml)\n"
		"\t-x\tshow display list\n"
		"\t-5\tshow md5 checksums\n"
		"\t-R -\trotate clockwise by given number of degrees\n"
		"\t-G -\tgamma correct output\n"
		"\t-I\tinvert output\n"
		"\t-l\tprint outline\n"
		"\t-i\tignore errors and continue with the next file\n"
		"\tpages\tcomma separated list of ranges\n";
    write(1, usage_str, strlen(usage_str));
	exit(1);
}

static int gettime(void)
{
	static struct timeval first;
	static int once = 1;
	struct timeval now;
	if (once) {
		gettimeofday(&first, NULL);
		once = 0;
	}
	gettimeofday(&now, NULL);
	return (now.tv_sec - first.tv_sec) * 1000 + (now.tv_usec - first.tv_usec) / 1000;
}

static int isrange(char *s)
{
	while (*s)
	{
		if ((*s < '0' || *s > '9') && *s != '-' && *s != ',')
			return 0;
		s++;
	}
	return 1;
}

static void
pdf_run_page_internal(fz_context *ctx, pdf_document *doc, pdf_page *page, const fz_matrix *transform, fz_cookie *cookie)
{
	fz_rect mediabox;
	pdf_bound_page(doc, page, &mediabox);
	fz_begin_page(ctx, dev, &mediabox, transform);
	pdf_run_page_contents(ctx, doc, page, dev, transform, cookie);
	fz_end_page(ctx, dev);
}

static inline void fz_gamma_pixmap_internal(fz_pixmap *pix, float gamma)
{
	unsigned char gamma_map[256];
	unsigned char *s = pix->samples;
	int k, x, y;

#ifdef BEZIERGAMMA
    //float t = (sqrtf(1.0f+(gamma*gamma-1.0)f*x) - 1.0f) / (gamma - 1.0f);
    //x = ((1.0f - gamma)*t + 2.0f * gamma ) * t / (gamma + 1.0f);
#else
    //x = powf(x, 1.0/gamma);        
#endif

	for (k = 0; k < 256; k++)
		gamma_map[k] = powf(k / 255.0f, gamma) * 255;

	for (y = 0; y < pix->h; y++) {
		for (x = 0; x < pix->w; x++) {
			for (k = 0; k < pix->n - 1; k++)
				s[k] = gamma_map[s[k]];
			s += pix->n;
		}
	}
}

static void drawpage(fz_context *ctx, pdf_document *doc, int pagenum)
{
	pdf_page *page;
	fz_buffer *buf = NULL;
	int start = gettime();
	fz_cookie cookie = { 0 };

	fz_try(ctx)
	{
		page = pdf_load_page(doc, pagenum - 1);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot load page %d in file '%s'", pagenum, filename);
	}

	if (showxml)
	{
		fz_try(ctx)
		{
            buf = fz_new_buffer(1024);
			dev = fz_new_trace_device(ctx, buf);
			pdf_run_page_internal(ctx, doc, page, &fz_identity, &cookie);
		}
		fz_always(ctx)
		{
            unsigned char *data = NULL;
            unsigned int data_size = fz_buffer_storage(buf, &data);
            write(1, data, data_size);
			fz_free_device(ctx, dev);
			dev = NULL;
            fz_drop_buffer(buf);
		}
		fz_catch(ctx)
		{
			pdf_free_page(doc, page);
            fz_drop_buffer(buf);
			fz_rethrow(ctx);
		}
	}

	if (showtext)
	{
		fz_stext_page *text = NULL;

		fz_try(ctx)
		{
			text = fz_new_stext_page(ctx);
			dev = fz_new_stext_device(ctx, sheet, text);
			pdf_run_page_internal(ctx, doc, page, &fz_identity, &cookie);
			fz_free_device(ctx, dev);
			dev = NULL;
			//fz_analyze_stext(ctx, sheet, text);
			if (showtext == TEXT_XML) {
				fz_print_stext_page_xml(ctx, outbuf, text);
			} else if (showtext == TEXT_PLAIN) {
				fz_print_stext_page(ctx, outbuf, text);
				fz_buffer_printf(outbuf, "\f\n");
			}
		}
		fz_always(ctx)
		{
			fz_free_device(ctx, dev);
			dev = NULL;
			fz_free_stext_page(ctx, text);
		}
		fz_catch(ctx)
		{
			pdf_free_page(doc, page);
			fz_rethrow(ctx);
		}
	}

	if (showmd5 || showtime)
		fz_warn(ctx->warn, "page %s %d", filename, pagenum);

	if (pdfout)
	{
		fz_matrix ctm;
		fz_rect bounds, tbounds;
		pdf_page *newpage;

		pdf_bound_page(doc, page, &bounds);
		fz_rotate(&ctm, rotation);
		tbounds = bounds;
		fz_transform_rect(&tbounds, &ctm);

		newpage = pdf_create_page(pdfout, bounds, 72, 0);

		fz_try(ctx)
		{
			dev = pdf_page_write(ctx, pdfout, newpage);
			pdf_run_page_internal(ctx, doc, page, &fz_identity, &cookie);
			dev = NULL;
		}
		fz_always(ctx)
		{
			dev = NULL;
		}
		fz_catch(ctx)
		{
			dev = NULL;
			pdf_free_page(doc, newpage);
			fz_rethrow(ctx);
		}
		pdf_insert_page(pdfout, newpage, INT_MAX);
		pdf_free_page(pdfout, newpage);
	}

	if (output && output_format == OUT_SVG)
	{
		float zoom;
		fz_matrix ctm;
		fz_rect bounds, tbounds;
		char filename_buf[256];
        int fd = -1;
        fz_buffer *buf;

		fz_snprintf(filename_buf, 255, output, pagenum);
		fd = open(filename_buf, O_WRONLY|O_CREAT|O_APPEND, 0644);
		if (fd < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot open file '%s': %s", filename_buf, strerror(errno));
		buf = fz_new_buffer(1024);

		pdf_bound_page(doc, page, &bounds);
		zoom = resolution / 72;
		fz_pre_rotate(fz_scale(&ctm, zoom, zoom), rotation);
		tbounds = bounds;
		fz_transform_rect(&tbounds, &ctm);

		fz_try(ctx)
		{
			dev = fz_new_svg_device(ctx, buf, tbounds.x1-tbounds.x0, tbounds.y1-tbounds.y0);
			pdf_run_page_internal(ctx, doc, page, &ctm, &cookie);
			fz_free_device(ctx, dev);
			dev = NULL;
		}
		fz_always(ctx)
		{
            unsigned char *data = NULL;
            unsigned int data_size = fz_buffer_storage(buf, &data);
            write(fd, data, data_size);
            close(fd);
			fz_free_device(ctx, dev);
			dev = NULL;
		    fz_drop_buffer(buf);
		}
		fz_catch(ctx)
		{
			pdf_free_page(doc, page);
			fz_rethrow(ctx);
		}
	}

	if ((output && output_format != OUT_SVG && output_format != OUT_PDF)|| showmd5 || showtime)
	{
		float zoom;
		fz_matrix ctm;
		fz_rect bounds, tbounds;
		fz_irect ibounds;
		fz_pixmap *pix = NULL;
		int w, h;

		pdf_bound_page(doc, page, &bounds);
		zoom = resolution / 72;
		fz_pre_scale(fz_rotate(&ctm, rotation), zoom, zoom);
		tbounds = bounds;
		fz_round_rect(&ibounds, fz_transform_rect(&tbounds, &ctm));

		/* Make local copies of our width/height */
		w = width;
		h = height;

		/* If a resolution is specified, check to see whether w/h are
		 * exceeded; if not, unset them. */
		if (res_specified)
		{
			int t;
			t = ibounds.x1 - ibounds.x0;
			if (w && t <= w)
				w = 0;
			t = ibounds.y1 - ibounds.y0;
			if (h && t <= h)
				h = 0;
		}

		/* Now w or h will be 0 unless they need to be enforced. */
		if (w || h)
		{
			float scalex = w / (tbounds.x1 - tbounds.x0);
			float scaley = h / (tbounds.y1 - tbounds.y0);
			fz_matrix scale_mat;

			if (fit)
			{
				if (w == 0)
					scalex = 1.0f;
				if (h == 0)
					scaley = 1.0f;
			}
			else
			{
				if (w == 0)
					scalex = scaley;
				if (h == 0)
					scaley = scalex;
			}
			if (!fit)
			{
				if (scalex > scaley)
					scalex = scaley;
				else
					scaley = scalex;
			}
			fz_scale(&scale_mat, scalex, scaley);
			fz_concat(&ctm, &ctm, &scale_mat);
			tbounds = bounds;
			fz_transform_rect(&tbounds, &ctm);
		}
		fz_round_rect(&ibounds, &tbounds);
		fz_rect_from_irect(&tbounds, &ibounds);

		/* TODO: banded rendering and multi-page ppm */

		fz_try(ctx)
		{
			int savealpha = (out_cs == CS_RGBA || out_cs == CS_GRAYALPHA);

			pix = fz_new_pixmap_with_bbox(ctx, colorspace, &ibounds);
			fz_pixmap_set_resolution(pix, resolution);

			if (savealpha)
				fz_clear_pixmap(pix);
			else
				fz_clear_pixmap_with_value(ctx, pix, 255);

			dev = fz_new_draw_device(ctx, pix);
			pdf_run_page_internal(ctx, doc, page, &ctm, &cookie);
			fz_free_device(ctx, dev);
			dev = NULL;

			if (invert)
				fz_invert_pixmap(ctx, pix);
			if (gamma_value != 1)
				fz_gamma_pixmap_internal(pix, gamma_value);

			if (savealpha)
				fz_unmultiply_pixmap(ctx, pix);

			if (output)
			{
                unsigned char *data = NULL;
                unsigned int data_size;
                int fd = -1;
				char fbuf[512];
				sprintf(fbuf, output, pagenum);
				if (output_format == OUT_PGM || output_format == OUT_PPM || output_format == OUT_PNM)
					buf = fz_write_pnm(ctx, pix);
				else if (output_format == OUT_PAM)
					buf = fz_write_pam(ctx, pix, savealpha);
				else if (output_format == OUT_PNG)
					buf = fz_write_png(ctx, pix, savealpha);
				else if (output_format == OUT_TGA)
					buf = fz_write_tga(ctx, pix, savealpha);
				else if (output_format == OUT_PWG)
				{
					if (strstr(output, "%d") != NULL)
						append = 0;
					if (out_cs == CS_MONO)
					{
						fz_bitmap *bit = fz_halftone_pixmap(ctx, pix, NULL);
						buf = fz_write_pwg_bitmap(ctx, bit, append, NULL);
						fz_drop_bitmap(ctx, bit);
					}
					else
						buf = fz_write_pwg(ctx, pix, append, NULL);
					append = 1;
				}
				else if (output_format == OUT_PCL)
				{
					fz_pcl_options options;

					fz_pcl_preset(ctx, &options, "ljet4");

					if (strstr(output, "%d") != NULL)
						append = 0;
					if (out_cs == CS_MONO)
					{
						fz_bitmap *bit = fz_halftone_pixmap(ctx, pix, NULL);
						buf = fz_write_pcl_bitmap(ctx, bit, append, &options);
						fz_drop_bitmap(ctx, bit);
					}
					else
						buf = fz_write_pcl(ctx, pix, append, &options);
					append = 1;
				}
				else if (output_format == OUT_PBM) {
					fz_bitmap *bit = fz_halftone_pixmap(ctx, pix, NULL);
					buf = fz_write_pbm(ctx, bit);
					fz_drop_bitmap(ctx, bit);
				}
		        fd = open(fbuf, O_WRONLY|O_CREAT|O_APPEND, 0644);
		        if (fd < 0)
			        fz_throw(ctx, FZ_ERROR_GENERIC, "cannot open file '%s': %s", fbuf, strerror(errno));
                data_size = fz_buffer_storage(buf, &data);
                write(fd, data, data_size);
                close(fd);
                fz_drop_buffer(buf);
			}

			if (showmd5)
			{
                unsigned char digest[16];
                char md5buf[35];

                fz_md5_pixmap(ctx, pix, digest);
                md5buf[0] = ' ';
                fz_md5_stringify(md5buf+1, digest);
                write(1, md5buf, 33);
			}
		}
		fz_always(ctx)
		{
			fz_free_device(ctx, dev);
			dev = NULL;
			fz_drop_pixmap(ctx, pix);
		}
		fz_catch(ctx)
		{
			pdf_free_page(doc, page);
			fz_rethrow(ctx);
		}
	}

	pdf_free_page(doc, page);

	if (showtime)
	{
		int end = gettime();
		int diff = end - start;

		if (diff < timing.min)
		{
			timing.min = diff;
			timing.minpage = pagenum;
			timing.minfilename = filename;
		}
		if (diff > timing.max)
		{
			timing.max = diff;
			timing.maxpage = pagenum;
			timing.maxfilename = filename;
		}
		timing.total += diff;
		timing.count ++;

		fz_warn(ctx->warn, " %dms", diff);
	}

	if (showmd5 || showtime)
		fz_warn(ctx->warn, "\n");

	fz_flush_warnings(ctx->warn);

	if (cookie.errors)
		errored = 1;
}

static void drawrange(fz_context *ctx, pdf_document *doc, char *range)
{
	int page, spage, epage, pagecount;
	char *spec, *dash;

	pagecount = pdf_count_pages(doc);
	spec = fz_strsep(&range, ",");
	while (spec)
	{
		dash = strchr(spec, '-');

		if (dash == spec)
			spage = epage = pagecount;
		else
			spage = epage = atoi(spec);

		if (dash)
		{
			if (strlen(dash) > 1)
				epage = atoi(dash + 1);
			else
				epage = pagecount;
		}

		spage = fz_clampi(spage, 1, pagecount);
		epage = fz_clampi(epage, 1, pagecount);

		if (spage < epage)
			for (page = spage; page <= epage; page++)
				drawpage(ctx, doc, page);
		else
			for (page = spage; page >= epage; page--)
				drawpage(ctx, doc, page);

		spec = fz_strsep(&range, ",");
	}
}

static void drawoutline(fz_context *ctx, pdf_document *doc)
{
	fz_outline *outline = pdf_load_outline(doc);
    fz_buffer *outlinebuf = NULL;

	fz_try(ctx)
	{
		outlinebuf = fz_new_buffer(1024);
		if (showoutline > 1)
			fz_print_outline_xml(outlinebuf, outline);
		else
			fz_print_outline(outlinebuf, outline);
	}
	fz_always(ctx)
	{
        unsigned char *data = NULL;
        unsigned int data_size = fz_buffer_storage(outlinebuf, &data);
        write(1, data, data_size);
		fz_drop_buffer(outlinebuf);
		fz_free_outline(ctx, outline);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}
}

static int
parse_colorspace(const char *name)
{
	int i;

	for (i = 0; i < nelem(cs_name_table); i++)
	{
		if (!strcmp(name, cs_name_table[i].name))
			return cs_name_table[i].colorspace;
	}
	return -1;
}

int main(int argc, char **argv)
{
	char *password = "";
	pdf_document *doc = NULL;
	int c;
	fz_context *ctx;
    fz_warn_context wctx;
    //fz_buffer *outbuf = NULL;

	while ((c = fz_getopt(argc, argv, "lo:p:r:R:b:c:dgmtx5G:Iw:h:fi")) != -1)
	{
		switch (c)
		{
		case 'o': output = fz_optarg; break;
		case 'p': password = fz_optarg; break;
		case 'r': resolution = atof(fz_optarg); res_specified = 1; break;
		case 'R': rotation = atof(fz_optarg); break;
		case 'b': alphabits = atoi(fz_optarg); break;
		case 'l': showoutline++; break;
		case 'm': showtime++; break;
		case 't': showtext++; break;
		case 'x': showxml++; break;
		case '5': showmd5++; break;
		case 'g': out_cs = CS_GRAY; break;
		case 'c': {
                      out_cs = parse_colorspace(fz_optarg);
                      if (out_cs < 0) {
	                      fz_warn(NULL, "Unknown colorspace \"%s\"\n", fz_optarg);
                          exit(-1);
                      }
                  }
                  break;
		case 'G': gamma_value = atof(fz_optarg); break;
		case 'w': width = atof(fz_optarg); break;
		case 'h': height = atof(fz_optarg); break;
		case 'f': fit = 1; break;
		case 'I': invert++; break;
		case 'i': ignore_errors = 1; break;
		case '?': fz_warn(&wctx, "%s: unknown option -%c\n", argv[0], c); usage(); break;
        case ':': fz_warn(&wctx, "%s: option requires argument -%c\n", argv[0], c); usage(); break;
		default: usage(); break;
		}
	}

	if (fz_optind == argc)
		usage();

	if (!showtext && !showxml && !showtime && !showmd5 && !showoutline && !output)
	{
		write(2, "mudraw: nothing to do, exiting\n", 31);
		exit(0);
	}

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_DEFAULT);
	if (!ctx)
	{
		write(2, "mudraw: error: cannot initialise context\n", 41);
		exit(1);
	}

	/* Determine output type */
	output_format = OUT_PNG;
	if (output)
	{
		char *suffix = output;
		int i;

		for (i = 0; i < nelem(suffix_table); i++)
		{
			char *s = strstr(suffix, suffix_table[i].suffix);

			if (s != NULL)
			{
				suffix = s+1;
				output_format = suffix_table[i].format;
				i = 0;
			}
		}
	}

	{
		int i, j;

		for (i = 0; i < nelem(format_cs_table); i++)
		{
			if (format_cs_table[i].format == output_format)
			{
				if (out_cs == CS_UNSET)
					out_cs = format_cs_table[i].default_cs;
				for (j = 0; j < nelem(format_cs_table[i].permitted_cs); j++)
				{
					if (format_cs_table[i].permitted_cs[j] == out_cs)
						break;
				}
				if (j == nelem(format_cs_table[i].permitted_cs))
				{
					fz_warn(ctx->warn, "Unsupported colorspace for this format\n");
					exit(1);
				}
			}
		}
	}

	switch (out_cs)
	{
	case CS_MONO:
	case CS_GRAY:
	case CS_GRAYALPHA:
		colorspace = fz_device_gray;
		break;
	case CS_RGB:
	case CS_RGBA:
		colorspace = fz_device_rgb;
		break;
	default:
		fz_warn(ctx->warn, "Unknown colorspace!\n");
		exit(1);
		break;
	}

    if (output_format == OUT_PDF)
        pdfout = pdf_create_document(ctx);

	timing.count = 0;
	timing.total = 0;
	timing.min = 1 << 30;
	timing.max = 0;
	timing.minpage = 0;
	timing.maxpage = 0;
	timing.minfilename = "";
	timing.maxfilename = "";

	outbuf = fz_new_buffer(65536);
	if (showxml || showtext == TEXT_XML)
		fz_buffer_printf(outbuf, "<?xml version=\"1.0\"?>\n");

	if (showtext)
		sheet = fz_new_stext_sheet(ctx);

    if (showxml)
		dev = fz_new_trace_device(ctx, fz_new_buffer(1024));

	fz_try(ctx)
	{
		while (fz_optind < argc)
		{
			fz_try(ctx)
			{
                unsigned char *inbuf;
                size_t insz;
				filename = argv[fz_optind++];
				files++;

				fz_try(ctx)
				{
                    inbuf = slurp_file(filename, &insz);
			        doc = pdf_open_document_no_run(ctx, inbuf, insz);
				}
				fz_catch(ctx)
				{
					fz_rethrow_message(ctx, "cannot open document: %s", filename);
				}

				if (pdf_needs_password(doc))
				{
					if (!pdf_authenticate_password(doc, password))
						fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", filename);
				}

#ifdef EPUB
				fz_layout_document(ctx, doc, layout_w, layout_h, layout_em);
#endif

				if (showxml || showtext == TEXT_XML)
					fz_buffer_printf(outbuf, "<document name=\"%s\">\n", filename);

				if (showoutline)
					drawoutline(ctx, doc);

				if (showtext || showxml || showtime || showmd5 || output)
				{
					if (fz_optind == argc || !isrange(argv[fz_optind]))
						drawrange(ctx, doc, "1-");
					if (fz_optind < argc && isrange(argv[fz_optind]))
						drawrange(ctx, doc, argv[fz_optind++]);
				}

				if (showxml || showtext == TEXT_XML)
					fz_buffer_printf(outbuf, "</document>\n");

				pdf_close_document(doc);
				doc = NULL;
			}
			fz_catch(ctx)
			{
				if (!ignore_errors)
					fz_rethrow(ctx);

				pdf_close_document(doc);
				doc = NULL;
				fz_warn(ctx->warn, "ignoring error in '%s'", filename);
			}
		}
	}
	fz_catch(ctx)
	{
		pdf_close_document(doc);
		fz_warn(ctx->warn, "error: cannot draw '%s'\n", filename);
		errored = 1;
	}

	if (pdfout) {
        fz_buffer *pdfoutbuf;
		fz_write_options opts = { 0 };
        unsigned char *data = NULL;
        unsigned int data_size;
		int fd = open(output, O_WRONLY|O_CREAT|O_APPEND, 0644);
		if (fd < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot open file '%s': %s", output, strerror(errno));
		fz_free_device(ctx, dev);
		pdfoutbuf = pdf_write_document(pdfout, &opts);
        data_size = fz_buffer_storage(pdfoutbuf, &data);
        write(fd, data, data_size);
		pdf_close_document(pdfout);
        close(fd);
		fz_drop_buffer(pdfoutbuf);
	}

	if (showtext)
		fz_free_stext_sheet(ctx, sheet);

	if (showxml || showtext) {
        unsigned char *data = NULL;
        unsigned int data_size = fz_buffer_storage(outbuf, &data);
        write(1, data, data_size);
		fz_drop_buffer(outbuf);
	}

	if (showtime && timing.count > 0)
	{
		if (files == 1)
		{
			printf("total %dms / %d pages for an average of %dms\n",
				timing.total, timing.count, timing.total / timing.count);
			printf("fastest page %d: %dms\n", timing.minpage, timing.min);
			printf("slowest page %d: %dms\n", timing.maxpage, timing.max);
		}
		else
		{
			printf("total %dms / %d pages for an average of %dms in %d files\n",
				timing.total, timing.count, timing.total / timing.count, files);
			printf("fastest page %d: %dms (%s)\n", timing.minpage, timing.min, timing.minfilename);
			printf("slowest page %d: %dms (%s)\n", timing.maxpage, timing.max, timing.maxfilename);
		}
	}

	fz_free_context(ctx);
	return (errored != 0);
}

#ifdef _MSC_VER
int wmain(int argc, wchar_t *wargv[])
{
	char **argv = fz_argv_from_wargv(argc, wargv);
	int ret = main(argc, argv);
	fz_free_argv(argc, argv);
	return ret;
}
#endif
