/*
 * pdfextract -- the ultimate way to extract images and fonts from pdfs
 */

#include "mupdf/pdf.h"

static pdf_document *doc = NULL;
static fz_context *ctx = NULL;
static unsigned int dumpraw = 0;
unsigned char *slurp_file(const char *filename, size_t *filesize);

struct fz_font_context_s {
	int ctx_refs;
};

void fz_new_font_context(fz_context *ctx)
{
	ctx->font = fz_malloc_struct(ctx, fz_font_context);
	ctx->font->ctx_refs = 1;
}

fz_font_context *
fz_keep_font_context(fz_context *ctx)
{
	if (!ctx || !ctx->font)
		return NULL;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	ctx->font->ctx_refs++;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	return ctx->font;
}

void fz_drop_font_context(fz_context *ctx)
{
	int drop;
	if (!ctx || !ctx->font)
		return;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	drop = --ctx->font->ctx_refs;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	if (drop == 0) {
		fz_free(ctx, ctx->font);
    }
}

static void usage(void)
{
	const char *usage_str = "usage: pdfextract [options] file.pdf [object numbers]\n"
	"\t-p\tpassword\n"
	"\t-r\tconvert images to rgb\n";
    write(1, usage_str, strlen(usage_str));
	exit(1);
}

static int isfontdesc(pdf_obj *obj)
{
	pdf_obj *type = pdf_dict_gets(obj, "Type");
	return pdf_is_name(type) && !strcmp(pdf_to_name(type), "FontDescriptor");
}

static void dumpstream(const char *filename, unsigned int num, pdf_obj *strm)
{
    int n, fd = -1;
	unsigned char *data;
    unsigned int len;
	fz_buffer *buf = NULL;

    if (dumpraw)
        buf = pdf_load_raw_stream(doc, num, pdf_to_gen(strm));
    else
        buf = pdf_load_stream(doc, num, pdf_to_gen(strm));

	fd = open(filename, O_WRONLY|O_CREAT|O_APPEND, 0644);
	if (fd < 0)
		fz_throw(ctx, FZ_ERROR_GENERIC, "Error creating font file");

	len = fz_buffer_storage(buf, &data);
    n = write(fd, data, len);
	if (n < len)
		fz_throw(ctx, FZ_ERROR_GENERIC, "Error writing font file");

	if (close(fd) < 0)
		fz_throw(ctx, FZ_ERROR_GENERIC, "Error closing font file");

	fz_drop_buffer(buf);
}

static void savefont(pdf_obj *dict, int num)
{
	char namebuf[1024];
	char *subtype;
	pdf_obj *stream = NULL;
	pdf_obj *obj;
	char *ext = "";
	char *fontname = "font";

	obj = pdf_dict_gets(dict, "FontName");
	if (obj)
		fontname = pdf_to_name(obj);

	obj = pdf_dict_gets(dict, "FontFile");
	if (obj)
	{
		stream = obj;
		ext = "pfa";
	}

	obj = pdf_dict_gets(dict, "FontFile2");
	if (obj)
	{
		stream = obj;
		ext = "ttf";
	}

	obj = pdf_dict_gets(dict, "FontFile3");
	if (obj)
	{
		stream = obj;

		obj = pdf_dict_gets(obj, "Subtype");
		if (obj && !pdf_is_name(obj))
			fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid font descriptor subtype");

		subtype = pdf_to_name(obj);
		if (!strcmp(subtype, "Type1C"))
			ext = "cff";
		else if (!strcmp(subtype, "CIDFontType0C"))
			ext = "cid";
		else if (!strcmp(subtype, "OpenType"))
			ext = "otf";
		else
			fz_throw(ctx, FZ_ERROR_GENERIC, "Unhandled font type '%s'", subtype);
	}

	if (!stream)
	{
		fz_warn(ctx->warn, "Unhandled font type");
		return;
	}

	snprintf(namebuf, sizeof(namebuf), "%s-%04d.%s", fontname, num, ext);
	fz_warn(ctx->warn, "Extracting font %s...", namebuf);
    dumpstream(namebuf, pdf_to_num(stream), stream);
}

static void showobject(int num)
{
	pdf_obj *obj;

	if (!doc)
		fz_throw(ctx, FZ_ERROR_GENERIC, "no file specified");

	obj = pdf_load_object(doc, num, 0);

	if (isfontdesc(obj)) {
		savefont(obj, num);
    } else {
	    pdf_obj *obj2 = pdf_dict_gets(obj, "Filter");
        if (obj2) {
	        char namebuf[1024];
	        unsigned int len = pdf_to_int(pdf_dict_gets(obj, "Length"));
            fz_warn(ctx->warn, "Filter: %s, Length: %u", pdf_to_name(obj2), len);
            fz_warn(ctx->warn, "Dumping stream %u...", num);
	        snprintf(namebuf, sizeof(namebuf), "stream-%04d", num);
            dumpstream(namebuf, num, obj);
        } else {
            fz_warn(ctx->warn, "obj %u is of type %s", num, pdf_objkindstr(obj));
        }
    }

	pdf_drop_obj(obj);
}

int main(int argc, char **argv)
{
	char *infile;
	char *password = "";
	int c, o;
    unsigned char *inbuf;
    size_t insz;
    fz_warn_context wctx;

	while ((c = fz_getopt(argc, argv, "p:r")) != -1)
	{
		switch (c)
		{
		case 'p': password = fz_optarg; break;
		case 'r': dumpraw = 1; break;
		default: usage(); break;
		}
	}

	if (fz_optind == argc)
		usage();

	infile = argv[fz_optind++];

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_DEFAULT);
	if (!ctx)
	{
		write(2, "Cannot initialise context\n", 27);
		exit(1);
	}

    inbuf = slurp_file(infile, &insz);
	doc = pdf_open_document_no_run(ctx, inbuf, insz);
	if (pdf_needs_password(doc))
		if (!pdf_authenticate_password(doc, password))
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", infile);

	if (fz_optind == argc)
	{
		int len = pdf_count_objects(doc);
		for (o = 0; o < len; o++)
			showobject(o);
	}
	else
	{
		while (fz_optind < argc)
		{
			showobject(atoi(argv[fz_optind]));
			fz_optind++;
		}
	}

	pdf_close_document(doc);
	fz_flush_warnings(ctx->warn);
	fz_free_context(ctx);
	return 0;
}

