/*
 * pdfshow -- the ultimate pdf debugging tool
 */

#include "mupdf/pdf.h"

static pdf_document *doc = NULL;
static fz_context *ctx = NULL;
static fz_buffer *pdfshow_outbuf = NULL;
static int showbinary = 0;
static int showdecode = 1;
static int showcolumn;
unsigned char *slurp_file(const char *filename, size_t *filesize);

struct fz_font_context_s {
	int ctx_refs;
};

void fz_new_font_context(fz_context *ctx)
{
	ctx->font = fz_malloc_struct(ctx, fz_font_context);
	ctx->font->ctx_refs = 1;
}

fz_font_context *
fz_keep_font_context(fz_context *ctx)
{
	if (!ctx || !ctx->font)
		return NULL;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	ctx->font->ctx_refs++;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	return ctx->font;
}

void fz_drop_font_context(fz_context *ctx)
{
	int drop;
	if (!ctx || !ctx->font)
		return;
	fz_lock(ctx, FZ_LOCK_ALLOC);
	drop = --ctx->font->ctx_refs;
	fz_unlock(ctx, FZ_LOCK_ALLOC);
	if (drop == 0) {
		fz_free(ctx, ctx->font);
    }
}

static void usage(void)
{
	const char *usage_str = "usage: pdfshow [options] file.pdf [grep] [xref] [trailer] [pages] [object numbers]\n\t-p\tpassword\n";
    write(1, usage_str, strlen(usage_str));
	exit(1);
}

static void showtrailer(void)
{
	fz_buffer_puts(pdfshow_outbuf, "trailer:\n");
	pdf_bufprint_obj(pdfshow_outbuf, pdf_trailer(doc));
    fz_write_buffer_byte(pdfshow_outbuf, '\n');
}

static void showencrypt(void)
{
	pdf_obj *encrypt = pdf_dict_gets(pdf_trailer(doc), "Encrypt");
	if (!encrypt)
		fz_throw(ctx, FZ_ERROR_GENERIC, "document not encrypted");
	fz_buffer_puts(pdfshow_outbuf, "encryption dictionary\n");
	pdf_bufprint_obj(pdfshow_outbuf, pdf_resolve_indirect(encrypt));
    fz_write_buffer_byte(pdfshow_outbuf, '\n');
}

static void showxref(void)
{
	pdf_bufprint_xref(pdfshow_outbuf, doc);
}

static void showpagetree(void)
{
	pdf_obj *ref;
	int count;
	int i;

	count = pdf_count_pages(doc);
	for (i = 0; i < count; i++) {
		ref = pdf_lookup_page_obj(doc, i);
		fz_buffer_printf(pdfshow_outbuf, "page %d = %d %d R\n", i + 1, pdf_to_num(ref), pdf_to_gen(ref));
	}
    fz_write_buffer_byte(pdfshow_outbuf, '\n');
}

static void showobject_no_stm(int num, int gen)
{
	pdf_obj *obj = pdf_load_object(doc, num, gen);

	if (!pdf_is_stream(doc, num, gen)) {
		fz_buffer_printf(pdfshow_outbuf, "%u 0 obj\n", num);
        pdfshow_outbuf->data[pdfshow_outbuf->len++] = '\n';
	    pdf_bufprint_obj(pdfshow_outbuf, obj);
        fz_buffer_puts(pdfshow_outbuf, "endobj\n\n");
	}

	pdf_drop_obj(obj);
}

static void showgrep(char *filename)
{
	pdf_obj *obj = NULL;
	unsigned int i, len = pdf_count_objects(doc);

	for (i = 0; i < len; i++)
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, i);
		if (entry->type == 'n' || entry->type == 'o')
		{
			fz_try(ctx)
			{
				obj = pdf_load_object(doc, i, 0);
			}
			fz_catch(ctx)
			{
				fz_warn(ctx->warn, "skipping object (%d 0 R)", i);
				continue;
			}

			pdf_sort_dict(obj);

	        fz_buffer_printf(pdfshow_outbuf, "%s:%d: ", filename, i);
	        pdf_bufprint_obj(pdfshow_outbuf, obj);
			pdf_drop_obj(obj);
		}
	}

	fz_buffer_printf(pdfshow_outbuf, "%s:", filename);
    showtrailer();
}

int main(int argc, char **argv)
{
	char *password = NULL; /* don't throw errors if encrypted */
	char *filename;
	int c;
    unsigned char *inbuf;
    size_t insz;
    fz_warn_context wctx;

	while ((c = fz_getopt(argc, argv, "p:be")) != -1)
	{
		switch (c)
		{
		case 'p': password = fz_optarg; break;
		case 'b': showbinary = 1; break;
		case 'e': showdecode = 0; break;
		case '?': fz_warn(&wctx, "%s: unknown option -%c\n", argv[0], c); usage(); break;
        case ':': fz_warn(&wctx, "%s: option requires argument -%c\n", argv[0], c); usage(); break;
		default: usage(); break;
		}
	}

	if (fz_optind == argc)
		usage();

	filename = argv[fz_optind++];

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_UNLIMITED);
	if (!ctx)
	{
		write(2, "Cannot initialise context\n", 27);
		exit(1);
	}

    pdfshow_outbuf = fz_new_buffer(65536);

	fz_try(ctx)
	{
        inbuf = slurp_file(filename, &insz);
	    doc = pdf_open_document_no_run(ctx, inbuf, insz);
		if (pdf_needs_password(doc))
			if (!pdf_authenticate_password(doc, password))
				fz_warn(ctx->warn, "cannot authenticate password: %s", filename);

		if (fz_optind == argc)
			showtrailer();

		while (fz_optind < argc)
		{
	        if (!doc && (argv[fz_optind][0] != 'g'))
		        fz_throw(ctx, FZ_ERROR_GENERIC, "no file specified");
			switch (argv[fz_optind][0])
			{
			case 't': showtrailer(); break;
			case 'e': showencrypt(); break;
			case 'x': showxref(); break;
			case 'p': showpagetree(); break;
			case 'g': showgrep(filename); break;
			default: {
                unsigned int k, uplim = fz_atoi(argv[fz_optind]);
                for (k = 0; k < uplim && k < pdf_xref_len(doc); k++) {
                    showobject_no_stm(k, 0);
                }
            }
            break;
			}
			fz_optind++;
		}
	}
	fz_catch(ctx)
	{
	}

    {
        unsigned char *data = NULL;
        unsigned int data_size = fz_buffer_storage(pdfshow_outbuf, &data);
        write(1, data, data_size);
	    write(1, "\n", 1);
    }
	fz_drop_buffer(pdfshow_outbuf);

	pdf_close_document(doc);
	fz_free_context(ctx);
	return 0;
}

