/*
 * PDF cleaning tool: general purpose pdf syntax washer.
 *
 * Rewrite PDF with pretty printed objects.
 * Garbage collect unreachable objects.
 * Inflate compressed streams.
 * Create subset documents.
 *
 * TODO: linearize document for fast web view
 */

#include "mupdf/pdf.h"
unsigned char *slurp_file(const char *filename, size_t *filesize);

typedef struct globals_s
{
	pdf_document *doc;
	fz_context *ctx;
} globals;
static globals glo;

static void usage(void)
{
    printf("usage: mutool clean [options] input.pdf [output.pdf] [pages]\n"
		"\t-p -\tpassword\n"
		"\t-g\tgarbage collect unused objects\n"
		"\t-f\ttoggle decompression of font streams\n"
		"\tpages\tcomma separated list of ranges\n");
	exit(1);
}

static int
string_in_names_list(fz_context *ctx, pdf_obj *p, pdf_obj *names_list)
{
	unsigned int i, n = pdf_array_len(names_list);
	char *str = pdf_to_str_buf(p);

	for (i = 0; i < n ; i += 2) {
		if (!strcmp(pdf_to_str_buf(pdf_array_get(names_list, i)), str))
			return 1;
	}
	return 0;
}

/*
 * Recreate page tree to only retain specified pages.
 */

static void retainpage(pdf_document *doc, pdf_obj *parent, pdf_obj *kids, int page)
{
	pdf_obj *pageref = pdf_lookup_page_obj(doc, page-1);
	pdf_obj *pageobj = pdf_resolve_indirect(pageref);
	pdf_dict_puts(pageobj, "Parent", parent);
	/* Store page object in new kids array */
	pdf_array_push(kids, pageref);
}

static void retainpages(fz_context *ctx, globals *glo, int argc, char **argv)
{
	pdf_obj *oldroot, *root, *pages, *kids, *countobj, *parent, *olddests;
	pdf_document *doc = glo->doc;
	int argidx = 0;
	pdf_obj *names_list = NULL;
	int pagecount;
	int i;

	/* Keep only pages/type and (reduced) dest entries to avoid
	 * references to unretained pages */
	oldroot = pdf_dict_gets(pdf_trailer(doc), "Root");
	pages = pdf_dict_gets(oldroot, "Pages");
	olddests = pdf_load_name_tree(doc, "Dests");

	root = pdf_new_dict(doc, 2);
	pdf_dict_puts(root, "Type", pdf_dict_gets(oldroot, "Type"));
	pdf_dict_puts(root, "Pages", pdf_dict_gets(oldroot, "Pages"));
	pdf_update_object(doc, pdf_to_num(oldroot), root);
	pdf_drop_obj(root);

	/* Create a new kids array with only the pages we want to keep */
	parent = pdf_new_indirect(doc, pdf_to_num(pages), pdf_to_gen(pages));
	kids = pdf_new_array(doc, 1);

	/* Retain pages specified */
	while (argc - argidx)
	{
		int page, spage, epage;
		char *spec, *dash;
		char *pagelist = argv[argidx];

		pagecount = pdf_count_pages(doc);
		spec = fz_strsep(&pagelist, ",");
		while (spec)
		{
			dash = strchr(spec, '-');

			if (dash == spec)
				spage = epage = pagecount;
			else
				spage = epage = fz_atoi(spec);

			if (dash)
			{
				if (strlen(dash) > 1)
					epage = fz_atoi(dash + 1);
				else
					epage = pagecount;
			}

			spage = fz_clampi(spage, 1, pagecount);
			epage = fz_clampi(epage, 1, pagecount);

	        fz_try(ctx)
	        {
			    if (spage < epage)
				    for (page = spage; page <= epage; ++page)
					    retainpage(doc, parent, kids, page);
			    else
				    for (page = spage; page >= epage; --page)
					    retainpage(doc, parent, kids, page);
	        }
	        fz_catch(ctx)
	        {
            }

			spec = fz_strsep(&pagelist, ",");
		}

		argidx++;
	}

	pdf_drop_obj(parent);

	/* Update page count and kids array */
	countobj = pdf_new_int(doc, pdf_array_len(kids));
	pdf_dict_puts(pages, "Count", countobj);
	pdf_drop_obj(countobj);
	pdf_dict_puts(pages, "Kids", kids);
	pdf_drop_obj(kids);

	/* Also preserve the (partial) Dests name tree */
	if (olddests)
	{
		pdf_obj *names = pdf_new_dict(doc, 1);
		pdf_obj *dests = pdf_new_dict(doc, 1);
		int len = pdf_dict_len(olddests);

		names_list = pdf_new_array(doc, 32);

		for (i = 0; i < len; i++)
		{
			pdf_obj *key = pdf_dict_get_key(olddests, i);
			pdf_obj *val = pdf_dict_get_val(olddests, i);
			pdf_obj *dest = pdf_dict_gets(val, "D");

			dest = pdf_array_get(dest ? dest : val, 0);
			if (pdf_array_contains(pdf_dict_gets(pages, "Kids"), dest))
			{
				pdf_obj *key_str = pdf_new_string(doc, pdf_to_name(key), strlen(pdf_to_name(key)));
				pdf_array_push(names_list, key_str);
				pdf_array_push(names_list, val);
				pdf_drop_obj(key_str);
			}
		}

		root = pdf_dict_gets(pdf_trailer(doc), "Root");
		pdf_dict_puts(dests, "Names", names_list);
		pdf_dict_puts(names, "Dests", dests);
		pdf_dict_puts(root, "Names", names);

		pdf_drop_obj(names);
		pdf_drop_obj(dests);
		pdf_drop_obj(names_list);
		pdf_drop_obj(olddests);
	}

	/* Force the next call to pdf_count_pages to recount */
	glo->doc->page_count = 0;
}

int main(int argc, char **argv)
{
    unsigned char *inbuf;
    size_t insz;
	char *infile = NULL, *outfile = "out.pdf";
	char *password = "";
	int c;
	fz_write_options opts;
	int errors = 0;
	fz_context *ctx;
    fz_warn_context wctx;

	opts.do_garbage = 0;
	opts.do_expand = 0;
    opts.do_deflate = 0;
	opts.continue_on_error = 1;
	opts.errors = &errors;

	while ((c = fz_getopt(argc, argv, "fglp:z")) != -1)
	{
		switch (c)
		{
		case 'p': password = fz_optarg; break;
		case 'g': opts.do_garbage++; break;
		case 'f': opts.do_expand ^= fz_expand_fonts; break;
		case 'l': break;
		case 'z': opts.do_deflate++; break;
		default: usage(); break;
		}
	}

	if (argc - fz_optind < 1)
		usage();

	infile = argv[fz_optind++];

	if (argc - fz_optind > 0) {
		outfile = argv[fz_optind++];
	}

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_UNLIMITED);
	if (!ctx)
	{
		write(2, "Cannot initialise context\n", 27);
		exit(1);
	}

	fz_try(ctx)
	{
        fz_buffer *pdfoutbuf;
        unsigned char *data = NULL;
        unsigned int data_size;
        inbuf = slurp_file(infile, &insz);
	    glo.doc = pdf_open_document_no_run(ctx, inbuf, insz);
		if (pdf_needs_password(glo.doc))
			if (!pdf_authenticate_password(glo.doc, password))
				fz_throw(glo.ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", infile);

		/* Only retain the specified subset of the pages */
		if (argc)
			retainpages(ctx, &glo, argc - fz_optind, &argv[fz_optind]);

		pdfoutbuf = pdf_write_document(glo.doc, &opts);
        fprintf(stderr, "asdf\n");
        {
		    int fd = open(outfile, O_WRONLY|O_CREAT|O_APPEND, 0644);
		    if (fd < 0)
			    fz_throw(ctx, FZ_ERROR_GENERIC, "cannot open file '%s': %s", outfile, strerror(errno));
            data_size = fz_buffer_storage(pdfoutbuf, &data);
            write(fd, data, data_size);
            close(fd);
        }
		fz_drop_buffer(pdfoutbuf);
	}
	fz_always(ctx)
	{
		pdf_close_document(glo.doc);
	}
	fz_catch(ctx)
	{
		if (opts.errors)
			opts.errors++;
        errors++;
	}
	fz_free_context(ctx);
	return errors == 0;
}

