/*
 * mudraw -- command line tool for drawing pdf/xps/cbz documents
 */

#include "mupdf/fitz.h"
#include "mupdf/pdf.h" /* for mujstest */
#include <sys/time.h>
unsigned char *slurp_file(const char *filename, size_t *filesize);
void fz_stext_begin_page(fz_device *dev, const fz_rect *mediabox, const fz_matrix *ctm);
void fz_stext_end_page(fz_device *dev);
void fz_stext_free_user(fz_device *dev);

enum { TEXT_PLAIN = 1, TEXT_XML = 2 };

/*
	A useful bit of bash script to call this to generate mjs files:
	for f in tests_private/pdf/forms/v1.3/ *.pdf ; do g=${f%.*} ; echo $g ; ../mupdf.git/win32/debug/mudraw.exe -j $g.mjs $g.pdf ; done

	Remove the space from "/ *.pdf" before running - can't leave that
	in here, as it causes a warning about a possibly malformed comment.
*/

static char lorem[] =
"Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum "
"vehicula augue id est lobortis mollis. Aenean vestibulum metus sed est "
"gravida non tempus lacus aliquet. Nulla vehicula lobortis tincidunt. "
"Donec malesuada nisl et lacus condimentum nec tincidunt urna gravida. "
"Sed dapibus magna eu velit ultrices non rhoncus risus lacinia. Fusce "
"vitae nulla volutpat elit dictum ornare at eu libero. Maecenas felis "
"enim, tempor a tincidunt id, commodo consequat lectus.\n"
"Morbi tincidunt adipiscing lacus eu dignissim. Pellentesque augue elit, "
"ultrices vitae fermentum et, faucibus et purus. Nam ante libero, lacinia "
"id tincidunt at, ultricies a lorem. Donec non neque at purus condimentum "
"eleifend quis sit amet libero. Sed semper, mi ut tempus tincidunt, lacus "
"eros pellentesque lacus, id vehicula est diam eu quam. Integer tristique "
"fringilla rhoncus. Phasellus convallis, justo ut mollis viverra, dui odio "
"euismod ante, nec fringilla nisl mi ac diam.\n"
"Maecenas mi urna, ornare commodo feugiat id, cursus in massa. Vivamus "
"augue augue, aliquam at varius eu, venenatis fermentum felis. Sed varius "
"turpis a felis ultrices quis aliquet nunc tincidunt. Suspendisse posuere "
"commodo nunc non viverra. Praesent condimentum varius quam, vel "
"consectetur odio volutpat in. Sed malesuada augue ut lectus commodo porta. "
"Vivamus eget mauris sit amet diam ultrices sollicitudin. Cras pharetra leo "
"non elit lacinia vulputate.\n"
"Donec ac enim justo, ornare scelerisque diam. Ut vel ante at lorem "
"placerat bibendum ultricies mattis metus. Phasellus in imperdiet odio. "
"Proin semper lacinia libero, sed rutrum eros blandit non. Duis tincidunt "
"ligula est, non pellentesque mauris. Aliquam in erat scelerisque lacus "
"dictum suscipit eget semper magna. Nullam luctus imperdiet risus a "
"semper.\n"
"Curabitur sit amet tempor sapien. Quisque et tortor in lacus dictum "
"pulvinar. Nunc at nisl ut velit vehicula hendrerit. Mauris elementum "
"sollicitudin leo ac ullamcorper. Proin vel leo nec justo tempus aliquet "
"nec ut mi. Pellentesque vel nisl id dui hendrerit fermentum nec quis "
"tortor. Proin eu sem luctus est consequat euismod. Vestibulum ante ipsum "
"primis in faucibus orci luctus et ultrices posuere cubilia Curae; Fusce "
"consectetur ultricies nisl ornare dictum. Cras sagittis consectetur lorem "
"sed posuere. Mauris accumsan laoreet arcu, id molestie lorem faucibus eu. "
"Vivamus commodo, neque nec imperdiet pretium, lorem metus viverra turpis, "
"malesuada vulputate justo eros sit amet neque. Nunc quis justo elit, non "
"rutrum mauris. Maecenas blandit condimentum nibh, nec vulputate orci "
"pulvinar at. Proin sed arcu vel odio tempus lobortis sed posuere ipsum. Ut "
"feugiat pellentesque tortor nec ornare.\n";

static int errored = 0;

static char *filename;
static int files = 0;
fz_buffer *outbuf = NULL;

static void usage(void)
{
    const char *usage_str = "usage: pdftotext [options] input [pages]\n"
		"\t-p -\tpassword\n"
		"\t-t\tshow text (-tt for xml, -ttt for more verbose xml)\n"
		"\t-l\tprint outline\n"
		"\tpages\tcomma separated list of ranges\n";
    write(1, usage_str, strlen(usage_str));
}

static int isrange(char *s)
{
	while (*s)
	{
		if ((*s < '0' || *s > '9') && *s != '-' && *s != ',')
			return 0;
		s++;
	}
	return 1;
}

static void pdf_run_page_internal(pdf_document *doc, pdf_page *page, fz_device *dev, const fz_matrix *transform, fz_cookie *cookie)
{
	//fz_annot *annot;
	fz_rect mediabox;

	pdf_bound_page(doc, page, &mediabox);
	fz_stext_begin_page(dev, &mediabox, transform);
	pdf_run_page_contents(doc->ctx, doc, page, dev, transform, cookie);
	fz_stext_end_page(dev);
}

static void dump_buffer(fz_context *ctx, fz_buffer *outbuf)   
{
    unsigned char *data = NULL;
    unsigned int data_size = fz_buffer_storage(outbuf, &data);
    write(1, data, data_size);
}

static void drawpage(fz_context *ctx, pdf_document *doc, int pagenum, fz_stext_sheet *sheet, int do_analyze)
{
	pdf_page *page;
	fz_device *dev = NULL;
	fz_cookie cookie = { 0 };

	fz_try(ctx)
	{
		page = pdf_load_page(doc, pagenum - 1);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot load page %d in file '%s'", pagenum, filename);
	}

	{
		fz_stext_page *text = NULL;

		fz_try(ctx)
		{
			text = fz_new_stext_page(ctx);
			dev = fz_new_stext_device(ctx, sheet, text);
			pdf_run_page_internal(doc, page, dev, &fz_identity, &cookie);
            fz_stext_free_user(dev);
            fz_free(ctx, dev);
			dev = NULL;
            if (do_analyze)
			    fz_analyze_stext(sheet, text);
			fz_print_stext_page(ctx, outbuf, text);
			fz_write_buffer(outbuf, "\f\n", 2);
            dump_buffer(ctx, outbuf);
		}
		fz_always(ctx)
		{
            fz_stext_free_user(dev);
            //fz_free(ctx, dev);
			dev = NULL;
			fz_free_stext_page(ctx, text);
		}
		fz_catch(ctx)
		{
			pdf_free_page(doc, page);
			fz_rethrow(ctx);
		}
	}

	pdf_free_page(doc, page);
	fz_flush_warnings(ctx->warn);
	if (cookie.errors)
		errored = 1;
}

static char *
fz_strsep_(char **stringp, const char delim)
{
	char *ret = *stringp;
	if (!ret) return NULL;
	if ((*stringp = strchr(*stringp, delim)))
		*((*stringp)++) = '\0';
	return ret;
}

#if 0
static void drawrange(fz_context *ctx, pdf_document *doc, char *range, fz_stext_sheet *sheet, int do_analyze)
{
    unsigned int page, pagecount;
	pagecount = pdf_count_pages(doc);
	for (page = 0; page < pagecount; page++)
		drawpage(ctx, doc, page+1, sheet, do_analyze);
}
#endif

static void drawrange(fz_context *ctx, pdf_document *doc, char *range, fz_stext_sheet *sheet, int do_analyze)
{
	int page, spage, epage, pagecount;
	char *spec, *dash;

	pagecount = pdf_count_pages(doc);
	spec = fz_strsep(&range, ",");
	while (spec)
	{
		dash = strchr(spec, '-');

		if (dash == spec)
			spage = epage = pagecount;
		else
			spage = epage = fz_atoi(spec);

		if (dash)
		{
			if (strlen(dash) > 1)
				epage = fz_atoi(dash + 1);
			else
				epage = pagecount;
		}

		spage = fz_clampi(spage, 1, pagecount);
		epage = fz_clampi(epage, 1, pagecount);

		if (spage < epage)
			for (page = spage; page <= epage; page++)
		        drawpage(ctx, doc, page, sheet, do_analyze);
		else
			for (page = spage; page >= epage; page--)
		        drawpage(ctx, doc, page, sheet, do_analyze);

		spec = fz_strsep(&range, ",");
	}
}

static void drawoutline(fz_context *ctx, pdf_document *doc, int showoutline)
{
	fz_outline *outline = pdf_load_outline(doc);
    fz_buffer *outlinebuf = NULL;

	outlinebuf = fz_new_buffer(1024);
	if (showoutline > 1)
		fz_print_outline_xml(outlinebuf, outline);
	else
		fz_print_outline(outlinebuf, outline);

	{
        unsigned char *data = NULL;
        unsigned int data_size = fz_buffer_storage(outlinebuf, &data);
        write(1, data, data_size);
		fz_drop_buffer(outlinebuf);
		fz_free_outline(ctx, outline);
	}
}

int main(int argc, char **argv)
{
	char *password = "";
	pdf_document *doc = NULL;
	int c;
	fz_context *ctx;
    fz_stext_sheet *sheet = NULL;
    int showtext = 1, showoutline = 0, showxml = 0, do_analyze = 0;
    fz_warn_context wctx;


	while ((c = fz_getopt(argc, argv, "Alo:p:r:R:b:c:dgmtx5G:Iw:h:fi")) != -1)
	{
		switch (c)
		{
		case 'A': do_analyze++; break;
		case 'p': password = fz_optarg; break;
		case 'l': showoutline++; break;
		case 't': showtext++; break;
		case 'x': showxml++; break;
		default: usage(); exit(1); break;
		}
	}

	if (fz_optind == argc) {
		usage();
        exit(1);
    }

	ctx = fz_new_context(NULL, &wctx, FZ_STORE_DEFAULT);
	if (!ctx)
	{
		write(2, "Cannot initialise context\n", 27);
		exit(1);
	}

	outbuf = fz_new_buffer(1024);
	if (showxml || showtext == TEXT_XML)
		fz_buffer_printf(outbuf, "<?xml version=\"1.0\"?>\n");

	sheet = fz_new_stext_sheet(ctx);
	fz_try(ctx)
	{
		while (fz_optind < argc)
		{
			fz_try(ctx)
			{
                unsigned char *inbuf;
                size_t insz;
				filename = argv[fz_optind++];
				files++;

				fz_try(ctx)
				{
                    inbuf = slurp_file(filename, &insz);
			        doc = pdf_open_document_no_run(ctx, inbuf, insz);
				}
				fz_catch(ctx)
				{
					fz_rethrow_message(ctx, "cannot open document: %s", filename);
				}

				if (pdf_needs_password(doc))
				{
					if (!pdf_authenticate_password(doc, password))
						fz_throw(ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", filename);
				}

				if (showxml || showtext == TEXT_XML)
					fz_buffer_printf(outbuf, "<document name=\"%s\">\n", filename);

				if (showoutline)
					drawoutline(ctx, doc, showoutline);

				if (fz_optind == argc || !isrange(argv[fz_optind]))
					drawrange(ctx, doc, "1-", sheet, do_analyze);
				if (fz_optind < argc && isrange(argv[fz_optind]))
					drawrange(ctx, doc, argv[fz_optind++], sheet, do_analyze);

				if (showxml || showtext == TEXT_XML)
					fz_buffer_printf(outbuf, "</document>\n");

				pdf_close_document(doc);
				doc = NULL;
			}
			fz_catch(ctx)
			{
				pdf_close_document(doc);
				doc = NULL;
				fz_warn(ctx->warn, "ignoring error in '%s'", filename);
			}
		}
	}
	fz_catch(ctx)
	{
		pdf_close_document(doc);
		write(2, "Error: cannot draw \"", 19);
		write(2, filename, strlen(filename));
		write(2, "\"\n", 2);
		errored = 1;
	}

	fz_free_stext_sheet(ctx, sheet);
    dump_buffer(ctx, outbuf);
	fz_drop_buffer(outbuf);

	fz_free_context(ctx);
	return (errored != 0);
}

#ifdef _MSC_VER
int wmain(int argc, wchar_t *wargv[])
{
	char **argv = fz_argv_from_wargv(argc, wargv);
	int ret = main(argc, argv);
	fz_free_argv(argc, argv);
	return ret;
}
#endif
