#include <stdint.h>
#include <math.h>

typedef union _f32 {
    float f;
    uint32_t i;
} _f32;

static const float
C1  = -5.0000000000e-01,
C2  =  4.1666667908e-02, /* 0x3d2aaaab */
C3  = -1.3888889225e-03, /* 0xbab60b61 */
C4  =  2.4801587642e-05, /* 0x37d00d01 */
C5  = -2.7557314297e-07; /* 0xb493f27c */

// Differs from libc cosf on [0, pi/2] at only the point 0x3fc460f8 (~1.5342093) by 0.0000001005828f
static inline float k_cosf(float x)
{
    float z,r;
    z  = x*x;
    r  = z*(C2+z*(C3+z*(C4+z*C5)));
    return 1.0f + (z*r + C1*z);
}

static const float
S1 =  0x1.7d3bbcp-19f,
S2 = -0x1.a06bbap-13f,
S3 =  0x1.11119ap-07f,
S4 = -0x1.555556p-03f;

/* Approximate sine on [-PI/4,+PI/4]. Maximum ulp error = 0.64721
   Returns -0.0f for an argument of -0.0f
   Polynomial approximation based on unpublished work by T. Myklebust
*/
static inline float k_sinf(float x)
{
    float z,r;
    z = x*x;
    r = z*(S4+z*(S3+z*(S2+z*S1)));
    return x+x*r;
}

static const float inv_pi  =  0.3183098733;  /* 0x3ea2f984 */
static const float invpio2 =  0.63661980629; /* 0x3f22f984 */

/*
 * pio2_1:   first  17 bit of pi/2
 * pio2_1t:  pi/2 - pio2_1
 */

static const float
pio2_1  =  1.5707855225e+00, /* 0x3fc90f80 */
pio2_1t =  1.0804334124e-05; /* 0x37354443 */

uint32_t
__ieee754_rem_pio2f(float x, float *y)
{
    float fn, t = fabsf(x);
    uint32_t n  = (uint32_t)(t*invpio2);
    fn = (float)n;
    *y = t-fn*pio2_1-fn*pio2_1t; /* 1st round good to 40 bit */
    return n;
}

float cosf(float x) {
    float y, z;
    unsigned int n = __ieee754_rem_pio2f(x, &y);
    z    = ((n&1) ? -k_cosf((float)M_PI_2 - y) : k_cosf(y));
    return ((n&2) ? -z : z);
}

/* 
 * uint32_t n = (uint32_t)(y*inv_pi);
 * z = k_sinf(y - ((float)M_PI * (float)n));
 */
float sinf(float x) {
    float y, z;
    unsigned int n = __ieee754_rem_pio2f(x, &y);
    z = ((n&1) ? k_cosf(y) : k_sinf(y));
    if (x < 0.0f) x = -x;
    return ((n&2) ? -z : z);
}

void sincosf(float x, float *sinx, float *cosx)
{
    float y;
    unsigned int n = __ieee754_rem_pio2f(x, &y);
    float s = k_sinf(y);
    float c = k_cosf(y);
    if (n & 2) {
        c = -c;
        s = -s;
    }
    *sinx = ((n & 1) ? c : s);
    *cosx = ((n & 1) ? -s : c);
}

