#include "mupdf/pdf.h"
#include <zlib.h>
void
pdf_compute_hash_r5(fz_context *ctx, unsigned char *password, int pwlen, unsigned char salt[16], unsigned char *ownerkey,
                    unsigned char *validationkey, unsigned char *hash);
void
pdf_compute_hardened_hash_r6(fz_context *ctx, unsigned char *password, int pwlen, unsigned char salt[16], unsigned char *ownerkey, unsigned char hash[32]);

enum
{
	PDF_FLAGS_MARKED = 1,
	PDF_FLAGS_SORTED = 2,
	PDF_FLAGS_MEMO = 4,
	PDF_FLAGS_MEMO_BOOL = 8
};

static inline int iswhite(int ch)
{
	return ch == '\000' || ch == '\011' || ch == '\012' || ch == '\014' || ch == '\015' || ch == '\040';
}

pdf_obj *
pdf_new_obj(pdf_document *doc)
{
	pdf_obj *obj = fz_malloc(doc->ctx, sizeof(pdf_obj));
	obj->doc = doc;
	obj->refs = 1;
	obj->flags = 0;
	obj->parent_num = 0;
    return obj;
}

pdf_obj *
pdf_new_string(pdf_document *doc, const char *str, unsigned int len)
{
	pdf_obj *obj = fz_malloc(doc->ctx, sizeof(struct pdf_obj_s) + len + 1);
	obj->doc = doc;
	obj->refs = 1;
	obj->kind = PDF_STRING;
	obj->flags = 0;
	obj->parent_num = 0;
	obj->u.s.len = len;
	memcpy(obj->u.s.buf, str, len);
	obj->u.s.buf[len] = '\0';
	return obj;
}

pdf_obj *
pdf_new_name(pdf_document *doc, const char *str)
{
    size_t len = strlen(str);
	pdf_obj *obj = fz_malloc(doc->ctx, sizeof(pdf_obj) + len + 1);
	obj->doc = doc;
	obj->refs = 1;
	obj->kind = PDF_NAME;
	obj->flags = 0;
	obj->parent_num = 0;
	memcpy(obj->u.n, str, len+1);
	return obj;
}

pdf_obj *
pdf_keep_obj(pdf_obj *obj)
{
	if (obj)
		obj->refs ++;
	return obj;
}

#define RESOLVE(obj) \
	do { \
		if (obj && obj->kind == PDF_INDIRECT) { \
			obj = pdf_resolve_indirect(obj); \
		} \
	} while (0)

/* for use by pdf_crypt_obj_imp to decrypt AES string in place */
void pdf_set_str_len(pdf_obj *obj, int newlen)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_STRING)
		return; /* This should never happen */
	if (newlen > obj->u.s.len)
		return; /* This should never happen */
	obj->u.s.len = newlen;
}

pdf_document *pdf_get_indirect_document(pdf_obj *obj)
{
	if (!obj || obj->kind != PDF_INDIRECT)
		return NULL;
	return obj->doc;
}

int
pdf_objcmp(pdf_obj *a, pdf_obj *b)
{
	int i;

	if (a == b)
		return 0;

	if (!a || !b)
		return 1;

	if (a->kind != b->kind)
		return 1;

	switch (a->kind)
	{
	case PDF_NULL:
		return 0;

	case PDF_BOOL:
		return a->u.b - b->u.b;

	case PDF_INT:
		return a->u.i - b->u.i;

	case PDF_REAL:
		if (a->u.f < b->u.f)
			return -1;
		if (a->u.f > b->u.f)
			return 1;
		return 0;

	case PDF_STRING:
		if (a->u.s.len < b->u.s.len)
		{
			if (memcmp(a->u.s.buf, b->u.s.buf, a->u.s.len) <= 0)
				return -1;
			return 1;
		}
		if (a->u.s.len > b->u.s.len)
		{
			if (memcmp(a->u.s.buf, b->u.s.buf, b->u.s.len) >= 0)
				return 1;
			return -1;
		}
		return memcmp(a->u.s.buf, b->u.s.buf, a->u.s.len);

	case PDF_NAME:
		return strcmp(a->u.n, b->u.n);

	case PDF_INDIRECT:
		if (a->u.r.num == b->u.r.num)
			return a->u.r.gen - b->u.r.gen;
		return a->u.r.num - b->u.r.num;

	case PDF_ARRAY:
		if (a->u.a.len != b->u.a.len)
			return a->u.a.len - b->u.a.len;
		for (i = 0; i < a->u.a.len; i++)
			if (pdf_objcmp(a->u.a.items[i], b->u.a.items[i]))
				return 1;
		return 0;

	case PDF_DICT:
		if (a->u.d.len != b->u.d.len)
			return a->u.d.len - b->u.d.len;
		for (i = 0; i < a->u.d.len; i++)
		{
			if (pdf_objcmp(a->u.d.items[i].k, b->u.d.items[i].k))
				return 1;
			if (pdf_objcmp(a->u.d.items[i].v, b->u.d.items[i].v))
				return 1;
		}
		return 0;

	}
	return 1;
}

char *
pdf_objkindstr(pdf_obj *obj)
{
	if (!obj)
		return "<NULL>";
	switch (obj->kind)
	{
	case PDF_NULL: return "null";
	case PDF_BOOL: return "boolean";
	case PDF_INT: return "integer";
	case PDF_REAL: return "real";
	case PDF_STRING: return "string";
	case PDF_NAME: return "name";
	case PDF_ARRAY: return "array";
	case PDF_DICT: return "dictionary";
	case PDF_INDIRECT: return "reference";
	}
	return "<unknown>";
}

pdf_obj *
pdf_new_array(pdf_document *doc, int initialcap)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *obj = pdf_new_obj(doc);
	int i;

	obj->kind = PDF_ARRAY;
	obj->u.a.len = 0;
	obj->u.a.cap = initialcap > 1 ? initialcap : 6;
	obj->u.a.items = fz_malloc_array(ctx, obj->u.a.cap, sizeof(pdf_obj*));
	for (i = 0; i < obj->u.a.cap; i++)
		obj->u.a.items[i] = NULL;

	return obj;
}

static void
pdf_array_grow(pdf_obj *obj)
{
	int i;
	int new_cap = (obj->u.a.cap * 3) / 2;

	obj->u.a.items = fz_resize_array(obj->doc->ctx, obj->u.a.items, new_cap, sizeof(pdf_obj*));
	obj->u.a.cap = new_cap;

	for (i = obj->u.a.len ; i < obj->u.a.cap; i++)
		obj->u.a.items[i] = NULL;
}

pdf_obj *
pdf_copy_array(pdf_obj *obj)
{
	pdf_obj *arr;
	int i;
	int n;
	fz_context *ctx = obj->doc->ctx;

	RESOLVE(obj);
	if (!obj)
		return NULL; /* Can't warn :( */
	if (obj->kind != PDF_ARRAY)
		fz_warn(ctx->warn, "assert: not an array (%s)", pdf_objkindstr(obj));

	arr = pdf_new_array(obj->doc, pdf_array_len(obj));
	n = pdf_array_len(obj);
	for (i = 0; i < n; i++)
		pdf_array_push(arr, pdf_array_get(obj, i));

	return arr;
}

static void object_altered(pdf_obj *obj, pdf_obj *val)
{
	/*
	parent_num = 0 while an object is being parsed from the file.
	No further action is necessary.
	*/
	if (obj->parent_num == 0 || obj->doc->freeze_updates)
		return;

	/*
	Otherwise we need to ensure that the containing hierarchy of objects
	has been moved to the incremental xref section and the newly linked
	object needs to record the parent_num
	*/
	pdf_xref_ensure_incremental_object(obj->doc, obj->parent_num);
	pdf_set_obj_parent(val, obj->parent_num);
}

void
pdf_array_put(pdf_obj *obj, unsigned int i, pdf_obj *item)
{
	RESOLVE(obj);

	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_ARRAY)
		fz_warn(obj->doc->ctx->warn, "assert: not an array (%s)", pdf_objkindstr(obj));
	else if (i >= obj->u.a.len)
		fz_warn(obj->doc->ctx->warn, "assert: index %d > length %d", i, obj->u.a.len);
	else
	{
		pdf_drop_obj(obj->u.a.items[i]);
		obj->u.a.items[i] = pdf_keep_obj(item);
	}

	object_altered(obj, item);
}

void
pdf_array_push(pdf_obj *obj, pdf_obj *item)
{
	RESOLVE(obj);

	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_ARRAY)
		fz_warn(obj->doc->ctx->warn, "assert: not an array (%s)", pdf_objkindstr(obj));
	else
	{
		if (obj->u.a.len + 1 > obj->u.a.cap)
			pdf_array_grow(obj);
		obj->u.a.items[obj->u.a.len] = pdf_keep_obj(item);
		obj->u.a.len++;
	}

	object_altered(obj, item);
}

void
pdf_array_insert(pdf_obj *obj, pdf_obj *item, unsigned int i)
{
	RESOLVE(obj);

	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_ARRAY)
		fz_warn(obj->doc->ctx->warn, "assert: not an array (%s)", pdf_objkindstr(obj));
	else
	{
		if (i > obj->u.a.len)
			fz_throw(obj->doc->ctx, FZ_ERROR_GENERIC, "attempt to insert object %d in array of length %d", i, obj->u.a.len);
		if (obj->u.a.len + 1 > obj->u.a.cap)
			pdf_array_grow(obj);
		memmove(obj->u.a.items + i + 1, obj->u.a.items + i, (obj->u.a.len - i) * sizeof(pdf_obj*));
		obj->u.a.items[i] = pdf_keep_obj(item);
		obj->u.a.len++;
	}

	object_altered(obj, item);
}

void
pdf_array_delete(pdf_obj *obj, unsigned int i)
{
	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_ARRAY)
		fz_warn(obj->doc->ctx->warn, "assert: not an array (%s)", pdf_objkindstr(obj));
	else
	{
		pdf_drop_obj(obj->u.a.items[i]);
		obj->u.a.items[i] = 0;
		obj->u.a.len--;
		memmove(obj->u.a.items + i, obj->u.a.items + i + 1, (obj->u.a.len - i) * sizeof(pdf_obj*));
	}
}

int
pdf_array_contains(pdf_obj *arr, pdf_obj *obj)
{
	int i, len;

	len = pdf_array_len(arr);
	for (i = 0; i < len; i++)
		if (!pdf_objcmp(pdf_array_get(arr, i), obj))
			return 1;

	return 0;
}

static inline void
pdf_array_push_internal(pdf_obj *obj, pdf_obj *item)
{
	//RESOLVE(obj);
	obj->u.a.items[obj->u.a.len] = pdf_keep_obj(item);
	obj->u.a.len++;
	object_altered(obj, item);
}

fz_rect *
pdf_bound_page(pdf_document *doc, pdf_page *page, fz_rect *bounds)
{
	fz_matrix mtx;
	fz_rect mediabox = page->mediabox;
	fz_transform_rect(&mediabox, fz_rotate(&mtx, page->rotate));
	bounds->x0 = bounds->y0 = 0;
	bounds->x1 = mediabox.x1 - mediabox.x0;
	bounds->y1 = mediabox.y1 - mediabox.y0;
	return bounds;
}

pdf_obj *pdf_new_rect(pdf_document *doc, const fz_rect *rect)
{
	pdf_obj *arr = NULL;
	pdf_obj *item = NULL;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		arr = pdf_new_array(doc, 4);

		item = pdf_new_real(doc, rect->x0);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;

		item = pdf_new_real(doc, rect->y0);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;

		item = pdf_new_real(doc, rect->x1);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;

		item = pdf_new_real(doc, rect->y1);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(item);
		pdf_drop_obj(arr);
		fz_rethrow(ctx);
	}

	return arr;
}

pdf_obj *pdf_new_matrix(pdf_document *doc, const fz_matrix *mtx)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *arr = NULL;
	pdf_obj *item = NULL;
    unsigned int i;

	fz_try(ctx)
	{
		arr = pdf_new_array(doc, 6);

        for (i = 0; i < 4; i++) {
		    item = pdf_new_real(doc, mtx->x[i]);
		    pdf_array_push_internal(arr, item);
		    pdf_drop_obj(item);
		    item = NULL;
        }

		item = pdf_new_real(doc, mtx->e);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;

		item = pdf_new_real(doc, mtx->f);
		pdf_array_push_internal(arr, item);
		pdf_drop_obj(item);
		item = NULL;
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(item);
		pdf_drop_obj(arr);
		fz_rethrow(ctx);
	}

	return arr;
}

/* dicts may only have names as keys! */

static int keyvalcmp(const void *ap, const void *bp)
{
	const struct keyval *a = ap;
	const struct keyval *b = bp;
	return strcmp(pdf_to_name(a->k), pdf_to_name(b->k));
}

pdf_obj *
pdf_new_dict(pdf_document *doc, int initialcap)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *obj = pdf_new_obj(doc);
	int i;

	obj->kind = PDF_DICT;
	obj->u.d.len = 0;
	obj->u.d.cap = initialcap > 1 ? initialcap : 10;
	obj->u.d.items = fz_malloc_array(ctx, obj->u.d.cap, sizeof(struct keyval));
	for (i = 0; i < obj->u.d.cap; i++) {
		obj->u.d.items[i].k = NULL;
		obj->u.d.items[i].v = NULL;
	}

	return obj;
}

static void
pdf_dict_grow(pdf_obj *obj)
{
	int i;
	int new_cap = (obj->u.d.cap * 3) / 2;

	obj->u.d.items = fz_resize_array(obj->doc->ctx, obj->u.d.items, new_cap, sizeof(struct keyval));
	obj->u.d.cap = new_cap;
	for (i = obj->u.d.len; i < obj->u.d.cap; i++) {
		obj->u.d.items[i].k = NULL;
		obj->u.d.items[i].v = NULL;
	}
}

pdf_obj *
pdf_copy_dict(pdf_obj *obj)
{
	pdf_obj *dict;
	int i, n;
	pdf_document *doc;

	if (!obj)
		return NULL; /* Can't warn :( */
	RESOLVE(obj);
	if (!obj)
		return NULL; /* Can't warn :( */
	doc = obj->doc;
	if (obj->kind != PDF_DICT)
		fz_warn(doc->ctx->warn, "assert: not a dict (%s)", pdf_objkindstr(obj));

	n = pdf_dict_len(obj);
	dict = pdf_new_dict(doc, n);
	for (i = 0; i < n; i++)
		pdf_dict_put(dict, pdf_dict_get_key(obj, i), pdf_dict_get_val(obj, i));

	return dict;
}

int
pdf_dict_len(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_DICT)
		return 0;
	return obj->u.d.len;
}

pdf_obj *
pdf_dict_get_key(pdf_obj *obj, int i)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_DICT)
		return NULL;

	if (i < 0 || i >= obj->u.d.len)
		return NULL;

	return obj->u.d.items[i].k;
}

pdf_obj *
pdf_dict_get_val(pdf_obj *obj, int i)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_DICT)
		return NULL;

	if (i < 0 || i >= obj->u.d.len)
		return NULL;

	return obj->u.d.items[i].v;
}

static int
pdf_dict_finds(pdf_obj *obj, const char *key, int *location)
{
	if ((obj->flags & PDF_FLAGS_SORTED) && obj->u.d.len > 0)
	{
		int l = 0;
		int r = obj->u.d.len - 1;

		if (strcmp(pdf_to_name(obj->u.d.items[r].k), key) < 0)
		{
			if (location)
				*location = r + 1;
			return -1;
		}

		while (l <= r)
		{
			int m = (l + r) >> 1;
			int c = -strcmp(pdf_to_name(obj->u.d.items[m].k), key);
			if (c < 0)
				r = m - 1;
			else if (c > 0)
				l = m + 1;
			else
				return m;

			if (location)
				*location = l;
		}
	}

	else
	{
		int i;
		for (i = 0; i < obj->u.d.len; i++)
			if (strcmp(pdf_to_name(obj->u.d.items[i].k), key) == 0)
				return i;

		if (location)
			*location = obj->u.d.len;
	}

	return -1;
}

pdf_obj *
pdf_dict_gets(pdf_obj *obj, const char *key)
{
	int i;

	RESOLVE(obj);
	if (!obj || obj->kind != PDF_DICT)
		return NULL;

	i = pdf_dict_finds(obj, key, NULL);
	if (i >= 0)
		return obj->u.d.items[i].v;

	return NULL;
}

pdf_obj *
pdf_dict_getp(pdf_obj *obj, const char *keys)
{
	char buf[256];
	char *k, *e;
    unsigned int len = strlen(keys);

	if (len > 255) {
		fz_warn(obj->doc->ctx->warn, "buffer overflow in pdf_dict_getp");
        len = 255;
    }

	memcpy(buf, keys, len);
    buf[len] = 0;

	e = buf;
	while (*e && obj)
	{
		k = e;
		while (*e != '/' && *e != '\0')
			e++;

		if (*e == '/')
		{
			*e = '\0';
			e++;
		}

		obj = pdf_dict_gets(obj, k);
	}

	return obj;
}

void
pdf_dict_put(pdf_obj *obj, pdf_obj *key, pdf_obj *val)
{
	int location;
	char *s;
	int i;

	RESOLVE(obj);
	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_DICT)
	{
		fz_warn(obj->doc->ctx->warn, "assert: not a dict (%s)", pdf_objkindstr(obj));
		return;
	}

	RESOLVE(key);
	if (!key || key->kind != PDF_NAME)
	{
		fz_warn(obj->doc->ctx->warn, "assert: key is not a name (%s)", pdf_objkindstr(obj));
		return;
	}
	else
		s = pdf_to_name(key);

	if (!val)
	{
		fz_warn(obj->doc->ctx->warn, "assert: val does not exist for key (%s)", s);
		return;
	}

	if (obj->u.d.len > 100 && !(obj->flags & PDF_FLAGS_SORTED))
		pdf_sort_dict(obj);

	i = pdf_dict_finds(obj, s, &location);
	if (i >= 0 && i < obj->u.d.len)
	{
		if (obj->u.d.items[i].v != val)
		{
			pdf_drop_obj(obj->u.d.items[i].v);
			obj->u.d.items[i].v = pdf_keep_obj(val);
		}
	}
	else
	{
		if (obj->u.d.len + 1 > obj->u.d.cap)
			pdf_dict_grow(obj);

		i = location;
		if ((obj->flags & PDF_FLAGS_SORTED) && obj->u.d.len > 0)
			memmove(&obj->u.d.items[i + 1],
				&obj->u.d.items[i],
				(obj->u.d.len - i) * sizeof(struct keyval));

		obj->u.d.items[i].k = pdf_keep_obj(key);
		obj->u.d.items[i].v = pdf_keep_obj(val);
		obj->u.d.len ++;
	}

	object_altered(obj, val);
}

void
pdf_dict_putp(pdf_obj *obj, const char *keys, pdf_obj *val)
{
	fz_context *ctx = obj->doc->ctx;
	char buf[256];
	char *k, *e;
	pdf_obj *cobj = NULL;
    unsigned int len = strlen(keys);

	if (len > 255) {
		fz_warn(ctx->warn, "buffer overflow in pdf_dict_getp");
        len = 255;
    }

	memcpy(buf, keys, len);
    buf[len] = 0;

	e = buf;
	while (*e)
	{
		k = e;
		while (*e != '/' && *e != '\0')
			e++;

		if (*e == '/')
		{
			*e = '\0';
			e++;
		}

		if (*e)
		{
			/* Not the last key in the key path. Create subdict if not already there. */
			cobj = pdf_dict_gets(obj, k);
			if (cobj == NULL)
			{
				cobj = pdf_new_dict(obj->doc, 1);
				pdf_dict_puts(obj, k, cobj);
				pdf_drop_obj(cobj);
			}
			/* Move to subdict */
			obj = cobj;
		}
		else
		{
			/* Last key. Use it to store the value */
			/* Use val = NULL to request delete */
			if (val)
				pdf_dict_puts(obj, k, val);
			else
				pdf_dict_dels(obj, k);
		}
	}
}

void
pdf_dict_dels(pdf_obj *obj, const char *key)
{
	RESOLVE(obj);

	if (!obj)
		return; /* Can't warn :( */
	if (obj->kind != PDF_DICT)
		fz_warn(obj->doc->ctx->warn, "assert: not a dict (%s)", pdf_objkindstr(obj));
	else
	{
		int i = pdf_dict_finds(obj, key, NULL);
		if (i >= 0)
		{
			pdf_drop_obj(obj->u.d.items[i].k);
			pdf_drop_obj(obj->u.d.items[i].v);
			obj->flags &= ~PDF_FLAGS_SORTED;
			obj->u.d.items[i] = obj->u.d.items[obj->u.d.len-1];
			obj->u.d.len --;
		}
	}

	object_altered(obj, NULL);
}

void
pdf_dict_del(pdf_obj *obj, pdf_obj *key)
{
	RESOLVE(key);
	if (!key || key->kind != PDF_NAME)
		fz_warn(obj->doc->ctx->warn, "assert: key is not a name (%s)", pdf_objkindstr(obj));
	else
		pdf_dict_dels(obj, key->u.n);
}

void
pdf_sort_dict(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj || obj->kind != PDF_DICT)
		return;
	if (!(obj->flags & PDF_FLAGS_SORTED))
	{
		qsort(obj->u.d.items, obj->u.d.len, sizeof(struct keyval), keyvalcmp);
		obj->flags |= PDF_FLAGS_SORTED;
	}
}

const unsigned short pdf_doc_encoding[256] =
{
	/* 0x0 to 0x17 except \t, \n and \r are really undefined */
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x02d8, 0x02c7, 0x02c6, 0x02d9, 0x02dd, 0x02db, 0x02da, 0x02dc,
	0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
	0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
	0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
	0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
	0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
	0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
	0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
	0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
	0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
	0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
	0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
	0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x0000,
	0x2022, 0x2020, 0x2021, 0x2026, 0x2014, 0x2013, 0x0192, 0x2044,
	0x2039, 0x203a, 0x2212, 0x2030, 0x201e, 0x201c, 0x201d, 0x0022,
	0x0022, 0x201a, 0x2122, 0xfb01, 0xfb02, 0x0141, 0x0152, 0x0160,
	0x0178, 0x017d, 0x0131, 0x0142, 0x0153, 0x0161, 0x017e, 0x0000,
	0x20ac, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
	0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x0000, 0x00ae, 0x00af,
	0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
	0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
	0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
	0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
	0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
	0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
	0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
	0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
	0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
	0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
};

/* Convert Unicode/PdfDocEncoding string into utf-8 */
char *
pdf_docenc_to_utf8(fz_context *ctx, unsigned char *srcptr, unsigned int srclen)
{
	char *dstptr, *dst;
	unsigned int i, ucs, dstlen = 0;

	if (srclen >= 2 && srcptr[0] == 254 && srcptr[1] == 255) {
        i = 2;
		while (i + 1 < srclen) {
			i += fz_rune_from_utf16be(&ucs, srcptr + i, srcptr + srclen);
			dstlen += fz_runelen(ucs);
		}

		dstptr = dst = fz_malloc(ctx, dstlen + 1);

        i = 2;
		while (i + 1 < srclen) {
			i += fz_rune_from_utf16be(&ucs, srcptr + i, srcptr + srclen);
			dstptr += fz_runetochar(dstptr, ucs);
		}
	} else if (srclen >= 2 && srcptr[0] == 255 && srcptr[1] == 254) {
        i = 2;
		while (i + 1 < srclen) {
			i += fz_rune_from_utf16be(&ucs, srcptr + i, srcptr + srclen);
			dstlen += fz_runelen(ucs);
		}

		dstptr = dst = fz_malloc(ctx, dstlen + 1);

        i = 2;
		while (i + 1 < srclen) {
			i += fz_rune_from_utf16be(&ucs, srcptr + i, srcptr + srclen);
			dstptr += fz_runetochar(dstptr, ucs);
		}
	} else { 
		for (i = 0; i < srclen; i++)
			dstlen += fz_runelen(pdf_doc_encoding[srcptr[i]]);

		dstptr = dst = fz_malloc(ctx, dstlen + 1);

		for (i = 0; i < srclen; i++) { 
			ucs = pdf_doc_encoding[srcptr[i]];
			dstptr += fz_runetochar(dstptr, ucs);
		}
	}

	*dstptr = '\0';
	return dst;
}

/* Convert Unicode/PdfDocEncoding string into ucs-2 */
unsigned short *
pdf_to_ucs2(pdf_document *doc, pdf_obj *src)
{
	fz_context *ctx = doc->ctx;
	unsigned char *srcptr = (unsigned char *) pdf_to_str_buf(src);
	unsigned short *dstptr, *dst;
	int srclen = pdf_to_str_len(src);
	int i;

	if (srclen >= 2 && srcptr[0] == 254 && srcptr[1] == 255)
	{
		dstptr = dst = fz_malloc_array(ctx, (srclen - 2) / 2 + 1, sizeof(short));
		for (i = 2; i + 1 < srclen; i += 2)
			*dstptr++ = srcptr[i] << 8 | srcptr[i+1];
	}
	else if (srclen >= 2 && srcptr[0] == 255 && srcptr[1] == 254)
	{
		dstptr = dst = fz_malloc_array(ctx, (srclen - 2) / 2 + 1, sizeof(short));
		for (i = 2; i + 1 < srclen; i += 2)
			*dstptr++ = srcptr[i] | srcptr[i+1] << 8;
	}
	else
	{
		dstptr = dst = fz_malloc_array(ctx, srclen + 1, sizeof(short));
		for (i = 0; i < srclen; i++)
			*dstptr++ = pdf_doc_encoding[srcptr[i]];
	}

	*dstptr = '\0';
	return dst;
}

/* allow to convert to UCS-2 without the need for an fz_context */
/* (buffer must be at least (fz_to_str_len(src) + 1) * 2 bytes in size) */
void
pdf_to_ucs2_buf(unsigned short *buffer, pdf_obj *src)
{
	unsigned char *srcptr = (unsigned char *) pdf_to_str_buf(src);
	unsigned short *dstptr = buffer;
	int srclen = pdf_to_str_len(src);
	int i;

	if (srclen >= 2 && srcptr[0] == 254 && srcptr[1] == 255)
	{
		for (i = 2; i + 1 < srclen; i += 2)
			*dstptr++ = srcptr[i] << 8 | srcptr[i+1];
	}
	else if (srclen >= 2 && srcptr[0] == 255 && srcptr[1] == 254)
	{
		for (i = 2; i + 1 < srclen; i += 2)
			*dstptr++ = srcptr[i] | srcptr[i+1] << 8;
	}
	else
	{
		for (i = 0; i < srclen; i++)
			*dstptr++ = pdf_doc_encoding[srcptr[i]];
	}

	*dstptr = '\0';
}

fz_link *
pdf_new_link(fz_context *ctx, const fz_rect *bbox, fz_link_dest dest)
{
	fz_link *link;
	link = fz_malloc_struct(ctx, fz_link);
	link->refs = 1;
	link->dest = dest;
	link->rect = *bbox;
	link->next = NULL;
	return link;
}

fz_link *
pdf_keep_link(fz_context *ctx, fz_link *link)
{
	if (link)
		link->refs++;
	return link;
}

void
pdf_drop_link(fz_context *ctx, fz_link *link)
{
	while (link && --link->refs == 0)
	{
		fz_link *next = link->next;
		fz_free_link_dest(ctx, &link->dest);
		fz_free(ctx, link);
		link = next;
	}
}

fz_link *
pdf_load_links(pdf_document *doc, pdf_page *page)
{
	return pdf_keep_link(doc->ctx, page->links);
}

static pdf_obj *
resolve_dest_rec(pdf_document *doc, pdf_obj *dest, int depth)
{
	if (depth > 10) /* Arbitrary to avoid infinite recursion */
		return NULL;

	if (pdf_is_name(dest) || pdf_is_string(dest))
	{
		dest = pdf_lookup_dest(doc, dest);
		return resolve_dest_rec(doc, dest, depth+1);
	}

	else if (pdf_is_array(dest))
	{
		return dest;
	}

	else if (pdf_is_dict(dest))
	{
		dest = pdf_dict_gets(dest, "D");
		return resolve_dest_rec(doc, dest, depth+1);
	}

	else if (pdf_is_indirect(dest))
		return dest;

	return NULL;
}

static pdf_obj *
resolve_dest(pdf_document *doc, pdf_obj *dest)
{
	return resolve_dest_rec(doc, dest, 0);
}

fz_link_dest
pdf_parse_link_dest(pdf_document *doc, pdf_obj *dest)
{
	fz_link_dest ld;
	pdf_obj *obj;

	int l_from_2 = 0;
	int b_from_3 = 0;
	int r_from_4 = 0;
	int t_from_5 = 0;
	int t_from_3 = 0;
	int t_from_2 = 0;
	int z_from_4 = 0;

	dest = resolve_dest(doc, dest);
	if (dest == NULL || !pdf_is_array(dest))
	{
		ld.kind = FZ_LINK_NONE;
		return ld;
	}

	obj = pdf_array_get(dest, 0);
	if (pdf_is_int(obj))
		ld.ld.gotor.page = pdf_to_int(obj);
	else
	{
		fz_try(doc->ctx)
		{
			ld.ld.gotor.page = pdf_lookup_page_number(doc, obj);
		}
		fz_catch(doc->ctx)
		{
			ld.kind = FZ_LINK_NONE;
			return ld;
		}
	}

	ld.kind = FZ_LINK_GOTO;
	ld.ld.gotor.flags = 0;
	ld.ld.gotor.lt.x = 0;
	ld.ld.gotor.lt.y = 0;
	ld.ld.gotor.rb.x = 0;
	ld.ld.gotor.rb.y = 0;
	ld.ld.gotor.file_spec = NULL;
	ld.ld.gotor.new_window = 0;

	obj = pdf_array_get(dest, 1);
	if (!pdf_is_name(obj))
		return ld;

	if (!strcmp("XYZ", pdf_to_name(obj)))
	{
		l_from_2 = t_from_3 = z_from_4 = 1;
		ld.ld.gotor.flags |= fz_link_flag_r_is_zoom;
	}
	else if ((!strcmp("Fit", pdf_to_name(obj))) || (!strcmp("FitB", pdf_to_name(obj))))
	{
		ld.ld.gotor.flags |= fz_link_flag_fit_h;
		ld.ld.gotor.flags |= fz_link_flag_fit_v;
	}
	else if ((!strcmp("FitH", pdf_to_name(obj))) || (!strcmp("FitBH", pdf_to_name(obj))))
	{
		t_from_2 = 1;
		ld.ld.gotor.flags |= fz_link_flag_fit_h;
	}
	else if ((!strcmp("FitV", pdf_to_name(obj))) || (!strcmp("FitBV", pdf_to_name(obj))))
	{
		l_from_2 = 1;
		ld.ld.gotor.flags |= fz_link_flag_fit_v;
	}
	else if (!strcmp("FitR", pdf_to_name(obj)))
	{
		l_from_2 = b_from_3 = r_from_4 = t_from_5 = 1;
		ld.ld.gotor.flags |= fz_link_flag_fit_h;
		ld.ld.gotor.flags |= fz_link_flag_fit_v;
	}

	if (l_from_2)
	{
		obj = pdf_array_get(dest, 2);
		if (pdf_is_int(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_l_valid;
			ld.ld.gotor.lt.x = pdf_to_int(obj);
		}
		else if (pdf_is_real(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_l_valid;
			ld.ld.gotor.lt.x = pdf_to_real(obj);
		}
	}
	if (b_from_3)
	{
		obj = pdf_array_get(dest, 3);
		if (pdf_is_int(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_b_valid;
			ld.ld.gotor.rb.y = pdf_to_int(obj);
		}
		else if (pdf_is_real(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_b_valid;
			ld.ld.gotor.rb.y = pdf_to_real(obj);
		}
	}
	if (r_from_4)
	{
		obj = pdf_array_get(dest, 4);
		if (pdf_is_int(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_r_valid;
			ld.ld.gotor.rb.x = pdf_to_int(obj);
		}
		else if (pdf_is_real(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_r_valid;
			ld.ld.gotor.rb.x = pdf_to_real(obj);
		}
	}
	if (t_from_5 || t_from_3 || t_from_2)
	{
		if (t_from_5)
			obj = pdf_array_get(dest, 5);
		else if (t_from_3)
			obj = pdf_array_get(dest, 3);
		else
			obj = pdf_array_get(dest, 2);
		if (pdf_is_int(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_t_valid;
			ld.ld.gotor.lt.y = pdf_to_int(obj);
		}
		else if (pdf_is_real(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_t_valid;
			ld.ld.gotor.lt.y = pdf_to_real(obj);
		}
	}
	if (z_from_4)
	{
		obj = pdf_array_get(dest, 4);
		if (pdf_is_int(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_r_valid;
			ld.ld.gotor.rb.x = pdf_to_int(obj);
		}
		else if (pdf_is_real(obj))
		{
			ld.ld.gotor.flags |= fz_link_flag_r_valid;
			ld.ld.gotor.rb.x = pdf_to_real(obj);
		}
	}

	/* Duplicate the values out for the sake of stupid clients */
	if ((ld.ld.gotor.flags & (fz_link_flag_l_valid | fz_link_flag_r_valid)) == fz_link_flag_l_valid)
		ld.ld.gotor.rb.x = ld.ld.gotor.lt.x;
	if ((ld.ld.gotor.flags & (fz_link_flag_l_valid | fz_link_flag_r_valid | fz_link_flag_r_is_zoom)) == fz_link_flag_r_valid)
		ld.ld.gotor.lt.x = ld.ld.gotor.rb.x;
	if ((ld.ld.gotor.flags & (fz_link_flag_t_valid | fz_link_flag_b_valid)) == fz_link_flag_t_valid)
		ld.ld.gotor.rb.y = ld.ld.gotor.lt.y;
	if ((ld.ld.gotor.flags & (fz_link_flag_t_valid | fz_link_flag_b_valid)) == fz_link_flag_b_valid)
		ld.ld.gotor.lt.y = ld.ld.gotor.rb.y;

	return ld;
}

static char *
pdf_parse_file_spec(pdf_document *doc, pdf_obj *file_spec)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *filename;

	if (pdf_is_string(file_spec))
		return pdf_to_utf8(doc, file_spec);

	if (pdf_is_dict(file_spec)) {
		filename = pdf_dict_gets(file_spec, "UF");
		if (!filename)
			filename = pdf_dict_gets(file_spec, "F");
		if (!filename)
			filename = pdf_dict_gets(file_spec, "Unix");
		if (!filename)
			filename = pdf_dict_gets(file_spec, "Mac");
		if (!filename)
			filename = pdf_dict_gets(file_spec, "DOS");

		return pdf_to_utf8(doc, filename);
	}

	fz_warn(ctx->warn, "cannot parse file specification");
	return NULL;
}

fz_link_dest
pdf_parse_action(pdf_document *doc, pdf_obj *action)
{
	fz_link_dest ld;
	pdf_obj *obj, *dest;
	fz_context *ctx = doc->ctx;

	UNUSED(ctx);

	ld.kind = FZ_LINK_NONE;

	if (!action)
		return ld;

	obj = pdf_dict_gets(action, "S");
	if (!strcmp(pdf_to_name(obj), "GoTo"))
	{
		dest = pdf_dict_gets(action, "D");
		ld = pdf_parse_link_dest(doc, dest);
	}
	else if (!strcmp(pdf_to_name(obj), "URI"))
	{
		ld.kind = FZ_LINK_URI;
		ld.ld.uri.is_map = pdf_to_bool(pdf_dict_gets(action, "IsMap"));
		ld.ld.uri.uri = pdf_to_utf8(doc, pdf_dict_gets(action, "URI"));
	}
	else if (!strcmp(pdf_to_name(obj), "Launch"))
	{
		ld.kind = FZ_LINK_LAUNCH;
		dest = pdf_dict_gets(action, "F");
		ld.ld.launch.file_spec = pdf_parse_file_spec(doc, dest);
		ld.ld.launch.new_window = pdf_to_int(pdf_dict_gets(action, "NewWindow"));
	}
	else if (!strcmp(pdf_to_name(obj), "Named"))
	{
		ld.kind = FZ_LINK_NAMED;
		ld.ld.named.named = pdf_to_utf8(doc, pdf_dict_gets(action, "N"));
	}
	else if (!strcmp(pdf_to_name(obj), "GoToR"))
	{
		dest = pdf_dict_gets(action, "D");
		ld = pdf_parse_link_dest(doc, dest);
		ld.kind = FZ_LINK_GOTOR;
		dest = pdf_dict_gets(action, "F");
		ld.ld.gotor.file_spec = pdf_parse_file_spec(doc, dest);
		ld.ld.gotor.new_window = pdf_to_int(pdf_dict_gets(action, "NewWindow"));
	}
	return ld;
}

static fz_link *
pdf_load_link(pdf_document *doc, pdf_obj *dict, const fz_matrix *page_ctm)
{
	pdf_obj *dest = NULL;
	pdf_obj *action;
	pdf_obj *obj;
	fz_rect bbox;
	fz_context *ctx = doc->ctx;
	fz_link_dest ld;

	obj = pdf_dict_gets(dict, "Rect");
	if (obj)
		pdf_to_rect(ctx, obj, &bbox);
	else
		bbox = fz_empty_rect;

	fz_transform_rect(&bbox, page_ctm);

	obj = pdf_dict_gets(dict, "Dest");
	if (obj)
	{
		dest = resolve_dest(doc, obj);
		ld = pdf_parse_link_dest(doc, dest);
	}
	else
	{
		action = pdf_dict_gets(dict, "A");
		/* fall back to additional action button's down/up action */
		if (!action)
			action = pdf_dict_getsa(pdf_dict_gets(dict, "AA"), "U", "D");

		ld = pdf_parse_action(doc, action);
	}
	if (ld.kind == FZ_LINK_NONE)
		return NULL;
	return pdf_new_link(ctx, &bbox, ld);
}

fz_link *
pdf_load_link_annots(pdf_document *doc, pdf_obj *annots, const fz_matrix *page_ctm)
{
	fz_link *link, *head, *tail;
	pdf_obj *obj;
	int i, n;

	head = tail = NULL;
	link = NULL;

	n = pdf_array_len(annots);
	fz_try(doc->ctx)
	{
    	for (i = 0; i < n; i++) { 
    		obj = pdf_array_get(annots, i);
    		link = pdf_load_link(doc, obj, page_ctm);
    		if (link) {
    			if (!head) {
    				head = tail = link;
    			} else {
    				tail->next = link;
    				tail = link;
    			}
    		}
    	}
    }
	fz_catch(doc->ctx)
	{
		//fz_rethrow_if(doc->ctx, FZ_ERROR_TRYLATER);
	}

	return head;
}

int
pdf_obj_marked(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj)
		return 0;
	return !!(obj->flags & PDF_FLAGS_MARKED);
}

int
pdf_mark_obj(pdf_obj *obj)
{
	int marked;
	RESOLVE(obj);
	if (!obj)
		return 0;
	marked = !!(obj->flags & PDF_FLAGS_MARKED);
	obj->flags |= PDF_FLAGS_MARKED;
	return marked;
}

void
pdf_unmark_obj(pdf_obj *obj)
{
	RESOLVE(obj);
	if (!obj)
		return;
	obj->flags &= ~PDF_FLAGS_MARKED;
}

void
pdf_set_obj_memo(pdf_obj *obj, int memo)
{
	obj->flags |= PDF_FLAGS_MEMO;
	if (memo)
		obj->flags |= PDF_FLAGS_MEMO_BOOL;
	else
		obj->flags &= ~PDF_FLAGS_MEMO_BOOL;
}

int
pdf_obj_memo(pdf_obj *obj, int *memo)
{
	if (!(obj->flags & PDF_FLAGS_MEMO))
		return 0;
	*memo = !!(obj->flags & PDF_FLAGS_MEMO_BOOL);
	return 1;
}

static void
pdf_free_array(pdf_obj *obj)
{
	fz_context *ctx = obj->doc->ctx;
	int i;

	for (i = 0; i < obj->u.a.len; i++)
		pdf_drop_obj(obj->u.a.items[i]);

	fz_free(ctx, obj->u.a.items);
	fz_free(ctx, obj);
}

static void
pdf_free_dict(pdf_obj *obj)
{
	int i;
	fz_context *ctx = obj->doc->ctx;

	for (i = 0; i < obj->u.d.len; i++) {
		pdf_drop_obj(obj->u.d.items[i].k);
		pdf_drop_obj(obj->u.d.items[i].v);
	}

	fz_free(ctx, obj->u.d.items);
	fz_free(ctx, obj);
}

void
pdf_drop_obj(pdf_obj *obj)
{
	if (!obj)
		return;
	if (--obj->refs)
		return;
	if (obj->kind == PDF_ARRAY)
		pdf_free_array(obj);
	else if (obj->kind == PDF_DICT)
		pdf_free_dict(obj);
	else
		fz_free(obj->doc->ctx, obj);
}

void
pdf_set_obj_parent(pdf_obj *obj, int num)
{
	int n, i;

	if (!obj)
		return;

	obj->parent_num = num;

	switch(obj->kind)
	{
	case PDF_ARRAY:
		n = pdf_array_len(obj);
		for (i = 0; i < n; i++)
			pdf_set_obj_parent(pdf_array_get(obj, i), num);
		break;
	case PDF_DICT:
		n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			pdf_set_obj_parent(pdf_dict_get_val(obj, i), num);
		break;
	}
}

pdf_obj *pdf_new_obj_from_str(pdf_document *doc, const char *src)
{
	pdf_obj *result;
	pdf_lexbuf lexbuf;
	fz_context *ctx = doc->ctx;
	fz_stream *stream = fz_open_memory(ctx, (unsigned char *)src, strlen(src));

	pdf_lexbuf_init(ctx, &lexbuf, PDF_LEXBUF_SMALL);
	fz_try(ctx)
	{
		result = pdf_parse_stm_obj(doc, stream, &lexbuf);
	}
	fz_always(ctx)
	{
		pdf_lexbuf_fin(ctx, &lexbuf);
		fz_close(stream);
	}
	fz_catch(ctx)
	{
		//fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
		return NULL;
	}

	return result;
}

static int
pdf_make_hash_key(fz_store_hash *hash, void *key_)
{
	pdf_obj *key = (pdf_obj *)key_;

	if (!pdf_is_indirect(key))
		return 0;
	hash->u.i.i0 = pdf_to_num(key);
	hash->u.i.i1 = pdf_to_gen(key);
	return 1;
}

static void *
pdf_keep_key(fz_context *ctx, void *key)
{
	return (void *)pdf_keep_obj((pdf_obj *)key);
}

static void
pdf_drop_key(fz_context *ctx, void *key)
{
	pdf_drop_obj((pdf_obj *)key);
}

static int
pdf_cmp_key(void *k0, void *k1)
{
	return pdf_objcmp((pdf_obj *)k0, (pdf_obj *)k1);
}

#ifdef PDF_DEBUG
static void
pdf_debug_key(FILE *out, void *key_)
{
	pdf_obj *key = (pdf_obj *)key_;

	if (pdf_is_indirect(key))
	{
		fprintf(out, "(%d %d R) ", pdf_to_num(key), pdf_to_gen(key));
	} else
		pdf_fprint_obj(out, key, 0);
}
#endif

static fz_store_type pdf_obj_store_type =
{
	pdf_make_hash_key,
	pdf_keep_key,
	pdf_drop_key,
	pdf_cmp_key,
#ifdef PDF_DEBUG
	pdf_debug_key
#endif
};

void
pdf_store_item(fz_context *ctx, pdf_obj *key, void *val, unsigned int itemsize)
{
	void *existing;
	existing = fz_store_item(ctx, key, val, itemsize, &pdf_obj_store_type);
	assert(existing == NULL);
}

void *
pdf_find_item(fz_context *ctx, fz_store_free_fn *free, pdf_obj *key)
{
	return fz_find_item(ctx, free, key, &pdf_obj_store_type);
}

void
pdf_remove_item(fz_context *ctx, fz_store_free_fn *free, pdf_obj *key)
{
	fz_remove_item(ctx, free, key, &pdf_obj_store_type);
}

fz_rect *
pdf_to_rect(fz_context *ctx, pdf_obj *array, fz_rect *r)
{
	float a = pdf_to_real(pdf_array_get(array, 0));
	float b = pdf_to_real(pdf_array_get(array, 1));
	float c = pdf_to_real(pdf_array_get(array, 2));
	float d = pdf_to_real(pdf_array_get(array, 3));
	r->x0 = fz_min(a, c);
	r->y0 = fz_min(b, d);
	r->x1 = fz_max(a, c);
	r->y1 = fz_max(b, d);
	return r;
}

fz_matrix *
pdf_to_matrix(fz_context *ctx, pdf_obj *array, fz_matrix *m)
{
	m->x[0] = pdf_to_real(pdf_array_get(array, 0));
	m->x[1] = pdf_to_real(pdf_array_get(array, 1));
	m->x[2] = pdf_to_real(pdf_array_get(array, 2));
	m->x[3] = pdf_to_real(pdf_array_get(array, 3));
	m->e = pdf_to_real(pdf_array_get(array, 4));
	m->f = pdf_to_real(pdf_array_get(array, 5));
	return m;
}

pdf_obj *
pdf_parse_array(pdf_document *doc, fz_stream *file, pdf_lexbuf *buf)
{
	pdf_obj *ary = NULL;
	pdf_obj *obj = NULL;
	int a = 0, b = 0, n = 0;
	pdf_token tok;
	fz_context *ctx = file->ctx;
	pdf_obj *op;

	ary = pdf_new_array(doc, 4);

	fz_try(ctx)
	{
		while (1)
		{
			tok = pdf_lex(ctx, file, buf);

			if (tok != PDF_TOK_INT && tok != PDF_TOK_R)
			{
				if (n > 0)
				{
					obj = pdf_new_int(doc, a);
					pdf_array_push(ary, obj);
					pdf_drop_obj(obj);
					obj = NULL;
				}
				if (n > 1)
				{
					obj = pdf_new_int(doc, b);
					pdf_array_push(ary, obj);
					pdf_drop_obj(obj);
					obj = NULL;
				}
				n = 0;
			}

			if (tok == PDF_TOK_INT && n == 2)
			{
				obj = pdf_new_int(doc, a);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				a = b;
				n --;
			}

			switch (tok)
			{
			case PDF_TOK_CLOSE_ARRAY:
				op = ary;
				goto end;

			case PDF_TOK_INT:
				if (n == 0)
					a = buf->i;
				if (n == 1)
					b = buf->i;
				n ++;
				break;

			case PDF_TOK_R:
				if (n != 2)
					fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse indirect reference in array");
				obj = pdf_new_indirect(doc, a, b);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				n = 0;
				break;

			case PDF_TOK_OPEN_ARRAY:
				obj = pdf_parse_array(doc, file, buf);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;

			case PDF_TOK_OPEN_DICT:
				obj = pdf_parse_dict(doc, file, buf);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;

			case PDF_TOK_NAME:
				obj = pdf_new_name(doc, buf->scratch);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;
			case PDF_TOK_REAL:
				obj = pdf_new_real(doc, buf->f);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;
			case PDF_TOK_STRING:
				obj = pdf_new_string(doc, buf->scratch, buf->len);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;
			case PDF_TOK_TRUE:
				obj = pdf_new_bool(doc, 1);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;
			case PDF_TOK_FALSE:
				obj = pdf_new_bool(doc, 0);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;
			case PDF_TOK_NULL:
				obj = pdf_new_null(doc);
				pdf_array_push(ary, obj);
				pdf_drop_obj(obj);
				obj = NULL;
				break;

			default:
				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse token in array");
			}
		}
end:
		{}
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(obj);
		pdf_drop_obj(ary);
		fz_rethrow_message(ctx, "cannot parse array");
	}
	return op;
}

pdf_obj *
pdf_parse_dict(pdf_document *doc, fz_stream *file, pdf_lexbuf *buf)
{
	pdf_obj *dict;
	pdf_obj *key = NULL;
	pdf_obj *val = NULL;
	pdf_token tok;
	int a, b;
	fz_context *ctx = file->ctx;

	dict = pdf_new_dict(doc, 8);

	fz_try(ctx)
	{
		while (1)
		{
			tok = pdf_lex(ctx, file, buf);
	skip:
			if (tok == PDF_TOK_CLOSE_DICT)
				break;

			/* for BI .. ID .. EI in content streams */
			if (tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "ID"))
				break;

			if (tok != PDF_TOK_NAME)
				fz_throw(ctx, FZ_ERROR_GENERIC, "invalid key %d in dict", tok);

			key = pdf_new_name(doc, buf->scratch);

			tok = pdf_lex(ctx, file, buf);

			switch (tok)
			{
			case PDF_TOK_OPEN_ARRAY:
				val = pdf_parse_array(doc, file, buf);
				break;

			case PDF_TOK_OPEN_DICT:
				val = pdf_parse_dict(doc, file, buf);
				break;

			case PDF_TOK_NAME: val = pdf_new_name(doc, buf->scratch); break;
			case PDF_TOK_REAL: val = pdf_new_real(doc, buf->f); break;
			case PDF_TOK_STRING: val = pdf_new_string(doc, buf->scratch, buf->len); break;
			case PDF_TOK_TRUE: val = pdf_new_bool(doc, 1); break;
			case PDF_TOK_FALSE: val = pdf_new_bool(doc, 0); break;
			case PDF_TOK_NULL: val = pdf_new_null(doc); break;

			case PDF_TOK_INT:
				/* 64-bit to allow for numbers > INT_MAX and overflow */
				a = buf->i;
				tok = pdf_lex(ctx, file, buf);
				if (tok == PDF_TOK_CLOSE_DICT || tok == PDF_TOK_NAME ||
					(tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "ID")))
				{
					val = pdf_new_int(doc, a);
					pdf_dict_put(dict, key, val);
					pdf_drop_obj(val);
					val = NULL;
					pdf_drop_obj(key);
					key = NULL;
					goto skip;
				}
				if (tok == PDF_TOK_INT)
				{
					b = buf->i;
					tok = pdf_lex(ctx, file, buf);
					if (tok == PDF_TOK_R)
					{
						val = pdf_new_indirect(doc, a, b);
						break;
					}
				}
				fz_throw(ctx, FZ_ERROR_GENERIC, "invalid indirect reference in dict");

			default:
				fz_throw(ctx, FZ_ERROR_GENERIC, "unknown token in dict");
			}

			pdf_dict_put(dict, key, val);
			pdf_drop_obj(val);
			val = NULL;
			pdf_drop_obj(key);
			key = NULL;
		}
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(dict);
		pdf_drop_obj(key);
		pdf_drop_obj(val);
		fz_rethrow_message(ctx, "cannot parse dict");
	}
	return dict;
}

pdf_obj *
pdf_parse_stm_obj(pdf_document *doc, fz_stream *file, pdf_lexbuf *buf)
{
	pdf_token tok;
	fz_context *ctx = file->ctx;

	tok = pdf_lex(ctx, file, buf);

	switch (tok)
	{
	case PDF_TOK_OPEN_ARRAY:
		return pdf_parse_array(doc, file, buf);
	case PDF_TOK_OPEN_DICT:
		return pdf_parse_dict(doc, file, buf);
	case PDF_TOK_NAME: return pdf_new_name(doc, buf->scratch); break;
	case PDF_TOK_REAL: return pdf_new_real(doc, buf->f); break;
	case PDF_TOK_STRING: return pdf_new_string(doc, buf->scratch, buf->len); break;
	case PDF_TOK_TRUE: return pdf_new_bool(doc, 1); break;
	case PDF_TOK_FALSE: return pdf_new_bool(doc, 0); break;
	case PDF_TOK_NULL: return pdf_new_null(doc); break;
	case PDF_TOK_INT: return pdf_new_int(doc, buf->i); break;
	default: fz_throw(ctx, FZ_ERROR_GENERIC, "unknown token in object stream");
	}
	return NULL; /* Stupid MSVC */
}

pdf_obj *
pdf_parse_ind_obj(pdf_document *doc,
	fz_stream *file, pdf_lexbuf *buf,
	int *onum, int *ogen, int *ostmofs)
{
	pdf_obj *obj = NULL;
	int num = 0, gen = 0, stm_ofs;
	pdf_token tok;
	int a, b;
	fz_context *ctx = file->ctx;

	tok = pdf_lex(ctx, file, buf);
	if (tok != PDF_TOK_INT)
		fz_throw(ctx, FZ_ERROR_GENERIC, "expected object number");
	num = buf->i;

	tok = pdf_lex(ctx, file, buf);
	if (tok != PDF_TOK_INT)
		fz_throw(ctx, FZ_ERROR_GENERIC, "expected generation number (%d ? obj)", num);
	gen = buf->i;

	tok = pdf_lex(ctx, file, buf);
	if (tok != PDF_TOK_OBJ)
		fz_throw(ctx, FZ_ERROR_GENERIC, "expected 'obj' keyword (%d %d ?)", num, gen);

	tok = pdf_lex(ctx, file, buf);

	switch (tok)
	{
	case PDF_TOK_OPEN_ARRAY:
		obj = pdf_parse_array(doc, file, buf);
		break;

	case PDF_TOK_OPEN_DICT:
		obj = pdf_parse_dict(doc, file, buf);
		break;

	case PDF_TOK_NAME: obj = pdf_new_name(doc, buf->scratch); break;
	case PDF_TOK_REAL: obj = pdf_new_real(doc, buf->f); break;
	case PDF_TOK_STRING: obj = pdf_new_string(doc, buf->scratch, buf->len); break;
	case PDF_TOK_TRUE: obj = pdf_new_bool(doc, 1); break;
	case PDF_TOK_FALSE: obj = pdf_new_bool(doc, 0); break;
	case PDF_TOK_NULL: obj = pdf_new_null(doc); break;

	case PDF_TOK_INT:
		a = buf->i;
		tok = pdf_lex(ctx, file, buf);

		if (tok == PDF_TOK_STREAM || tok == PDF_TOK_ENDOBJ)
		{
			obj = pdf_new_int(doc, a);
			goto skip;
		}
		if (tok == PDF_TOK_INT)
		{
			b = buf->i;
			tok = pdf_lex(ctx, file, buf);
			if (tok == PDF_TOK_R)
			{
				obj = pdf_new_indirect(doc, a, b);
				break;
			}
		}
		fz_throw(ctx, FZ_ERROR_GENERIC, "expected 'R' keyword (%d %d R)", num, gen);

	case PDF_TOK_ENDOBJ:
		obj = pdf_new_null(doc);
		goto skip;

	default:
		fz_throw(ctx, FZ_ERROR_GENERIC, "syntax error in object (%d %d R)", num, gen);
	}

	fz_try(ctx)
	{
		tok = pdf_lex(ctx, file, buf);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(obj);
		fz_rethrow_message(ctx, "cannot parse indirect object (%d %d R)", num, gen);
	}

skip:
	if (tok == PDF_TOK_STREAM)
	{
		int c = fz_read_byte(file);
		while (c == ' ')
			c = fz_read_byte(file);
		if (c == '\r')
		{
			c = fz_peek_byte(file);
			if (c != '\n')
				fz_warn(ctx->warn, "line feed missing after stream begin marker (%d %d R)", num, gen);
			else
				fz_read_byte(file);
		}
		stm_ofs = fz_tell(file);
	}
	else if (tok == PDF_TOK_ENDOBJ)
	{
		stm_ofs = 0;
	}
	else
	{
		fz_warn(ctx->warn, "expected 'endobj' or 'stream' keyword (%d %d R)", num, gen);
		stm_ofs = 0;
	}

	if (onum) *onum = num;
	if (ogen) *ogen = gen;
	if (ostmofs) *ostmofs = stm_ofs;
	return obj;
}

/* Convert Unicode/PdfDocEncoding string into utf-8 */
char *
pdf_to_utf8(pdf_document *doc, pdf_obj *src)
{
	fz_context *ctx = doc->ctx;
	fz_buffer *strmbuf = NULL;
	unsigned char *srcptr;
	char *dst;
	int srclen;

	fz_try(ctx)
	{
		if (pdf_is_string(src))
		{
			srcptr = (unsigned char *) pdf_to_str_buf(src);
			srclen = pdf_to_str_len(src);
		}
		else if (pdf_is_stream(doc, pdf_to_num(src), pdf_to_gen(src)))
		{
			strmbuf = pdf_load_stream(doc, pdf_to_num(src), pdf_to_gen(src));
			srclen = fz_buffer_storage(strmbuf, (unsigned char **)&srcptr);
		}
		else
		{
			srclen = 0;
		}

        dst = pdf_docenc_to_utf8(ctx, srcptr, srclen);
	}
	fz_always(ctx)
	{
		fz_drop_buffer(strmbuf);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	return dst;
}

pdf_obj *
pdf_to_utf8_name(pdf_document *doc, pdf_obj *src)
{
	char *buf = pdf_to_utf8(doc, src);
	pdf_obj *dst = pdf_new_name(doc, buf);
	fz_free(doc->ctx, buf);
	return dst;
}

/*
 * Compute an encryption key (PDF 1.7 algorithm 3.2)
 */

static const unsigned char padding[32] =
{
	0x28, 0xbf, 0x4e, 0x5e, 0x4e, 0x75, 0x8a, 0x41,
	0x64, 0x00, 0x4e, 0x56, 0xff, 0xfa, 0x01, 0x08,
	0x2e, 0x2e, 0x00, 0xb6, 0xd0, 0x68, 0x3e, 0x80,
	0x2f, 0x0c, 0xa9, 0xfe, 0x64, 0x53, 0x69, 0x7a
};

static void
pdf_compute_encryption_key(pdf_crypt *crypt, unsigned char *password, int pwlen, unsigned char *key)
{
	unsigned char buf[32];
	unsigned int p;
	int i, n;
	fz_md5 md5;

	n = crypt->length / 8;

	/* Step 1 - copy and pad password string */
	if (pwlen > 32)
		pwlen = 32;
	memcpy(buf, password, pwlen);
	memcpy(buf + pwlen, padding, 32 - pwlen);

	/* Step 2 - init md5 and pass value of step 1 */
	fz_md5_init(&md5);
	fz_md5_update(&md5, buf, 32);

	/* Step 3 - pass O value */
	fz_md5_update(&md5, crypt->o, 32);

	/* Step 4 - pass P value as unsigned int, low-order byte first */
	p = (unsigned int) crypt->p;
	buf[0] = (p) & 0xFF;
	buf[1] = (p >> 8) & 0xFF;
	buf[2] = (p >> 16) & 0xFF;
	buf[3] = (p >> 24) & 0xFF;
	fz_md5_update(&md5, buf, 4);

	/* Step 5 - pass first element of ID array */
	fz_md5_update(&md5, (unsigned char *)pdf_to_str_buf(crypt->id), pdf_to_str_len(crypt->id));

	/* Step 6 (revision 4 or greater) - if metadata is not encrypted pass 0xFFFFFFFF */
	if (crypt->r >= 4)
	{
		if (!crypt->encrypt_metadata)
		{
			buf[0] = 0xFF;
			buf[1] = 0xFF;
			buf[2] = 0xFF;
			buf[3] = 0xFF;
			fz_md5_update(&md5, buf, 4);
		}
	}

	/* Step 7 - finish the hash */
	fz_md5_final(&md5, buf);

	/* Step 8 (revision 3 or greater) - do some voodoo 50 times */
	if (crypt->r >= 3)
	{
		for (i = 0; i < 50; i++)
		{
			fz_md5_init(&md5);
			fz_md5_update(&md5, buf, n);
			fz_md5_final(&md5, buf);
		}
	}

	/* Step 9 - the key is the first 'n' bytes of the result */
	memcpy(key, buf, n);
}

static void pdf_compute_encryption_key_r5r6(fz_context *ctx, pdf_crypt *crypt, unsigned char *password, int pwlen, int ownerkey, unsigned char *validationkey)
{
	unsigned char hash[32], *input = (ownerkey ? crypt->oe : crypt->ue);
    unsigned int i;
	fz_aes aes;

	/* Step 2 - truncate UTF-8 password to 127 characters */

	if (pwlen > 127)
		pwlen = 127;

	if (crypt->r == 5) {
	    pdf_compute_hash_r5(ctx, password, pwlen, (ownerkey ? crypt->o : crypt->u) + 32, ownerkey ? crypt->u : NULL, validationkey, hash);
	} else if (crypt->r == 6) {
	    pdf_compute_hardened_hash_r6(ctx, password, pwlen, (ownerkey ? crypt->o : crypt->u) + 32, ownerkey ? crypt->u : NULL, validationkey);
	    pdf_compute_hardened_hash_r6(ctx, password, pwlen, crypt->u + 40, NULL, hash);
    }

	if (aes_setup(&aes, hash, 32))
		fz_throw(ctx, FZ_ERROR_GENERIC, "AES key init failed (keylen=256)");
	aes_ecb_decrypt(&aes, input, crypt->key);
	aes_ecb_decrypt(&aes, input, crypt->key+16);
	for(i = 0; i < 16; i++)
		crypt->key[i+16] = (unsigned char)(crypt->key[i+16] ^ input[i]);
}

/*
 * Computing the user password (PDF 1.7 algorithm 3.4 and 3.5)
 * Also save the generated key for decrypting objects and streams in crypt->key.
 */

static void
pdf_compute_user_password(fz_context *ctx, pdf_crypt *crypt, unsigned char *password, int pwlen, unsigned char *output)
{
    unsigned int i;

	if (crypt->r == 2)
	{
		fz_arc4 arc4;

		pdf_compute_encryption_key(crypt, password, pwlen, crypt->key);
		fz_arc4_init(&arc4, crypt->key, crypt->length / 8);
        for (i = 0; i < 32; i++) {
            unsigned char x = fz_arc4_next(&arc4);
            output[i] = padding[i] ^ x;
        }
	}

	if (crypt->r == 3 || crypt->r == 4)
	{
		unsigned char xor[32];
		unsigned char digest[16];
		fz_md5 md5;
		fz_arc4 arc4;
		int i, x, n;

		n = crypt->length / 8;

		pdf_compute_encryption_key(crypt, password, pwlen, crypt->key);

		fz_md5_init(&md5);
		fz_md5_update(&md5, padding, 32);
		fz_md5_update(&md5, (unsigned char*)pdf_to_str_buf(crypt->id), pdf_to_str_len(crypt->id));
		fz_md5_final(&md5, digest);

		fz_arc4_init(&arc4, crypt->key, n);
        for (i = 0; i < 16; i++) {
            unsigned char x = fz_arc4_next(&arc4);
            output[i] = digest[i] ^ x;
        }

		for (x = 1; x <= 19; x++)
		{
			for (i = 0; i < n; i++)
				xor[i] = crypt->key[i] ^ x;
			fz_arc4_init(&arc4, xor, n);
            for (i = 0; i < 16; i++) {
                unsigned char x = fz_arc4_next(&arc4);
                output[i] ^= x;
            }
		}

		memcpy(output + 16, padding, 16);
	}

#ifndef DISABLE_PDF_CRYPT_R56
    if (crypt->r == 5 || crypt->r == 6) {
	    pdf_compute_encryption_key_r5r6(ctx, crypt, password, pwlen, 0, output);
    }
#endif
}

/*
 * Authenticating the user password (PDF 1.7 algorithm 3.6
 * and ExtensionLevel 3 algorithm 3.11)
 * This also has the side effect of saving a key generated
 * from the password for decrypting objects and streams.
 */

static int
pdf_authenticate_user_password(fz_context *ctx, pdf_crypt *crypt, unsigned char *password, int pwlen)
{
	unsigned char output[32];
	pdf_compute_user_password(ctx, crypt, password, pwlen, output);
	if (crypt->r == 2 || crypt->r == 5 || crypt->r == 6)
		return memcmp(output, crypt->u, 32) == 0;
	if (crypt->r == 3 || crypt->r == 4)
		return memcmp(output, crypt->u, 16) == 0;
	return 0;
}

/*
 * Authenticating the owner password (PDF 1.7 algorithm 3.7
 * and ExtensionLevel 3 algorithm 3.12)
 * Generates the user password from the owner password
 * and calls pdf_authenticate_user_password.
 */

static int
pdf_authenticate_owner_password(fz_context *ctx, pdf_crypt *crypt, unsigned char *ownerpass, int pwlen)
{
	unsigned char pwbuf[32];
	unsigned char key[32];
	unsigned char xor[32];
	unsigned char userpass[32];
	int i, n, x;
	fz_md5 md5;
	fz_arc4 arc4;

#ifndef DISABLE_PDF_CRYPT_R56
    if (crypt->r == 5 || crypt->r == 6) {
		/* r5: PDF 1.7 ExtensionLevel 3 algorithm 3.12 */
		/* r6: PDF 1.7 ExtensionLevel 8 algorithm */
		pdf_compute_encryption_key_r5r6(ctx, crypt, ownerpass, pwlen, 1, key);
	    return !memcmp(key, crypt->o, 32);
    }
#endif

	n = crypt->length / 8;

	/* Step 1 -- steps 1 to 4 of PDF 1.7 algorithm 3.3 */

	/* copy and pad password string */
	if (pwlen > 32)
		pwlen = 32;
	memcpy(pwbuf, ownerpass, pwlen);
	memcpy(pwbuf + pwlen, padding, 32 - pwlen);

	/* take md5 hash of padded password */
	fz_md5_init(&md5);
	fz_md5_update(&md5, pwbuf, 32);
	fz_md5_final(&md5, key);

	/* do some voodoo 50 times (Revision 3 or greater) */
	if (crypt->r >= 3)
	{
		for (i = 0; i < 50; i++)
		{
			fz_md5_init(&md5);
			fz_md5_update(&md5, key, 16);
			fz_md5_final(&md5, key);
		}
	}

	/* Step 2 (Revision 2) */
	if (crypt->r == 2)
	{
		fz_arc4_init(&arc4, key, n);
        for (i = 0; i < 32; i++) {
            unsigned char x = fz_arc4_next(&arc4);
            userpass[i] = crypt->o[i] ^ x;
        }
	}

	/* Step 2 (Revision 3 or greater) */
	if (crypt->r >= 3)
	{
		memcpy(userpass, crypt->o, 32);
		for (x = 0; x < 20; x++)
		{
			for (i = 0; i < n; i++)
				xor[i] = key[i] ^ (19 - x);
			fz_arc4_init(&arc4, xor, n);
            for (i = 0; i < 32; i++) {
                unsigned char x = fz_arc4_next(&arc4);
                userpass[i] ^= x;
            }
		}
	}

	return pdf_authenticate_user_password(ctx, crypt, userpass, 32);
}

static const unsigned short pdf_doc_encoding_80[32] =
{
	0x2022, 0x2020, 0x2021, 0x2026, 0x2014, 0x2013, 0x0192, 0x2044,
	0x2039, 0x203a, 0x2212, 0x2030, 0x201e, 0x201c, 0x201d, 0x2018,
	0x2019, 0x201a, 0x2122, 0xfb01, 0xfb02, 0x0141, 0x0152, 0x0160,
	0x0178, 0x017d, 0x0131, 0x0142, 0x0153, 0x0161, 0x017e, 0x00,
};

static const unsigned char docenc_18_to_1f[8] = { 0x1a, 0x19, 0x18, 0x1b, 0x1e, 0x1d, 0x1f, 0x1c };

static void pdf_docenc_from_utf8(char *password, const char *utf8, int n)
{
	int i = 0, k;
    unsigned int c;
	while (*utf8 && i + 1 < n) {
		utf8 += fz_chartorune(&c, utf8);
        if ((c < 0x100) && ((c < 0x18) || ((c >= 0x20) && (c < 0x80)) || (c >= 0xa0))) {
		    password[i++] = c;
        } else {
            if (((c|0x10) - 0x2d6) < 8) password[i++] = docenc_18_to_1f[(c|0x10) - 0x02d6];
            else if (c == 0x20ac) password[i++] = 0xa0;
            else {
		        for (k = 0; k < 32; k++) {
			        if (c == pdf_doc_encoding_80[k]) {
				        password[i++] = k+128;
				        break;
			        }
		        }
            }
        }
		/* FIXME: drop characters that can't be encoded or return an error? */
	}
	password[i] = 0;
}

int
pdf_authenticate_password(pdf_document *doc, const char *pwd_utf8)
{
	char password[2048];

	if (doc->crypt) {
		password[0] = 0;
		if (pwd_utf8) {
			if (doc->crypt->r <= 4)
				pdf_docenc_from_utf8(password, pwd_utf8, sizeof(password));
			else
	            /* TODO: stringprep with SALSprep profile */
	            fz_strlcpy(password, pwd_utf8, sizeof(password));
		}

		if (pdf_authenticate_user_password(doc->ctx, doc->crypt, (unsigned char *)password, strlen(password)))
			return 1;
		if (pdf_authenticate_owner_password(doc->ctx, doc->crypt, (unsigned char *)password, strlen(password)))
			return 1;
		return 0;
	}
	return 1;
}

int
pdf_needs_password(pdf_document *doc)
{
	if (!doc->crypt)
		return 0;
	if (pdf_authenticate_password(doc, ""))
		return 0;
	return 1;
}

int
pdf_has_permission(pdf_document *doc, int p)
{
	if (!doc->crypt)
		return 1;
	return doc->crypt->p & p;
}

unsigned char *
pdf_crypt_key(pdf_document *doc)
{
	if (doc->crypt)
		return doc->crypt->key;
	return NULL;
}

int
pdf_crypt_version(pdf_document *doc)
{
	if (doc->crypt)
		return doc->crypt->v;
	return 0;
}

int pdf_crypt_revision(pdf_document *doc)
{
	if (doc->crypt)
		return doc->crypt->r;
	return 0;
}

char *
pdf_crypt_method(pdf_document *doc)
{
	if (doc->crypt)
	{
		switch (doc->crypt->strf.method)
		{
		case PDF_CRYPT_NONE: return "None";
		case PDF_CRYPT_RC4: return "RC4";
		case PDF_CRYPT_AESV2: return "AES";
		case PDF_CRYPT_AESV3: return "AES";
		case PDF_CRYPT_UNKNOWN: return "Unknown";
		}
	}
	return "None";
}

int
pdf_crypt_length(pdf_document *doc)
{
	if (doc->crypt)
		return doc->crypt->length;
	return 0;
}

/*
 * Create crypt object for decrypting strings and streams
 * given the Encryption and ID objects.
 */

pdf_crypt *
pdf_new_crypt(fz_context *ctx, pdf_obj *dict, pdf_obj *id)
{
	pdf_crypt *crypt;
	pdf_obj *obj;

	crypt = fz_malloc_struct(ctx, pdf_crypt);

	/* Common to all security handlers (PDF 1.7 table 3.18) */

	obj = pdf_dict_gets(dict, "Filter");
	if (!pdf_is_name(obj))
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "unspecified encryption handler");
	}
	if (strcmp(pdf_to_name(obj), "Standard") != 0)
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "unknown encryption handler: '%s'", pdf_to_name(obj));
	}

	crypt->v = 0;
	obj = pdf_dict_gets(dict, "V");
	if (pdf_is_int(obj))
		crypt->v = pdf_to_int(obj);
	if (crypt->v != 1 && crypt->v != 2 && crypt->v != 4 && crypt->v != 5)
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "unknown encryption version");
	}

	/* Standard security handler (PDF 1.7 table 3.19) */
	obj = pdf_dict_gets(dict, "R");
	if (pdf_is_int(obj))
		crypt->r = pdf_to_int(obj);
	else if (crypt->v <= 4)
	{
		fz_warn(ctx->warn, "encryption dictionary missing revision value, guessing...");
		if (crypt->v < 2)
			crypt->r = 2;
		else if (crypt->v == 2)
			crypt->r = 3;
		else if (crypt->v == 4)
			crypt->r = 4;
	}
	else
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "encryption dictionary missing version and revision value");
	}

	obj = pdf_dict_gets(dict, "O");
	if (pdf_is_string(obj) && pdf_to_str_len(obj) == 32)
		memcpy(crypt->o, pdf_to_str_buf(obj), 32);
	/* /O and /U are supposed to be 48 bytes long for revision 5 and 6, they're often longer, though */
	else if (crypt->r >= 5 && pdf_is_string(obj) && pdf_to_str_len(obj) >= 48)
		memcpy(crypt->o, pdf_to_str_buf(obj), 48);
	else
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "encryption dictionary missing owner password");
	}

	obj = pdf_dict_gets(dict, "U");
	if (pdf_is_string(obj) && pdf_to_str_len(obj) == 32)
		memcpy(crypt->u, pdf_to_str_buf(obj), 32);
	/* /O and /U are supposed to be 48 bytes long for revision 5 and 6, they're often longer, though */
	else if (crypt->r >= 5 && pdf_is_string(obj) && pdf_to_str_len(obj) >= 48)
		memcpy(crypt->u, pdf_to_str_buf(obj), 48);
	else if (pdf_is_string(obj) && pdf_to_str_len(obj) < 32)
	{
		fz_warn(ctx->warn, "encryption password key too short (%d)", pdf_to_str_len(obj));
		memcpy(crypt->u, pdf_to_str_buf(obj), pdf_to_str_len(obj));
	}
	else
	{
		pdf_free_crypt(ctx, crypt);
		fz_throw(ctx, FZ_ERROR_GENERIC, "encryption dictionary missing user password");
	}

	obj = pdf_dict_gets(dict, "P");
	if (pdf_is_int(obj))
		crypt->p = pdf_to_int(obj);
	else
	{
		fz_warn(ctx->warn, "encryption dictionary missing permissions");
		crypt->p = 0xfffffffc;
	}

	crypt->encrypt_metadata = 1;
	obj = pdf_dict_gets(dict, "EncryptMetadata");
	if (pdf_is_bool(obj))
		crypt->encrypt_metadata = pdf_to_bool(obj);

	/* Extract file identifier string */
	if (pdf_is_array(id) && pdf_array_len(id) == 2)
	{
		obj = pdf_array_get(id, 0);
		if (pdf_is_string(obj))
			crypt->id = pdf_keep_obj(obj);
	}
	else
		fz_warn(ctx->warn, "missing file identifier, may not be able to do decryption");

	/* Determine encryption key length */
	crypt->length = 40;
	if (crypt->v == 2 || crypt->v == 4)
	{
		obj = pdf_dict_gets(dict, "Length");
		if (pdf_is_int(obj))
			crypt->length = pdf_to_int(obj);

		/* work-around for pdf generators that assume length is in bytes */
		if (crypt->length < 40)
			crypt->length = crypt->length * 8;

		if (crypt->length % 8 != 0)
		{
			pdf_free_crypt(ctx, crypt);
			fz_throw(ctx, FZ_ERROR_GENERIC, "invalid encryption key length");
		}
		if (crypt->length < 0 || crypt->length > 256)
		{
			pdf_free_crypt(ctx, crypt);
			fz_throw(ctx, FZ_ERROR_GENERIC, "invalid encryption key length");
		}
	}

	if (crypt->v == 5)
		crypt->length = 256;

	if (crypt->v == 1 || crypt->v == 2)
	{
		crypt->stmf.method = PDF_CRYPT_RC4;
		crypt->stmf.length = crypt->length;

		crypt->strf.method = PDF_CRYPT_RC4;
		crypt->strf.length = crypt->length;
	}

	if (crypt->v == 4 || crypt->v == 5)
	{
		crypt->stmf.method = PDF_CRYPT_NONE;
		crypt->stmf.length = crypt->length;

		crypt->strf.method = PDF_CRYPT_NONE;
		crypt->strf.length = crypt->length;

		obj = pdf_dict_gets(dict, "CF");
		if (pdf_is_dict(obj))
		{
			crypt->cf = pdf_keep_obj(obj);
		}
		else
		{
			crypt->cf = NULL;
		}

		fz_try(ctx)
		{
			obj = pdf_dict_gets(dict, "StmF");
			if (pdf_is_name(obj))
				pdf_parse_crypt_filter(ctx, &crypt->stmf, crypt, pdf_to_name(obj));

			obj = pdf_dict_gets(dict, "StrF");
			if (pdf_is_name(obj))
				pdf_parse_crypt_filter(ctx, &crypt->strf, crypt, pdf_to_name(obj));
		}
		fz_catch(ctx)
		{
			pdf_free_crypt(ctx, crypt);
			fz_rethrow_message(ctx, "cannot parse string crypt filter (%d %d R)", pdf_to_num(obj), pdf_to_gen(obj));
		}

		/* in crypt revision 4, the crypt filter determines the key length */
		if (crypt->strf.method != PDF_CRYPT_NONE)
			crypt->length = crypt->stmf.length;
	}

	return crypt;
}

void
pdf_free_crypt(fz_context *ctx, pdf_crypt *crypt)
{
	pdf_drop_obj(crypt->id);
	pdf_drop_obj(crypt->cf);
	fz_free(ctx, crypt);
}

#ifndef NDEBUG
void pdf_print_crypt(pdf_crypt *crypt)
{
	int i;

	printf("crypt {\n");

	printf("\tv=%d length=%d\n", crypt->v, crypt->length);
	printf("\tstmf method=%d length=%d\n", crypt->stmf.method, crypt->stmf.length);
	printf("\tstrf method=%d length=%d\n", crypt->strf.method, crypt->strf.length);
	printf("\tr=%d\n", crypt->r);

	printf("\to=<");
	for (i = 0; i < 32; i++)
		printf("%02X", crypt->o[i]);
	printf(">\n");

	printf("\tu=<");
	for (i = 0; i < 32; i++)
		printf("%02X", crypt->u[i]);
	printf(">\n");

	printf("}\n");
}
#endif

#undef DEBUG_PROGESSIVE_ADVANCE

#ifdef DEBUG_PROGESSIVE_ADVANCE
#define DEBUGMESS(A) do { fz_warn A; } while (0)
#else
#define DEBUGMESS(A) do { } while (0)
#endif

/*
 * xref tables
 */

static void pdf_free_xref_sections(pdf_document *doc)
{
	fz_context *ctx = doc->ctx;
	int x, e;

	for (x = 0; x < doc->num_xref_sections; x++)
	{
		pdf_xref *xref = &doc->xref_sections[x];

		for (e = 0; e < xref->len; e++)
		{
			pdf_xref_entry *entry = &xref->table[e];

			if (entry->obj)
			{
				pdf_drop_obj(entry->obj);
				fz_drop_buffer(entry->stm_buf);
			}
		}

		fz_free(ctx, xref->table);
		pdf_drop_obj(xref->trailer);
	}

	fz_free(ctx, doc->xref_sections);
	doc->xref_sections = NULL;
	doc->num_xref_sections = 0;
}

static void pdf_resize_xref(fz_context *ctx, pdf_xref *xref, int newlen)
{
	int i;

	xref->table = fz_resize_array(ctx, xref->table, newlen, sizeof(pdf_xref_entry));
	for (i = xref->len; i < newlen; i++)
	{
		xref->table[i].type = 0;
		xref->table[i].ofs = 0;
		xref->table[i].gen = 0;
		xref->table[i].stm_ofs = 0;
		xref->table[i].stm_buf = NULL;
		xref->table[i].obj = NULL;
	}
	xref->len = newlen;
}

static void pdf_populate_next_xref_level(pdf_document *doc)
{
	pdf_xref *xref;
	doc->xref_sections = fz_resize_array(doc->ctx, doc->xref_sections, doc->num_xref_sections + 1, sizeof(pdf_xref));
	doc->num_xref_sections++;

	xref = &doc->xref_sections[doc->num_xref_sections - 1];
	xref->len = 0;
	xref->table = NULL;
	xref->trailer = NULL;
}

pdf_obj *pdf_trailer(pdf_document *doc)
{
	/* Return the document's final trailer */
	pdf_xref *xref = &doc->xref_sections[0];

	return xref->trailer;
}

void pdf_set_populating_xref_trailer(pdf_document *doc, pdf_obj *trailer)
{
	/* Update the trailer of the xref section being populated */
	pdf_xref *xref = &doc->xref_sections[doc->num_xref_sections - 1];
	pdf_drop_obj(xref->trailer);
	xref->trailer = pdf_keep_obj(trailer);
}

int pdf_xref_len(pdf_document *doc)
{
	/* Return the length of the document's final xref section */
	pdf_xref *xref = &doc->xref_sections[0];

	return xref->len;
}

/* Used while reading the individual xref sections from a file */
pdf_xref_entry *pdf_get_populating_xref_entry(pdf_document *doc, int num)
{
	/* Return an entry within the xref currently being populated */
	pdf_xref *xref;
	int i;

	if (doc->num_xref_sections == 0)
	{
		doc->xref_sections = fz_calloc(doc->ctx, 1, sizeof(pdf_xref));
		doc->num_xref_sections = 1;
	}

	/* Ensure all xref sections map this entry */
	for (i = doc->num_xref_sections - 1; i >= 0; i--)
	{
		xref = &doc->xref_sections[i];

		if (num >= xref->len)
			pdf_resize_xref(doc->ctx, xref, num+1);
		else
			break; /* Remaining sections already of sufficient size */
	}

	/* Loop leaves xref pointing at the populating section */
	return &doc->xref_sections[doc->num_xref_sections-1].table[num];
}

/* Used after loading a document to access entries */
pdf_xref_entry *pdf_get_xref_entry(pdf_document *doc, int i)
{
	int j;

	/* Find the first xref section where the entry is defined. */
	for (j = 0; j < doc->num_xref_sections; j++)
	{
		pdf_xref *xref = &doc->xref_sections[j];

		if (i >= 0 && i < xref->len)
		{
			pdf_xref_entry *entry = &xref->table[i];

			if (entry->type)
				return entry;
		}
	}

	/*
		Didn't find the entry in any section. Return the entry from the final
		section.
	*/
	return &doc->xref_sections[0].table[i];
}

/*
		Ensure we have an incremental xref section where we can store
		updated versions of indirect objects
*/
static void ensure_incremental_xref(pdf_document *doc)
{
	fz_context *ctx = doc->ctx;

	if (!doc->xref_altered)
	{
		pdf_xref *xref = &doc->xref_sections[0];
		pdf_xref *pxref;
		pdf_xref_entry *new_table = fz_calloc(ctx, xref->len, sizeof(pdf_xref_entry));
		pdf_obj *trailer = NULL;

		fz_try(ctx)
		{
			trailer = pdf_copy_dict(xref->trailer);
			doc->xref_sections = fz_resize_array(ctx, doc->xref_sections, doc->num_xref_sections + 1, sizeof(pdf_xref));
			xref = &doc->xref_sections[0];
			pxref = &doc->xref_sections[1];
			memmove(pxref, xref, doc->num_xref_sections * sizeof(pdf_xref));
			/* xref->len is already correct */
			xref->table = new_table;
			xref->trailer = trailer;
			doc->num_xref_sections++;
			doc->xref_altered = 1;
		}
		fz_catch(ctx)
		{
			fz_free(ctx, new_table);
			pdf_drop_obj(trailer);
			fz_rethrow(ctx);
		}
	}
}

/* Used when altering a document */
static pdf_xref_entry *pdf_get_incremental_xref_entry(pdf_document *doc, int i)
{
	fz_context *ctx = doc->ctx;
	pdf_xref *xref;

	/* Make a new final xref section if we haven't already */
	ensure_incremental_xref(doc);

	xref = &doc->xref_sections[0];
	if (i >= xref->len)
		pdf_resize_xref(ctx, xref, i + 1);

	return &xref->table[i];
}

int pdf_xref_is_incremental(pdf_document *doc, int num)
{
	pdf_xref *xref = &doc->xref_sections[0];

	return doc->xref_altered && num < xref->len && xref->table[num].type;
}

/* Ensure that an object has been cloned into the incremental xref section */
void pdf_xref_ensure_incremental_object(pdf_document *doc, int num)
{
	pdf_xref_entry *new_entry, *old_entry;
	int i;

	/* Make sure we have created an xref section for incremental updates */
	ensure_incremental_xref(doc);

	/* Search for the section that contains this object */
	for (i = 0; i < doc->num_xref_sections; i++)
	{
		pdf_xref *xref = &doc->xref_sections[i];
		if (num >= 0 && num < xref->len && xref->table[num].type)
			break;
	}

	/* If we don't find it, or it's already in the incremental section, return */
	if (i == 0 || i == doc->num_xref_sections)
		return;

	/* Move the object to the incremental section */
	old_entry = &doc->xref_sections[i].table[num];
	new_entry = pdf_get_incremental_xref_entry(doc, num);
	*new_entry = *old_entry;
	old_entry->obj = NULL;
	old_entry->stm_buf = NULL;
}

void pdf_replace_xref(pdf_document *doc, pdf_xref_entry *entries, int n)
{
	fz_context *ctx = doc->ctx;
	pdf_xref *xref;
	pdf_obj *trailer = pdf_keep_obj(pdf_trailer(doc));

	/* The new table completely replaces the previous separate sections */
	pdf_free_xref_sections(doc);

	xref = fz_calloc(ctx, 1, sizeof(pdf_xref));
	xref->table = entries;
	xref->len = n;
	xref->trailer = trailer;
	trailer = NULL;

	doc->xref_sections = xref;
	doc->num_xref_sections = 1;
}

/*
 * magic version tag and startxref
 */

static void
pdf_load_version(pdf_document *doc)
{
	char buf[20];

	fz_seek(doc->file, 0, SEEK_SET);
	fz_read_line(doc->file, buf, sizeof buf);
	if (memcmp(buf, "%PDF-", 5) != 0)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot recognize version marker");

	doc->version = ((buf[5] - 0x30) * 10) + (buf[7] - 0x30);
}

static void
pdf_read_start_xref(pdf_document *doc)
{
	unsigned char buf[1024];
	int t, n;
	int i;

	fz_seek(doc->file, 0, SEEK_END);

	doc->file_size = fz_tell(doc->file);

	t = fz_maxi(0, doc->file_size - (int)sizeof buf);
	fz_seek(doc->file, t, SEEK_SET);

	n = fz_read(doc->file, buf, sizeof buf);
	if (n < 0)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot read from file");

	for (i = n - 9; i >= 0; i--)
	{
		if (memcmp(buf + i, "startxref", 9) == 0)
		{
			i += 9;
			while (iswhite(buf[i]) && i < n)
				i ++;
			doc->startxref = fz_atoi((char*)(buf + i));
			if (doc->startxref != 0)
				return;
			break;
		}
	}

	fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot find startxref");
}

static char *
fz_strsep_(char **stringp, const char delim)
{
	char *ret = *stringp;
	if (!ret) return NULL;
	if ((*stringp = strchr(*stringp, delim)))
		*((*stringp)++) = '\0';
	return ret;
}

static void
fz_skip_space(fz_stream *stm)
{
	do
	{
		int c = fz_peek_byte(stm);
		if (c > 32 && c != EOF)
			return;
		(void)fz_read_byte(stm);
	}
	while (1);
}

static int
fz_skip_string(fz_stream *stm, const char *str)
{
	while (*str)
	{
		int c = fz_peek_byte(stm);
		if (c == EOF || c != *str++)
			return 1;
		(void)fz_read_byte(stm);
	}
	return 0;
}

/*
 * trailer dictionary
 */

static int
pdf_xref_size_from_old_trailer(pdf_document *doc, pdf_lexbuf *buf)
{
	int len;
	char *s;
	int t;
	pdf_token tok;
	int c;
	int size;
	int ofs;

	/* Record the current file read offset so that we can reinstate it */
	ofs = fz_tell(doc->file);

	fz_skip_space(doc->file);
	if (fz_skip_string(doc->file, "xref"))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot find xref marker");
	fz_skip_space(doc->file);

	while (1)
	{
		c = fz_peek_byte(doc->file);
		if (!(c >= '0' && c <= '9'))
			break;

		fz_read_line(doc->file, buf->scratch, buf->size);
		s = buf->scratch;
		fz_strsep_(&s, ' '); /* ignore ofs */
		if (!s)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "invalid range marker in xref");
		len = fz_atoi(fz_strsep_(&s, ' '));

		/* broken pdfs where the section is not on a separate line */
		if (s && *s != '\0')
			fz_seek(doc->file, -(2 + (int)strlen(s)), SEEK_CUR);

		t = fz_tell(doc->file);
		if (t < 0)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot tell in file");

		fz_seek(doc->file, t + 20 * len, SEEK_SET);
	}

	fz_try(doc->ctx)
	{
		pdf_obj *trailer;
		tok = pdf_lex(doc->ctx, doc->file, buf);
		if (tok != PDF_TOK_TRAILER)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "expected trailer marker, got: %s", buf->scratch);

		tok = pdf_lex(doc->ctx, doc->file, buf);
		if (tok != PDF_TOK_OPEN_DICT)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "expected trailer dictionary");

		trailer = pdf_parse_dict(doc, doc->file, buf);

		size = pdf_to_int(pdf_dict_gets(trailer, "Size"));
		if (!size)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "trailer missing Size entry");

		pdf_drop_obj(trailer);
	}
	fz_catch(doc->ctx)
	{
		fz_rethrow_message(doc->ctx, "cannot parse trailer");
	}

	fz_seek(doc->file, ofs, SEEK_SET);

	return size;
}

pdf_obj *
pdf_new_ref(pdf_document *doc, pdf_obj *obj)
{
	int num = pdf_create_object(doc);
	pdf_update_object(doc, num, obj);
	return pdf_new_indirect(doc, num, 0);
}

static pdf_obj *
pdf_read_old_xref(pdf_document *doc, pdf_lexbuf *buf)
{
	int ofs, len;
	char *s;
	int n;
	pdf_token tok;
	int i;
	int c;
	pdf_obj *trailer;
	int xref_len = pdf_xref_size_from_old_trailer(doc, buf);

	/* Access last entry to ensure xref size up front and avoid reallocs */
	(void)pdf_get_populating_xref_entry(doc, xref_len - 1);

	fz_skip_space(doc->file);
	if (fz_skip_string(doc->file, "xref"))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot find xref marker");
	fz_skip_space(doc->file);

	while (1)
	{
		c = fz_peek_byte(doc->file);
		if (!(c >= '0' && c <= '9'))
			break;

		fz_read_line(doc->file, buf->scratch, buf->size);
		s = buf->scratch;
		ofs = fz_atoi(fz_strsep_(&s, ' '));
		len = fz_atoi(fz_strsep_(&s, ' '));

		/* broken pdfs where the section is not on a separate line */
		if (s && *s != '\0')
		{
			fz_warn(doc->ctx->warn, "broken xref section. proceeding anyway.");
			fz_seek(doc->file, -(2 + (int)strlen(s)), SEEK_CUR);
		}

		if (ofs < 0)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "out of range object num in xref: %d", ofs);

		/* broken pdfs where size in trailer undershoots entries in xref sections */
		if (ofs + len > xref_len)
		{
			fz_warn(doc->ctx->warn, "broken xref section, proceeding anyway.");
			/* Access last entry to ensure size */
			(void)pdf_get_populating_xref_entry(doc, ofs + len - 1);
		}

		for (i = ofs; i < ofs + len; i++)
		{
			pdf_xref_entry *entry = pdf_get_populating_xref_entry(doc, i);
			n = fz_read(doc->file, (unsigned char *) buf->scratch, 20);
			if (n < 0)
				fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot read xref table");
			if (!entry->type)
			{
				s = buf->scratch;

				/* broken pdfs where line start with white space */
				while (*s != '\0' && iswhite(*s))
					s++;

				entry->ofs = fz_atoi(s);
				entry->gen = fz_atoi(s + 11);
				entry->type = s[17];
				if (s[17] != 'f' && s[17] != 'n' && s[17] != 'o')
					fz_throw(doc->ctx, FZ_ERROR_GENERIC, "unexpected xref type: %#x (%d %d R)", s[17], i, entry->gen);
			}
		}
	}

	fz_try(doc->ctx)
	{
		tok = pdf_lex(doc->ctx, doc->file, buf);
		if (tok != PDF_TOK_TRAILER)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "expected trailer marker, got: %s", buf->scratch);

		tok = pdf_lex(doc->ctx, doc->file, buf);
		if (tok != PDF_TOK_OPEN_DICT)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "expected trailer dictionary");

		trailer = pdf_parse_dict(doc, doc->file, buf);
	}
	fz_catch(doc->ctx)
	{
		fz_rethrow_message(doc->ctx, "cannot parse trailer");
	}
	return trailer;
}

static void
pdf_read_new_xref_section(pdf_document *doc, fz_stream *stm, int i0, int i1, int w0, int w1, int w2)
{
	int i, n;

	if (i0 < 0 || i1 < 0)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "negative xref stream entry index");
	if (i0 + i1 > pdf_xref_len(doc))
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "xref stream has too many entries");

	for (i = i0; i < i0 + i1; i++)
	{
		pdf_xref_entry *entry = pdf_get_populating_xref_entry(doc, i);
		int a = 0;
		int b = 0;
		int c = 0;

		if (fz_is_eof(stm))
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "truncated xref stream");

		for (n = 0; n < w0; n++)
			a = (a << 8) + fz_read_byte(stm);
		for (n = 0; n < w1; n++)
			b = (b << 8) + fz_read_byte(stm);
		for (n = 0; n < w2; n++)
			c = (c << 8) + fz_read_byte(stm);

		if (!entry->type)
		{
			int t = w0 ? a : 1;
			entry->type = t == 0 ? 'f' : t == 1 ? 'n' : t == 2 ? 'o' : 0;
			entry->ofs = w1 ? b : 0;
			entry->gen = w2 ? c : 0;
		}
	}
}

/* Entered with file locked, remains locked throughout. */
static pdf_obj *
pdf_read_new_xref(pdf_document *doc, pdf_lexbuf *buf)
{
	fz_stream *stm = NULL;
	pdf_obj *trailer = NULL;
	pdf_obj *index = NULL;
	pdf_obj *obj = NULL;
	int num, gen, stm_ofs;
	int size, w0, w1, w2;
	int t;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		pdf_xref_entry *entry;
		int ofs = fz_tell(doc->file);
		trailer = pdf_parse_ind_obj(doc, doc->file, buf, &num, &gen, &stm_ofs);
		entry = pdf_get_populating_xref_entry(doc, num);
		entry->ofs = ofs;
		entry->gen = gen;
		entry->stm_ofs = stm_ofs;
		pdf_drop_obj(entry->obj);
		entry->obj = pdf_keep_obj(trailer);
		entry->type = 'n';
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot parse compressed xref stream object");
	}

	fz_try(ctx)
	{
		obj = pdf_dict_gets(trailer, "Size");
		if (!obj)
			fz_throw(ctx, FZ_ERROR_GENERIC, "xref stream missing Size entry (%d %d R)", num, gen);

		size = pdf_to_int(obj);
		/* Access xref entry to assure table size */
		(void)pdf_get_populating_xref_entry(doc, size-1);

		if (num < 0 || num >= pdf_xref_len(doc))
			fz_throw(ctx, FZ_ERROR_GENERIC, "object id (%d %d R) out of range (0..%d)", num, gen, pdf_xref_len(doc) - 1);

		obj = pdf_dict_gets(trailer, "W");
		if (!obj)
			fz_throw(ctx, FZ_ERROR_GENERIC, "xref stream missing W entry (%d %d R)", num, gen);
		w0 = pdf_to_int(pdf_array_get(obj, 0));
		w1 = pdf_to_int(pdf_array_get(obj, 1));
		w2 = pdf_to_int(pdf_array_get(obj, 2));

		if (w0 < 0)
			fz_warn(ctx->warn, "xref stream objects have corrupt type");
		if (w1 < 0)
			fz_warn(ctx->warn, "xref stream objects have corrupt offset");
		if (w2 < 0)
			fz_warn(ctx->warn, "xref stream objects have corrupt generation");

		w0 = w0 < 0 ? 0 : w0;
		w1 = w1 < 0 ? 0 : w1;
		w2 = w2 < 0 ? 0 : w2;

		index = pdf_dict_gets(trailer, "Index");

		stm = pdf_open_stream_with_offset(doc, num, gen, trailer, stm_ofs);

		if (!index)
		{
			pdf_read_new_xref_section(doc, stm, 0, size, w0, w1, w2);
		}
		else
		{
			int n = pdf_array_len(index);
			for (t = 0; t < n; t += 2)
			{
				int i0 = pdf_to_int(pdf_array_get(index, t + 0));
				int i1 = pdf_to_int(pdf_array_get(index, t + 1));
				pdf_read_new_xref_section(doc, stm, i0, i1, w0, w1, w2);
			}
		}
	}
	fz_always(ctx)
	{
		fz_close(stm);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(trailer);
		fz_rethrow(ctx);
	}

	return trailer;
}

/* File is locked on entry, and exit (but may be dropped in the middle) */
static pdf_obj *
pdf_read_xref(pdf_document *doc, int ofs, pdf_lexbuf *buf)
{
	int c;
	fz_context *ctx = doc->ctx;
	pdf_obj *trailer;

	fz_seek(doc->file, ofs, SEEK_SET);

	while (iswhite(fz_peek_byte(doc->file)))
		fz_read_byte(doc->file);

	fz_try(ctx)
	{
		c = fz_peek_byte(doc->file);
		if (c == 'x')
			trailer = pdf_read_old_xref(doc, buf);
		else if (c >= '0' && c <= '9')
			trailer = pdf_read_new_xref(doc, buf);
		else
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot recognize xref format");
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot read xref (ofs=%d)", ofs);
	}
	return trailer;
}

typedef struct ofs_list_s ofs_list;

struct ofs_list_s
{
	int max;
	int len;
	int *list;
};

static int
read_xref_section(pdf_document *doc, int ofs, pdf_lexbuf *buf, ofs_list *offsets)
{
	pdf_obj *trailer = NULL;
	fz_context *ctx = doc->ctx;
	int xrefstmofs = 0;
	int prevofs = 0;

	fz_try(ctx)
	{
		int i;
		/* Avoid potential infinite recursion */
		for (i = 0; i < offsets->len; i ++)
		{
			if (offsets->list[i] == ofs)
				break;
		}
		if (i < offsets->len)
		{
			fz_warn(ctx->warn, "ignoring xref recursion with offset %d", ofs);
			break;
		}
		if (offsets->len == offsets->max)
		{
			offsets->list = fz_resize_array(ctx, offsets->list, offsets->max*2, sizeof(int));
			offsets->max *= 2;
		}
		offsets->list[offsets->len++] = ofs;

		trailer = pdf_read_xref(doc, ofs, buf);

		pdf_set_populating_xref_trailer(doc, trailer);

		/* FIXME: do we overwrite free entries properly? */
		/* FIXME: Does this work properly with progression? */
		xrefstmofs = pdf_to_int(pdf_dict_gets(trailer, "XRefStm"));
		if (xrefstmofs)
		{
			if (xrefstmofs < 0)
				fz_throw(ctx, FZ_ERROR_GENERIC, "negative xref stream offset");

			/*
				Read the XRefStm stream, but throw away the resulting trailer. We do not
				follow any Prev tag therein, as specified on Page 108 of the PDF reference
				1.7
			*/
			pdf_drop_obj(pdf_read_xref(doc, xrefstmofs, buf));
		}

		prevofs = pdf_to_int(pdf_dict_gets(trailer, "Prev"));
		if (prevofs < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "negative xref stream offset for previous xref stream");
	}
	fz_always(ctx)
	{
		pdf_drop_obj(trailer);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot read xref at offset %d", ofs);
	}

	return prevofs;
}

static void
pdf_read_xref_sections(pdf_document *doc, int ofs, pdf_lexbuf *buf, int read_previous)
{
	fz_context *ctx = doc->ctx;
	ofs_list list;

	list.len = 0;
	list.max = 12;
	list.list = fz_malloc_array(ctx, 12, sizeof(int));
	fz_try(ctx)
	{
		while(ofs) {
			pdf_populate_next_xref_level(doc);
			ofs = read_xref_section(doc, ofs, buf, &list);
			if (!read_previous)
				break;
		}
	}
	fz_always(ctx)
	{
		fz_free(ctx, list.list);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}
}

/*
 * load xref tables from pdf
 *
 * File locked on entry, throughout and on exit.
 */

static void
pdf_load_xref(pdf_document *doc, pdf_lexbuf *buf)
{
	int i;
	int xref_len;
	fz_context *ctx = doc->ctx;

	pdf_read_start_xref(doc);

	pdf_read_xref_sections(doc, doc->startxref, buf, 1);

	if (pdf_xref_len(doc) == 0)
		fz_throw(ctx, FZ_ERROR_GENERIC, "found xref was empty");

	/* broken pdfs where first object is not free */
	if (pdf_get_xref_entry(doc, 0)->type != 'f')
		fz_warn(ctx->warn, "first object in xref is not free");

	/* broken pdfs where object offsets are out of range */
	xref_len = pdf_xref_len(doc);
	for (i = 0; i < xref_len; i++)
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, i);
		if (entry->type == 'n')
		{
			/* Special case code: "0000000000 * n" means free,
			 * according to some producers (inc Quartz) */
			if (entry->ofs == 0)
				entry->type = 'f';
			else if (entry->ofs <= 0 || entry->ofs >= doc->file_size)
				fz_throw(ctx, FZ_ERROR_GENERIC, "object offset out of range: %d (%d 0 R)", entry->ofs, i);
		}
		if (entry->type == 'o')
			if (entry->ofs <= 0 || entry->ofs >= xref_len || pdf_get_xref_entry(doc, entry->ofs)->type != 'n')
				fz_warn(ctx->warn, "invalid reference to an objstm that does not exist: %d (%d 0 R)", entry->ofs, i);
	}
}

int pdf_obj_refs(pdf_obj *ref)
{
	return (ref ? ref->refs : 0);
}

void pdf_mark_xref(pdf_document *doc)
{
	int x, e;

	for (x = 0; x < doc->num_xref_sections; x++)
	{
		pdf_xref *xref = &doc->xref_sections[x];

		for (e = 0; e < xref->len; e++)
		{
			pdf_xref_entry *entry = &xref->table[e];

			if (entry->obj)
			{
				entry->flags |= PDF_OBJ_FLAG_MARK;
			}
		}
	}
}

void pdf_clear_xref(pdf_document *doc)
{
	int x, e;

	for (x = 0; x < doc->num_xref_sections; x++)
	{
		pdf_xref *xref = &doc->xref_sections[x];

		for (e = 0; e < xref->len; e++)
		{
			pdf_xref_entry *entry = &xref->table[e];

			if (entry->obj)
			{
				if (pdf_obj_refs(entry->obj) == 1)
				{
					pdf_drop_obj(entry->obj);
					entry->obj = NULL;
				}
			}
		}
	}
}

void pdf_clear_xref_to_mark(pdf_document *doc)
{
	int x, e;

	for (x = 0; x < doc->num_xref_sections; x++)
	{
		pdf_xref *xref = &doc->xref_sections[x];

		for (e = 0; e < xref->len; e++)
		{
			pdf_xref_entry *entry = &xref->table[e];

			if (entry->obj)
			{
				if ((entry->flags & PDF_OBJ_FLAG_MARK) == 0 && pdf_obj_refs(entry->obj) == 1)
				{
					pdf_drop_obj(entry->obj);
					entry->obj = NULL;
				}
			}
		}
	}
}

/* Scan file for objects and reconstruct xref table */

/* Define in PDF 1.7 to be 8388607, but mupdf is more lenient. */
#define MAX_OBJECT_NUMBER (10 << 20)

struct entry
{
	int num;
	int gen;
	int ofs;
	int stm_ofs;
	int stm_len;
};

int
pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int *stmofsp, int *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, int *tmpofs)
{
	pdf_token tok;
	int stm_len;
	int n;
	fz_stream *file = doc->file;

	*stmofsp = 0;
	if (stmlenp)
		*stmlenp = -1;

	stm_len = 0;

	/* On entry to this function, we know that we've just seen
	 * '<int> <int> obj'. We expect the next thing we see to be a
	 * pdf object. Regardless of the type of thing we meet next
	 * we only need to fully parse it if it is a dictionary. */
	tok = pdf_lex(ctx, file, buf);

	if (tok == PDF_TOK_OPEN_DICT)
	{
		pdf_obj *dict, *obj;

		/* Send NULL xref so we don't try to resolve references */
		fz_try(ctx)
		{
			dict = pdf_parse_dict(doc, file, buf);
		}
		fz_catch(ctx)
		{
			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
			/* Don't let a broken object at EOF overwrite a good one */
			if (file->eof)
				fz_rethrow_message(ctx, "broken object at EOF ignored");
			/* Silently swallow the error */
			dict = pdf_new_dict(doc, 2);
		}

		if (encrypt && id)
		{
			obj = pdf_dict_gets(dict, "Type");
			if (pdf_is_name(obj) && !strcmp(pdf_to_name(obj), "XRef"))
			{
				obj = pdf_dict_gets(dict, "Encrypt");
				if (obj)
				{
					pdf_drop_obj(*encrypt);
					*encrypt = pdf_keep_obj(obj);
				}

				obj = pdf_dict_gets(dict, "ID");
				if (obj)
				{
					pdf_drop_obj(*id);
					*id = pdf_keep_obj(obj);
				}
			}
		}

		obj = pdf_dict_gets(dict, "Length");
		if (!pdf_is_indirect(obj) && pdf_is_int(obj))
			stm_len = pdf_to_int(obj);

		if (doc->file_reading_linearly && page)
		{
			obj = pdf_dict_gets(dict, "Type");
			if (!strcmp(pdf_to_name(obj), "Page"))
			{
				pdf_drop_obj(*page);
				*page = pdf_keep_obj(dict);
			}
		}

		pdf_drop_obj(dict);
	}

	while ( tok != PDF_TOK_STREAM &&
		tok != PDF_TOK_ENDOBJ &&
		tok != PDF_TOK_ERROR &&
		tok != PDF_TOK_EOF &&
		tok != PDF_TOK_INT )
	{
		*tmpofs = fz_tell(file);
		if (*tmpofs < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
		tok = pdf_lex(ctx, file, buf);
	}

	if (tok == PDF_TOK_STREAM)
	{
		int c = fz_read_byte(file);
		if (c == '\r') {
			c = fz_peek_byte(file);
			if (c == '\n')
				fz_read_byte(file);
		}

		*stmofsp = fz_tell(file);
		if (*stmofsp < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot seek in file");

		if (stm_len > 0)
		{
			fz_seek(file, *stmofsp + stm_len, 0);
			tok = pdf_lex(ctx, file, buf);
			if (tok == PDF_TOK_ENDSTREAM) {
				goto atobjend;
			} else {
				fz_warn(ctx->warn, "cannot find endstream token, falling back to scanning");
			}
			fz_seek(file, *stmofsp, 0);
		}

		n = fz_read(file, (unsigned char *) buf->scratch, 9);
		if (n < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot read from file");

		while (memcmp(buf->scratch, "endstream", 9) != 0)
		{
			c = fz_read_byte(file);
			if (c == EOF)
				break;
			memmove(&buf->scratch[0], &buf->scratch[1], 8);
			buf->scratch[8] = c;
		}

		if (stmlenp)
			*stmlenp = fz_tell(file) - *stmofsp - 9;

atobjend:
		*tmpofs = fz_tell(file);
		if (*tmpofs < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
		tok = pdf_lex(ctx, file, buf);
		if (tok != PDF_TOK_ENDOBJ)
			fz_warn(ctx->warn, "object missing 'endobj' token");
		else
		{
			/* Read another token as we always return the next one */
			*tmpofs = fz_tell(file);
			if (*tmpofs < 0)
				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");
			tok = pdf_lex(ctx, file, buf);
		}
	}
	return tok;
}

void
pdf_repair_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf)
{
	pdf_obj *dict, *obj = NULL;
	pdf_obj *length;

	pdf_obj *encrypt = NULL;
	pdf_obj *id = NULL;
	pdf_obj *root = NULL;
	pdf_obj *info = NULL;

	struct entry *list = NULL;
	int listlen;
	int listcap;
	int maxnum = 0;

	int num = 0;
	int gen = 0;
	int tmpofs, numofs = 0, genofs = 0;
	int stm_len, stm_ofs = 0;
	pdf_token tok;
	int next;
	int i, n, c;

	doc->dirty = 1;
	/* Can't support incremental update after repair */
	doc->freeze_updates = 1;

	fz_seek(doc->file, 0, 0);

	fz_try(ctx)
	{
		pdf_xref_entry *entry;
		listlen = 0;
		listcap = 1024;
		list = fz_malloc_array(ctx, listcap, sizeof(struct entry));

		/* look for '%PDF' version marker within first kilobyte of file */
		n = fz_read(doc->file, (unsigned char *)buf->scratch, fz_mini(buf->size, 1024));
		if (n < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot read from file");

		fz_seek(doc->file, 0, 0);
		for (i = 0; i < n - 4; i++)
		{
			if (memcmp(&buf->scratch[i], "%PDF", 4) == 0)
			{
				fz_seek(doc->file, i + 8, 0); /* skip "%PDF-X.Y" */
				break;
			}
		}

		/* skip comment line after version marker since some generators
		 * forget to terminate the comment with a newline */
		c = fz_read_byte(doc->file);
		while (c >= 0 && (c == ' ' || c == '%'))
			c = fz_read_byte(doc->file);
		fz_unread_byte(doc->file);

		while (1)
		{
			tmpofs = fz_tell(doc->file);
			if (tmpofs < 0)
				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot tell in file");

			tok = pdf_lex(ctx, doc->file, buf);

			/* If we have the next token already, then we'll jump
			 * back here, rather than going through the top of
			 * the loop. */
		have_next_token:

			if (tok == PDF_TOK_INT)
			{
				numofs = genofs;
				num = gen;
				genofs = tmpofs;
				gen = buf->i;
			}

			else if (tok == PDF_TOK_OBJ)
			{
				fz_try(ctx)
				{
					tok = pdf_repair_obj(ctx, doc, buf, &stm_ofs, &stm_len, &encrypt, &id, NULL, &tmpofs);
				}
				fz_catch(ctx)
				{
					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
					/* If we haven't seen a root yet, there is nothing
					 * we can do, but give up. Otherwise, we'll make
					 * do. */
					if (!root)
						fz_rethrow(ctx);
					fz_warn(ctx->warn, "cannot parse object (%d %d R) - ignoring rest of file", num, gen);
					break;
				}

				if (num <= 0 || num > MAX_OBJECT_NUMBER)
				{
					fz_warn(ctx->warn, "ignoring object with invalid object number (%d %d R)", num, gen);
					goto have_next_token;
				}

				gen = fz_clampi(gen, 0, 65535);

				if (listlen + 1 == listcap)
				{
					listcap = (listcap * 3) / 2;
					list = fz_resize_array(ctx, list, listcap, sizeof(struct entry));
				}

				list[listlen].num = num;
				list[listlen].gen = gen;
				list[listlen].ofs = numofs;
				list[listlen].stm_ofs = stm_ofs;
				list[listlen].stm_len = stm_len;
				listlen ++;

				if (num > maxnum)
					maxnum = num;

				goto have_next_token;
			}

			/* trailer dictionary */
			else if (tok == PDF_TOK_OPEN_DICT)
			{
				fz_try(ctx)
				{
					dict = pdf_parse_dict(doc, doc->file, buf);
				}
				fz_catch(ctx)
				{
					/* If we haven't seen a root yet, there is nothing
					 * we can do, but give up. Otherwise, we'll make
					 * do. */
					if (!root)
						fz_rethrow(ctx);
					fz_warn(ctx->warn, "cannot parse trailer dictionary - ignoring rest of file");
					break;
				}

				obj = pdf_dict_gets(dict, "Encrypt");
				if (obj)
				{
					pdf_drop_obj(encrypt);
					encrypt = pdf_keep_obj(obj);
				}

				obj = pdf_dict_gets(dict, "ID");
				if (obj)
				{
					pdf_drop_obj(id);
					id = pdf_keep_obj(obj);
				}

				obj = pdf_dict_gets(dict, "Root");
				if (obj)
				{
					pdf_drop_obj(root);
					root = pdf_keep_obj(obj);
				}

				obj = pdf_dict_gets(dict, "Info");
				if (obj)
				{
					pdf_drop_obj(info);
					info = pdf_keep_obj(obj);
				}

				pdf_drop_obj(dict);
				obj = NULL;
			}

			else if (tok == PDF_TOK_ERROR)
				fz_read_byte(doc->file);

			else if (tok == PDF_TOK_EOF)
				break;
		}

		/* make xref reasonable */

		/*
			Dummy access to entry to assure sufficient space in the xref table
			and avoid repeated reallocs in the loop
		*/
		(void)pdf_get_populating_xref_entry(doc, maxnum);

		for (i = 0; i < listlen; i++)
		{
			entry = pdf_get_populating_xref_entry(doc, list[i].num);
			entry->type = 'n';
			entry->ofs = list[i].ofs;
			entry->gen = list[i].gen;

			entry->stm_ofs = list[i].stm_ofs;

			/* correct stream length for unencrypted documents */
			if (!encrypt && list[i].stm_len >= 0)
			{
				dict = pdf_load_object(doc, list[i].num, list[i].gen);

				length = pdf_new_int(doc, list[i].stm_len);
				pdf_dict_puts(dict, "Length", length);
				pdf_drop_obj(length);

				pdf_drop_obj(dict);
			}
		}

		entry = pdf_get_populating_xref_entry(doc, 0);
		entry->type = 'f';
		entry->ofs = 0;
		entry->gen = 65535;
		entry->stm_ofs = 0;
		entry->obj = NULL;

		next = 0;
		for (i = pdf_xref_len(doc) - 1; i >= 0; i--)
		{
			entry = pdf_get_populating_xref_entry(doc, i);
			if (entry->type == 'f')
			{
				entry->ofs = next;
				if (entry->gen < 65535)
					entry->gen ++;
				next = i;
			}
		}

		/* create a repaired trailer, Root will be added later */

		obj = pdf_new_dict(doc, 5);
		/* During repair there is only a single xref section */
		pdf_set_populating_xref_trailer(doc, obj);
		pdf_drop_obj(obj);
		obj = NULL;

		obj = pdf_new_int(doc, maxnum + 1);
		pdf_dict_puts(pdf_trailer(doc), "Size", obj);
		pdf_drop_obj(obj);
		obj = NULL;

		if (root)
		{
			pdf_dict_puts(pdf_trailer(doc), "Root", root);
			pdf_drop_obj(root);
			root = NULL;
		}
		if (info)
		{
			pdf_dict_puts(pdf_trailer(doc), "Info", info);
			pdf_drop_obj(info);
			info = NULL;
		}

		if (encrypt)
		{
			if (pdf_is_indirect(encrypt))
			{
				/* create new reference with non-NULL xref pointer */
				obj = pdf_new_indirect(doc, pdf_to_num(encrypt), pdf_to_gen(encrypt));
				pdf_drop_obj(encrypt);
				encrypt = obj;
				obj = NULL;
			}
			pdf_dict_puts(pdf_trailer(doc), "Encrypt", encrypt);
			pdf_drop_obj(encrypt);
			encrypt = NULL;
		}

		if (id)
		{
			if (pdf_is_indirect(id))
			{
				/* create new reference with non-NULL xref pointer */
				obj = pdf_new_indirect(doc, pdf_to_num(id), pdf_to_gen(id));
				pdf_drop_obj(id);
				id = obj;
				obj = NULL;
			}
			pdf_dict_puts(pdf_trailer(doc), "ID", id);
			pdf_drop_obj(id);
			id = NULL;
		}

		fz_free(ctx, list);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(encrypt);
		pdf_drop_obj(id);
		pdf_drop_obj(root);
		pdf_drop_obj(obj);
		pdf_drop_obj(info);
		fz_free(ctx, list);
		fz_rethrow(ctx);
	}
}

static void
pdf_repair_obj_stm(pdf_document *doc, int num, int gen)
{
	pdf_obj *obj;
	fz_stream *stm = NULL;
	pdf_token tok;
	int i, n, count;
	fz_context *ctx = doc->ctx;
	pdf_lexbuf buf;

	pdf_lexbuf_init(ctx, &buf, PDF_LEXBUF_SMALL);

	fz_try(ctx)
	{
		obj = pdf_load_object(doc, num, gen);

		count = pdf_to_int(pdf_dict_gets(obj, "N"));

		pdf_drop_obj(obj);

		stm = pdf_open_stream(doc, num, gen);

		for (i = 0; i < count; i++)
		{
			pdf_xref_entry *entry;

			tok = pdf_lex(ctx, stm, &buf);
			if (tok != PDF_TOK_INT)
				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d %d R)", num, gen);

			n = buf.i;
			if (n < 0)
			{
				fz_warn(ctx->warn, "ignoring object with invalid object number (%d %d R)", n, i);
				continue;
			}
			else if (n > MAX_OBJECT_NUMBER)
			{
				fz_warn(ctx->warn, "ignoring object with invalid object number (%d %d R)", n, i);
				continue;
			}

			entry = pdf_get_populating_xref_entry(doc, n);
			entry->ofs = num;
			entry->gen = i;
			entry->stm_ofs = 0;
			pdf_drop_obj(entry->obj);
			entry->obj = NULL;
			entry->type = 'o';

			tok = pdf_lex(ctx, stm, &buf);
			if (tok != PDF_TOK_INT)
				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d %d R)", num, gen);
		}
	}
	fz_always(ctx)
	{
		fz_close(stm);
		pdf_lexbuf_fin(ctx, &buf);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot load object stream object (%d %d R)", num, gen);
	}
}

void
pdf_repair_obj_stms(fz_context *ctx, pdf_document *doc)
{
	pdf_obj *dict;
	int i;
	int xref_len = pdf_xref_len(doc);

	for (i = 0; i < xref_len; i++) {
		pdf_xref_entry *entry = pdf_get_populating_xref_entry(doc, i);

		if (entry->stm_ofs) {
			dict = pdf_load_object(doc, i, 0);
			fz_try(ctx)
			{
				if (!strcmp(pdf_to_name(pdf_dict_gets(dict, "Type")), "ObjStm"))
					pdf_repair_obj_stm(doc, i, 0);
			}
 			fz_catch(ctx)
 			{
				fz_warn(ctx->warn, "ignoring broken object stream (%d 0 R)", i);
 			}
			pdf_drop_obj(dict);
		}
	}

	/* Ensure that streamed objects reside inside a known non-streamed object */
	for (i = 0; i < xref_len; i++) {
		pdf_xref_entry *entry = pdf_get_populating_xref_entry(doc, i);
		if (entry->type == 'o' && pdf_get_populating_xref_entry(doc, entry->ofs)->type != 'n')
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "invalid reference to non-object-stream: %d (%d 0 R)", entry->ofs, i);
	}
}

static void
pdf_load_linear(pdf_document *doc)
{
	pdf_obj *dict = NULL;
	pdf_obj *hint = NULL;
	pdf_obj *o;
	int num, gen, stmofs, lin, len;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		pdf_xref_entry *entry;

		dict = pdf_parse_ind_obj(doc, doc->file, &doc->lexbuf.base, &num, &gen, &stmofs);
		if (!pdf_is_dict(dict))
			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to read linearized dictionary");
		o = pdf_dict_gets(dict, "Linearized");
		if (o == NULL)
			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to read linearized dictionary");
		lin = pdf_to_int(o);
		if (lin != 1)
			fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected version of Linearized tag (%d)", lin);
		len = pdf_to_int(pdf_dict_gets(dict, "L"));
		if (len != doc->file_length)
			fz_throw(ctx, FZ_ERROR_GENERIC, "File has been updated since linearization");

		pdf_read_xref_sections(doc, fz_tell(doc->file), &doc->lexbuf.base, 0);

		doc->page_count = pdf_to_int(pdf_dict_gets(dict, "N"));
		doc->linear_page_refs = fz_resize_array(ctx, doc->linear_page_refs, doc->page_count, sizeof(pdf_obj *));
		memset(doc->linear_page_refs, 0, doc->page_count * sizeof(pdf_obj*));
		doc->linear_obj = dict;
		doc->linear_pos = fz_tell(doc->file);
		doc->linear_page1_obj_num = pdf_to_int(pdf_dict_gets(dict, "O"));
		doc->linear_page_refs[0] = pdf_new_indirect(doc, doc->linear_page1_obj_num, 0);
		doc->linear_page_num = 0;
		hint = pdf_dict_gets(dict, "H");
		doc->hint_object_offset = pdf_to_int(pdf_array_get(hint, 0));
		doc->hint_object_length = pdf_to_int(pdf_array_get(hint, 1));

		entry = pdf_get_populating_xref_entry(doc, 0);
		entry->type = 'f';
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(dict);
		/* Drop back to non linearized reading mode */
		doc->file_reading_linearly = 0;
	}
}

void
pdf_ocg_set_config(pdf_document *doc, int config)
{
	int i, j, len, len2;
	pdf_ocg_descriptor *desc = doc->ocg;
	pdf_obj *obj, *cobj;
	char *name;

	obj = pdf_dict_gets(pdf_dict_gets(pdf_trailer(doc), "Root"), "OCProperties");
	if (!obj)
	{
		if (config == 0)
			return;
		else
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "Unknown OCG config (None known!)");
	}
	if (config == 0)
	{
		cobj = pdf_dict_gets(obj, "D");
		if (!cobj)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "No default OCG config");
	}
	else
	{
		cobj = pdf_array_get(pdf_dict_gets(obj, "Configs"), config);
		if (!cobj)
			fz_throw(doc->ctx, FZ_ERROR_GENERIC, "Illegal OCG config");
	}

	pdf_drop_obj(desc->intent);
	desc->intent = pdf_dict_gets(cobj, "Intent");
	if (desc->intent)
		pdf_keep_obj(desc->intent);

	len = desc->len;
	name = pdf_to_name(pdf_dict_gets(cobj, "BaseState"));
	if (strcmp(name, "Unchanged") == 0)
	{
		/* Do nothing */
	}
	else if (strcmp(name, "OFF") == 0)
	{
		for (i = 0; i < len; i++)
		{
			desc->ocgs[i].state = 0;
		}
	}
	else /* Default to ON */
	{
		for (i = 0; i < len; i++)
		{
			desc->ocgs[i].state = 1;
		}
	}

	obj = pdf_dict_gets(cobj, "ON");
	len2 = pdf_array_len(obj);
	for (i = 0; i < len2; i++)
	{
		pdf_obj *o = pdf_array_get(obj, i);
		int n = pdf_to_num(o);
		int g = pdf_to_gen(o);
		for (j=0; j < len; j++)
		{
			if (desc->ocgs[j].num == n && desc->ocgs[j].gen == g)
			{
				desc->ocgs[j].state = 1;
				break;
			}
		}
	}

	obj = pdf_dict_gets(cobj, "OFF");
	len2 = pdf_array_len(obj);
	for (i = 0; i < len2; i++)
	{
		pdf_obj *o = pdf_array_get(obj, i);
		int n = pdf_to_num(o);
		int g = pdf_to_gen(o);
		for (j=0; j < len; j++)
		{
			if (desc->ocgs[j].num == n && desc->ocgs[j].gen == g)
			{
				desc->ocgs[j].state = 0;
				break;
			}
		}
	}

	/* FIXME: Should make 'num configs' available in the descriptor. */
	/* FIXME: Should copy out 'Intent' here into the descriptor, and remove
	 * csi->intent in favour of that. */
	/* FIXME: Should copy 'AS' into the descriptor, and visibility
	 * decisions should respect it. */
	/* FIXME: Make 'Order' available via the descriptor (when we have an
	 * app that needs it) */
	/* FIXME: Make 'ListMode' available via the descriptor (when we have
	 * an app that needs it) */
	/* FIXME: Make 'RBGroups' available via the descriptor (when we have
	 * an app that needs it) */
	/* FIXME: Make 'Locked' available via the descriptor (when we have
	 * an app that needs it) */
}

static void
pdf_read_ocg(pdf_document *doc)
{
	pdf_obj *obj, *ocg;
	int len, i;
	pdf_ocg_descriptor *desc;
	fz_context *ctx = doc->ctx;

	obj = pdf_dict_gets(pdf_dict_gets(pdf_trailer(doc), "Root"), "OCProperties");
	if (!obj)
		return;
	ocg = pdf_dict_gets(obj, "OCGs");
	if (!ocg || !pdf_is_array(ocg))
		/* Not ever supposed to happen, but live with it. */
		return;
	len = pdf_array_len(ocg);
	fz_try(ctx)
	{
		desc = fz_calloc(ctx, 1, sizeof(*desc));
		desc->len = len;
		desc->ocgs = fz_calloc(ctx, len, sizeof(*desc->ocgs));
		desc->intent = NULL;
		for (i=0; i < len; i++)
		{
			pdf_obj *o = pdf_array_get(ocg, i);
			desc->ocgs[i].num = pdf_to_num(o);
			desc->ocgs[i].gen = pdf_to_gen(o);
			desc->ocgs[i].state = 0;
		}
		doc->ocg = desc;
	}
	fz_catch(ctx)
	{
		if (desc)
			fz_free(ctx, desc->ocgs);
		fz_free(ctx, desc);
		fz_rethrow(ctx);
	}

	pdf_ocg_set_config(doc, 0);
}

static void
pdf_free_ocg(fz_context *ctx, pdf_ocg_descriptor *desc)
{
	if (!desc)
		return;

	pdf_drop_obj(desc->intent);
	fz_free(ctx, desc->ocgs);
	fz_free(ctx, desc);
}

/*
 * Initialize and load xref tables.
 * If password is not null, try to decrypt.
 */

static void
pdf_init_document(pdf_document *doc)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *encrypt, *id;
	pdf_obj *dict = NULL;
	pdf_obj *obj;
	pdf_obj *nobj = NULL;
	int i, repaired = 0;

	fz_try(ctx)
	{
		pdf_load_version(doc);

		doc->file_length = fz_stream_meta(doc->file, FZ_STREAM_META_LENGTH, 0, NULL);
		if (doc->file_length < 0)
			doc->file_length = 0;
#if 0
		/* Check to see if we should work in progressive mode */
		if (fz_stream_meta(doc->file, FZ_STREAM_META_PROGRESSIVE, 0, NULL) > 0)
			doc->file_reading_linearly = 1;

		/* Try to load the linearized file if we are in progressive
		 * mode. */
		if (doc->file_reading_linearly)
			pdf_load_linear(doc);

		/* If we aren't in progressive mode (or the linear load failed
		 * and has set us back to non-progressive mode), load normally.
		 */
		if (!doc->file_reading_linearly)
#endif
			pdf_load_xref(doc, &doc->lexbuf.base);
	}
	fz_catch(ctx)
	{
		pdf_free_xref_sections(doc);
		fz_warn(ctx->warn, "trying to repair broken xref");
		repaired = 1;
	}

	fz_try(ctx)
	{
		int hasroot, hasinfo;

		if (repaired)
			pdf_repair_xref(ctx, doc, &doc->lexbuf.base);

		encrypt = pdf_dict_gets(pdf_trailer(doc), "Encrypt");
		id = pdf_dict_gets(pdf_trailer(doc), "ID");
		if (pdf_is_dict(encrypt))
			doc->crypt = pdf_new_crypt(ctx, encrypt, id);

		/* Allow lazy clients to read encrypted files with a blank password */
		pdf_authenticate_password(doc, "");

		if (repaired)
		{
			int xref_len = pdf_xref_len(doc);
			pdf_repair_obj_stms(ctx, doc);

			hasroot = (pdf_dict_gets(pdf_trailer(doc), "Root") != NULL);
			hasinfo = (pdf_dict_gets(pdf_trailer(doc), "Info") != NULL);

			for (i = 1; i < xref_len; i++)
			{
				pdf_xref_entry *entry = pdf_get_xref_entry(doc, i);
				if (entry->type == 0 || entry->type == 'f')
					continue;

				fz_try(ctx)
				{
					dict = pdf_load_object(doc, i, 0);
				}
				fz_catch(ctx)
				{
					fz_warn(ctx->warn, "ignoring broken object (%d 0 R)", i);
					continue;
				}

				if (!hasroot)
				{
					obj = pdf_dict_gets(dict, "Type");
					if (pdf_is_name(obj) && !strcmp(pdf_to_name(obj), "Catalog"))
					{
						nobj = pdf_new_indirect(doc, i, 0);
						pdf_dict_puts(pdf_trailer(doc), "Root", nobj);
						pdf_drop_obj(nobj);
						nobj = NULL;
					}
				}

				if (!hasinfo)
				{
					if (pdf_dict_gets(dict, "Creator") || pdf_dict_gets(dict, "Producer"))
					{
						nobj = pdf_new_indirect(doc, i, 0);
						pdf_dict_puts(pdf_trailer(doc), "Info", nobj);
						pdf_drop_obj(nobj);
						nobj = NULL;
					}
				}

				pdf_drop_obj(dict);
				dict = NULL;
			}

            /* ensure that strings are not used in their repaired, non-decrypted form */
            if (doc->crypt)
                pdf_clear_xref(doc);
		}
		doc->js = NULL;
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(dict);
		pdf_drop_obj(nobj);
		fz_rethrow_message(ctx, "cannot open document");
	}

	fz_try(ctx)
	{
		pdf_read_ocg(doc);
	}
	fz_catch(ctx)
	{
		fz_warn(ctx->warn, "Ignoring Broken Optional Content");
	}
}

void
pdf_close_document(pdf_document *doc)
{
	fz_context *ctx;
	int i;

	if (!doc)
		return;
	ctx = doc->ctx;

	pdf_free_xref_sections(doc);

	if (doc->focus_obj)
		pdf_drop_obj(doc->focus_obj);
	if (doc->file)
		fz_close(doc->file);
	if (doc->crypt)
		pdf_free_crypt(ctx, doc->crypt);

	pdf_drop_obj(doc->linear_obj);
	if (doc->linear_page_refs)
	{
		for (i=0; i < doc->page_count; i++)
		{
			pdf_drop_obj(doc->linear_page_refs[i]);
		}
		fz_free(ctx, doc->linear_page_refs);
	}
	fz_free(ctx, doc->hint_page);
	fz_free(ctx, doc->hint_shared_ref);
	fz_free(ctx, doc->hint_shared);
	fz_free(ctx, doc->hint_obj_offsets);

	pdf_free_ocg(ctx, doc->ocg);

	fz_empty_store(ctx);

	pdf_lexbuf_fin(ctx, &doc->lexbuf.base);

	fz_free(ctx, doc);
}

void
pdf_bufprint_xref(fz_buffer *outbuf, pdf_document *doc)
{
	unsigned int i, xref_len = pdf_xref_len(doc);
	fz_write_buffer(outbuf, "xref\n0 ", 8);
    fz_buffer_itoa10(outbuf, xref_len);
    fz_write_buffer_byte(outbuf, '\n');
	for (i = 0; i < xref_len; i++)
	{
		pdf_xref_entry *entry = pdf_get_xref_entry(doc, i);
		fz_buffer_printf(outbuf, "%05d: %010d %05d %c (stm_ofs=%d; stm_buf=%p)\n", i,
			              entry->ofs, entry->gen, entry->type ? entry->type : '-',
                          entry->stm_ofs, entry->stm_buf);
	}
}

/*
 * compressed object streams
 */

static void
pdf_load_obj_stm(pdf_document *doc, int num, int gen, pdf_lexbuf *buf)
{
	fz_stream *stm = NULL;
	pdf_obj *objstm = NULL;
	int *numbuf = NULL;
	int *ofsbuf = NULL;

	pdf_obj *obj;
	int first;
	int count;
	int i;
	pdf_token tok;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		objstm = pdf_load_object(doc, num, gen);

		count = pdf_to_int(pdf_dict_gets(objstm, "N"));
		first = pdf_to_int(pdf_dict_gets(objstm, "First"));

		if (count < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "negative number of objects in object stream");
		if (first < 0)
			fz_throw(ctx, FZ_ERROR_GENERIC, "first object in object stream resides outside stream");

		numbuf = fz_calloc(ctx, count, sizeof(int));
		ofsbuf = fz_calloc(ctx, count, sizeof(int));

		stm = pdf_open_stream(doc, num, gen);
		for (i = 0; i < count; i++)
		{
			tok = pdf_lex(ctx, stm, buf);
			if (tok != PDF_TOK_INT)
				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d %d R)", num, gen);
			numbuf[i] = buf->i;

			tok = pdf_lex(ctx, stm, buf);
			if (tok != PDF_TOK_INT)
				fz_throw(ctx, FZ_ERROR_GENERIC, "corrupt object stream (%d %d R)", num, gen);
			ofsbuf[i] = buf->i;
		}

		fz_seek(stm, first, SEEK_SET);

		for (i = 0; i < count; i++)
		{
			int xref_len = pdf_xref_len(doc);
			pdf_xref_entry *entry;
			fz_seek(stm, first + ofsbuf[i], SEEK_SET);

			obj = pdf_parse_stm_obj(doc, stm, buf);

			if (numbuf[i] < 1 || numbuf[i] >= xref_len)
			{
				pdf_drop_obj(obj);
				fz_throw(ctx, FZ_ERROR_GENERIC, "object id (%d 0 R) out of range (0..%d)", numbuf[i], xref_len - 1);
			}

			entry = pdf_get_xref_entry(doc, numbuf[i]);

			if (entry->type == 'o' && entry->ofs == num)
			{
				/* If we already have an entry for this object,
				 * we'd like to drop it and use the new one -
				 * but this means that anyone currently holding
				 * a pointer to the old one will be left with a
				 * stale pointer. Instead, we drop the new one
				 * and trust that the old one is correct. */
				if (entry->obj) {
					if (pdf_objcmp(entry->obj, obj))
						fz_warn(ctx->warn, "Encountered new definition for object %d - keeping the original one", numbuf[i]);
					pdf_drop_obj(obj);
				} else
					entry->obj = obj;
			}
			else
			{
				pdf_drop_obj(obj);
			}
		}
	}
	fz_always(ctx)
	{
		fz_close(stm);
		fz_free(ctx, ofsbuf);
		fz_free(ctx, numbuf);
		pdf_drop_obj(objstm);
	}
	fz_catch(ctx)
	{
		fz_rethrow_message(ctx, "cannot open object stream (%d %d R)", num, gen);
	}
}

/*
 * object loading
 */
static int
pdf_obj_read(pdf_document *doc, int *offset, int *nump, pdf_obj **page)
{
	int num, numofs, gen, genofs, stmofs, tmpofs, tok;
	pdf_lexbuf *buf = &doc->lexbuf.base;
	fz_context *ctx = doc->ctx;
	int xref_len;
	pdf_xref_entry *entry;
	int newtmpofs;

	numofs = *offset;
	fz_seek(doc->file, numofs, SEEK_SET);

	/* We expect to read 'num' here */
	tok = pdf_lex(ctx, doc->file, buf);
	genofs = fz_tell(doc->file);
	if (tok != PDF_TOK_INT)
	{
		/* Failed! */
		DEBUGMESS((ctx, "skipping unexpected data (tok=%d) at %d", tok, *offset));
		*offset = genofs;
		return tok == PDF_TOK_EOF;
	}
	*nump = num = buf->i;

	/* We expect to read 'gen' here */
	tok = pdf_lex(ctx, doc->file, buf);
	tmpofs = fz_tell(doc->file);
	if (tok != PDF_TOK_INT)
	{
		/* Failed! */
		DEBUGMESS((ctx, "skipping unexpected data after \"%d\" (tok=%d) at %d", num, tok, *offset));
		*offset = tmpofs;
		return tok == PDF_TOK_EOF;
	}
	gen = buf->i;

	/* We expect to read 'obj' here */
	do
	{
		tmpofs = fz_tell(doc->file);
		tok = pdf_lex(ctx, doc->file, buf);
		if (tok == PDF_TOK_OBJ)
			break;
		if (tok != PDF_TOK_INT)
		{
			DEBUGMESS((ctx, "skipping unexpected data (tok=%d) at %d", tok, tmpofs));
			*offset = fz_tell(doc->file);
			return tok == PDF_TOK_EOF;
		}
		DEBUGMESS((ctx, "skipping unexpected int %d at %d", num, numofs));
		*nump = num = gen;
		numofs = genofs;
		gen = buf->i;
		genofs = tmpofs;
	}
	while (1);

	/* Now we read the actual object */
	xref_len = pdf_xref_len(doc);

	/* When we are reading a progressive file, we typically see:
	 *    File Header
	 *    obj m (Linearization params)
	 *    xref #1 (refers to objects m-n)
	 *    obj m+1
	 *    ...
	 *    obj n
	 *    obj 1
	 *    ...
	 *    obj n-1
	 *    xref #2
	 *
	 * The linearisation params are read elsewhere, hence
	 * whenever we read an object it should just go into the
	 * previous xref.
	 */
	tok = pdf_repair_obj(ctx, doc, buf, &stmofs, NULL, NULL, NULL, page, &newtmpofs);

	do /* So we can break out of it */
	{
		if (num <= 0 || num >= xref_len)
		{
			fz_warn(ctx->warn, "Not a valid object number (%d %d obj)", num, gen);
			break;
		}
		if (gen != 0)
		{
			fz_warn(ctx->warn, "Unexpected non zero generation number in linearized file");
		}
		entry = pdf_get_populating_xref_entry(doc, num);
		if (entry->type != 0)
		{
			DEBUGMESS((ctx, "Duplicate object found (%d %d obj)", num, gen));
			break;
		}
		if (page && *page)
		{
			DEBUGMESS((ctx, "Successfully read object %d @ %d - and found page %d!", num, numofs, doc->linear_page_num));
			if (!entry->obj)
				entry->obj = pdf_keep_obj(*page);

			if (doc->linear_page_refs[doc->linear_page_num] == NULL)
				doc->linear_page_refs[doc->linear_page_num] = pdf_new_indirect(doc, num, gen);
		}
		else
		{
			DEBUGMESS((ctx, "Successfully read object %d @ %d", num, numofs));
		}
		entry->type = 'n';
		entry->gen = 0;
		entry->ofs = numofs;
		entry->stm_ofs = stmofs;
	}
	while (0);
	if (page && *page)
		doc->linear_page_num++;

	if (tok == PDF_TOK_ENDOBJ)
	{
		*offset = fz_tell(doc->file);
	}
	else
	{
		*offset = newtmpofs;
	}
	return 0;
}

static int
read_hinted_object(pdf_document *doc, int num)
{
	/* Try to find the object using our hint table. Find the closest
	 * object <= the one we want that has a hint and read forward from
	 * there. */
	fz_context *ctx = doc->ctx;
	int expected = num;
	int curr_pos;
	int start, offset;

	while (doc->hint_obj_offsets[expected] == 0 && expected > 0)
		expected--;
	if (expected != num)
		DEBUGMESS((ctx, "object %d is unhinted, will search forward from %d", expected, num));
	if (expected == 0)	/* No hints found, just bale */
		return 0;

	curr_pos = fz_tell(doc->file);
	offset = doc->hint_obj_offsets[expected];

	fz_try(ctx)
	{
		int found;

		/* Try to read forward from there */
		do
		{
			start = offset;
			DEBUGMESS((ctx, "Searching for object %d @ %d", expected, offset));
			pdf_obj_read(doc, &offset, &found, 0);
			DEBUGMESS((ctx, "Found object %d - next will be @ %d", found, offset));
			if (found <= expected)
			{
				/* We found the right one (or one earlier than
				 * we expected). Update the hints. */
				doc->hint_obj_offsets[expected] = offset;
				doc->hint_obj_offsets[found] = start;
				doc->hint_obj_offsets[found+1] = offset;
				/* Retry with the next one */
				expected = found+1;
			}
			else
			{
				/* We found one later than we expected. */
				doc->hint_obj_offsets[expected] = 0;
				doc->hint_obj_offsets[found] = start;
				doc->hint_obj_offsets[found+1] = offset;
				while (doc->hint_obj_offsets[expected] == 0 && expected > 0)
					expected--;
				if (expected == 0)	/* No hints found, just bale */
					return 0;
			}
		}
		while (found != num);
	}
	fz_always(ctx)
	{
		fz_seek(doc->file, curr_pos, SEEK_SET);
	}
	fz_catch(ctx)
	{
		/* FIXME: Currently we ignore the hint. Perhaps we should
		 * drop back to non-hinted operation here. */
		doc->hint_obj_offsets[expected] = 0;
		fz_rethrow(ctx);
	}
	return 1;
}

/*
 * PDF 1.7 algorithm 3.1 and ExtensionLevel 3 algorithm 3.1a
 *
 * Using the global encryption key that was generated from the
 * password, create a new key that is used to decrypt individual
 * objects and streams. This key is based on the object and
 * generation numbers.
 */

static int
pdf_compute_object_key(pdf_crypt *crypt, pdf_crypt_filter *cf, int num, int gen, unsigned char *key, int max_len)
{
	fz_md5 md5;
	unsigned char message[5];
	int key_len = crypt->length / 8;

	if (key_len > max_len)
		key_len = max_len;

	if (cf->method == PDF_CRYPT_AESV3)
	{
		memcpy(key, crypt->key, key_len);
		return key_len;
	}

	fz_md5_init(&md5);
	fz_md5_update(&md5, crypt->key, key_len);
	message[0] = (num) & 0xFF;
	message[1] = (num >> 8) & 0xFF;
	message[2] = (num >> 16) & 0xFF;
	message[3] = (gen) & 0xFF;
	message[4] = (gen >> 8) & 0xFF;
	fz_md5_update(&md5, message, 5);

	if (cf->method == PDF_CRYPT_AESV2)
		fz_md5_update(&md5, (unsigned char *)"sAlT", 4);

	fz_md5_final(&md5, key);

	if (key_len + 5 > 16)
		return 16;
	return key_len + 5;
}

static void aes_cbc_decrypt(fz_aes *ctx, unsigned char iv[16], unsigned char *output,
				     const unsigned char *input, unsigned int length)
{
    unsigned int i;
	unsigned char temp[16];

	while(length > 0) {
		memcpy(temp, input, 16);
		aes_ecb_decrypt(ctx, input, output);
		for(i = 0; i < 16; i++)
			output[i] = (unsigned char)(output[i] ^ iv[i]);
		memcpy(iv, temp, 16);
		input += 16;
		output += 16;
		length -= 16;
	}
}

/*
 * PDF 1.7 algorithm 3.1 and ExtensionLevel 3 algorithm 3.1a
 *
 * Decrypt all strings in obj modifying the data in-place.
 * Recurse through arrays and dictionaries, but do not follow
 * indirect references.
 */

static void
pdf_crypt_obj_imp(fz_context *ctx, pdf_crypt *crypt, pdf_obj *obj, unsigned char *key, int keylen)
{
	unsigned char *s;
	int i, n;

	if (pdf_is_indirect(obj))
		return;

	if (pdf_is_string(obj))
	{
		s = (unsigned char *)pdf_to_str_buf(obj);
		n = pdf_to_str_len(obj);

		if (crypt->strf.method == PDF_CRYPT_RC4)
		{
			fz_arc4 arc4;
			fz_arc4_init(&arc4, key, keylen);
            for (i = 0; i < n; i++) {
                unsigned char x = fz_arc4_next(&arc4);
                s[i] ^= x;
            }
		}

		if (crypt->strf.method == PDF_CRYPT_AESV2 || crypt->strf.method == PDF_CRYPT_AESV3)
		{
			if (n == 0)
			{
				/* Empty strings are permissible */
			}
			else if (n & 15)
				fz_warn(ctx->warn, "invalid string length %d for aes encryption", n);
			else
			{
				unsigned char iv[16];
				fz_aes aes;
				memcpy(iv, s, 16);
				if (aes_setup(&aes, key, keylen))
					fz_throw(ctx, FZ_ERROR_GENERIC, "AES key init failed (keylen=%d)", keylen * 8);
				aes_cbc_decrypt(&aes, iv, s, s + 16, n - 16);
				/* delete space used for iv and padding bytes at end */
				if (s[n - 17] < 1 || s[n - 17] > 16)
					fz_warn(ctx->warn, "aes padding out of range");
				else
					pdf_set_str_len(obj, n - 16 - s[n - 17]);
			}
		}
	}

	else if (pdf_is_array(obj))
	{
		n = pdf_array_len(obj);
		for (i = 0; i < n; i++)
		{
			pdf_crypt_obj_imp(ctx, crypt, pdf_array_get(obj, i), key, keylen);
		}
	}

	else if (pdf_is_dict(obj))
	{
		n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
		{
			pdf_crypt_obj_imp(ctx, crypt, pdf_dict_get_val(obj, i), key, keylen);
		}
	}
}

void
pdf_crypt_obj(fz_context *ctx, pdf_crypt *crypt, pdf_obj *obj, int num, int gen)
{
	unsigned char key[32];
	unsigned int len = pdf_compute_object_key(crypt, &crypt->strf, num, gen, key, 32);
	pdf_crypt_obj_imp(ctx, crypt, obj, key, len);
}

void
pdf_cache_object(pdf_document *doc, unsigned int num, int gen)
{
	pdf_xref_entry *x;
	int rnum, rgen;
	fz_context *ctx = doc->ctx;

	if (num >= pdf_xref_len(doc))
		fz_throw(ctx, FZ_ERROR_GENERIC, "object out of range (%d %d R); xref size %d", num, gen, pdf_xref_len(doc));

object_updated:
	x = pdf_get_xref_entry(doc, num);

	if (x->obj)
		return;

	if (x->type == 'f')
	{
		x->obj = pdf_new_null(doc);
	}
	else if (x->type == 'n')
	{
		fz_seek(doc->file, x->ofs, SEEK_SET);

		fz_try(ctx)
		{
			x->obj = pdf_parse_ind_obj(doc, doc->file, &doc->lexbuf.base,
					&rnum, &rgen, &x->stm_ofs);
		}
		fz_catch(ctx)
		{
			fz_warn(ctx->warn, "cannot parse object (%d %d R)", num, gen);
            return;
		}

		if (rnum != num)
		{
			pdf_drop_obj(x->obj);
			x->obj = NULL;
			fz_rethrow_message(ctx, "found object (%d %d R) instead of (%d %d R)", rnum, rgen, num, gen);
		}

		if (doc->crypt)
			pdf_crypt_obj(ctx, doc->crypt, x->obj, num, gen);
	}
	else if (x->type == 'o')
	{
		if (!x->obj)
		{
			fz_try(ctx)
			{
				pdf_load_obj_stm(doc, x->ofs, 0, &doc->lexbuf.base);
			}
			fz_catch(ctx)
			{
				fz_rethrow_message(ctx, "cannot load object stream containing object (%d %d R)", num, gen);
			}
			if (!x->obj)
				fz_throw(ctx, FZ_ERROR_GENERIC, "object (%d %d R) was not found in its object stream", num, gen);
		}
	}
	else if (doc->hint_obj_offsets && read_hinted_object(doc, num))
	{
		goto object_updated;
	}
	else if (doc->file_length && doc->linear_pos < doc->file_length)
	{
		fz_throw(ctx, FZ_ERROR_TRYLATER, "cannot find object in xref (%d %d R) - not loaded yet?", num, gen);
	}
	else
	{
		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find object in xref (%d %d R)", num, gen);
	}

	pdf_set_obj_parent(x->obj, num);
}

pdf_obj *
pdf_load_object(pdf_document *doc, int num, int gen)
{
	fz_context *ctx = doc->ctx;
	pdf_xref_entry *entry;

	fz_try(ctx)
	{
		pdf_cache_object(doc, num, gen);
	}
	fz_catch(ctx)
	{
		fz_warn(ctx->warn, "cannot load object (%d %d R) into cache", num, gen);
	}

	entry = pdf_get_xref_entry(doc, num);

	assert(entry->obj);

	return pdf_keep_obj(entry->obj);
}

pdf_obj *
pdf_resolve_indirect(pdf_obj *ref)
{
	int sanity = 10;
	int num;
	int gen;
	fz_context *ctx = NULL; /* Avoid warning for stupid compilers */
	pdf_document *doc;
	pdf_xref_entry *entry;

	while (pdf_is_indirect(ref))
	{
		if (--sanity == 0)
		{
			fz_warn(ctx->warn, "Too many indirections (possible indirection cycle involving %d %d R)", num, gen);
			return NULL;
		}
		doc = pdf_get_indirect_document(ref);
		if (!doc)
			return NULL;
		ctx = doc->ctx;
		num = pdf_to_num(ref);
		gen = pdf_to_gen(ref);
		fz_try(ctx)
		{
			pdf_cache_object(doc, num, gen);
		}
		fz_catch(ctx)
		{
			fz_warn(ctx->warn, "cannot load object (%d %d R) into cache", num, gen);
			return NULL;
		}
		entry = pdf_get_xref_entry(doc, num);
		if (!entry->obj)
			return NULL;
		ref = entry->obj;
	}

	return ref;
}

int
pdf_count_objects(pdf_document *doc)
{
	return pdf_xref_len(doc);
}

int
pdf_create_object(pdf_document *doc)
{
	/* TODO: reuse free object slots by properly linking free object chains in the ofs field */
	pdf_xref_entry *entry;
	int num = pdf_xref_len(doc);
	entry = pdf_get_incremental_xref_entry(doc, num);
	entry->type = 'f';
	entry->ofs = -1;
	entry->gen = 0;
	entry->stm_ofs = 0;
	entry->stm_buf = NULL;
	entry->obj = NULL;
	return num;
}

void
pdf_delete_object(pdf_document *doc, int num)
{
	pdf_xref_entry *x;

	if (num < 0 || num >= pdf_xref_len(doc))
	{
		fz_warn(doc->ctx->warn, "object out of range (%d 0 R); xref size %d", num, pdf_xref_len(doc));
		return;
	}

	x = pdf_get_incremental_xref_entry(doc, num);

	fz_drop_buffer(x->stm_buf);
	pdf_drop_obj(x->obj);

	x->type = 'f';
	x->ofs = 0;
	x->gen = 0;
	x->stm_ofs = 0;
	x->stm_buf = NULL;
	x->obj = NULL;
}

void
pdf_update_object(pdf_document *doc, int num, pdf_obj *newobj)
{
	pdf_xref_entry *x;

	if (num < 0 || num >= pdf_xref_len(doc))
	{
		fz_warn(doc->ctx->warn, "object out of range (%d 0 R); xref size %d", num, pdf_xref_len(doc));
		return;
	}

	x = pdf_get_incremental_xref_entry(doc, num);

	pdf_drop_obj(x->obj);

	x->type = 'n';
	x->ofs = 0;
	x->obj = pdf_keep_obj(newobj);
}

void
pdf_update_stream(pdf_document *doc, pdf_obj *ref, fz_buffer *newbuf, int compressed)
{
	pdf_xref_entry *x;
    unsigned int num = pdf_to_num(ref);

	if (num >= pdf_xref_len(doc)) {
		fz_warn(doc->ctx->warn, "object out of range (%d 0 R); xref size %d", num, pdf_xref_len(doc));
		return;
	}

	x = pdf_get_xref_entry(doc, num);

	fz_drop_buffer(x->stm_buf);
	x->stm_buf = fz_keep_buffer(newbuf);

    pdf_dict_puts_drop(ref, "Length", pdf_new_int(doc, newbuf->len));
    if (!compressed) {
        pdf_dict_dels(ref, "Filter");
        pdf_dict_dels(ref, "DecodeParms");
    }
}

int
pdf_meta(pdf_document *doc, int key, void *ptr, int size)
{
	switch (key)
	{
	/*
		ptr: Pointer to block (uninitialised on entry)
		size: Size of block (at least 64 bytes)
		Returns: Document format as a brief text string.
	*/
	case FZ_META_FORMAT_INFO:
		fz_snprintf((char *)ptr, 63, "PDF %d.%d", doc->version/10, doc->version % 10);
		return FZ_META_OK;
	case FZ_META_CRYPT_INFO:
		if (doc->crypt)
			fz_snprintf((char *)ptr, 63, "Standard V%d R%d %d-bit %s",
                    doc->crypt->v, doc->crypt->r, doc->crypt->length,
                    pdf_crypt_method(doc));
		else
			fz_strlcpy((char *)ptr, "None", 63);
		return FZ_META_OK;
	case FZ_META_HAS_PERMISSION:
	{
		int i;
		switch (size)
		{
		case FZ_PERMISSION_PRINT:
			i = PDF_PERM_PRINT;
			break;
		case FZ_PERMISSION_CHANGE:
			i = PDF_PERM_CHANGE;
			break;
		case FZ_PERMISSION_COPY:
			i = PDF_PERM_COPY;
			break;
		case FZ_PERMISSION_NOTES:
			i = PDF_PERM_NOTES;
			break;
		default:
			return 0;
		}
		return pdf_has_permission(doc, i);
	}
	case FZ_META_INFO:
	{
		pdf_obj *info = pdf_dict_gets(pdf_trailer(doc), "Info");
		if (!info)
		{
			if (ptr)
				*(char *)ptr = 0;
			return 0;
		}
		info = pdf_dict_gets(info, *(char **)ptr);
		if (!info)
		{
			if (ptr)
				*(char *)ptr = 0;
			return 0;
		}
		if (info && ptr && size)
		{
			char *utf8 = pdf_to_utf8(doc, info);
			fz_strlcpy(ptr, utf8, size);
			fz_free(doc->ctx, utf8);
		}
		return 1;
	}
	default:
		return FZ_META_UNKNOWN_KEY;
	}
}

static pdf_obj *
pdf_lookup_page_loc_imp(pdf_document *doc, pdf_obj *node, int *skip, pdf_obj **parentp, int *indexp)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *kids, *hit;
	int i, len;

	kids = pdf_dict_gets(node, "Kids");
	len = pdf_array_len(kids);

	if (pdf_mark_obj(node))
		fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree");

	hit = NULL;

	fz_try(ctx)
	{
		for (i = 0; i < len; i++)
		{
			pdf_obj *kid = pdf_array_get(kids, i);
			char *type = pdf_to_name(pdf_dict_gets(kid, "Type"));
			if (!strcmp(type, "Page"))
			{
				if (*skip == 0)
				{
					if (parentp) *parentp = node;
					if (indexp) *indexp = i;
					hit = kid;
					break;
				}
				else
				{
					(*skip)--;
				}
			}
			else if (!strcmp(type, "Pages"))
			{
				int count = pdf_to_int(pdf_dict_gets(kid, "Count"));
				if (*skip < count)
				{
					hit = pdf_lookup_page_loc_imp(doc, kid, skip, parentp, indexp);
					if (hit)
						break;
				}
				else
				{
					*skip -= count;
				}
			}
			else
			{
				fz_throw(ctx, FZ_ERROR_GENERIC, "non-page object in page tree");
			}
		}
	}
	fz_always(ctx)
	{
		pdf_unmark_obj(node);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	return hit;
}

pdf_obj *
pdf_lookup_page_loc(pdf_document *doc, int needle, pdf_obj **parentp, int *indexp)
{
	pdf_obj *root = pdf_dict_gets(pdf_trailer(doc), "Root");
	pdf_obj *node = pdf_dict_gets(root, "Pages");
	int skip = needle;
	pdf_obj *hit = pdf_lookup_page_loc_imp(doc, node, &skip, parentp, indexp);
	if (!hit)
		fz_throw(doc->ctx, FZ_ERROR_GENERIC, "cannot find page %d in page tree", needle);
	return hit;
}

pdf_obj *
pdf_lookup_page_obj(pdf_document *doc, int needle)
{
	return pdf_lookup_page_loc(doc, needle, NULL, NULL);
}

static int
pdf_count_pages_before_kid(pdf_document *doc, pdf_obj *parent, int kid_num)
{
	pdf_obj *kids = pdf_dict_gets(parent, "Kids");
	int i, total = 0, len = pdf_array_len(kids);
	for (i = 0; i < len; i++)
	{
		pdf_obj *kid = pdf_array_get(kids, i);
		if (pdf_to_num(kid) == kid_num)
			return total;
		if (!strcmp(pdf_to_name(pdf_dict_gets(kid, "Type")), "Pages"))
		{
			pdf_obj *count = pdf_dict_gets(kid, "Count");
			int n = pdf_to_int(count);
			if (count == NULL || n <= 0)
				fz_throw(doc->ctx, FZ_ERROR_GENERIC, "illegal or missing count in pages tree");
			total += n;
		}
		else
			total++;
	}
	fz_throw(doc->ctx, FZ_ERROR_GENERIC, "kid not found in parent's kids array");
}

int
pdf_lookup_page_number(pdf_document *doc, pdf_obj *node)
{
	fz_context *ctx = doc->ctx;
	int needle = pdf_to_num(node);
	int total = 0;
	pdf_obj *parent, *parent2;

	if (strcmp(pdf_to_name(pdf_dict_gets(node, "Type")), "Page") != 0)
		fz_throw(ctx, FZ_ERROR_GENERIC, "invalid page object");

	parent2 = parent = pdf_dict_gets(node, "Parent");
	fz_try(ctx)
	{
		while (pdf_is_dict(parent))
		{
			if (pdf_mark_obj(parent))
				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree (parents)");
			total += pdf_count_pages_before_kid(doc, parent, needle);
			needle = pdf_to_num(parent);
			parent = pdf_dict_gets(parent, "Parent");
		}
	}
	fz_always(ctx)
	{
		/* Run back and unmark */
		while (parent2)
		{
			pdf_unmark_obj(parent2);
			if (parent2 == parent)
				break;
			parent2 = pdf_dict_gets(parent2, "Parent");
		}
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	return total;
}

static pdf_obj *
pdf_lookup_inherited_page_item(pdf_document *doc, pdf_obj *node, const char *key)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *node2 = node;
	pdf_obj *val;

	fz_try(ctx)
	{
		do
		{
			val = pdf_dict_gets(node, key);
			if (val)
				break;
			if (pdf_mark_obj(node))
				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree (parents)");
			node = pdf_dict_gets(node, "Parent");
		}
		while (node);
	}
	fz_always(ctx)
	{
		do
		{
			pdf_unmark_obj(node2);
			if (node2 == node)
				break;
			node2 = pdf_dict_gets(node2, "Parent");
		}
		while (node2);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	return val;
}

/* We need to know whether to install a page-level transparency group */

static int pdf_resources_use_blending(pdf_document *doc, pdf_obj *rdb);

static int
pdf_extgstate_uses_blending(pdf_document *doc, pdf_obj *dict)
{
	pdf_obj *obj = pdf_dict_gets(dict, "BM");
	if (pdf_is_name(obj) && strcmp(pdf_to_name(obj), "Normal"))
		return 1;
	return 0;
}

static int
pdf_pattern_uses_blending(pdf_document *doc, pdf_obj *dict)
{
	pdf_obj *obj;
	obj = pdf_dict_gets(dict, "Resources");
	if (pdf_resources_use_blending(doc, obj))
		return 1;
	obj = pdf_dict_gets(dict, "ExtGState");
	return pdf_extgstate_uses_blending(doc, obj);
}

static int
pdf_xobject_uses_blending(pdf_document *doc, pdf_obj *dict)
{
	pdf_obj *obj = pdf_dict_gets(dict, "Resources");
	return pdf_resources_use_blending(doc, obj);
}

static int
pdf_resources_use_blending(pdf_document *doc, pdf_obj *rdb)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *obj;
	int i, n, useBM = 0;

	if (!rdb)
		return 0;

	/* Have we been here before and remembered an answer? */
	if (pdf_obj_memo(rdb, &useBM))
		return useBM;

	/* stop on cyclic resource dependencies */
	if (pdf_mark_obj(rdb))
		return 0;

	fz_try(ctx)
	{
		obj = pdf_dict_gets(rdb, "ExtGState");
		n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			if (pdf_extgstate_uses_blending(doc, pdf_dict_get_val(obj, i)))
				goto found;

		obj = pdf_dict_gets(rdb, "Pattern");
		n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			if (pdf_pattern_uses_blending(doc, pdf_dict_get_val(obj, i)))
				goto found;

		obj = pdf_dict_gets(rdb, "XObject");
		n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			if (pdf_xobject_uses_blending(doc, pdf_dict_get_val(obj, i)))
				goto found;
		if (0)
		{
found:
			useBM = 1;
		}
	}
	fz_always(ctx)
	{
		pdf_unmark_obj(rdb);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	pdf_set_obj_memo(rdb, useBM);
	return useBM;
}

pdf_page *
pdf_load_page(pdf_document *doc, int number)
{
	fz_context *ctx = doc->ctx;
	pdf_page *page;
	pdf_obj *pageobj, *pageref, *obj;
	fz_rect mediabox, cropbox, realbox;
	float userunit;
	fz_matrix mat;

	pageref = pdf_lookup_page_obj(doc, number);
	pageobj = pdf_resolve_indirect(pageref);

	page = fz_malloc_struct(ctx, pdf_page);
	page->resources = NULL;
	page->contents = NULL;
	page->transparency = 0;
	page->links = NULL;
	page->annots = NULL;
	page->deleted_annots = NULL;
	page->tmp_annots = NULL;
	page->me = pdf_keep_obj(pageobj);
	page->incomplete = 0;

	obj = pdf_dict_gets(pageobj, "UserUnit");
	if (pdf_is_real(obj))
		userunit = pdf_to_real(obj);
	else
		userunit = 1;

	pdf_to_rect(ctx, pdf_lookup_inherited_page_item(doc, pageobj, "MediaBox"), &mediabox);
	if (fz_is_empty_rect(&mediabox))
	{
		fz_warn(ctx->warn, "cannot find page size for page %d", number + 1);
		mediabox.x0 = 0;
		mediabox.y0 = 0;
		mediabox.x1 = 612;
		mediabox.y1 = 792;
	}

	pdf_to_rect(ctx, pdf_lookup_inherited_page_item(doc, pageobj, "CropBox"), &cropbox);
	if (!fz_is_empty_rect(&cropbox))
		fz_intersect_rect(&mediabox, &cropbox);

	page->mediabox.x0 = fz_min(mediabox.x0, mediabox.x1) * userunit;
	page->mediabox.y0 = fz_min(mediabox.y0, mediabox.y1) * userunit;
	page->mediabox.x1 = fz_max(mediabox.x0, mediabox.x1) * userunit;
	page->mediabox.y1 = fz_max(mediabox.y0, mediabox.y1) * userunit;

	if (page->mediabox.x1 - page->mediabox.x0 < 1 || page->mediabox.y1 - page->mediabox.y0 < 1)
	{
		fz_warn(ctx->warn, "invalid page size in page %d", number + 1);
		page->mediabox = fz_unit_rect;
	}

	page->rotate = pdf_to_int(pdf_lookup_inherited_page_item(doc, pageobj, "Rotate"));
	/* Snap page->rotate to 0, 90, 180 or 270 */
	if (page->rotate < 0)
		page->rotate = 360 - ((-page->rotate) % 360);
	if (page->rotate >= 360)
		page->rotate = page->rotate % 360;
	page->rotate = 90*((page->rotate + 45)/90);
	if (page->rotate > 360)
		page->rotate = 0;

	fz_pre_rotate(fz_scale(&page->ctm, 1, -1), -page->rotate);
	realbox = page->mediabox;
	fz_transform_rect(&realbox, &page->ctm);
	fz_pre_scale(fz_translate(&mat, -realbox.x0, -realbox.y0), userunit, userunit);
	fz_concat(&page->ctm, &page->ctm, &mat);

	fz_try(ctx)
	{
		obj = pdf_dict_gets(pageobj, "Annots");
		if (obj)
		{
			page->links = pdf_load_link_annots(doc, obj, &page->ctm);
		}
	}
	fz_catch(ctx)
	{
		page->incomplete |= PDF_PAGE_INCOMPLETE_ANNOTS;
		pdf_drop_link(ctx, page->links);
		page->links = NULL;
	}

	page->duration = pdf_to_real(pdf_dict_gets(pageobj, "Dur"));

	// TODO: inherit
	page->resources = pdf_lookup_inherited_page_item(doc, pageobj, "Resources");
	if (page->resources)
		pdf_keep_obj(page->resources);

	obj = pdf_dict_gets(pageobj, "Contents");
	fz_try(ctx)
	{
		page->contents = pdf_keep_obj(obj);
		if (pdf_resources_use_blending(doc, page->resources))
			page->transparency = 1;
	}
	fz_catch(ctx)
	{
		if (fz_caught(ctx) != FZ_ERROR_TRYLATER)
		{
			pdf_free_page(doc, page);
			fz_rethrow_message(ctx, "cannot load page %d contents (%d 0 R)", number + 1, pdf_to_num(pageref));
		}
		page->incomplete |= PDF_PAGE_INCOMPLETE_CONTENTS;
	}

	return page;
}

void
pdf_free_page(pdf_document *doc, pdf_page *page)
{
	if (page == NULL)
		return;
	pdf_drop_obj(page->resources);
	pdf_drop_obj(page->contents);
	if (page->links)
		pdf_drop_link(doc->ctx, page->links);
	/* doc->focus, when not NULL, refers to one of
	 * the annotations and must be NULLed when the
	 * annotations are destroyed. doc->focus_obj
	 * keeps track of the actual annotation object. */
	doc->focus = NULL;
	pdf_drop_obj(page->me);
	fz_free(doc->ctx, page);
}

void
pdf_delete_page(pdf_document *doc, int at)
{
	pdf_obj *parent, *kids;
	int i;

	pdf_lookup_page_loc(doc, at, &parent, &i);
	kids = pdf_dict_gets(parent, "Kids");
	pdf_array_delete(kids, i);

	while (parent)
	{
		int count = pdf_to_int(pdf_dict_gets(parent, "Name"));
		pdf_dict_puts_drop(parent, "Name", pdf_new_int(doc, count - 1));
		parent = pdf_dict_gets(parent, "Parent");
	}

	doc->page_count = 0; /* invalidate cached value */
}

void
pdf_insert_page(pdf_document *doc, pdf_page *page, int at)
{
	fz_context *ctx = doc->ctx;
	int count = pdf_count_pages(doc);
	pdf_obj *parent, *kids;
	pdf_obj *page_ref;
	int i;

	page_ref = pdf_new_ref(doc, page->me);

	fz_try(ctx)
	{
		if (count == 0)
		{
			pdf_obj *root = pdf_dict_gets(pdf_trailer(doc), "Root");
			parent = pdf_dict_gets(root, "Pages");
			if (!parent)
				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page tree");

			kids = pdf_dict_gets(parent, "Kids");
			if (!kids)
				fz_throw(ctx, FZ_ERROR_GENERIC, "malformed page tree");

			pdf_array_insert(kids, page_ref, 0);
		}
		else if (at >= count)
		{
			if (at == INT_MAX)
				at = count;

			if (at > count)
				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot insert page beyond end of page tree");

			/* append after last page */
			pdf_lookup_page_loc(doc, count - 1, &parent, &i);
			kids = pdf_dict_gets(parent, "Kids");
			pdf_array_insert(kids, page_ref, i + 1);
		}
		else
		{
			/* insert before found page */
			pdf_lookup_page_loc(doc, at, &parent, &i);
			kids = pdf_dict_gets(parent, "Kids");
			pdf_array_insert(kids, page_ref, i);
		}

		pdf_dict_puts(page->me, "Parent", parent);

		/* Adjust page counts */
		while (parent)
		{
			int count = pdf_to_int(pdf_dict_gets(parent, "Count"));
			pdf_dict_puts_drop(parent, "Count", pdf_new_int(doc, count + 1));
			parent = pdf_dict_gets(parent, "Parent");
		}

	}
	fz_always(ctx)
	{
		pdf_drop_obj(page_ref);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	doc->page_count = 0; /* invalidate cached value */
}

void
pdf_delete_page_range(pdf_document *doc, int start, int end)
{
	while (start < end)
		pdf_delete_page(doc, start++);
}

pdf_page *
pdf_create_page(pdf_document *doc, fz_rect mediabox, int res, int rotate)
{
	fz_context *ctx = doc->ctx;
	pdf_page *page = NULL;
	pdf_obj *obj, *pageobj;
	float userunit = 1;
	fz_matrix ctm, tmp;
	fz_rect realbox;

    page = fz_malloc_struct(ctx, pdf_page);

	fz_try(ctx)
	{
        page->resources = NULL;
        page->contents = NULL;
        page->transparency = 0;
        page->links = NULL;
        page->annots = NULL;
		page->me = pageobj = pdf_new_dict(doc, 4);

        pdf_dict_puts_drop(pageobj, "Type", pdf_new_name(doc, "Page"));

		page->mediabox.x0 = fz_min(mediabox.x0, mediabox.x1) * userunit;
		page->mediabox.y0 = fz_min(mediabox.y0, mediabox.y1) * userunit;
		page->mediabox.x1 = fz_max(mediabox.x0, mediabox.x1) * userunit;
		page->mediabox.y1 = fz_max(mediabox.y0, mediabox.y1) * userunit;
		pdf_dict_puts_drop(pageobj, "MediaBox", pdf_new_rect(doc, &page->mediabox));

		/* Snap page->rotate to 0, 90, 180 or 270 */
		if (page->rotate < 0)
			page->rotate = 360 - ((-page->rotate) % 360);
		if (page->rotate >= 360)
			page->rotate = page->rotate % 360;
		page->rotate = 90*((page->rotate + 45)/90);
		if (page->rotate > 360)
			page->rotate = 0;
		pdf_dict_puts_drop(pageobj, "Rotate", pdf_new_int(doc, page->rotate));

		fz_pre_rotate(fz_scale(&ctm, 1, -1), -page->rotate);
		realbox = page->mediabox;
		fz_transform_rect(&realbox, &ctm);
		fz_pre_scale(fz_translate(&tmp, -realbox.x0, -realbox.y0), userunit, userunit);
		fz_concat(&ctm, &ctm, &tmp);
		page->ctm = ctm;
		/* Do not create a Contents, as an empty Contents dict is not valid. See Bug 694712 */
        obj = pdf_new_dict(doc, 4);
        page->contents = pdf_new_ref(doc, obj);
        pdf_drop_obj(obj);
        obj = NULL;
        pdf_dict_puts(pageobj, "Contents", page->contents);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(page->me);
		fz_free(ctx, page);
		fz_rethrow_message(ctx, "Failed to create page");
	}

	return page;
}

static fz_outline *
pdf_load_outline_imp(pdf_document *doc, pdf_obj *dict)
{
	fz_context *ctx = doc->ctx;
	fz_outline *node, **prev, *first;
	pdf_obj *obj;
	pdf_obj *odict = dict;

	fz_try(ctx)
	{
		first = NULL;
		prev = &first;
		while (dict && pdf_is_dict(dict))
		{
			if (pdf_mark_obj(dict))
				break;
			node = fz_malloc_struct(ctx, fz_outline);
			node->title = NULL;
			node->dest.kind = FZ_LINK_NONE;
			node->down = NULL;
			node->next = NULL;
			*prev = node;
			prev = &node->next;

			obj = pdf_dict_gets(dict, "Title");
			if (obj)
				node->title = pdf_to_utf8(doc, obj);

			if ((obj = pdf_dict_gets(dict, "Dest")))
				node->dest = pdf_parse_link_dest(doc, obj);
			else if ((obj = pdf_dict_gets(dict, "A")))
				node->dest = pdf_parse_action(doc, obj);

			obj = pdf_dict_gets(dict, "First");
			if (obj)
				node->down = pdf_load_outline_imp(doc, obj);

			dict = pdf_dict_gets(dict, "Next");
		}
	}
	fz_always(ctx)
	{
		for (dict = odict; dict && pdf_obj_marked(dict); dict = pdf_dict_gets(dict, "Next"))
			pdf_unmark_obj(dict);
	}
	fz_catch(ctx)
	{
		fz_free_outline(ctx, first);
		fz_rethrow(ctx);
	}

	return first;
}

fz_outline *
pdf_load_outline(pdf_document *doc)
{
	pdf_obj *root, *obj, *first;

	root = pdf_dict_gets(pdf_trailer(doc), "Root");
	obj = pdf_dict_gets(root, "Outlines");
	first = pdf_dict_gets(obj, "First");
	if (first)
		return pdf_load_outline_imp(doc, first);

	return NULL;
}

int pdf_count_pages(pdf_document *doc)
{
	if (doc->page_count == 0) {
		pdf_obj *count = pdf_dict_getp(pdf_trailer(doc), "Root/Pages/Count");
		doc->page_count = pdf_to_int(count);
	}
	return doc->page_count;
}

/*
	Initializers for the fz_document interface.

	The functions are split across two files to allow calls to a
	version of the constructor that does not link in the interpreter.
	The interpreter references the built-in font and cmap resources
	which are quite big. Not linking those into the mubusy binary
	saves roughly 6MB of space.
*/

static pdf_document *
pdf_new_document(fz_context *ctx, fz_stream *file)
{
	pdf_document *doc = fz_malloc_struct(ctx, pdf_document);

	doc->super.close = (void*)pdf_close_document;
	doc->super.needs_password = (void*)pdf_needs_password;
	doc->super.authenticate_password = (void*)pdf_authenticate_password;
	doc->super.load_outline = (void*)pdf_load_outline;
	doc->super.count_pages = (void*)pdf_count_pages;
	doc->super.load_page = (void*)pdf_load_page;
	doc->super.load_links = (void*)pdf_load_links;
	doc->super.bound_page = (void*)pdf_bound_page;
	doc->super.run_page_contents = NULL; /* see pdf_xref_aux.c */
	doc->super.free_page = (void*)pdf_free_page;
	doc->super.meta = (void*)pdf_meta;

	pdf_lexbuf_init(ctx, &doc->lexbuf.base, PDF_LEXBUF_LARGE);
	doc->file = fz_keep_stream(file);
	doc->ctx = ctx;

	return doc;
}

pdf_document *
pdf_open_document_no_run_with_stream(fz_context *ctx, fz_stream *file)
{
	pdf_document *doc = pdf_new_document(ctx, file);

	fz_try(ctx)
	{
		pdf_init_document(doc);
	}
	fz_catch(ctx)
	{
		pdf_close_document(doc);
		fz_rethrow_message(ctx, "cannot load document from stream");
	}
	return doc;
}

pdf_document *
pdf_open_document_no_run(fz_context *ctx, unsigned char *data, unsigned int datasize)
{
	fz_stream *file = NULL;
	pdf_document *doc = NULL;
	file = fz_open_memory(ctx, data, datasize);
	doc = pdf_new_document(ctx, file);
	pdf_init_document(doc);
	return doc;
}

pdf_document *pdf_specifics(fz_document *doc)
{
	return (pdf_document *)((doc && doc->close == (void *)pdf_close_document) ? doc : NULL);
}

pdf_document *pdf_create_document(fz_context *ctx)
{
	pdf_document *doc;
	pdf_obj *o = NULL;
	pdf_obj *root;
	pdf_obj *pages;
	pdf_obj *trailer = NULL;

	doc = pdf_new_document(ctx, NULL);
	fz_try(ctx)
	{
		doc->version = 14;
		doc->file_size = 0;
		doc->startxref = 0;
		doc->num_xref_sections = 0;
		pdf_get_populating_xref_entry(doc, 0);
		doc->xref_altered = 1;
		trailer = pdf_new_dict(doc, 2);
		pdf_dict_puts_drop(trailer, "Size", pdf_new_int(doc, 3));
		o = root = pdf_new_dict(doc, 2);
		pdf_dict_puts_drop(trailer, "Root", pdf_new_ref(doc, o));
		pdf_drop_obj(o);
		o = NULL;
		pdf_dict_puts_drop(root, "Type", pdf_new_name(doc, "Catalog"));
		o = pages = pdf_new_dict(doc, 3);
		pdf_dict_puts_drop(root, "Pages", pdf_new_ref(doc, o));
		pdf_drop_obj(o);
		o = NULL;
		pdf_dict_puts_drop(pages, "Type", pdf_new_name(doc, "Pages"));
		pdf_dict_puts_drop(pages, "Count", pdf_new_int(doc, 0));
		pdf_dict_puts_drop(pages, "Kids", pdf_new_array(doc, 1));
		pdf_set_populating_xref_trailer(doc, trailer);
		pdf_drop_obj(trailer);
	}
	fz_catch(ctx)
	{
		pdf_drop_obj(trailer);
		pdf_drop_obj(o);
		fz_rethrow_message(ctx, "Failed to create empty document");
	}
	return doc;
}

/* Pretty printing objects */
struct fmt
{
    fz_buffer *buf;
	int indent;
};

static void fmt_obj(struct fmt *fmt, pdf_obj *obj);
unsigned int fz_ftoa(char *s, float v, int* K);

static inline int isdelim(int ch)
{
	return
		ch == '(' || ch == ')' ||
		ch == '<' || ch == '>' ||
		ch == '[' || ch == ']' ||
		ch == '{' || ch == '}' ||
		ch == '/' ||
		ch == '%';
}

static inline void fmt_indent(struct fmt *fmt)
{
	unsigned int i = fmt->indent;
	while (i--) {
		fz_write_buffer(fmt->buf, "  ", 2);
	}
}

static void fmt_str(struct fmt *fmt, pdf_obj *obj)
{
	char *s = pdf_to_str_buf(obj);
	int n = pdf_to_str_len(obj);
	int i, c;

	fz_write_buffer_byte(fmt->buf, '(');
	for (i = 0; i < n; i++)
	{
		c = (unsigned char)s[i];
        if (strchr("\n\t\r\f\b\\()", c))
	        fz_write_buffer_byte(fmt->buf, '\\');
		if (c == '\n')
	        fz_write_buffer_byte(fmt->buf, 'n');
		else if (c == '\r')
	        fz_write_buffer_byte(fmt->buf, 'r');
		else if (c == '\t')
	        fz_write_buffer_byte(fmt->buf, 't');
		else if (c == '\b')
	        fz_write_buffer_byte(fmt->buf, 'b');
		else if (c == '\f')
	        fz_write_buffer_byte(fmt->buf, 'f');
		else if (c == '(')
	        fz_write_buffer_byte(fmt->buf, '(');
		else if (c == ')')
	        fz_write_buffer_byte(fmt->buf, ')');
		else if (c == '\\')
	        fz_write_buffer_byte(fmt->buf, '\\');
		else if (c < 32) {
		}
		else
	        fz_write_buffer_byte(fmt->buf, c);
	}
	fz_write_buffer(fmt->buf, ")\n", 2);
}

static void fmt_hex(struct fmt *fmt, pdf_obj *obj)
{
    static const unsigned char hex[16] = "0123456789abcdef";
	char *s = pdf_to_str_buf(obj);
	unsigned int i, n = pdf_to_str_len(obj);

    fz_write_buffer_byte(fmt->buf, '<');
	for (i = 0; i < n; i++) {
		unsigned char b = (unsigned char) s[i];
		fz_write_buffer_byte(fmt->buf, hex[b>>4]);
		fz_write_buffer_byte(fmt->buf, hex[b&0x0f]);
	}
    fz_write_buffer(fmt->buf, ">\n", 2);
}

static void fmt_name(struct fmt *fmt, pdf_obj *obj)
{
    static const unsigned char hex[16] = "0123456789ABCDEF";
	unsigned char *s = (unsigned char *) pdf_to_name(obj);
	int i;

	fz_write_buffer_byte(fmt->buf, '/');

	for (i = 0; s[i]; i++)
	{
		if (isdelim(s[i]) || iswhite(s[i]) ||
			s[i] == '#' || s[i] < 32 || s[i] >= 127)
		{
            fz_write_buffer_byte(fmt->buf, '#');
			fz_write_buffer_byte(fmt->buf, hex[s[i] >> 4]);
			fz_write_buffer_byte(fmt->buf, hex[s[i] & 0x0f]);
		}
		else
		{
			fz_write_buffer_byte(fmt->buf, s[i]);
		}
	}
}

static void fmt_array(struct fmt *fmt, pdf_obj *obj)
{
	unsigned int i, n = pdf_array_len(obj);
    fz_write_buffer_byte(fmt->buf, '[');
	for (i = 0; i < n; i++) {
		fmt_obj(fmt, pdf_array_get(obj, i));
        fz_write_buffer_byte(fmt->buf, ' ');
	}
    fz_write_buffer(fmt->buf, "]\n", 2);
}

static void fmt_dict(struct fmt *fmt, pdf_obj *obj)
{
	unsigned int i, n = pdf_dict_len(obj);
    fz_write_buffer(fmt->buf, "<<", 2);
	for (i = 0; i < n; i++) {
		fmt_obj(fmt, pdf_dict_get_key(obj, i));
        fz_write_buffer_byte(fmt->buf, ' ');
		fmt_obj(fmt, pdf_dict_get_val(obj, i));
	}
    fz_write_buffer(fmt->buf, ">>\n", 3);
}

static void fmt_obj(struct fmt *fmt, pdf_obj *obj)
{
	if (!obj)
		fz_write_buffer(fmt->buf, "<NULL>", 6);
	else if (pdf_is_indirect(obj))
	{
		fz_buffer_itoa10(fmt->buf, pdf_to_num(obj)); fz_write_buffer_byte(fmt->buf, ' ');
		fz_buffer_itoa10(fmt->buf, pdf_to_gen(obj)); fz_write_buffer_byte(fmt->buf, ' ');
        fz_write_buffer_byte(fmt->buf, 'R');
	}
	else if (pdf_is_null(obj))
		fz_write_buffer(fmt->buf, "null", 4);
	else if (pdf_is_bool(obj))
		fz_buffer_puts(fmt->buf, pdf_to_bool(obj) ? "true" : "false");
	else if (pdf_is_int(obj))
	{
        unsigned int tmp = pdf_to_int(obj);
        if ((int)tmp < 0) {
            fz_write_buffer_byte(fmt->buf, '-');
            tmp = -(int)tmp;
        }
		fz_buffer_itoa10(fmt->buf, tmp);
	}
	else if (pdf_is_real(obj))
	{
        char tmpbuf[1024];
        unsigned int tmplen = fz_ftoa(tmpbuf, pdf_to_real(obj), NULL);
		fz_write_buffer(fmt->buf, tmpbuf, tmplen);
	}
	else if (pdf_is_string(obj))
	{
		char *str = pdf_to_str_buf(obj);
		int len = pdf_to_str_len(obj);
		int added = 0;
		int i, c;
		for (i = 0; i < len; i++) {
			c = (unsigned char)str[i];
			if (strchr("()\\\n\r\t\b\f", c))
				added ++;
			else if (c < 32)
				added += 3;
		}
		if (added < len)
			fmt_str(fmt, obj);
		else
			fmt_hex(fmt, obj);
	}
	else if (pdf_is_name(obj))
		fmt_name(fmt, obj);
	else if (pdf_is_array(obj))
		fmt_array(fmt, obj);
	else if (pdf_is_dict(obj))
		fmt_dict(fmt, obj);
	else
		fz_buffer_puts(fmt->buf, "<unknown object>");
}

void
pdf_bufprint_obj(fz_buffer *outbuf, pdf_obj *obj)
{
	struct fmt fmt;
	fmt.indent = 0;
	fmt.buf = outbuf;
	fmt_obj(&fmt, obj);
}

void
pdf_bufprint_ref(fz_buffer *outbuf, pdf_obj *ref)
{
	return pdf_bufprint_obj(outbuf, pdf_resolve_indirect(ref));
}

typedef struct pdf_write_options_s
{
    fz_buffer *out;
	int do_expand;
	int do_deflate;
	int do_garbage;
	int *use_list;
	int *ofs_list;
	int *gen_list;
	int continue_on_error;
	int *errors;
	int file_len;
	int page_count;
} pdf_write_options;

void heap_sort(int *list, int n, const int *val, int (*ge)(int, int))
{
	int i, j;

	/* Step 1: Make a heap */
	/* Invariant: Valid heap in [0..i), unsorted elements in [i..n) */
	for (i = 1; i < n; i++) {
		/* Now bubble backwards to maintain heap invariant */
		j = i;
		while (j != 0)
		{
			int tmp;
			int k = (j-1)>>1;
			if (ge(val[list[k]], val[list[j]]))
				break;
			tmp = list[k];
			list[k] = list[j];
			list[j] = tmp;
			j = k;
		}
	}

	/* Step 2: Heap sort */
	/* Invariant: valid heap in [0..i), sorted list in [i..n) */
	/* Initially: i = n */
	for (i = n-1; i > 0; i--) {
		/* Swap the maximum (0th) element from the page_objects into its place
		 * in the sorted list (position i). */
		int tmp = list[0];
		list[0] = list[i];
		list[i] = tmp;
		/* Now, the page_objects is invalid because the 0th element is out
		 * of place. Bubble it until the page_objects is valid. */
		j = 0;
		while (1) {
			/* Children are k and k+1 */
			int k = (j+1)*2-1;
			/* If both children out of the page_objects, we're done */
			if (k > i-1)
				break;
			/* If both are in the page_objects, pick the larger one */
			if (k < i-1 && ge(val[list[k+1]], val[list[k]]))
				k++;
			/* If j is bigger than k (i.e. both of it's children),
			 * we're done */
			if (ge(val[list[j]], val[list[k]]))
				break;
			tmp = list[k];
			list[k] = list[j];
			list[j] = tmp;
			j = k;
		}
	}
}

/*
 * Garbage collect objects not reachable from the trailer.
 */

static pdf_obj *sweepref(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj)
{
	int num = pdf_to_num(obj);
	int gen = pdf_to_gen(obj);
	fz_context *ctx = doc->ctx;

	if (num < 0 || num >= pdf_xref_len(doc))
		return NULL;
	if (opts->use_list[num])
		return NULL;

	opts->use_list[num] = 1;

	/* Bake in /Length in stream objects */
	fz_try(ctx)
	{
		if (pdf_is_stream(doc, num, gen))
		{
			pdf_obj *len = pdf_dict_gets(obj, "Length");
			if (pdf_is_indirect(len))
			{
				opts->use_list[pdf_to_num(len)] = 0;
				len = pdf_resolve_indirect(len);
				pdf_dict_puts(obj, "Length", len);
			}
		}
	}
	fz_catch(ctx)
	{
		/* Leave broken */
	}

	return pdf_resolve_indirect(obj);
}

static void sweepobj(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj)
{
	int i;

	if (pdf_is_indirect(obj))
		obj = sweepref(doc, opts, obj);

	if (pdf_is_dict(obj))
	{
		int n = pdf_dict_len(obj);
		for (i = 0; i < n; i++)
			sweepobj(doc, opts, pdf_dict_get_val(obj, i));
	}

	else if (pdf_is_array(obj))
	{
		int n = pdf_array_len(obj);
		for (i = 0; i < n; i++)
			sweepobj(doc, opts, pdf_array_get(obj, i));
	}
}

/*
 * Make sure we have loaded objects from object streams.
 */

static void preloadobjstms(pdf_document *doc)
{
	unsigned int num, xref_len = pdf_xref_len(doc);

	for (num = 0; num < xref_len; num++) {
		if (pdf_get_xref_entry(doc, num)->type == 'o') {
	        fz_try(doc->ctx) {
			    pdf_obj *obj = pdf_load_object(doc, num, 0);
			    pdf_drop_obj(obj);
            }
	        fz_catch(doc->ctx)
	        {
            }
		}
	}
}

/*
 * Save streams and objects to the output
 */

static fz_buffer *deflatebuf(fz_context *ctx, unsigned char *p, int n)
{
	fz_buffer *buf;
    uLong csize;
    int t;

    buf = fz_new_buffer(compressBound(n));
    csize = buf->cap;
    t = compress(buf->data, &csize, p, n);
    if (t != Z_OK) {
        fz_drop_buffer(buf);
        fz_throw(ctx, FZ_ERROR_GENERIC, "cannot deflate buffer");
    }
    buf->len = csize;
    return buf;
}

static void copystream(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj_orig, int num, int gen)
{
 	fz_buffer *buf, *tmp;
	pdf_obj *obj;

	buf = pdf_load_raw_renumbered_stream(doc, num, gen, num, gen);
	obj = pdf_copy_dict(obj_orig);

    if (opts->do_deflate && !pdf_dict_gets(obj, "Filter")) {
        pdf_dict_puts(obj, "Filter", pdf_new_name(doc, "FlateDecode"));
        tmp = deflatebuf(doc->ctx, buf->data, buf->len);
        fz_drop_buffer(buf);
        buf = tmp;
    }

	fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	pdf_bufprint_obj(opts->out, obj);
	fz_buffer_puts(opts->out, "stream\n");
	fz_write_buffer(opts->out, buf->data, buf->len);
	fz_buffer_puts(opts->out, "endstream\nendobj\n");

	fz_drop_buffer(buf);
	pdf_drop_obj(obj);
}

static void expandstream(pdf_document *doc, pdf_write_options *opts, pdf_obj *obj_orig, int num, int gen)
{
 	fz_buffer *buf, *tmp;
	pdf_obj *newlen;
	pdf_obj *obj;
	fz_context *ctx = doc->ctx;
	int truncated = 0;

	buf = pdf_load_renumbered_stream(doc, num, gen, num, gen, (opts->continue_on_error ? &truncated : NULL));
	if (truncated && opts->errors)
		(*opts->errors)++;

	obj = pdf_copy_dict(obj_orig);
	pdf_dict_dels(obj, "Filter");
	pdf_dict_dels(obj, "DecodeParms");

    if (opts->do_deflate && !pdf_dict_gets(obj, "Filter")) {
        pdf_dict_puts(obj, "Filter", pdf_new_name(doc, "FlateDecode"));
        tmp = deflatebuf(ctx, buf->data, buf->len);
        fz_drop_buffer(buf);
        buf = tmp;
    }

	newlen = pdf_new_int(doc, buf->len);
	pdf_dict_puts(obj, "Length", newlen);
	pdf_drop_obj(newlen);

	fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	pdf_bufprint_obj(opts->out, obj);
	fz_buffer_puts(opts->out, "stream\n");
	fz_write_buffer(opts->out, buf->data, buf->len);
	fz_buffer_puts(opts->out, "endstream\nendobj\n");

	fz_drop_buffer(buf);
	pdf_drop_obj(obj);
}

static int is_image_filter(char *s)
{
	if (!strcmp(s, "CCITTFaxDecode") || !strcmp(s, "CCF") ||
		!strcmp(s, "DCTDecode") || !strcmp(s, "DCT") ||
		!strcmp(s, "RunLengthDecode") || !strcmp(s, "RL") ||
		!strcmp(s, "JBIG2Decode"))
		return 1;
	return 0;
}

static int filter_implies_image(pdf_document *doc, pdf_obj *o)
{
	if (!o)
		return 0;
	if (pdf_is_name(o))
		return is_image_filter(pdf_to_name(o));
	if (pdf_is_array(o))
	{
		int i, len;
		len = pdf_array_len(o);
		for (i = 0; i < len; i++)
			if (is_image_filter(pdf_to_name(pdf_array_get(o, i))))
				return 1;
	}
	return 0;
}

static void writeobject(pdf_document *doc, pdf_write_options *opts, int num, int gen, int skip_xrefs)
{
	pdf_xref_entry *entry;
	pdf_obj *obj;
	pdf_obj *type;
	fz_context *ctx = doc->ctx;

	fz_try(ctx)
	{
		obj = pdf_load_object(doc, num, gen);
	}
	fz_catch(ctx)
	{
		fz_buffer_printf(opts->out, "%d %d obj\nnull\nendobj\n", num, gen);
		if (opts->errors)
			(*opts->errors)++;
		fz_warn(ctx->warn, "caught error: %s", fz_caught_message(ctx));
		//return;
	}

	/* skip ObjStm and XRef objects */
	if (pdf_is_dict(obj))
	{
		type = pdf_dict_gets(obj, "Type");
		if (pdf_is_name(type) && !strcmp(pdf_to_name(type), "ObjStm"))
		{
			opts->use_list[num] = 0;
			pdf_drop_obj(obj);
			return;
		}
		if (skip_xrefs && pdf_is_name(type) && !strcmp(pdf_to_name(type), "XRef"))
		{
			opts->use_list[num] = 0;
			pdf_drop_obj(obj);
			return;
		}
	}

	entry = pdf_get_xref_entry(doc, num);
	if (!pdf_is_stream(doc, num, gen))
	{
		fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	    pdf_bufprint_obj(opts->out, obj);
		fz_buffer_puts(opts->out, "endobj\n");
	}
	else if (entry->stm_ofs < 0 && entry->stm_buf == NULL)
	{
		fz_buffer_printf(opts->out, "%d %d obj\n", num, gen);
	    pdf_bufprint_obj(opts->out, obj);
		fz_buffer_puts(opts->out, "stream\nendstream\nendobj\n");
	}
	else
	{
		int dontexpand = 0;
		if (opts->do_expand != 0 && opts->do_expand != fz_expand_all)
		{
			pdf_obj *o;

			if ((o = pdf_dict_gets(obj, "Type"), !strcmp(pdf_to_name(o), "XObject")) &&
				(o = pdf_dict_gets(obj, "Subtype"), !strcmp(pdf_to_name(o), "Image")))
				dontexpand = !(opts->do_expand & fz_expand_images);
			if (o = pdf_dict_gets(obj, "Filter"), filter_implies_image(doc, o))
				dontexpand = !(opts->do_expand & fz_expand_images);
			if (pdf_dict_gets(obj, "Width") != NULL && pdf_dict_gets(obj, "Height") != NULL)
				dontexpand = !(opts->do_expand & fz_expand_images);
		}
		fz_try(ctx)
		{
			if (opts->do_expand && !dontexpand)
				expandstream(doc, opts, obj, num, gen);
			else
				copystream(doc, opts, obj, num, gen);
		}
		fz_catch(ctx)
		{
			fz_buffer_printf(opts->out, "%d %d obj\nnull\nendobj\n", num, gen);
			if (opts->errors)
				(*opts->errors)++;
			fz_warn(ctx->warn, "caught error: %s", fz_caught_message(ctx));
		}
	}

	pdf_drop_obj(obj);
}

static void writexrefsubsect(fz_context *ctx, pdf_write_options *opts, int from, int to)
{
	int num;

	fz_buffer_printf(opts->out, "%d %d\n", from, to - from);
	for (num = from; num < to; num++)
	{
		if (opts->use_list[num])
			fz_buffer_printf(opts->out, "%010d %05d n \n", opts->ofs_list[num], opts->gen_list[num]);
		else
			fz_buffer_printf(opts->out, "%010d %05d f \n", opts->ofs_list[num], opts->gen_list[num]);
	}
}

static void writexref(pdf_document *doc, pdf_write_options *opts, int to)
{
	fz_context *ctx = doc->ctx;
	pdf_obj *obj, *nobj = NULL, *trailer = NULL;
	fz_md5 md5;

	fz_buffer_puts(opts->out, "xref\n");
	writexrefsubsect(doc->ctx, opts, 0, to);
	fz_write_buffer_byte(opts->out, '\n');

	fz_try(ctx)
	{
		{
            fz_buffer *idbuf = fz_new_buffer(1024);
            unsigned char md5buf[17];
			trailer = pdf_new_dict(doc, 5);

			nobj = pdf_new_int(doc, to);
			pdf_dict_puts(trailer, "Size", nobj);
			pdf_drop_obj(nobj);
			nobj = NULL;

            fz_buffer_itoa10(idbuf, time(NULL));
            fz_buffer_itoa10(idbuf, to);
			obj = pdf_dict_gets(pdf_trailer(doc), "Info");
	        if (obj)
                pdf_bufprint_obj(idbuf, pdf_resolve_indirect(obj));

			fz_md5_init(&md5);
			fz_md5_update(&md5, idbuf->data, idbuf->len);
			fz_md5_final(&md5, md5buf);
            fz_md5_stringify(idbuf->data, md5buf);
            idbuf->len = 32;

			{
				obj = pdf_dict_gets(pdf_trailer(doc), "Info");
				if (obj)
					pdf_dict_puts(trailer, "Info", obj);

				obj = pdf_dict_gets(pdf_trailer(doc), "Root");
				if (obj)
					pdf_dict_puts(trailer, "Root", obj);

				obj = pdf_dict_gets(pdf_trailer(doc), "ID");
				if (obj) {
                    pdf_obj *obj2 = pdf_new_string(doc, idbuf->data, idbuf->len);
                    pdf_array_insert(obj, obj2, 1);
					pdf_dict_puts(trailer, "ID", obj);
                    pdf_drop_obj(obj2);
                }
			}
            fz_drop_buffer(idbuf);
		}
	}
	fz_always(ctx)
	{
		pdf_drop_obj(nobj);
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}

	fz_buffer_printf(opts->out, "trailer\n");
	pdf_bufprint_obj(opts->out, trailer);
    fz_write_buffer_byte(opts->out, '\n');

	pdf_drop_obj(trailer);

    write(1, opts->out->data, opts->out->len);

	//doc->has_xref_streams = 0;
}

static void
dowriteobject(pdf_document *doc, pdf_write_options *opts, int num, int pass)
{
	pdf_xref_entry *entry = pdf_get_xref_entry(doc, num);
	if (entry->type == 'f')
		opts->gen_list[num] = entry->gen;
	if (entry->type == 'n')
		opts->gen_list[num] = entry->gen;
	if (entry->type == 'o')
		opts->gen_list[num] = 0;

	if (opts->do_garbage && !opts->use_list[num])
		return;

	if (entry->type == 'n' || entry->type == 'o') {
		opts->ofs_list[num] = opts->out->len;
		writeobject(doc, opts, num, opts->gen_list[num], 1);
	} else {
		opts->use_list[num] = 0;
    }
}

static void
writeobjects(pdf_document *doc, pdf_write_options *opts)
{
	int num;
	int xref_len = pdf_xref_len(doc);

	fz_buffer_printf(opts->out, "%%PDF-%d.%d\n", doc->version / 10, doc->version % 10);
	fz_buffer_puts(opts->out, "%%\316\274\341\277\246\n");
	for (num = 0; num < xref_len; num++)
		dowriteobject(doc, opts, num, 0);
}

fz_buffer *pdf_write_document(pdf_document *doc, fz_write_options *fz_opts)
{
	int lastfree, num, xref_offs = 0, xref_len = 0;
	pdf_write_options opts = { 0 };
	fz_context *ctx;
	fz_write_options fz_opts_defaults = { 0 };

	if (!doc)
		return NULL;

	if (!fz_opts)
		fz_opts = &fz_opts_defaults;

	ctx = doc->ctx;

	pdf_finish_edit(doc);

	xref_len = pdf_xref_len(doc);

    opts.out = fz_new_buffer(1048576);

	fz_try(ctx)
	{
		opts.do_expand = fz_opts->do_expand;
		opts.do_garbage = fz_opts->do_garbage;
        opts.do_deflate = fz_opts->do_deflate;
		/* We deliberately make these arrays long enough to cope with
		 * 1 to n access rather than 0..n-1, and add space for 2 new
		 * extra entries that may be required for linearization. */
		opts.use_list = fz_malloc_array(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.ofs_list = fz_malloc_array(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.gen_list = fz_malloc_array(ctx, pdf_xref_len(doc) + 3, sizeof(int));
		opts.continue_on_error = fz_opts->continue_on_error;
		opts.errors = fz_opts->errors;

		for (num = 0; num < xref_len; num++)
		{
			opts.use_list[num] = 0;
			opts.ofs_list[num] = 0;
			opts.gen_list[num] = 0;
		}

		/* Make sure any objects hidden in compressed streams have been loaded */
		preloadobjstms(doc);

		/* Sweep & mark objects from the trailer */
		if (opts.do_garbage >= 1)
			sweepobj(doc, &opts, pdf_trailer(doc));
		else
			for (num = 0; num < xref_len; num++)
				opts.use_list[num] = 1;

		writeobjects(doc, &opts);

			/* Construct linked list of free object slots */
			lastfree = 0;
			for (num = 0; num < xref_len; num++)
			{
				if (!opts.use_list[num])
				{
					opts.gen_list[num]++;
					opts.ofs_list[lastfree] = num;
					lastfree = num;
				}
			}

#if 0
        fz_buffer_printf(opts.out, "%d 0 obj\n<< /Type /Catalog\n   /Pages %d 0 R\n>>\nendobj\n",
				         catalog.id, surface->pages_resource.id);
#endif
        xref_offs = opts.out->len;
		writexref(doc, &opts, xref_len);
	    fz_buffer_printf(opts.out, "startxref\n%d\n%%%%EOF\n", xref_offs);
		doc->dirty = 0;
	}
	fz_always(ctx)
	{
		fz_free(ctx, opts.use_list);
		fz_free(ctx, opts.ofs_list);
		fz_free(ctx, opts.gen_list);
		doc->freeze_updates = 0;
	}
	fz_catch(ctx)
	{
		fz_rethrow(ctx);
	}
    return opts.out;
}

static void
pdf_rebalance_page_tree(pdf_document *doc)
{
}

void pdf_finish_edit(pdf_document *doc)
{
	if (!doc)
		return;
	pdf_rebalance_page_tree(doc);
}

